#include "../include/Adlsys.h"
#include "../include/Adldev.h"
#include "../include/Adllib.h"
#include "../include/AdlS593X.h"
#include "../include/Adlplx90.h"

U32 dwAdlPioKeepMask[5] = {0xffffffff, 0x000000ff, 0x0000ffff, 0x00ffffff, 0xffffffff};
U32 dwAdlPioDropMask[5] = {0xffffffff, 0xffffff00, 0xffff0000, 0xff000000, 0x00000000};


void S593FlushInTrace(U32 dwPort, PPCI_DBX pDbx)
{
#ifdef __SMP__
    //unsigned long irq_flags;
#endif
    U16 wIndex1, wIndex2;

#ifdef __SMP__
    //KeAcquireSpinLock(&pDbx->kSpinLock, &kIrql);
    //...spin_lock_irqsave(&pDbx->kSpinLock, irq_flags);
#endif

    if(pDbx->xFlag.f.DblBuf){
        wIndex1 = pDbx->xFlag.f.DrvHfRdyH ? 1:0;
        wIndex2 = pDbx->xFlag.f.DrvHfRdyH ? 0:1;
    }
    else{
        wIndex1 = 0;
        wIndex2 = 0;
    }
    pDbx->dwAccessPos = S593_InFlushFIFO(dwPort, pDbx->dwAccessCount,
        pDbx->xAddDbx[wIndex1].dwAddMap, pDbx->xAddDbx[wIndex2].dwAddMap
    );

#ifdef __SMP__
    //KeReleaseSpinLock(&pDbx->kSpinLock, kIrql);
    //...spin_unlock_irqrestore( &pDbx->kSpinLock, irq_flags );
#endif
}


BOOLEAN AdlIocGetDesp(PPCI_DEVEXT pDevExt, DAS_IOT_CARD_DESP *pSbuf)
{
    PPCI_DRVEXT pDrvExt = pDevExt->pDrvExt;
    PPCI_DBX pDbx;
    DAS_IOT_FUNC_DESP *pFdsp;
    DAS_IOT_PVIEW VAddr[4];
    int  vi;

    /*copy general description to user buffer*/
    memcpy(pSbuf, &(pDrvExt->CardDesp), sizeof(DAS_IOT_CARD_DESP));
    pSbuf->wErrorCode = DAS_ERR_NO;
    pSbuf->wBus = pDevExt->wBus;
    pSbuf->wSlot = (U16)pDevExt->dwSlot;

    /*get each FnctionBlock return data*/
    for(vi=0; vi<DAS_FUNC_COUNT; vi++){
        pFdsp = &pSbuf->mFunc[vi];
        pDbx = pDevExt->pDbx[vi];
        if(pDbx){
            //set buffer size
            pFdsp->dwBufferSize = pDbx->dwBufferSize;
            //set view address
            VAddr[vi].ViewAddr[0] = pDbx->hMapView[0];
            VAddr[vi].ViewAddr[1] = pDbx->hMapView[1];
            //for Get View-Address in MultiOpen
            //but map-view address is returned by mmap through System
            /*
            if(pDevExt->Reference){
                if(!AdlOpenDbxView(pDbx))
                    return 0;
            }
            */
            pFdsp->pView[0] = pDbx->hMapView[0];
            pFdsp->pView[1] = pDbx->hMapView[1];
            pDbx->hMapView[0] = VAddr[vi].ViewAddr[0];
            pDbx->hMapView[1] = VAddr[vi].ViewAddr[1];
        }
        else{
            pFdsp->dwBufferSize = 0L;
            pFdsp->pView[0] = NULL;
            pFdsp->pView[1] = NULL;
        }
    }

    //dual interrupt notify
    if(!pDevExt->pDin){
        pSbuf->fDualInterruptEventNotify = 0;
    }
    //trace event notify
    if(!pDevExt->pTcn){
        pSbuf->fTraceEventNotify = 0;
    }
    //printk("[Drv] pSbuf->mFunc[0].wFlag.bFunction: %d\n", pSbuf->mFunc[0].wFlag.bFunction);

    pSbuf->Reference = pDevExt->Reference;
    pSbuf->wErrorCode = DAS_ERR_NO;
    return TRUE;
}


/*
 * wEnable; wReady; wViewIndex;
 */
/*DAS_IOC_XX_DBF_MODE*/
BOOLEAN AdlIocDbfMode(PPCI_DEVEXT pDevExt, PPCI_DBX pDbx, DAS_IOT_DBF *pSbuf)
{
#ifdef __SMP__
    unsigned long irq_flags;
#endif

    if(!pDbx){
        pSbuf->wErrorCode = DAS_ERR_NO_FUNCTION;
        return FALSE;
    }

#ifdef __SMP__
    //KeAcquireSpinLock(&pDbx->kSpinLock, &kIrql);
    spin_lock_irqsave(&pDbx->kSpinLock, irq_flags);
#endif

    if(pDbx->xFlag.f.Running){
        pSbuf->wErrorCode = DAS_ERR_RUNNING;
    #ifdef __SMP__
        //KeReleaseSpinLock(&pDbx->kSpinLock, kIrql);
        spin_unlock_irqrestore(&pDbx->kSpinLock, irq_flags);
    #endif
        return FALSE;
    }

    pDbx->xFlag.f.DblBuf = pSbuf->wEnable? 1:0;
    pDbx->xFlag.f.UsrHfRdyH = 0;

#ifdef __SMP__
    //KeReleaseSpinLock(&pDbx->kSpinLock, kIrql);
    spin_unlock_irqrestore(&pDbx->kSpinLock, irq_flags);
#endif

    pSbuf->wErrorCode = DAS_ERR_NO;
    return TRUE;
}


/*DAS_IOC_XX_DBF_HALF_READY*/
BOOLEAN AdlIocDbfHalfReady(PPCI_DEVEXT pDevExt, PPCI_DBX pDbx, DAS_IOT_DBF *pSbuf)
{
#ifdef __SMP__
    unsigned long irq_flags;
#endif

    /*error check*/
    if(!pDbx){
        pSbuf->wErrorCode = DAS_ERR_NO_FUNCTION;
        return FALSE;
    }

#ifdef __SMP__
    //KeAcquireSpinLock(&pDbx->kSpinLock, &kIrql);
    spin_lock_irqsave(&pDbx->kSpinLock, irq_flags);
#endif

    /*not double buffer mode ==> return error*/
    if(!pDbx->xFlag.f.DblBuf){
        pSbuf->wErrorCode = DAS_ERR_DOUBLE_BUFFER_MODE;
    #ifdef __SMP__
        //KeReleaseSpinLock(&pDbx->kSpinLock, kIrql);
        spin_unlock_irqrestore(&pDbx->kSpinLock, irq_flags);
    #endif
        return FALSE;
    }
    /*check ready seqence according UsrHfRdyH flag*/
    /*set ready flag according HalfReady flag*/
    if(pDbx->xFlag.f.UsrHfRdyH){
        /*High ready*/
        pSbuf->wViewIndex = 1;
        pSbuf->wReady = (U16)pDbx->xFlag.f.HalfRdyH;
    }
    else{
        /*Low ready*/
        pSbuf->wViewIndex = 0;
        pSbuf->wReady = (U16)pDbx->xFlag.f.HalfRdyL;
    }

    if(pSbuf->wReady){
        /*Any buffer ready ==> modify check seqence*/
        pDbx->xFlag.f.UsrHfRdyH = !pDbx->xFlag.f.UsrHfRdyH;
    }

#ifdef __SMP__
    //KeReleaseSpinLock(&pDbx->kSpinLock, kIrql);
    spin_unlock_irqrestore(&pDbx->kSpinLock, irq_flags);
#endif

    pSbuf->wErrorCode = DAS_ERR_NO;
    return TRUE;
}


/*
 * dwAccessCount; wStopped;
 */
/*DAS_IOC_XX_ASY_CHECK*/
BOOLEAN AdlIocAsyCheck(PPCI_DEVEXT pDevExt, PPCI_DBX pDbx, DAS_IOT_ASYNC *pSbuf)
{
#ifdef __SMP__
    unsigned long irq_flags;
#endif

    /*error check*/
    if(!pDbx){
        pSbuf->wErrorCode = DAS_ERR_NO_FUNCTION;
        return FALSE;
    }

#ifdef __SMP__
    //KeAcquireSpinLock(&pDbx->kSpinLock, &kIrql);
    spin_lock_irqsave(&pDbx->kSpinLock, irq_flags);
#endif

    /*SyncMode ==> error*/
    if(pDbx->xFlag.f.SyncMode){
        pSbuf->wErrorCode = DAS_ERR_SYNC_MODE;
    #ifdef __SMP__
        //KeReleaseSpinLock(&pDbx->kSpinLock, kIrql);
        spin_unlock_irqrestore(&pDbx->kSpinLock, irq_flags);
    #endif
        return FALSE;
    }

    /*set return STOP flag*/
    pSbuf->wStopped = pDbx->xFlag.f.Running? 0:1;
    /*PCI Controller dependent Parts*/
    switch(pDevExt->wPciController){
        case PCI_CTRL_S593:
            pSbuf->dwAccessCount = pDbx->dwAccessCount;
            if(pDbx->xFlag.f.Output){
                /*Output Function*/
                //read access count from controller
                pSbuf->dwAccessCount -= S593_OutCountLeft(OP_Port(pDevExt));
                //if DMA stop ==> wait FIFO clean
                if(!pDbx->xFlag.f.Running)
                    pSbuf->wStopped = S593_OutFlushFIFO(OP_Port(pDevExt));
            }
            else{
                /*Input Function*/
                if(pDbx->xFlag.f.Tracing){
                    //Trace Notify
                    if(!pDbx->xFlag.f.Running)
                        pSbuf->dwAccessCount = pDbx->dwAccessPos;
                    else
                        pSbuf->dwAccessCount = 0L;
                }
                else{
                    //non trace mode
                    if(pDbx->xFlag.f.Running)
                        /*DMA stop ==> read count from controller*/
                        pSbuf->dwAccessCount -= S593_InCountLeft(OP_Port(pDevExt));
                }
            }
            break;
        case PCI_CTRL_PLX90:
            /*PLX9050 not support DMA, driver manage access count internal*/
            pSbuf->dwAccessCount = pDbx->dwAccessPos;
            break;
        case PCI_CTRL_P9080:
            pSbuf->dwAccessCount = pDbx->dwAccessPos;
            break;
        default:
            break;
    }

    /*return access count(internal we save byte count)*/
    if(pDbx->wBytesPerData!=0)
        pSbuf->dwAccessCount /= pDbx->wBytesPerData;
    else
        pSbuf->dwAccessCount = 0;

#ifdef __SMP__
    //KeReleaseSpinLock(&pDbx->kSpinLock, kIrql);
    spin_unlock_irqrestore(&pDbx->kSpinLock, irq_flags);
#endif

    pSbuf->wErrorCode = DAS_ERR_NO;
    return TRUE;
}


/*DAS_IOC_XX_ASY_CLEAR*/
BOOLEAN AdlIocAsyClear(PPCI_DEVEXT pDevExt, PPCI_DBX pDbx, DAS_IOT_ASYNC *pSbuf)
{
    U32 dwPort;
#ifdef __SMP__
    unsigned long irq_flags;
#endif

    /*error check*/
    if(!pDbx){
        pSbuf->wErrorCode = DAS_ERR_NO_FUNCTION;
        return FALSE;
    }

#ifdef __SMP__
    //KeAcquireSpinLock(&pDbx->kSpinLock, &kIrql);
    spin_lock_irqsave(&pDbx->kSpinLock, irq_flags);
#endif

    /*SyncMode ==> error*/
    if(pDbx->xFlag.f.SyncMode){
        pSbuf->wErrorCode = DAS_ERR_SYNC_MODE;
    #ifdef __SMP__
        //KeReleaseSpinLock(&pDbx->kSpinLock, kIrql);
        spin_unlock_irqrestore(&pDbx->kSpinLock, irq_flags);
    #endif
        return FALSE;
    }

    /*Force return STOP flag*/
    pSbuf->wStopped = 1;
    dwPort = OP_Port(pDevExt);
    /*PCI Controller dependent Parts*/
    switch(pDevExt->wPciController){
        case PCI_CTRL_S593:
            pSbuf->dwAccessCount = pDbx->dwAccessCount;
            if(pDbx->xFlag.f.Output)
                pSbuf->dwAccessCount -= S593_StopODMA(dwPort);
            else{
                if(pDbx->xFlag.f.Tracing){
                    if(!pDbx->xFlag.f.Running)
                        pSbuf->dwAccessCount = pDbx->dwAccessPos;
                    else{
                        pSbuf->dwAccessCount -= S593_StopIDMA(dwPort);
                        S593_StopIMB(dwPort);
                    }
                }
                else{
                    if(pDbx->xFlag.f.Running){
                        //yuan add 06/05/03
                        OP_REG opr;
                        //yuan add 03/30/01
                        //...S593FlushInTrace(dwPort, pDbx);
                        //...pSbuf->dwAccessCount = pDbx->dwAccessPos;
                        /*Disable DMA complete interrupt*/
                        opr.r = OP_inpd(dwPort,INTCSR);
                        opr.intcsr.C_MWTC_IE = 0;
                        OP_outd(dwPort,INTCSR, opr.r);
                        /*Disable DMA transfer, Flush FIFO*/
                        opr.r = OP_inpd(dwPort, MCSR);
                        //...opr.mcsr.C_A2P_FLAG_RESET = 1;
                        opr.mcsr.C_A2P_EN_TRANSFER = 0;
                        OP_outd(dwPort, MCSR, opr.r);
                        S593FlushInTrace(dwPort, pDbx);
                        pSbuf->dwAccessCount = pDbx->dwAccessPos;
                        opr.r = OP_inpd(dwPort, MCSR);
                        opr.mcsr.C_A2P_FLAG_RESET = 1;
                        opr.mcsr.C_A2P_EN_TRANSFER = 0;
                        OP_outd(dwPort, MCSR, opr.r);
                    }
                }
            }
            break;
        case PCI_CTRL_PLX90:
            /*PLX9050 not support DMA, driver manage access count internal*/
            pSbuf->dwAccessCount = pDbx->dwAccessPos;
            break;
        case PCI_CTRL_P9080:
            pSbuf->dwAccessCount = pDbx->dwAccessPos;
            break;
        default:
            break;
    }

    /*return access count(internal we save byte count)*/
    if(pDbx->wBytesPerData!=0)
        pSbuf->dwAccessCount /= pDbx->wBytesPerData;
    else
        pSbuf->dwAccessCount = 0;
    /*clean DBX flags*/
    pDbx->xFlag.f.Running = 0;
    pDbx->xFlag.f.DrvHfRdyH = 0;
    //yuan remove 07/25/01
    pDbx->xFlag.f.HalfRdyL = pDbx->xFlag.f.HalfRdyH = 0;
    /*return*/

#ifdef __SMP__
    //KeReleaseSpinLock(&pDbx->kSpinLock, kIrql);
    spin_unlock_irqrestore( &pDbx->kSpinLock, irq_flags );
#endif

    pSbuf->wErrorCode = DAS_ERR_NO;
    return TRUE;
}



/*
 * dwAccessCount; dwPacerValue; wAutoScan;
 * wIterations; wSyncMode; wChannelCount; dwChannelGain[1];
 */
/*DAS_IOC_XX_CON_READ*/
BOOLEAN AdlIocContPreProc(PPCI_DEVEXT pDevExt, PPCI_DBX pDbx, DAS_IOT_CONT *pSbuf)
{
#ifdef __SMP__
    unsigned long irq_flags;
#endif

    /*error check*/
    if(!pDbx ){
        pSbuf->wErrorCode = DAS_ERR_NO_FUNCTION;
        return FALSE;
    }

#ifdef __SMP__
    //KeAcquireSpinLock(&pDbx->kSpinLock, &kIrql);
    spin_lock_irqsave(&pDbx->kSpinLock, irq_flags);
#endif

    /*check Channel Count*/
    if(pSbuf->wChannelCount>pDbx->wCGArraySize){
        pSbuf->wErrorCode = DAS_ERR_CHANNEL_GAIN_COUNT;
        goto PreContError;
    }
    /*check Pacer Value*/
    /*
    if(pSbuf->dwPacerValue<(U32)pDbx->wPacerMin){
        pSbuf->wErrorCode = DAS_ERR_PACER_COUNT;
        goto PreContError;
    }
    */
    /*check still running*/
    if(pDbx->xFlag.f.Running){
        pSbuf->wErrorCode = DAS_ERR_RUNNING;
        goto PreContError;
    }
    /*get internal access BYTE count*/
    pDbx->dwAccessCount = pSbuf->dwAccessCount*pDbx->wBytesPerData;
    /*decide SyncMode flag*/
    pDbx->xFlag.f.SyncMode = (pSbuf->wSyncMode==DAS_SYN_SYNCH)? 1:0;
    /*SyncMode error check*/
    if(pDbx->xFlag.f.SyncMode){
        if(pDbx->xFlag.f.Tracing||pDbx->xFlag.f.DblBuf){
            /*No Trace mode, no DoubleBufferMode*/
            pSbuf->wErrorCode = DAS_ERR_SYNC_MODE;
            goto PreContError;
        }
    }
    /*Chain DMA*/
    if(pDbx->xFlag.f.ChainDMA){
        /*No DoubleBuffer*/
        if(pDbx->xFlag.f.DblBuf){
            pSbuf->wErrorCode = DAS_ERR_DOUBLE_BUFFER_MODE;
            goto PreContError;
        }
    }
    else{
        /*Non Chain Mode*/
        /*DoubleBuffer Mode check*/
        if(pDbx->xFlag.f.DblBuf){
            if(pDbx->dwAccessCount>pDbx->dwBufferSize/2){
                /*Acees Count not exceed Half Buffer size*/
                pSbuf->wErrorCode = DAS_ERR_ACCESS_COUNT;
                goto PreContError;
            }
        }
        else{
            /*Single Buffer Mode*/
            /*Acees Count not exceed Full Buffer size*/
            if(pDbx->dwAccessCount>pDbx->dwBufferSize){
                pSbuf->wErrorCode = DAS_ERR_ACCESS_COUNT;
                goto PreContError;
            }
        }

        pDbx->xAddDbx[1].dwLength = pDbx->xAddDbx[0].dwLength = pDbx->dwAccessCount;
        pDbx->xFlag.f.HalfRdyH = pDbx->xFlag.f.UsrHfRdyH = pDbx->xFlag.f.Output;
    }

    //KeClearEvent(pDbx->pkEvent);
    pDbx->xFlag.f.DrvHfRdyH = 0;
    pDbx->xFlag.f.HalfRdyL = 0;
    pDbx->xFlag.f.AutoScan = pSbuf->wAutoScan? 1:0;
    pDbx->xFlag.f.Running = 1;
    pDbx->wIterations = pDbx->xFlag.f.Output? pSbuf->wIterations:1;
    pDbx->dwAccessPos = 0L;

#ifdef __SMP__
    //KeReleaseSpinLock(&pDbx->kSpinLock, kIrql);
    spin_unlock_irqrestore(&pDbx->kSpinLock, irq_flags);
#endif

    pSbuf->wErrorCode = DAS_ERR_NO;
    return TRUE;

PreContError:

#ifdef __SMP__
    //KeReleaseSpinLock(&pDbx->kSpinLock, kIrql);
    spin_unlock_irqrestore(&pDbx->kSpinLock, irq_flags);
#endif

    return FALSE;
}


int AdlIocContPostProc(PPCI_DEVEXT pDevExt, PPCI_DBX pDbx)
{
    U32 dwPort;
#ifdef __SMP__
    unsigned long irq_flags;
#endif

    /*error check*/
    if(!pDbx)
        return DAS_ERR_DOUBLE_BUFFER_MODE;

    dwPort = OP_Port(pDevExt);

#ifdef __SMP__
    //KeAcquireSpinLock(&pDbx->kSpinLock, &kIrql);
    spin_lock_irqsave(&pDbx->kSpinLock, irq_flags);
#endif

    if(pDevExt->wPciController==PCI_CTRL_S593){
        if(pDbx->xFlag.f.Output){
            S593_StartODMA(dwPort, pDbx->xAddDbx);
        }
        else{
            S593_StartIDMA(dwPort, pDbx->xAddDbx);
            if(pDbx->xFlag.f.Tracing)
                S593_StartIMB(dwPort, MAILBOX4, FIFOBYTE3);
        }
    }

#ifdef __SMP__
    //KeReleaseSpinLock(&pDbx->kSpinLock, kIrql);
    spin_unlock_irqrestore(&pDbx->kSpinLock, irq_flags);
#endif

    return (pDbx->xFlag.f.SyncMode)? DAS_STATUS_PENDING:DAS_ERR_NO;
}


BOOLEAN AdlIocPioCheck(U16 wBytesPerData, DAS_IOT_PIO *pSbuf)
{
    if(pSbuf->wByteCount>wBytesPerData){
        pSbuf->wErrorCode = DAS_ERR_CHANNEL_BYTE_COUNT;
        return FALSE;
    }

    if(pSbuf->wByteCount){
        pSbuf->dwValue &= dwAdlPioKeepMask[pSbuf->wByteCount];
    }
    else{
        if(pSbuf->wLine_Gain>wBytesPerData*8){
            pSbuf->wErrorCode = DAS_ERR_CHANNEL_LINE;
            return FALSE;
        }
    }

    pSbuf->wErrorCode = DAS_ERR_NO;
    return TRUE;
}


BOOLEAN AdlIocPioRead(U16 wBytesPerData, DAS_IOT_PIO *pSbuf)
{
    if(!AdlIocPioCheck(wBytesPerData, pSbuf))
        return FALSE;

    if(pSbuf->wByteCount==0){
        pSbuf->dwValue >>= pSbuf->wLine_Gain;
        pSbuf->dwValue &= 1L;
    }

    return TRUE;
}


BOOLEAN AdlIocPioWrite(U16 wBytesPerData, DAS_IOT_PIO *pSbuf, U32 dwOldIn)
{
    U32 dwBitMask;

    if(!AdlIocPioCheck(wBytesPerData, pSbuf))
        return FALSE;

    if(pSbuf->wByteCount){
        pSbuf->dwValue |= (dwOldIn&dwAdlPioDropMask[pSbuf->wByteCount]);
    }
    else{
        pSbuf->dwValue = (pSbuf->dwValue&1L)<<(pSbuf->wLine_Gain);
        dwBitMask = 1L<<(pSbuf->wLine_Gain);
        pSbuf->dwValue |= (dwOldIn&(~dwBitMask));
    }

    return TRUE;
}


void AdlIocPortIoWrite(PPCI_DEVEXT pDevExt, U16 wBytesPerData, DAS_IOT_PORTIO *pSbuf)
{
    U32 dwPort;

    if(pSbuf->wByteCount!=wBytesPerData){
        pSbuf->wErrorCode = DAS_ERR_PORTIO_BYTE_COUNT;
        //00/06/05 jeffrey added
        return;
    }

    dwPort = PT_Port(pDevExt)+pSbuf->wPort;
    switch(pSbuf->wByteCount){
        case 1:
            outB(dwPort, (U8)pSbuf->dwValue);
            break;
        case 2:
            outW(dwPort, (U16)pSbuf->dwValue);
            break;
        case 4:
            outD(dwPort, pSbuf->dwValue);
            break;
        default:
            break;
    }
}


void AdlIocPortIoRead(PPCI_DEVEXT pDevExt, U16 wBytesPerData, DAS_IOT_PORTIO *pSbuf)
{
    U32 dwPort;

    if(pSbuf->wByteCount!=wBytesPerData){
        pSbuf->wErrorCode = DAS_ERR_PORTIO_BYTE_COUNT;
        //00/06/05 jeffrey added
        return;
    }

    dwPort = PT_Port(pDevExt)+pSbuf->wPort;
    switch(pSbuf->wByteCount){
        case 1:
            pSbuf->dwValue = inpB(dwPort);
            break;
        case 2:
            pSbuf->dwValue = inpW(dwPort);
            break;
        case 4:
            pSbuf->dwValue = inpD(dwPort);
            break;
        default:
            break;
    }

    pSbuf->wErrorCode = DAS_ERR_NO;
}


void AdlIocGetBufSize(PPCI_DEVEXT pDevExt, DAS_IOT_FUNC_BUFSIZE *pSbuf)
{
    PPCI_DBX pDbx;
    int vi;

    for(vi=0; vi<DAS_FUNC_COUNT; vi++){
        pDbx = pDevExt->pDbx[vi];
        if(pDbx){
            /*set buffer size*/
            pSbuf->BufSize[vi] = pDbx->dwBufferSize;
        }
        else
            pSbuf->BufSize[vi] = 0;
    }
}


U32 AdlSplitePacer(U32 dwPacerIn)
{
    U32 dwV1, dwV2;

    if(dwPacerIn<4)
        return 0x00020002;

    dwV1 = dwPacerIn>>1;
    dwV2 = 2;
    while(dwV1&0xFFFF0000){
        dwV1 >>= 1;
        dwV2 <<= 1;
    }

    return (dwV1<<16)|dwV2;
}
