//#include <linux/config.h>
#include <linux/delay.h>
#include <linux/pagemap.h>

//#define __SMP__
#if defined(__SMP__)
#include <linux/smp.h>
#endif

#if defined(MODVERSIONS)
#if defined(__SMP__)
#include <linux/modversions-smp.h>
#else
#include <linux/modversions.h>
#endif
#endif

#include <asm/uaccess.h>
#include <linux/pagemap.h>

#include "../include/Adlsys.h"
#include "../include/Adldev.h"
#include "../include/Adllib.h"
#include "../include/Adlp9080.h"


#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,20)
#define PTE_OFFSET pte_offset
#else
#ifdef pte_offset_map
#define PTE_MAP
#define PTE_OFFSET  pte_offset_map
#else
#ifdef pte_offset_kernel
#define PTE_OFFSET  pte_offset_kernel
#else
#define PTE_OFFSET pte_offset
#endif
#endif
#endif
//#define PTE_OFFSET pte_offset_kernel
//#endif

#ifdef pte_offset_atomic
#define SUSE_ATOMIC_PTE
#define PTE_OFFSET pte_offset_atomic
#endif

//#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,20)
//#define PTE_OFFSET pte_offset
//#else
//#define PTE_OFFSET pte_offset_kernel
//#endif


void P9080_SoftReset(U32 dwPort)
{
    //U16 wIter = 100;
    OP_outb(dwPort,EEPROM,0xd0);
    /*
    while(wIter){
        OP_inpb(dwPort, EEPROM);
        wIter--;
    }
    */
    //11/30/07 modified
    mdelay(100); //100msec
    OP_outb(dwPort,EEPROM,0x90);
}


void P9080_SetDMAWidth(U32 dwPort, U32 dwChannel, U32 dwBytes)
{
    U32 dwRPT;
    OP_REG rOP;
    U32 dwTmp;

    dwRPT = dwPort+(dwChannel? ADL_OP_DMAMODE1:ADL_OP_DMAMODE0);
    rOP.r = inpD(dwRPT);
    switch(dwBytes){
        case 1:
            dwTmp = 0;
            break;
        case 2:
            dwTmp = 1;
            break;
        case 4:
            dwTmp = 2;
            break;
        default:
            return;
    }
    rOP.dmamode.LocalBusWidth = dwTmp;
    outD(dwRPT, rOP.r);
}


BOOLEAN P9080_StopDMA(U32 dwPort, U32 dwChannel)
{
    U32 dwRPT;
    OP_REG rOP;
    U16 wIter;
    BOOLEAN fOK = 0;

    /*Check channel*/
    if(dwChannel>1)
        return FALSE;
    /*Disable DMA0/DMA1 Interrupt Notify*/
    //09/13/00 jeffrey, mark this section with the new drivers from YUAN
    /*
    dwRPT = dwPort+ADL_OP_INTCSR;
    rOP.r = inpD(dwRPT);
    if(dwChannel)
        rOP.intcsr.IE_L_DMA1 = 0;
    else
        rOP.intcsr.IE_L_DMA0 = 0;
    outD(dwRPT, rOP.r);
    */
    /*Get DMACSR Port Address*/
    dwRPT = dwPort+(dwChannel? ADL_OP_DMACSR1:ADL_OP_DMACSR0);
    /*Get DMACSR Value*/
    rOP.b[0] = inpB(dwRPT);
    /*Disable DMACSR*/
    //09/13/00 jeffrey , mark this line with the new driver from YUAN
    //outB(dwRPT, 0);
    /*abort DMA if DMA not finish*/
    if(rOP.dmacsr.ChnlDone==0){
        fOK = FALSE;
        //yuan add 04/10/01
        rOP.r = 0;
        rOP.dmacsr.ChnlStart = 1;
        outB(dwRPT, rOP.b[0]);
        rOP.r = 0;
        rOP.dmacsr.ChnlStart = 1;
        rOP.dmacsr.ChnlAbort = 1;
        outB(dwRPT, rOP.b[0]);
        for(wIter=100; wIter; wIter--){
            rOP.b[0] = inpB(dwRPT);
            if(rOP.dmacsr.ChnlDone){
                fOK = TRUE;
                break;
            }
        }
    }
    /*else{
        //09/13/00 jeffrey, add Disable DMACSR with the new driver from YUAN
        outB(dwRPT, 0);
        fOK = TRUE;
    }*/
    //yuan test 07/20/01
    if(!fOK){
        P9080_ClearController(dwPort);
        fOK=1;
    }
    //Clear DMA Complete Interrupt
    rOP.r = 0;
    rOP.dmacsr.ClearInterrupt = 1;
    outB(dwRPT, rOP.b[0]);

    //return status
    return fOK;
}


BOOLEAN P9080_StartDMA(U32 dwPort, U32 dwChannel, PPCI_DBX pDbx)
{
    U32 dwRPT;
    OP_REG rOP;

    /*Check channel*/
    if(dwChannel>1)
        return FALSE;
    /*Enable Chain Mode*/
    dwRPT = dwPort+(dwChannel? ADL_OP_DMAMODE1:ADL_OP_DMAMODE0);
    rOP.r = inpD(dwRPT);
    rOP.dmamode.Chain = 1;
    //yuan add 02/19/04
    rOP.dmamode.DoneInterrupt = 1;
    rOP.dmamode.DmaInterrupt = 1;
    outD(dwRPT, rOP.r);
    /*set Descriptor Address Reg*/
    dwRPT = dwPort+(dwChannel? ADL_OP_DMADPR1:ADL_OP_DMADPR0);
    rOP.r = AdlDscrHeadPhy(pDbx)|(pDbx->xFlag.f.Output?0x1L:0x9L);
    outD(dwRPT, rOP.r);

    /*Enable DMA0/DMA1 interrupt*/
    dwRPT = dwPort+ADL_OP_INTCSR;
    rOP.r = inpD(dwRPT);
    rOP.intcsr.IE_P_Out = 1;
    rOP.intcsr.IE_P_Local = 1;
    if(dwChannel)
        rOP.intcsr.IE_L_DMA1 = 1;
    else
        rOP.intcsr.IE_L_DMA0 = 1;
    outD(dwRPT, rOP.r);
    /*Enable and Start DMA0/DMA1*/
    dwRPT = dwPort+(dwChannel? ADL_OP_DMACSR1:ADL_OP_DMACSR0);
    rOP.r = 0;
    rOP.dmacsr.ChnlEnable = 1;
    outB(dwRPT, rOP.b[0]);
    rOP.dmacsr.ChnlStart = 1;
    outB(dwRPT, rOP.b[0]);

    return TRUE;
}

/*
 *** LAS0RR
 *       b0 Memory space : PCI I/O space
 *       b1b2 Locate anywhere in 32 bit PCI address space
 *       b3..b31 PCI to space0 decoding range
 *** LAS0BA
 *       b0 space 0 enable
 *       b2..b31 space0 remap range
 *** LBRD0 --- Local Address Space 0__Exp ROM
 *       b0b1 : 32bit
 *       b6b7 ReadyInput Enable, BTERM enable
 *       b24 : memory space 0 burst enable
 *       b25 : load SubSystemID and local addr 1 reg
 *       b26 : expandion ROM space burst disable
 *       b27 : de-assert TRDY# when FIFO full
 *       b28..b31 : PCI target retry delay cloks
 */
void P9080_ClearController(U32 dwPort)
{
    //yuan remove 02/19/04
    P9080_SoftReset(dwPort);
    OP_outd(dwPort, LAS0RR, 0xffffffc1L); /*LAS0RR*/
    OP_outd(dwPort, LAS0BA, 0x00000001L); /*LAS0BA*/
    OP_outd(dwPort, LBRD0 , 0x4b0003c3L); /*LBRD0*/
    OP_outd(dwPort, INTCSR, 0x00000000L);
    /*PCI (memory)  read command code : 6*/
    /*PCI (memory) write command code : 7*/
    OP_outd(dwPort, CNTRL , 0x9000767cL); //Alex Lee 8/14/99 use read memory multiple command to increase PCI throughput;
    /*clear DMA channel 0*/
    OP_outb(dwPort, DMACSR0, 0x0);
    /*32bit local bus width,BTERM en, Ready en*/
    /*DemandMode, local addr fix, chain*/
    OP_outd(dwPort, DMAMODE0, 0x00021dc1L);
    /*clear DMA channel 1*/
    OP_outb(dwPort, DMACSR1, 0x0L);
    OP_outd(dwPort, DMAMODE1, 0x00021d81L);
    OP_outd(dwPort, DMATHR  , 0x44447a47L);
}


void P9080_ClearMBufDscr(PPCI_DBX pDbx)
{
    U32 vi;
    U8 notroot = 0; //pDbx->xFlag.f.DlyTrig;

    for(vi=0; vi<pDbx->wIterations; vi++){
        //yuan add 06/12/06
        notroot = ((U32) pDbx->hMapView[vi])&0x1;
        pDbx->hMapView[vi] = (void *)(((U32)pDbx->hMapView[vi])&0xfffffffe);
        if(notroot&&pDbx->hMapView[vi]){
            int i;
            struct page **pages = ((P_P9080_MDLADDR) pDbx->hMapView[vi])->pages;
            U32 nr_pages = ((P_P9080_MDLADDR)pDbx->hMapView[vi])->pagecnt;
            for(i=0; i<nr_pages; i++){
                if(!PageReserved(pages[i]))
                    SetPageDirty(pages[i]);
                /* FIXME: cache flush missing for rw==READ
                 * FIXME: call the correct reference counting function
                 */
                page_cache_release(pages[i]);
            }
            kfree(pages);
            kfree(pDbx->hMapView[vi]);
        }
        pDbx->hMapView[vi] = NULL;
        (pDbx->xAddMbx[vi]).dwAddPhy = 0;
        (pDbx->xAddMbx[vi]).dwAddMap = 0;
    }
    if(vi<8){
        (pDbx->xAddMbx[vi]).dwAddPhy = 0;
        (pDbx->xAddMbx[vi]).dwAddMap = 0;
    }

    pDbx->wIterations = 0;
    memset((void *)AdlDscrHeadMap(pDbx), 0x00, AdlDscrHeadLen(pDbx));
    AdlDscrLength(pDbx) = 0;
    pDbx->xAddMbx[0].dwAddPhy = AdlDscrHeadPhy(pDbx);
    pDbx->xAddMbx[0].dwAddMap = AdlDscrHeadMap(pDbx);
}


void P9080_ClearDmaDscr(PPCI_DBX pDbx)
{
    U8 notroot = pDbx->xFlag.f.DlyTrig;
    //yuan add 06/12/06
    if(notroot&&pDbx->hMapView[0]){
        int i;
        struct page **pages = ((P_P9080_MDLADDR)pDbx->hMapView[0])->pages;
        U32 nr_pages = ((P_P9080_MDLADDR)pDbx->hMapView[0])->pagecnt;
        for(i=0; i<nr_pages; i++){
            if (!PageReserved(pages[i]))
                SetPageDirty(pages[i]);
            /* FIXME: cache flush missing for rw==READ
             * FIXME: call the correct reference counting function
             */
            page_cache_release(pages[i]);
        }
        kfree(pages);
        kfree(pDbx->hMapView[0]);
        pDbx->hMapView[0] = 0;
    }
    memset((void *)AdlDscrHeadMap(pDbx), 0x00, AdlDscrHeadLen(pDbx));
    AdlDscrLength(pDbx) = 0;
    //yuan add 04/10/01
    pDbx->wIterations = 0;
}


BOOLEAN P9080_BuildDmaDscr(PPCI_DBX pDbx)
{
    P_P9080_DMADSCR pDscrBA;
    U32 dwDscrPA;
    U32 dwLockPA, dwLockPA_i;
    U32 dwLockBA, dwLockBA_i;
    U32 dwLockSize, pagecnt;
    U32 dwCbpOfst, dwCbpOfst_i;
    U32 dwCbpSize, times = 0;
    //yuan 04/02/01
    U16 wIters;
    //yuan add 06/12/06
    U8 needlock = (U8)pDbx->xFlag.f.DlyTrig; //checking
    struct page **pages;

    dwLockBA = (U32)AdlDscrUserBuffer(pDbx);
    dwLockSize = AdlDscrLength(pDbx);

    /*
     * 0 < User Buffer length < 10**23
     * user buffer must dword align
     */
    if((dwLockSize==0)||(dwLockSize&0x00000003L)||(BYTE_OFFSET(dwLockBA)&0x3L)){
        printk("[PLX9080] Address is not dword align\n");
        return FALSE;
    }

    //yuan add 04/02/01
    pagecnt = ADDRESS_AND_SIZE_TO_SPAN_PAGES(dwLockBA,dwLockSize);
    wIters = pDbx->wIterations? pDbx->wIterations:1;
    if((wIters*pagecnt)>pDbx->dwBufferSize/sizeof(T_P9080_DMADSCR)){
        //return FALSE;
        //if(ADDRESS_AND_SIZE_TO_SPAN_PAGES(dwLockBA,dwLockSize)>pDbx->dwBufferSize/sizeof(T_P9080_DMADSCR)){
        printk("[PLX9080] Memory for page_descriptor is too small\n");
        return FALSE;
    }

    //yuan add 06/16/06
    if(needlock){
        int res = 0;
        if((pages = kmalloc(pagecnt*sizeof(*pages), GFP_KERNEL))==NULL)
            //return -ENOMEM;
            return FALSE;
        //printk("need lock\n");
        /*Try to fault in all of the necessary pages*/
        down_read(&current->mm->mmap_sem);
        /*rw==READ means read from drive, write into memory area*/
        res = get_user_pages(
            current,
            current->mm,
            dwLockBA,
            pagecnt,
            1,
            0, /*don't force*/
            pages,
            NULL
        );
        up_read(&current->mm->mmap_sem);
        if(res<0){
            /*Try to fault in all of the necessary pages*/
            down_read(&current->mm->mmap_sem);
            /*rw==READ means read from drive, write into memory area*/
            res = get_user_pages(
                current,
                current->mm,
                dwLockBA,
                pagecnt,
                1,
                0, /*don't force*/
                pages,
                NULL
            );
            up_read(&current->mm->mmap_sem);
            if(res!=pagecnt){
                for( ; res; --res)
                    page_cache_release(pages[res-1]);
                kfree(pages);
                return FALSE;
            }
        }
        pDbx->hMapView[0] = (P_P9080_MDLADDR)(kmalloc(sizeof(T_P9080_MDLADDR), GFP_KERNEL));
        ((P_P9080_MDLADDR) pDbx->hMapView[0])->BufAddr = dwLockBA;
        ((P_P9080_MDLADDR) pDbx->hMapView[0])->pagecnt = pagecnt;
        ((P_P9080_MDLADDR) pDbx->hMapView[0])->pages = pages;
        //dwLockPA = (unsigned long) page_to_phys(pages[0]);
        //dwCbpOfst = dwLockBA & ~PAGE_MASK;
    }

    //printk("dwLockBA: %x, size: %d, iter: %d\n", dwLockBA, dwLockSize, wIters);
    dwLockPA = dwLockPA_i = translate_phys_addr((U32)dwLockBA, &dwCbpOfst); //only page_boundary address
    if(dwLockPA==0){ //check for translated address , 0 maybe caused by Un-Locked memory
        printk("[PLX9080] USER_VA to PCI_PA failed\n" );
        return FALSE;
    }
    dwCbpOfst_i = dwCbpOfst;
    pDbx->dwAccessCount = dwLockSize;
    pDbx->dwAccessPos = 0L;
    dwCbpSize = dwLockSize;
    dwDscrPA = AdlDscrHeadPhy(pDbx);
    pDscrBA = (P_P9080_DMADSCR)AdlDscrHeadMap(pDbx);
    //printk("pDscrBA: %x %x\n", pDscrBA, dwDscrPA);
    dwLockBA =  dwLockBA & PAGE_MASK; //page_align for next addr_translate
    dwLockBA_i =  dwLockBA;

    //yuan add 04/10/01
    while(wIters>0){
        dwCbpSize = dwLockSize;
        dwLockBA = dwLockBA_i;
        dwLockPA = dwLockPA_i;
        dwCbpOfst = dwCbpOfst_i;
        while(1){
            times++;
            pDscrBA->PciAddr = dwLockPA+dwCbpOfst;
            dwLockBA += PAGE_SIZE;
            pDscrBA->LocalAddr = AdlDscrLocalAddr(pDbx);
            //printk("times: %d\n", times);
            if((dwCbpOfst+dwCbpSize)>PAGE_SIZE){
                pDscrBA->DmaSize = PAGE_SIZE-dwCbpOfst;
                dwCbpOfst = 0L;
                dwCbpSize -= pDscrBA->DmaSize;
                dwDscrPA += sizeof(T_P9080_DMADSCR);
                pDscrBA->NextDscrAddr = dwDscrPA|(pDbx->xFlag.f.Output? 0x1L:0x9L);
                pDscrBA++;
                //translate to phys-addr for next page
                dwLockPA = translate_phys_addr((U32)dwLockBA, &dwCbpOfst); //only page_boundary address
                if(dwLockPA==0){ //check for translated address , 0 maybe caused by Un-Locked memory
                    printk("[PLX9080] USER_VA to PCI_PA failed in while_loop\n" );
                    return FALSE;
                }
                continue;
            }
            else{
                pDscrBA->DmaSize = dwCbpSize;
                //pDscrBA->NextDscrAddr = pDbx->xFlag.f.Output? 0x3L:0xbL;
                //yuan 04/02/01
                if(!pDbx->wIterations){
                    pDscrBA->NextDscrAddr = \
                        AdlDscrHeadPhy(pDbx)|(pDbx->xFlag.f.Output? 0x1L:0x9L);
                }
                else if(pDbx->wIterations==1){
                    pDscrBA->NextDscrAddr = pDbx->xFlag.f.Output? 0x3L:0xbL;
                }
                else{
                    if(wIters==1)
                        pDscrBA->NextDscrAddr = pDbx->xFlag.f.Output? 0x3L:0xbL;
                    else{
                        dwDscrPA += sizeof(T_P9080_DMADSCR);
                        pDscrBA->NextDscrAddr = \
                            dwDscrPA|(pDbx->xFlag.f.Output? 0x1L:0x9L);
                        pDscrBA++;
                    }
                }
                wIters -= 1;
                break;
            }//else
        }//while(1)
    }//while of wIter
    //printk("final times: %d\n", times);

    return TRUE;
}


//yuan 02/26/01 for multi-buffer setting
BOOLEAN P9080_BuildMBufDscr(DAS_IOT_MBUF *pSbuf, PPCI_DBX pDbx)
{
    P_P9080_DMADSCR pDscrBA, tmpDscrBA;
    U32 dwDscrPA;
    U32 dwLockPA;
    U32 dwLockBA;
    U32 dwLockSize;
    U32 dwCbpOfst;
    U32 dwCbpSize, pagecnt,dwValSize;
    //yuan add 06/12/06
    U8 needlock = (U8) pSbuf->wErrorCode;/*pDbx->xFlag.f.DlyTrig;*/ //checking
    struct page **pages;

    dwLockBA = pSbuf->dwBufferAddr;
    dwLockSize = pSbuf->dwAccessCount * pDbx->wBytesPerData;
    dwValSize = pDbx->dwBufferSize - ((pDbx->xAddMbx[pDbx->wIterations]).dwAddMap - AdlDscrHeadMap(pDbx));
    //yuan add 02/26/01
    if(pDbx->wIterations > MAXMBUFCNT)
        return FALSE;
    /*
     * 0 < User Buffer length < 10**23
     * user buffer must dword align
     */
    if((dwLockSize==0)||(dwLockSize&0x00000003L)||(BYTE_OFFSET(dwLockBA)&0x3L)){
        printk("[PLX9080] Address is not dword align\n" );
        return FALSE;
    }

    if((pagecnt = ADDRESS_AND_SIZE_TO_SPAN_PAGES(dwLockBA,dwLockSize))> dwValSize/sizeof(T_P9080_DMADSCR)){
        printk("[PLX9080] Memory for page_descriptor is too small\n");
        return FALSE;
    }
    //pDbx->hMapView[pDbx->wIterations] = (void *)dwLockBA;

    pDbx->dwAccessCount = dwLockSize;
    pDbx->dwAccessPos = 0L;
    //yuan add 06/16/06
    if(needlock){
        int res = 0;
        if((pages = kmalloc(pagecnt*sizeof(*pages), GFP_KERNEL))==NULL)
            return -ENOMEM;

        /*Try to fault in all of the necessary pages*/
        down_read(&current->mm->mmap_sem);
        /*rw==READ means read from drive, write into memory area*/
        res = get_user_pages(
            current,
            current->mm,
            dwLockBA,
            pagecnt,
            1,
            0, /*don't force*/
            pages,
            NULL
        );
        up_read(&current->mm->mmap_sem);
        if(res<0){
            /*Try to fault in all of the necessary pages*/
            down_read(&current->mm->mmap_sem);
            /*rw==READ means read from drive, write into memory area*/
            res = get_user_pages(
                current,
                current->mm,
                dwLockBA,
                pagecnt,
                1,
                0, /*don't force*/
                pages,
                NULL
            );
            up_read(&current->mm->mmap_sem);
            if(res!=pagecnt) {
                for( ; res; --res)
                    page_cache_release(pages[res-1]);
                kfree(pages);
                return FALSE;
            }
        }
        pDbx->hMapView[pDbx->wIterations] = (P_P9080_MDLADDR)(kmalloc(sizeof(T_P9080_MDLADDR), GFP_KERNEL));
        ((P_P9080_MDLADDR) pDbx->hMapView[pDbx->wIterations])->BufAddr = dwLockBA;
        ((P_P9080_MDLADDR) pDbx->hMapView[pDbx->wIterations])->pagecnt = pagecnt;
        ((P_P9080_MDLADDR) pDbx->hMapView[pDbx->wIterations])->pages = pages;
        pDbx->hMapView[pDbx->wIterations] = (void *)(((u32)pDbx->hMapView[pDbx->wIterations])|0x01);
        //dwLockPA = (unsigned long)page_to_phys(pages[0]);
        //dwCbpOfst = dwLockBA & ~PAGE_MASK;
    }
    else
        pDbx->hMapView[pDbx->wIterations] = 0;

    dwLockPA = translate_phys_addr((U32)dwLockBA, &dwCbpOfst); //only page_boundary address
    if(dwLockPA==0){ //check for translated address , 0 maybe caused by Un-Locked memory
        printk("[PLX9080] USER_VA to PCI_PA failed\n" );
        return FALSE;
    }

    dwCbpSize = dwLockSize;
    //dwDscrPA = AdlDscrHeadPhy(pDbx);
    //pDscrBA = (P_P9080_DMADSCR)AdlDscrHeadMap(pDbx);
    dwDscrPA = (pDbx->xAddMbx[pDbx->wIterations]).dwAddPhy;
    pDscrBA = (P_P9080_DMADSCR)((pDbx->xAddMbx[pDbx->wIterations]).dwAddMap);
    dwLockBA = dwLockBA&PAGE_MASK; //page_align for next addr_translate
    //yuan add 04/13/01
    if(pDbx->wIterations>0){
        tmpDscrBA = (P_P9080_DMADSCR)(((U32)pDscrBA)-sizeof(T_P9080_DMADSCR));
        tmpDscrBA->NextDscrAddr = dwDscrPA|(pDbx->xFlag.f.Output? 0x5L:0xdL);
    }
    while(1){
        pDscrBA->PciAddr = dwLockPA+dwCbpOfst;
        dwLockBA += PAGE_SIZE;
        pDscrBA->LocalAddr = AdlDscrLocalAddr(pDbx);

        if(dwCbpOfst+dwCbpSize>PAGE_SIZE){
            pDscrBA->DmaSize = PAGE_SIZE-dwCbpOfst;
            dwCbpOfst = 0L;
            dwCbpSize -= pDscrBA->DmaSize;
            dwDscrPA += sizeof(T_P9080_DMADSCR);
            pDscrBA->NextDscrAddr = dwDscrPA|(pDbx->xFlag.f.Output? 0x1L:0x9L);
            pDscrBA++;
            //translate to phys-addr for next page
            dwLockPA = translate_phys_addr((U32)dwLockBA, &dwCbpOfst); //only page_boundary address
            if(dwLockPA==0){ //check for translated address , 0 maybe caused by Un-Locked memory
                printk("[PLX9080] USER_VA to PCI_PA failed in while_loop\n" );
                return FALSE;
            }
            continue;
        }
        else{
            pDscrBA->DmaSize = dwCbpSize;
            //pDscrBA->NextDscrAddr = pDbx->xFlag.f.Output? 0x3L:0xbL;
            //yuan 04/13/01
            pDscrBA->NextDscrAddr = \
                (pDbx->xAddMbx[0]).dwAddPhy|(pDbx->xFlag.f.Output? 0x5L:0xdL);
            break;
        }
    }

    pSbuf->wBufferId = pDbx->wIterations;
    pDbx->wIterations = pDbx->wIterations+1;
    if(pDbx->wIterations<MAXMBUFCNT){
        (pDbx->xAddMbx[pDbx->wIterations]).dwAddPhy = dwDscrPA+sizeof(T_P9080_DMADSCR);
        (pDbx->xAddMbx[pDbx->wIterations]).dwAddMap = (U32)(((U32)pDscrBA)+sizeof(T_P9080_DMADSCR));
    }

    return TRUE;
}


#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)

U32 translate_phys_addr(U32 user_virt_addr, U32 *pOffset)
{
    U32 adr, offset;
    pgd_t *pgd; pmd_t *pmd; pte_t *ptep, pte;
    struct page *pp;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10)
    pud_t *pud;
#endif
    unsigned long ret = 0UL;

    offset =  user_virt_addr&(PAGE_SIZE-1);
    adr = user_virt_addr&PAGE_MASK;
    //printk("[pci-7300] adr %x\n", adr);
    pgd = pgd_offset(current->mm, adr);
    //printk("[pci-7300] pgd %x\n", pgd);
    if(!pgd_none(*pgd)){
        //yuan add 05/16/05
    #if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,10)
        pud = pud_offset(pgd, adr);
        pmd = pmd_offset(pud, adr);
    #else
        pmd = pmd_offset(pgd, adr);
    #endif
        //printk("[pci-7300] pmd %x\n", pmd);
        if(!pmd_none(*pmd)){
            if(pmd_bad(*pmd)){
                pmd_ERROR(*pmd);
                pmd_clear(pmd);
                return 0;
            }
            ptep = PTE_OFFSET(pmd, adr);
            //printk("[pci-7300] ptep %x\n", ptep);
            pte = *ptep;
            //printk("[pci-7300] pte %x\n", pte);
            if(pte_present(pte)){
                pp = pte_page(pte);
                //pte_unmap(ptep);
                if((PageHighMem(pp))||(!page_address(pp))) {
                    //printk("[pci-7300] enter PageHighMem(pp) %d, page_address(pp) %x\n", PageHighMem(pp), page_address(pp));
                #ifndef SUSE_ATOMIC_PTE
                    //ret = (unsigned long)((((pp)-page_zone(pp)->zone_mem_map)<<PAGE_SHIFT)+page_zone(pp)->zone_start_paddr);
                    //...ret = (unsigned long)((pp-mem_map)<<PAGE_SHIFT);
                    //yuan modify 06/12/06
                    ret = (unsigned long)page_to_phys(pp);
                #else
                    ret = pte.pte_low&PAGE_MASK;
                #endif
                    //printk("[pci-7300] ret mem: %x\n",ret);
                    //printk("high mem: %x %x\n",ret, mem_map);
                    //pte_unmap(ptep);
                }
                else{
                    //ret = (unsigned long)page_address(pte_page(pte));
                #ifndef SUSE_ATOMIC_PTE
                    ret = (unsigned long)page_address(pp);
                    ret = __pa(ret);
                #else
                    ret = pte.pte_low&PAGE_MASK;
                #endif
                    //printk("[pci-7300] ret mem: %x\n",ret);
                }
                #ifdef SUSE_ATOMIC_PTE
                    pte_kunmap(ptep); //release atomic lock on pte object
                #endif
                #ifdef PTE_MAP
                    pte_unmap(ptep);
                #endif
            }
            else{
            #ifdef SUSE_ATOMIC_PTE
                pte_kunmap(ptep);  //release atomic lock on pte object
            #endif
            #ifdef PTE_MAP
                pte_unmap(ptep);
            #endif
                //printk("[pci-7300] pte_present(pte)==FALSE, return 0\n");
                return (U32)(0);
            }
        }
        else{
        /*
        #ifdef SUSE_ATOMIC_PTE
            pte_kunmap(pte);  //release atomic lock on pte object
        #endif
        #ifdef PTE_MAP
            pte_unmap(ptep);
        #endif
        */
            //printk("[pci-7300] pmd_none(*pmd)==TRUE, return 0\n");
            return (U32)(0);
        }
    }
    //if(pte_present(pte)){
    //    ret = (unsigned long)page_address(pte_page(pte));
    //    if(ret!=0x00)
    //        ret |= (adr&(PAGE_SIZE-1));
    //}else
    //    return (U32)(0);
    //}
    //}
    //printk("ret %x\n", ret);
    *pOffset = offset;

    return ret; //__pa(ret);
}

#elif LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0)

U32 translate_phys_addr(U32 user_virt_addr, U32 *pOffset)
{
    U32 address, offset, phys_addr;
    pgd_t *pgd; pmd_t *pmd; pte_t *pte;

    offset =  user_virt_addr&(~PAGE_MASK);
    address = user_virt_addr&PAGE_MASK;

    pgd = pgd_offset(current->mm, address);
    pmd = pmd_offset(pgd, address);
    pte = PTE_OFFSET(pmd, address);

    //07/03/00 jeffrey, add the pte_present in translate_phys_addr , for next version
    if(pte_present(*pte))
        phys_addr = __pa(pte_page(*pte)); //the address of PAGE
    else
        return (U32)(0);

    *pOffset = offset;
    return phys_addr;
}

#endif
