#ifndef _ADL_LIB_H_
#define _ADL_LIB_H_

#include <asm/io.h>
#include <linux/fs.h> //for 'struct fasync_struct'
#include <linux/module.h>

MODULE_LICENSE("GPL");

int  AdlFindDevice(U16 VendorID, U16 DeviceID, U16 PciController, U16 wPort_Count);
int  AdlInitDevice(U16 VendorID, U16 DeviceID, U16 PciController, U16 wPort_Count, PPCI_DEVEXT pDevExt, struct pci_dev *pciptr, int bd);
void AdlInitModule(PCI_DRVEXT *pdrv_ext);

/*driver dispatch assist func*/
void adl_kill_fasync(struct fasync_struct *fa, int sig);
void AdlDpcDsen(register PPCI_DSEN pDsen);
void AdlDropDsen(register PPCI_DSEN pDsen);
void AdlDpcDbx(register PPCI_DBX pDbx);
void AdlCloseDsen(register PPCI_DSEN pDsen);
void AdlCloseDbx(register PPCI_DBX pDbx);
void AdlFlushInTrace(U32 dwPort, PPCI_DBX pDbx);

/*Driver IOCTL call assist func*/
void    AdlIocGetBufSize(PPCI_DEVEXT pDevExt, DAS_IOT_FUNC_BUFSIZE *pSbuf);
BOOLEAN AdlIocGetDesp(PPCI_DEVEXT pDevExt, DAS_IOT_CARD_DESP *pSbuf);
BOOLEAN AdlIocPioRead(U16 wBytesPerData, DAS_IOT_PIO *pSbuf);
BOOLEAN AdlIocPioWrite(U16 wBytesPerData, DAS_IOT_PIO *pSbuf, U32 dwOldIn);
void    AdlIocPortIoRead(PPCI_DEVEXT pDevExt, U16 wBytesPerData, DAS_IOT_PORTIO *pSbuf);
void    AdlIocPortIoWrite(PPCI_DEVEXT pDevExt, U16 wBytesPerData, DAS_IOT_PORTIO *pSbuf);
BOOLEAN AdlIocDbfMode(PPCI_DEVEXT pDevExt, PPCI_DBX pDbx, DAS_IOT_DBF *pIoBuf);
BOOLEAN AdlIocDbfHalfReady(PPCI_DEVEXT pDevExt, PPCI_DBX pDbx, DAS_IOT_DBF *pIoBuf);
BOOLEAN AdlIocAsyCheck(PPCI_DEVEXT pDevExt, PPCI_DBX pDbx, DAS_IOT_ASYNC *pIoBuf);
BOOLEAN AdlIocAsyClear(PPCI_DEVEXT pDevExt, PPCI_DBX pDbx, DAS_IOT_ASYNC *pIoBuf);
U32     AdlSplitePacer(U32 dwPacerIn);
//int     AdlIocCheckBufferLength(PIRP pIrp);
BOOLEAN AdlIocContPreProc(PPCI_DEVEXT pDevExt, PPCI_DBX pDbx, DAS_IOT_CONT *pSbuf);
int     AdlIocContPostProc(PPCI_DEVEXT pDevExt, PPCI_DBX pDbx);

/*Resversed memory allocate/release from module: Adl_Mem_Mgr*/
extern int adl_free_memory(U32 address);
extern U32 adl_request_memory(U32 kilobytes, int priority);

/*Funtion from ADLDISP.C*/
BOOLEAN AdlOpenDevice(PPCI_DEVEXT pDevExt);
void    AdlControllerReset(PPCI_DEVEXT pDevExt, U32 reset_9050);

/**/
BOOLEAN   AdlDbxAllocate(PPCI_DEVEXT pDevExt);
BOOLEAN   AdlSetupDbxBuffer(register PPCI_DBX pDbx, U16 wArraySize, U32 buffer_kbyte);
void      AdlDropDbx(register PPCI_DBX pDbx);
PPCI_DSEN AdlSetupDsen(void);
BOOLEAN   AdlSetupBuffer(PPCI_DEVEXT pDevExt);
/*Add for new 7300 memory allocation*/
BOOLEAN   AdlDbxAllocate2(PPCI_DEVEXT pDevExt);
void      AdlDropDbx2(register PPCI_DBX pDbx);
BOOLEAN   AdlSetupDbxBuf(register PPCI_DBX pDbx, U16 wArraySize, U32 buffer_kbyte);
/**/
int           adl_free_mem(unsigned long adr, unsigned long len);
unsigned long adl_req_memory(unsigned long bytecnt);


/*defination for debug output*/
#ifdef _DEBUG
#define AdlDump(STRING) DbgPrint STRING;
#else
#define AdlDump(STRING)
#endif
#define DAS_BIT(Val) Val? 1:0

typedef union Das4Bytes{
    U8  c[4];
    U16 w[2];
    U32 d;
}
DAS_4BYTE;


/*Macro for OP port I/O*/
#define OP_Port(pDevExt) (pDevExt->dwPortAddr[0])

#define OP_outb(Port,Ofst,Value) outb_p((U8)Value, (U32)(Port+ADL_OP_##Ofst))

#define OP_outw(Port,Ofst,Value) outw_p((U16)Value, (U32)(Port+ADL_OP_##Ofst))

#define OP_outd(Port,Ofst,Value) outl_p((U32)Value, (U32)(Port+ADL_OP_##Ofst))

#define OP_inpb(Port,Ofst) inb_p((U32)(Port+ADL_OP_##Ofst))

#define OP_inpw(Port,Ofst) inw_p((U32)(Port+ADL_OP_##Ofst))

#define OP_inpd(Port,Ofst) inl_p((U32)(Port+ADL_OP_##Ofst))


/*Macro for PT port I/O*/
#define PT_Port(pDevExt) (pDevExt->dwPortAddr[1])

#define PT_outb(Port,Ofst,Value) outb_p((U8)Value, (U32)(Port+ADL_PT_##Ofst))

#define PT_outw(Port,Ofst,Value) outw_p((U16)Value, (U32)(Port+ADL_PT_##Ofst))

#define PT_outd(Port,Ofst,Value) outl_p((U32)Value, (U32)(Port+ADL_PT_##Ofst))

#define PT_inpb(Port,Ofst) inb_p((U32)(Port+ADL_PT_##Ofst))

#define PT_inpw(Port,Ofst) inw_p((U32)(Port+ADL_PT_##Ofst))

#define PT_inpd(Port,Ofst) inl_p((U32)(Port+ADL_PT_##Ofst))


/*Macro for P2 port I/O*/
#define P2_Port(pDevExt) (pDevExt->dwPortAddr[2])

#define P2_outb(Port,Ofst,Value) outb_p((U8)Value, (U32)(Port+ADL_P2_##Ofst))

#define P2_outw(Port,Ofst,Value) outw_p((U16)Value, (U32)(Port+ADL_P2_##Ofst))

#define P2_outd(Port,Ofst,Value) outl_p((U32)Value, (U32)(Port+ADL_P2_##Ofst))

#define P2_inpb(Port,Ofst) inb_p((U32)(Port+ADL_P2_##Ofst))

#define P2_inpw(Port,Ofst) inw_p((U32)(Port+ADL_P2_##Ofst))

#define P2_inpd(Port,Ofst) inl_p((U32)(Port+ADL_P2_##Ofst))


/*Macro for P3 port I/O*/
#define P3_Port(pDevExt) (pDevExt->dwPortAddr[3])

#define P3_outb(Port,Ofst,Value) outb_p((U8)Value, (U32)(Port+ADL_P3_##Ofst))

#define P3_outw(Port,Ofst,Value) outw_p((U16)Value, (U32)(Port+ADL_P3_##Ofst))

#define P3_outd(Port,Ofst,Value) outl_p((U32)Value, (U32)(Port+ADL_P3_##Ofst))

#define P3_inpb(Port,Ofst) inb_p((U32)(Port+ADL_P3_##Ofst))

#define P3_inpw(Port,Ofst) inw_p((U32)(Port+ADL_P3_##Ofst))

#define P3_inpd(Port,Ofst) inl_p((U32)(Port+ADL_P3_##Ofst))


/*Macro for General port I/O ----*/
#define outB(Port,Value) outb_p((U8)(Value),(U32)(Port))

#define outW(Port,Value) outw_p((U16)(Value),(U32)(Port))

#define outD(Port,Value) outl_p((U32)(Value), (U32)(Port))

#define inpB(Port) inb_p((U32)(Port))

#define inpW(Port) inw_p((U32)(Port))

#define inpD(Port) inl_p((U32)(Port))


/*defination for IOCTL system buffer reference*/
#define ADL_UCS_MAX 256
#define ADL_UCS_NUM 16

/*----------------------------------------------------------------------------*/
#ifndef io_remap_page_range
#define io_remap_page_range(p1, p2, p3, p4, p5) io_remap_pfn_range(p1, p2, p3>>PAGE_SHIFT, p4, p5)
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 17)
#define MODULE_PARM(x,y) module_param_array(x, ulong, NULL, 0)
#endif

#ifndef pci_module_init
#define pci_module_init pci_register_driver
#endif

#ifndef SA_INTERRUPT
#define SA_INTERRUPT IRQF_DISABLED
#endif
#ifndef SA_SHIRQ
#define SA_SHIRQ IRQF_SHARED
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,13)
#define class_create class_simple_create
#define class_destroy class_simple_destroy
#define device_create(a,b,c,d,e,f) class_simple_device_add(a,c,d,e,f)
#define device_destroy(a,b) class_simple_device_remove(b)
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2,6,26)
#define device_create class_device_create
#define device_destroy class_device_destroy
#endif

#endif
