#include <asm/page.h>
#ifndef _ADLP9080_H_
#define _ADLP9080_H_

/*Function prototype*/
void    P9080_SetDMAWidth(U32 dwPort, U32 dwChannel, U32 dwBytes);
BOOLEAN P9080_StartDMA(U32 dwPort, U32 dwChannel, PPCI_DBX pDbx);
BOOLEAN P9080_StopDMA(U32 dwPort, U32 dwChannel);
void    P9080_ClearController(U32 dwPort);
void    P9080_SoftReset(U32 dwPort);
void    P9080_ClearDmaDscr(PPCI_DBX pDbx);
BOOLEAN P9080_BuildDmaDscr(PPCI_DBX pDbx);
U32     translate_phys_addr(U32 user_virt_addr, U32 *pOffset);
BOOLEAN P9080_BuildMBufDscr(DAS_IOT_MBUF *pSbuf, PPCI_DBX pDbx);
void    P9080_ClearMBufDscr(PPCI_DBX pDbx);

/*Save Data in DBX*/
#define BYTE_OFFSET(addr)                          (addr&(~PAGE_MASK))
#define ADDRESS_AND_SIZE_TO_SPAN_PAGES(addr, size) ((size+BYTE_OFFSET(addr)-1)/PAGE_SIZE+1)
#define AdlDscrUserBuffer(pDbx)                    (pDbx->xAddDbx[1].dwAddMap)
#define AdlDscrLocalAddr(pDbx)                     (pDbx->xAddDbx[1].dwAddPhy)
#define AdlDscrLength(pDbx)                        (pDbx->xAddDbx[1].dwLength)
#define AdlDscrHeadPhy(pDbx)                       (pDbx->xAddDbx[0].dwAddPhy)
#define AdlDscrHeadLen(pDbx)                       (pDbx->xAddDbx[0].dwLength)
#define AdlDscrHeadMap(pDbx)                       (pDbx->xAddDbx[0].dwAddMap)
#define AdlPMDL(pDbx)                              (pDbx->hMapView[0])

/*PLX9080 Register Port Address Offset*/
#define ADL_OP_LAS0RR       0x00
#define ADL_OP_LAS0BA       0x04
#define ADL_OP_MARBR        0x08
#define ADL_OP_BIGEND       0x0C
#define ADL_OP_EROMRR       0x10
#define ADL_OP_EROMBA       0x14
#define ADL_OP_LBRD0        0x18
#define ADL_OP_DMRR         0x1C
#define ADL_OP_DMLBAM       0x20
#define ADL_OP_DMLRAI       0x24
#define ADL_OP_DMPBAM       0x28
#define ADL_OP_DMCFGA       0x2C

#define ADL_OP_LAS1RR       0xF0
#define ADL_OP_LAS1BA       0xF4
#define ADL_OP_LBRD1        0xF8

#define ADL_OP_MBOX0        0x40
#define ADL_OP_MBOX1        0x44
#define ADL_OP_MBOX2        0x48
#define ADL_OP_MBOX3        0x4C
#define ADL_OP_MBOX4        0x50
#define ADL_OP_MBOX5        0x54
#define ADL_OP_MBOX6        0x58
#define ADL_OP_MBOX7        0x5C
#define ADL_OP_P2LDBELL     0x60
#define ADL_OP_L2PDBELL     0x64
#define ADL_OP_INTCSR       0x68
#define ADL_OP_CNTRL        0x6C
#define ADL_OP_EEPROM       0x6F
#define ADL_OP_PCIHIDR      0x70
#define ADL_OP_PCIHREV      0x74

#define ADL_OP_DMAMODE0     0x80
#define ADL_OP_DMAPADR0     0x84
#define ADL_OP_DMALADR0     0x88
#define ADL_OP_DMASIZ0      0x8C
#define ADL_OP_DMADPR0      0x90
#define ADL_OP_DMAMODE1     0x94
#define ADL_OP_DMAPADR1     0x98
#define ADL_OP_DMALADR1     0x9C
#define ADL_OP_DMASIZ1      0xA0
#define ADL_OP_DMADPR1      0xA4
#define ADL_OP_DMACSR0      0xA8
#define ADL_OP_DMACSR1      0xA9
#define ADL_OP_DMAARB       0xAC
#define ADL_OP_DMATHR       0xB0

#define ADL_OP_OPLFIS       0x30
#define ADL_OP_OPLFIM       0x34
#define ADL_OP_IQP          0x40
#define ADL_OP_OQP          0x44
#define ADL_OP_MQCR         0xC0
#define ADL_OP_QBAR         0xC4
#define ADL_OP_IFHPR        0xC8
#define ADL_OP_IFTPR        0xCC
#define ADL_OP_IPHPR        0xD0
#define ADL_OP_IPTPR        0xD4
#define ADL_OP_OFHPR        0xD8
#define ADL_OP_OFTPR        0xDC
#define ADL_OP_OPHPR        0xE0
#define ADL_OP_OPTPR        0xE4
#define ADL_OP_QSR          0xE8

#define P9080_DMA_INT_CLEAR 0x08

//#pragma pack(push,1)

/*Chain DMA Mode Descriptor*/
typedef struct{
    U32 PciAddr;
    U32 LocalAddr;
    U32 DmaSize;
    U32 NextDscrAddr;
}
T_P9080_DMADSCR, *P_P9080_DMADSCR;

/*PLX9080 Register structure defination*/
typedef struct{
    U32 LocalBusWidth      : 2;
    U32 InternalWaitStates : 4;
    U32 ReadyInput         : 1;
    U32 BTERM              : 1;
    U32 LocalBurst         : 1;
    U32 Chain              : 1;
    U32 DoneInterrupt      : 1;
    U32 LocalAddress       : 1;
    U32 DemandMode         : 1;
    U32 WriteInvalidate    : 1;
    U32 DmaEOT             : 1;
    U32 DmaSDT             : 1;
    U32 DmaClearCount      : 1;
    U32 DmaInterrupt       : 1;
    U32 Reserved           : 14;
}
T_P9080_DMAMODE, *P_P9080_DMAMODE;

typedef struct{
    U32 DscptLoc      : 1;
    U32 EndChain      : 1;
    U32 TermInterrupt : 1;
    U32 DirTransfer   : 1;
    U32 NextDscpt     : 28;
}
T_P9080_DMADPR, *P_P9080_DMADPR;

typedef struct{
    U32 C0PLAF : 4;
    U32 C0LPAE : 4;
    U32 C0LPAF : 4;
    U32 C0PLAE : 4;
    U32 C1PLAF : 4;
    U32 C1LPAE : 4;
    U32 C1LPAF : 4;
    U32 C1PLAE : 4;
}
T_P9080_DMATHR, *P_P9080_DMATHR;

typedef struct{
    U32 LSERR_Enable      : 1;
    U32 LSERR_ParityError : 1;
    U32 _SERR             : 1;
    U32 IE_L_MBox         : 1;
    /*7:4*/
    U32 Reserved          : 4;
    /*11:8*/
    U32 IE_P_Out          : 1; /*PCI Interrupt Enable*/
    U32 IE_P_DoorBell     : 1;
    U32 IE_P_Abort        : 1;
    U32 IE_P_Local        : 1;
    /*15:12*/
    U32 AE_TargetRetry    : 1; /*Target Abort Enable*/
    U32 IA_P_DoorBell     : 1; /*PCI Interrupt Active*/
    U32 IA_P_Abort        : 1;
    U32 IA_P_Local        : 1;
    /*19:16*/
    U32 IE_L_Out          : 1; /*Local Interrupt Enable*/
    U32 IE_L_DoorBell     : 1;
    U32 IE_L_DMA0         : 1;
    U32 IE_L_DMA1         : 1;
    /*Local Interrupt Active*/
    U32 IA_L_DoorBell     : 1;
    U32 IA_L_DMA0         : 1;
    U32 IA_L_DMA1         : 1;
    U32 IA_P_BIST         : 1;
    /*0 was Bus Master when Abort*/
    U32 _BMA_DirectMaster : 1;
    U32 _BMA_DMA0         : 1;
    U32 _BMA_DMA1         : 1;
    /*0 was Target Retry Abort when Abort*/
    U32 _TAA_Retry        : 1;
    /*PCI Wrote Data to MailBox*/
    U32 PWD_MBX0          : 1;
    U32 PWD_MBX1          : 1;
    U32 PWD_MBX2          : 1;
    U32 PWD_MBX3          : 1;
}
T_P9080_INTCSR, *P_P9080_INTCSR;

typedef struct{
    U32 ChnlEnable     : 1;
    U32 ChnlStart      : 1;
    U32 ChnlAbort      : 1;
    U32 ClearInterrupt : 1;
    U32 ChnlDone       : 1;
    U32 Reserved       : 3;
    U32 Unused         : 24;
}
T_P9080_DMACSR, *P_P9080_DMACSR;

typedef union _p9080_dwreg_
{
    T_P9080_DMAMODE dmamode;
    T_P9080_DMADPR  dmadpr;
    T_P9080_DMATHR  dmathr;
    T_P9080_INTCSR  intcsr;
    T_P9080_DMACSR  dmacsr;
    U32             r;
    U16             w[2];
    U8              b[4];
}
OP_REG,*POP_REG;

typedef struct
{
    U8 ClockToggle     : 1;
    U8 ChipSelect      : 1;
    U8 DataIn          : 1;
    U8 DataOut         : 1;
    U8 Present         : 1;
    U8 ReloadConfig    : 1;
    U8 SoftReset       : 1;
    U8 LocalInitStatus : 1;
}
T_P9080_EEPROM, *P_P9080_EEPROM;

#include <asm/io.h>
typedef struct
{
    U32         BufAddr;
    U32         pagecnt;
    struct page **pages;
    U32         LastESise;
    U32         NextAddr;
}
T_P9080_MDLADDR, *P_P9080_MDLADDR;

#endif
