#ifndef _ADLSYS_H_
#define _ADLSYS_H_

#include <linux/ioctl.h>

#ifndef __KERNEL__
#include <signal.h> //00/06/09 jeffrey , for signal function set in linux
#endif

#define MAGIC_NUM 'P'

#define SIGEVENT1 61
#define SIGEVENT2 62
#define SIGEVENT3 60
#define SIGEVENT4 59
#define SIGDBI    58
#define SIGDBO    57

#ifndef basic_type
#define basic_type

typedef unsigned char      U8;
typedef short              I16;
typedef unsigned short     U16;
typedef long               I32;
typedef unsigned long      U32;
typedef unsigned long long U64;
typedef float              F32;
typedef double             F64;
typedef enum{FALSE, TRUE}  BOOLEAN;

#define FIRSTBYTE(VALUE)  (VALUE&0x00FF)
#define SECONDBYTE(VALUE) ((VALUE>>8)&0x00FF)
#define THIRDBYTE(VALUE)  ((VALUE>>16)&0x00FF)
#define FOURTHBYTE(VALUE) ((VALUE>>24)&0x00FF)
#endif //basic_type

/*File type constant for create device*/
#define DAS_FILE_TYPE            0xADC0
#define DAS_FUNC_OUTPUT(wIndex)  (wIndex & 1)
#define DAS_FUNC_DIGITAL(wIndex) (wIndex & 2)


/*
 * Constant for API caller
 */
//Function Index -- AI, AO, DI, DO
#define DAS_FUNC_AI    0
#define DAS_FUNC_AO    1
#define DAS_FUNC_DI    2
#define DAS_FUNC_DO    3
#define DAS_FUNC_COUNT 4

/*Limitation for system-call data struct*/
#define DAS_AD_GAIN_COUNT      20
#define DAS_CHANNEL_GAIN_COUNT 256
#define DAS_EVENT_NAME_SIZE    16

#define DbxORunStructSize      4096

/*
 * DeviceControl IN/OUT parameter data structure
 */
//#pragma pack(push,1)
/*Wait Event*/
//jeffrey 00/04/14, mark the event structure in linux
/*
typedef struct _DAS_IOT_EVENT_ITEM
{
    struct{
        U16 fAI     : 1; //Analog Input Event
        U16 fAO     : 1; //Analog Output Event
        U16 fDI     : 1; //Digital Input Event
        U16 fDO     : 1; //Digital Output Event
        U16 fI1     : 1; //Dual Interrupt Source 1 Event
        U16 fI2     : 1; //Dual Interrupt Source 2 Event
        U16 fTC     : 1; //Trace Complete Event
        U16 fUnused : 9;
    }bit;
    U16 word;
}
DAS_IOT_EVENT_ITEM;

typedef union _DAS_IOT_EVENT_STYLE
{
    struct{
        U16 fNoWait   : 1; //Do not wait, return at once
        U16 fWaitAll  : 1; //Wait all event meets
        U16 fInfinite : 1; //Wait Infinite --> Wait Timer disable
        U16 fUnused   : 13;
    }bit;
    U16 word;
}
DAS_IOT_EVENT_STYLE;

typedef struct _DAS_IOT_WAITEVT
{
    short              wErrorCode;
    DAS_IOT_EVENT_ITEM wItem;
}
DAS_IOT_WAITEVT;
*/

/*Use for DAS_IOC_GET_DESP*/
typedef struct _DAS_IOT_FUNC_FLAG
{
    U16 bFunction       : 1;
    U16 bDMA            : 1;
    U16 bInterrupt      : 1;
    U16 bPolling        : 1;
    U16 bDirectionIn    : 1;
    U16 bAutoScan       : 1;
    U16 bScanEndingZero : 1;
    U16 bScanIncrease   : 1;
    U16 bSampleRate     : 1;
    U16 bScanRate       : 1;
    U16 bSingleEnd      : 1;
    U16 bDifferential   : 1;
    U16 bDigitalTrigger : 1;
    U16 bAnalogTrigger  : 1;
    U16 bChainDMA       : 1;
    U16 bUnused1        : 1;
}
DAS_IOT_FUNC_FLAG;

typedef struct _DAS_IOT_FUNC_DESP
{
    DAS_IOT_FUNC_FLAG wFlag;
    U16               wChannelCount;
    U16               wResoluation;
    U16               wArraySize;
    U32               dwBufferSize;
    void              *pView[2];
}
DAS_IOT_FUNC_DESP;

typedef struct _DAS_IOT_AD_GAIN
{
    U16 wRange;
    U16 wGain;
}
DAS_IOT_AD_GAIN;

typedef struct _DAS_IOT_CARD_DESP
{
    short             wErrorCode;
    U16               wVersion;
    U16               wBus;
    U16               wSlot;
    U16               wCtrChannelCount;
    U16               wCtrResoluation;
    U32               dwCtrKHz;
    U16               wAdGainCount;
    //jeffrey 00/04/14 , mark unused event structure for linux drivers
    BOOLEAN           fDualInterruptEventNotify;
    BOOLEAN           fTraceEventNotify;
    DAS_IOT_FUNC_DESP mFunc[DAS_FUNC_COUNT];
    DAS_IOT_AD_GAIN   mAdGain[DAS_AD_GAIN_COUNT];
    U16               Reference;
}
DAS_IOT_CARD_DESP;

//yuan add 07/25/01
typedef struct _DAS_IOT_RESRC
{
    short wErrorCode;
    U16   PortAddr[4];
}
DAS_IOT_RESRC;

typedef struct _DAS_IOT_FUNC_BUFSIZE
{
    U32 BufSize[4];
}
DAS_IOT_FUNC_BUFSIZE;

typedef struct _DAS_IOT_PVIEW
{
    void *ViewAddr[2];
}
DAS_IOT_PVIEW;

typedef struct _DAS_IOT_MEMMAP
{
    short         wErrorCode;
    DAS_IOT_PVIEW pMView[4];
}
DAS_IOT_MEMMAP;

typedef struct{
    U16 wCurrentRange;
    U16 wReserve1;
    U16 wReserve2;
    U16 wReserve3;
}
DAS_6208AO_CONFIG;

typedef struct{
    U16 wTrigSource;
    U16 fTimerCascade     : 1;
    U16 fInWaitTrigger    : 1;
    U16 fInTriggerRising  : 1;
    U16 fInRequestRising  : 1;
    U16 fOutRequestEnable : 1;
    U16 fOutTriggerHigh   : 1;
    U16 fUnused           : 10;
    U16 wReserved[2];
}
DAS_7200DIO_CONFIG;

typedef struct{ //for PCI-7300A Rev.A
    U16 wTrigSource;        //use PCI7300_TRG_xxx
    U16 fPortWidth     : 8; //use DAS_PORT_WIDTH_xxx
    U16 fStartWait     : 2; //use PCI7300_START_WAIT_xxx
    U16 fTerminator    : 1; //1: Term On, 0: Term Off, DI->PortA, DO->PortB
    U16 fRequestRising : 1; //1: rising, 0: falling, DI only
    U16 clear_fifo     : 1; //only for DI
    U16 disable_di     : 1; //only for DI
    U16 fConfig        : 1; //flag for setting User buffer
    U16 fReserved      : 1;
    U32 dwBufferAddr;       //User Buffer Address
}
DAS_7300DIO_CONFIG;

//hua+
typedef struct{ //for PCI-7300A Rev.B
    U16 wTrigSource;        //use PCI7300_TRG_xxx
    U16 fPortWidth     : 8; //use DAS_PORT_WIDTH_xxx
    U16 fStartWait     : 2; //use PCI7300_START_WAIT_xxx
    U16 fTerminator    : 1; //1: Term On, 0: Term Off, DI->PortA, DO->PortB
    U16 fConfig        : 1; //flag for setting User buffer
    U16 fReserved1     : 4;
    U32 dwBufferAddr;       //User Buffer Address
    U16 fRequestRising : 8; //1: rising, 0: falling, DI only
    U16 clear_fifo     : 1; //only for DI
    U16 disable_di     : 1; //only for DI
    U16 fReserved2     : 6;
    U32 fifo_threshold;     //1: rising, 0: falling, DI only
}
DAS_7300bDIO_CONFIG;

typedef struct{
    union{
        struct{
            U16 bCN1_SRC   : 2;
            U16 bCN2_SRC   : 2;
            U16 bCN3_SRC   : 2;
            U16 bCN4_SRC   : 2;
            U16 bCN5_SRC   : 2;
            U16 bCN6_SRC   : 2;
            U16 bCN7_SRC   : 2;
            U16 bCN8_SRC   : 2;
            U16 bCN9_SRC   : 2;
            U16 bCN10_SRC  : 2;
            U16 bCK1SEL    : 1;
            U16 bReserved2 : 1;
            U16 bDBCSEL    : 1;
            U16 bReserved3 : 1;
        }mCMCbit;
        struct{
            U16 bCMC1 :8;
            U16 bCMC2 :8;
            U16 bCMC3 :8;
        }mCMCbyte;
    }
    mCMCdata;
}
DAS_8554CTR_CONFIG;
//hua-

typedef struct{
    U16 wTrigSource;
    U16 wTraceEnable;
    U16 wTraceCount;
    U16 wEDO_Function;
}
DAS_9111AI_CONFIG;

typedef struct{
    U16 wTrigSource;
    U16 wReserve1;
    U16 wReserve2;
    U16 wReserve3;
}
DAS_9113AI_CONFIG;

typedef struct{
    U16 wTrigSource;
    U16 wReserve1;
    U16 wReserve2;
    U16 wReserve3;
}
DAS_9114AI_CONFIG;

typedef struct{
    U16 fU_CMMD           : 1;
    U16 fDifferential     : 1;
    U16 fUnipolar         : 1;
    U16 fExternalTrigger  : 1;
    U16 fTriggerMode      : 3;
    U16 fTriggerPositive  : 1;
    U16 fExternalTimeBase : 1;
    U16 fDLYInSamples     : 1;
    U16 fEnableReTrig     : 1;
    U16 fEnableMCounter   : 1;
    U16 fAdType           : 2;
    U16 fUnused           : 2;
    U16 wPostTriggerCounter;
    U16 wMCounter;
    U16 wRetrgCnt;
}
DAS_9116AI_CONFIG;

typedef struct{
    U16 fUnipolar                : 1;
    U16 fDifferential            : 1;
    U16 fExternalGateControl     : 1;
    U16 fExternalTrigger         : 1;
    U16 fDigitalTriggerPositive  : 1;
    U16 fExternalTriggerPositive : 1;
    U16 fBurstMode               : 1;
    U16 fSampleHold              : 1;
    U16 fPostTrigger             : 1;
    U16 fAboutTrigger            : 1;
    U16 fUnused                  : 6;
    U16 wBurstConversionCounter;
    U16 wPostTriggerCounter;
    U16 wReserved;
}
DAS_9118AI_CONFIG;

typedef struct{
        U16 bTriggerMode   : 3;
        U16 bTriggerSource : 3;
        U16 bTriggerSlope  : 1;
        U16 bFreqSelect    : 1;
        U16 bClockSource   : 2;
        U16 bReserved      : 6;
        U16 wTriggerLevel;
        U16 wPostTriggerCount;
        U16 wReserved;
}
DAS_9812AI_CONFIG;

typedef struct{
    U16 wOutputPolarity;
    U16 wReserve1;
    U16 wReserve2;
    U16 wReserve3;
}
DAS_9111AO_CONFIG;

typedef struct{
    U16 wIntOrExtRef;
    U16 wReserve1;
    U16 wReserve2;
    U16 wReserve3;
}
DAS_9112AO_CONFIG;

//06/11/06 for 7442
typedef struct{
    U16 wChannel;
    U16 wLine;
    U16 wDirectionIn;
    U16 fLine;
}
DAS_7442DIO_LINECONFIG;

typedef struct _DAS_IOT_CONFIG
{
    short wErrorCode;
    union{
        struct{
            U16 wTrigSource;
        }mAIO;
        struct{
            U16 wChannel;
            int fDirectionIn;
        }mDIO;
        DAS_6208AO_CONFIG      m6208AO;
        DAS_7200DIO_CONFIG     m7200DIO;
        DAS_7300DIO_CONFIG     m7300DIO;
//hua+
        DAS_7300bDIO_CONFIG    m7300bDIO;
        DAS_8554CTR_CONFIG     m8554CTR;
//hua-
        DAS_9111AI_CONFIG      m9111AI;
        DAS_9113AI_CONFIG      m9113AI;
        DAS_9114AI_CONFIG      m9114AI;
        DAS_9118AI_CONFIG      m9118AI;
        DAS_9812AI_CONFIG      m9812AI;
        DAS_9111AO_CONFIG      m9111AO;         //Alex Lee 6/1/99
        DAS_9112AO_CONFIG      m9112AO;         //Alex Lee 6/1/99
        DAS_9116AI_CONFIG      m9116AI;         //Yuan 10/02/00
        DAS_7442DIO_LINECONFIG m7442LINECONFIG; //06/11/06
        U16                    wValue[4];
    }
    mData;
}
DAS_IOT_CONFIG;


/*
 * Continuous Status parameter
 * used for DAS_IOC_XX_STATUS
 */
typedef struct{
    U16 I_OVR   : 1;
    U16 O_UND   : 1;
    U16 fUnused : 14;
}
DAS_7200DIO_STATUS;

typedef struct{
    U16 FF_EF   : 1;
    U16 FF_HF   : 1;
    U16 FF_FF   : 1;
    U16 AD_BUSY : 1;
    U16 fUnused : 12;
}
DAS_PLXAI_STATUS;

typedef struct{
    U16 AdReady   : 1;
    U16 AdOverRun : 1;
    U16 fUnused   : 14;
}
DAS_9112AI_STATUS;

typedef struct{
    U16 FF_ERR  : 1;
    U16 FF_FUL  : 1;
    U16 FF_EMP  : 1;
    U16 fUnused : 13;
}
DAS_7300DIO_STATUS;

typedef struct{
    U16 AdReady              : 1;
    U16 AdOverRun            : 1;
    U16 AdOverSpeed          : 1;
    U16 BurstModeOverRun     : 1;
    U16 DigialTriggerHappen  : 1;
    U16 AboutTriggerComplete : 1;
    U16 FifoEmpty            : 1;
    U16 FifoHalfFull         : 1;
    U16 FifoOverflow         : 1;
    U16 fUnused              : 7;
}
DAS_9118AI_STATUS;

typedef struct{
    U16 FifoInReady  : 1;
    U16 FifoHalfFull : 1;
    U16 FifoOutReady : 1;
    U16 PostTrigEnd  : 1;
    U16 fUnused      : 12;
}
DAS_9812AI_STATUS;

typedef struct _DAS_IOT_STATUS
{
    short wErrorCode;
    union{
        DAS_7200DIO_STATUS m7200STAT;
        DAS_PLXAI_STATUS   mPlxSTAT;
        DAS_9118AI_STATUS  m9118STAT;
        DAS_9812AI_STATUS  m9812STAT;
        DAS_7300DIO_STATUS m7300STAT;
        U16                wValue;
    }
    mStat;
}
DAS_IOT_STATUS;

/*
 * Asynchronous control parameter
 * Use for DAS_IOC_XX_ASY_CHECK
 * Use for DAS_IOC_XX_ASY_CLEAR
 */
typedef struct _DAS_IOT_ASYNC
{
    short wErrorCode;
    U32   dwAccessCount;
    U16   wStopped;
}
DAS_IOT_ASYNC;

/*
 * Double buffer tarnsfer control parameter
 * Use for DAS_IOC_XX_DBF_HALF_READY
 * Use for DAS_IOC_XX_DBF_MODE
 */
typedef struct _DAS_IOT_DBF
{
    short wErrorCode;
    U16   wEnable;
    U16   wReady;
    U16   wViewIndex;
}
DAS_IOT_DBF;

//yuan add 01/29/01
/*
 * Double buffer view information
 * Use for DAS_IOC_XX_DBF_VIEW
 */
typedef struct _DAS_IOT_DBF_VIEW
{
    short wErrorCode;
    U16   wViewIndex;
    U32   dwAccessCount;
}
DAS_IOT_DBF_VIEW;

//yuan add 02/26/01
/*
 * Double buffer view information
 * Use for DAS_IOC_XX_DBF_VIEW
 */
typedef struct _DAS_IOT_MBUF
{
    short wErrorCode;
    U32   dwBufferAddr;  /*User Buffer Address*/
    U32   dwAccessCount;
    U16   wBufferId;
}
DAS_IOT_MBUF;

/*
 * Continguous transfer mode control parameter
 * Use for DAS_IOC_XX_CON_READ
 * Use for DAS_IOC_XX_CON_WRITE
 *
 */
typedef struct _DAS_CHANNEL_GAIN
{
    U16 wChannel;
    U16 wGain;
}
DAS_CHANNEL_GAIN;

typedef struct _DAS_IOT_CONT
{
    short            wErrorCode;
    U32              dwAccessCount;
    U32              dwPacerValue;
    U16              wAutoScan;
    U16              wIterations;
    U16              wSyncMode;
    U16              wChannelCount;
    DAS_CHANNEL_GAIN dwChannelGain[255];
}
DAS_IOT_CONT;

typedef struct _DAS_IOT_PGEN
{
    short wErrorCode;
    U32   dwAccessCount;
    U32   dwPacerValue;
}
DAS_IOT_PGEN;

/*
 * One shot mode control parameter
 * Use for DAS_IOC_XX_PIO_READ
 * Use for DAS_IOC_XX_PIO_WRITE
 */
typedef struct _DAS_IOT_PIO
{
    short wErrorCode;
    U16   wChannel;
    U16   wLine_Gain;
    U16   wByteCount;
    U32   dwValue;
}
DAS_IOT_PIO;

typedef struct _DAS_IOT_SIMU_PIO
{
    short            wErrorCode;
    U16              wDirectory;
    U16              wChannelCount;
    DAS_CHANNEL_GAIN dwChannelGain[64];
}
DAS_IOT_SIMU_PIO;

/*
 * Timer control parameter
 * Use for DAS_IOC_CTR_XXXXXXX
 */
typedef struct _DAS_IOT_CTR
{
    short wErrorCode;
    U16   wCounterIndex;
    U16   wMode;
    U16   wState;
    U16   wBinBcd;
    U32   dwCounterValue;
}
DAS_IOT_CTR;

/*
 * GPTC control parameter
 * Use for DAS_IOC_GPTC_XXXXXXX
 */
typedef struct _DAS_IOT_GPTC
{
    short wErrorCode;
    U16   wCounterIndex;
    union{
        struct{
            U16 bMode: 2;
            U16 Reserved : 1;
            U16 fExternalClkSrc : 1;
            U16 fExternalGateSrc : 1;
            U16 fExternalUpDownSrc : 1;
            U16 fUpCounter : 1;
            U16 fEnableGPTC : 1;
            U16 bReserved2 : 8;
        }Ctrlbit;
        U16 wValue;
    }Ctrldata;
    U32 dwCounterValue;
}
DAS_IOT_GPTC;

/*
 * Scan Counter control parameter
 * Use for PCI9116 counter setting
 * Yuan add 10/02/00
 */
typedef struct _DAS_IOT_SI
{
    short wErrorCode;
    U32   dwScanInterval;
    U32   dwSampInterval;
}
DAS_IOT_SI;

/*
 * General Port Read/Write parameter
 * Use for DAS_IOC_PORT_XXXX
 */
//used for winnt
typedef struct _DAS_IOT_PORTIO
{
    short wErrorCode;
    U16   wPort;
    U16   wByteCount;
    U32   dwValue;
}
DAS_IOT_PORTIO;

//used for win98
typedef struct _DAS_IOT_DPT
{
    short wErrorCode;
    U16   wPort;
    U16   wByteCount;
    U32   dwValue;
}
DAS_IOT_DPT;

/*Dual-Interrupt-System Get/Set*/
typedef struct _DAS_IOT_DUALINT
{
    short wErrorCode;
    U8    bInt1Enable;
    U8    bInt1ActiveHigh;
    U8    bInt1Status; //Get Only
    U8    bInt1Mode;
    U8    bInt2Enable;
    U8    bInt2ActiveHigh;
    U8    bInt2Status; //Get Only
    U8    bInt2Mode;
}
DAS_IOT_DUALINT;

//yuan 10/18/99 for COS interrupt control
typedef struct
{
    U8 P_3_A   : 1;
    U8 P_3_B   : 1;
    U8 P_3_C   : 1;
    U8 bUnused : 5;
}
COS_REG;

typedef struct _DAS_IOT_COSINT
{
    short wErrorCode;
    U8    wPort;
    union{
        COS_REG cosRb;
        U8      wValue;
    }mCOS;
}
DAS_IOT_COSINT;

//yuan 12/22/05 for COS32 interrupt control
typedef struct _DAS_IOT_COSINT32
{
    short wErrorCode;
    U8    bPort;
    U32   dwValue;
}
DAS_IOT_COSINT32;

//yuan add 09/11/02
typedef union pci_dbx_overrun_flag
{
    struct{
        U16 lock:1;
        U16 overrun:1;
        U16 extended : 1;
        U16 bReserved : 13;
    }f;
    U16 w;
}
PCI_DBXORUNFLAG, *PPCI_DBXORUNFLAG;

//Double buffer overrun
typedef struct pci_dbx_overrun
{
    PCI_DBXORUNFLAG xFlag;
    U32             dwCtr_tfer[2];
    U32             dwCtr_isr[2];
}
PCI_DBXORUN, *PPCI_DBXORUN;

//yuan add 06/17/04
typedef struct pci7300_dbx_overrun
{
    PCI_DBXORUNFLAG xFlag;
    U32             dwCtr_tfer[8];
    U32             dwCtr_isr[8];
    U32             dwCtr_eisr[56];
}
PCI7300_DBXORUN, *PPCI7300_DBXORUN;

// #pragma pack(pop)

/*
 * DeviceControl IOCTL code
 */
/*---- Configuration ----*/
#define DAS_IOC_GET_BUFSIZE       _IOWR(MAGIC_NUM, 2, DAS_IOT_FUNC_BUFSIZE)
#define DAS_IOC_GET_DESP          _IOWR(MAGIC_NUM, 3, DAS_IOT_CARD_DESP)
#define DAS_IOC_AI_CONFIG         _IOWR(MAGIC_NUM, 4, DAS_IOT_CONFIG)
#define DAS_IOC_AO_CONFIG         _IOWR(MAGIC_NUM, 5, DAS_IOT_CONFIG)
#define DAS_IOC_DI_CONFIG         _IOWR(MAGIC_NUM, 6, DAS_IOT_CONFIG)
#define DAS_IOC_DO_CONFIG         _IOWR(MAGIC_NUM, 7, DAS_IOT_CONFIG)
#define DAS_IOC_DIO_CONFIG        _IOWR(MAGIC_NUM, 8, DAS_IOT_CONFIG)
#define DAS_IOC_8554CTR_CONFIG    _IOWR(MAGIC_NUM, 9, DAS_IOT_CONFIG)
/*---- Asynchronous mode ----*/
#define DAS_IOC_AI_ASY_CHECK      _IOWR(MAGIC_NUM, 10, DAS_IOT_ASYNC)
#define DAS_IOC_AO_ASY_CHECK      _IOWR(MAGIC_NUM, 11, DAS_IOT_ASYNC)
#define DAS_IOC_DI_ASY_CHECK      _IOWR(MAGIC_NUM, 12, DAS_IOT_ASYNC)
#define DAS_IOC_DO_ASY_CHECK      _IOWR(MAGIC_NUM, 13, DAS_IOT_ASYNC)
#define DAS_IOC_AI_ASY_CLEAR      _IOWR(MAGIC_NUM, 14, DAS_IOT_ASYNC)
#define DAS_IOC_AO_ASY_CLEAR      _IOWR(MAGIC_NUM, 15, DAS_IOT_ASYNC)
#define DAS_IOC_DI_ASY_CLEAR      _IOWR(MAGIC_NUM, 16, DAS_IOT_ASYNC)
#define DAS_IOC_DO_ASY_CLEAR      _IOWR(MAGIC_NUM, 17, DAS_IOT_ASYNC)
/*---- Double buffer transfer mode ----*/
#define DAS_IOC_AI_DBF_HALF_READY _IOWR(MAGIC_NUM, 18, DAS_IOT_DBF)
#define DAS_IOC_AO_DBF_HALF_READY _IOWR(MAGIC_NUM, 19, DAS_IOT_DBF)
#define DAS_IOC_DI_DBF_HALF_READY _IOWR(MAGIC_NUM, 20, DAS_IOT_DBF)
#define DAS_IOC_DO_DBF_HALF_READY _IOWR(MAGIC_NUM, 21, DAS_IOT_DBF)
#define DAS_IOC_AI_DBF_MODE       _IOWR(MAGIC_NUM, 22, DAS_IOT_DBF)
#define DAS_IOC_AO_DBF_MODE       _IOWR(MAGIC_NUM, 23, DAS_IOT_DBF)
#define DAS_IOC_DI_DBF_MODE       _IOWR(MAGIC_NUM, 24, DAS_IOT_DBF)
#define DAS_IOC_DO_DBF_MODE       _IOWR(MAGIC_NUM, 25, DAS_IOT_DBF)
/*---- Continguous mode ----*/
#define DAS_IOC_AI_CON_READ       _IOWR(MAGIC_NUM, 26, DAS_IOT_CONT)
#define DAS_IOC_AO_CON_WRITE      _IOWR(MAGIC_NUM, 27, DAS_IOT_CONT)
#define DAS_IOC_DI_CON_READ       _IOWR(MAGIC_NUM, 28, DAS_IOT_CONT)
#define DAS_IOC_DO_CON_WRITE      _IOWR(MAGIC_NUM, 29, DAS_IOT_CONT)
/*---- Pattern Generation ----*/
#define DAS_IOC_DO_PG_START       _IOWR(MAGIC_NUM, 30, DAS_IOT_PGEN)
#define DAS_IOC_DO_PG_STOP        _IOWR(MAGIC_NUM, 31, DAS_IOT_PGEN)
/*---- Continguous Status Get----*/
#define DAS_IOC_AI_CON_STATUS     _IOWR(MAGIC_NUM, 32, DAS_IOT_STATUS)
#define DAS_IOC_AO_CON_STATUS     _IOWR(MAGIC_NUM, 33, DAS_IOT_STATUS)
#define DAS_IOC_DI_CON_STATUS     _IOWR(MAGIC_NUM, 34, DAS_IOT_STATUS)
#define DAS_IOC_DO_CON_STATUS     _IOWR(MAGIC_NUM, 35, DAS_IOT_STATUS)
/*---- One shot mode ----*/
#define DAS_IOC_AI_PIO_READ       _IOWR(MAGIC_NUM, 36, DAS_IOT_PIO)
#define DAS_IOC_AO_PIO_WRITE      _IOWR(MAGIC_NUM, 37, DAS_IOT_PIO)
#define DAS_IOC_DI_PIO_READ       _IOWR(MAGIC_NUM, 38, DAS_IOT_PIO)
#define DAS_IOC_DO_PIO_WRITE      _IOWR(MAGIC_NUM, 39, DAS_IOT_PIO)
#define DAS_IOC_DO_PIO_READ       _IOWR(MAGIC_NUM, 40, DAS_IOT_PIO)
/*---- Timer control ----*/
#define DAS_IOC_CTR_CONFIG        _IOWR(MAGIC_NUM, 41, DAS_IOT_CTR)
#define DAS_IOC_CTR_RESET         _IOWR(MAGIC_NUM, 42, DAS_IOT_CTR)
#define DAS_IOC_CTR_READ          _IOWR(MAGIC_NUM, 43, DAS_IOT_CTR)
/*---- General Port Read/Write ----*/
/*
#define DAS_IOC_DPT_MAP             _IOWR(MAGIC_NUM, 44, DAS_IOT_CARD_DESP)
#define DAS_IOC_DPT_READ            _IOWR(MAGIC_NUM, 45, DAS_IOT_CARD_DESP)
#define DAS_IOC_DPT_WRITE           _IOWR(MAGIC_NUM, 46, DAS_IOT_CARD_DESP)
*/
#define DAS_IOC_PORT_READ         _IOWR(MAGIC_NUM, 47, DAS_IOT_PORTIO)
#define DAS_IOC_PORT_WRITE        _IOWR(MAGIC_NUM, 48, DAS_IOT_PORTIO)
/*---- Dual-Interrupt-System Set/Get ----*/
#define DAS_IOC_DUALINT_SET       _IOWR(MAGIC_NUM, 49, DAS_IOT_DUALINT)
#define DAS_IOC_DUALINT_GET       _IOWR(MAGIC_NUM, 50, DAS_IOT_DUALINT)
//#define DAS_IOC_DINT1_SET       _IOWR(MAGIC_NUM, 51, DAS_IOT_CARD_DESP)
//#define DAS_IOC_DINT2_SET       _IOWR(MAGIC_NUM, 52, DAS_IOT_CARD_DESP)
/*---- Wait Event -----*/
//#define DAS_IOC_WAIT_EVENT        DAS_IOCTL(WAITEVT,NO,NO)
//#define DAS_IOC_WAIT_EVENT2       DAS_IOCTL(WAITEVT2,NO,NO)
//yuan add 10/19/99 for COS interrupt setting
/*---- COS INT Set -----*/
#define DAS_IOC_COSINT_SET        _IOWR(MAGIC_NUM, 53, DAS_IOT_COSINT)
//yuan add 10/25/99 for Memory Mapping
#define DAS_IOC_MEM_MAP           _IOWR(MAGIC_NUM, 54, DAS_IOT_MEMMAP)
#define DAS_IOC_MEM_UMAP          _IOWR(MAGIC_NUM, 55, DAS_IOT_MEMMAP)
//yuan add 11/06/00 for SCan/Sample counter setting
#define DAS_IOC_9116SC_CONFIG     _IOWR(MAGIC_NUM, 56, DAS_IOT_SI)
/*---- GPTC control ----*/
#define DAS_IOC_GPTC_CONFIG       _IOWR(MAGIC_NUM, 57, DAS_IOT_GPTC)
#define DAS_IOC_GPTC_RESET        _IOWR(MAGIC_NUM, 58, DAS_IOT_GPTC)
#define DAS_IOC_GPTC_READ         _IOWR(MAGIC_NUM, 59, DAS_IOT_GPTC)
//yuan add for dbf view
#define DAS_IOC_AI_DBF_VIEW       _IOWR(MAGIC_NUM, 60, DAS_IOT_DBF_VIEW)
#define DAS_IOC_AO_DBF_VIEW       _IOWR(MAGIC_NUM, 61, DAS_IOT_DBF_VIEW)
#define DAS_IOC_DI_DBF_VIEW       _IOWR(MAGIC_NUM, 62, DAS_IOT_DBF_VIEW)
#define DAS_IOC_DO_DBF_VIEW       _IOWR(MAGIC_NUM, 63, DAS_IOT_DBF_VIEW)
//yuan add for dbf view
#define DAS_IOC_DI_MBF_SET        _IOWR(MAGIC_NUM, 64, DAS_IOT_MBUF)
//yuan add for resource getting
#define DAS_IOC_GET_RESRC         _IOWR(MAGIC_NUM, 65, DAS_IOT_RESRC)
//jeffrey add for 7300b Latency Timer setting
#define DAS_IOC_SET_LATENCY_TIMER _IOWR(MAGIC_NUM, 66, U8)
#define DAS_IOC_DI_SIGNAL_SET     _IOWR(MAGIC_NUM, 67, DAS_IOT_DUALINT)
#define DAS_IOC_COSINT_SET32      _IOWR(MAGIC_NUM, 68, DAS_IOT_COSINT32)
#define DAS_IOC_SIMU_AI_PIO       _IOWR(MAGIC_NUM, 69, DAS_IOT_SIMU_PIO)
//add for 9221
#define DAS_IOC_MBF_SET           _IOWR(MAGIC_NUM, 70, DAS_IOT_MBUF2)
#define DAS_IOC_MBF_RESET         _IOWR(MAGIC_NUM, 71, DAS_IOT_MBUF2)
#define DAS_IOC_AI_CON_READ_SG    _IOWR(MAGIC_NUM, 72, DAS_IOT_SG_CONT)
#define DAS_IOC_GPTC_CONFIG_9221  _IOWR(MAGIC_NUM, 73, DAS_IOT_GPTC_9221)
#define DAS_IOC_GPTC_RESET_9221   _IOWR(MAGIC_NUM, 74, DAS_IOT_GPTC_9221)
#define DAS_IOC_GPTC_READ_9221    _IOWR(MAGIC_NUM, 75, DAS_IOT_GPTC_9221)
#define DAS_IOC_CAL_ACT           _IOWR(MAGIC_NUM, 76, DAS_IOT_CTR)
//01/14/08 for syncmode timeout
#define DAS_IOC_AI_TIMEOUT_SET    _IOWR(MAGIC_NUM, 77, DAS_IOT_SI)
#define DAS_IOC_AO_TIMEOUT_SET    _IOWR(MAGIC_NUM, 78, DAS_IOT_SI)
#define DAS_IOC_DI_TIMEOUT_SET    _IOWR(MAGIC_NUM, 79, DAS_IOT_SI)
#define DAS_IOC_DO_TIMEOUT_SET    _IOWR(MAGIC_NUM, 80, DAS_IOT_SI)

/*
 * Constant Define for system-call argument
 */
/*Error Number return from system-call*/
#define DAS_ERR_NO                 0
#define DAS_ERR_POOL_ALLOC         -101
#define DAS_ERR_IOCTL_CODE         -102
#define DAS_ERR_IOCTL_INSIZE       -103
#define DAS_ERR_IOCTL_OUTSIZE      -104
#define DAS_ERR_NO_FUNCTION        -105
#define DAS_ERR_RUNNING            -106
#define DAS_ERR_DOUBLE_BUFFER_MODE -107
#define DAS_ERR_SYNC_MODE          -108
#define DAS_ERR_TRIGGER_SOURCE     -109
#define DAS_ERR_PACER_COUNT        -110
#define DAS_ERR_COUNTER_NUMBER     -111
#define DAS_ERR_COUNTER_MODE       -112
#define DAS_ERR_COUNTER_STATE      -113
#define DAS_ERR_ACCESS_COUNT       -114
#define DAS_ERR_CHANNEL_NUMBER     -115
#define DAS_ERR_CHANNEL_BYTE_COUNT -116
#define DAS_ERR_CHANNEL_LINE       -117
#define DAS_ERR_PORTIO_BYTE_COUNT  -118
#define DAS_ERR_CHANNEL_GAIN_COUNT -119
#define DAS_ERR_IO_TIMEOUT         -120
#define DAS_ERR_BURST_MODE         -121
#define DAS_ERR_TRIGGER_MODE       -122
#define DAS_ERR_MISALIGN           -123
#define DAS_ERR_NO_DEVICE          -124
#define DAS_STATUS_PENDING         101

/*Iteration Count*/
#define DAS_ITER_INFINIT 0

/*Synchronous Mode*/
#define DAS_SYN_SYNCH  1
#define DAS_SYN_ASYNCH 2

/*Trigger Source*/
#define DAS_TRG_SOFTWARE   0
#define DAS_TRG_INT_PACER  1
#define DAS_TRG_EXT_STROBE 2
#define DAS_TRG_HANDSHAKE  3

/*AD Range*/
#define DAS_RNG_B_10_V     1
#define DAS_RNG_B_5_V      2
#define DAS_RNG_B_2_5_V    3
#define DAS_RNG_B_1_25_V   4
#define DAS_RNG_B_0_625_V  5
#define DAS_RNG_B_0_3125_V 6
#define DAS_RNG_B_0_5_V    7
#define DAS_RNG_B_0_05_V   8
#define DAS_RNG_B_0_005_V  9
#define DAS_RNG_B_1_V      10
#define DAS_RNG_B_0_1_V    11
#define DAS_RNG_B_0_01_V   12
#define DAS_RNG_B_0_001_V  13
#define DAS_RNG_U_20_V     14
#define DAS_RNG_U_10_V     15
#define DAS_RNG_U_5_V      16
#define DAS_RNG_U_2_5_V    17
#define DAS_RNG_U_1_25_V   18
#define DAS_RNG_U_1_V      19
#define DAS_RNG_U_0_1_V    20
#define DAS_RNG_U_0_01_V   21
#define DAS_RNG_U_0_001_V  22
#define DAS_RNG_B_2_V      23
#define DAS_RNG_U_2_V      24

/*Channel Number for Parrell port device*/
#define DAS_CHN_P1A   0
#define DAS_CHN_P1B   1
#define DAS_CHN_P1C   2
#define DAS_CHN_P1CL  3
#define DAS_CHN_P1CH  4
#define DAS_CHN_P1AE 10
#define DAS_CHN_P1BE 11
#define DAS_CHN_P1CE 12
#define DAS_CHN_P2A   5
#define DAS_CHN_P2B   6
#define DAS_CHN_P2C   7
#define DAS_CHN_P2CL  8
#define DAS_CHN_P2CH  9
#define DAS_CHN_P2AE 15
#define DAS_CHN_P2BE 16
#define DAS_CHN_P2CE 17
#define DAS_CHN_P3A  10
#define DAS_CHN_P3B  11
#define DAS_CHN_P3C  12
#define DAS_CHN_P3CL 13
#define DAS_CHN_P3CH 14
#define DAS_CHN_P4A  15
#define DAS_CHN_P4B  16
#define DAS_CHN_P4C  17
#define DAS_CHN_P4CL 18
#define DAS_CHN_P4CH 19
#define DAS_CHN_P5A  20
#define DAS_CHN_P5B  21
#define DAS_CHN_P5C  22
#define DAS_CHN_P5CL 23
#define DAS_CHN_P5CH 24
#define DAS_CHN_P6A  25
#define DAS_CHN_P6B  26
#define DAS_CHN_P6C  27
#define DAS_CHN_P6CL 28
#define DAS_CHN_P6CH 29
#define DAS_CHN_P1T  30
#define DAS_CHN_P2T  31
#define DAS_CHN_P3T  32
#define DAS_CHN_P4T  33
#define DAS_CHN_P1E  34
#define DAS_CHN_P2E  35
#define DAS_CHN_P3E  36
#define DAS_CHN_P4E  37

/*Channel Width*/
#define DAS_PORT_WIDTH_0  0
#define DAS_PORT_WIDTH_4  1
#define DAS_PORT_WIDTH_8  2
#define DAS_PORT_WIDTH_16 3
#define DAS_PORT_WIDTH_24 4
#define DAS_PORT_WIDTH_32 5
#define DAS_PORT_WIDTH_64 6

/*COS INT Selection Port*/
#define COS_P1 0
#define COS_P2 1
#define COS_P3 2
#define COS_P4 3

/**/
#define DAS_DISABLE 0
#define DAS_ENABLE  1

/**/
#define PCI6208_CURRENT_0_20MA 0
#define PCI6208_CURRENT_4_20MA 3
#define PCI6208_CURRENT_5_25MA 1

/**/
#define PCI6308_CURRENT_0_20MA 0
#define PCI6308_CURRENT_4_20MA 3
#define PCI6308_CURRENT_5_25MA 1

/*Interrupt Source Control/Status for PCI7248*/
#define PCI7248_INT1_DISABLE       0 //INT1 Disable
#define PCI7248_INT1_FP1C0         1 //P1C0 Falling
#define PCI7248_INT1_RP1C0_FP1C3   2 //P1C0 Rising or P1C3 Falling
#define PCI7248_INT1_EVENT_COUNTER 3 //Event counter down to 0
#define PCI7248_INT2_DISABLE       0 //INT2 Disable
#define PCI7248_INT2_FP2C0         1 //P2C0 Falling
#define PCI7248_INT2_RP2C0_FP2C3   2 //P2C0 Rising or P2C3 Falling
#define PCI7248_INT2_TIMER_COUNTER 3 //Timer counter down to 0

/*Interrupt Source Control/Status for PCI7296*/
#define PCI7296_INT1_DISABLE       0 //INT1 Disable
#define PCI7296_INT1_FP1C0         1 //P1C0 Falling
#define PCI7296_INT1_RP1C0_FP1C3   2 //P1C0 Rising or P1C3 Falling
#define PCI7296_INT1_EVENT_COUNTER 3 //Event counter down to 0
#define PCI7296_INT2_DISABLE       0 //INT2 Disable
#define PCI7296_INT2_FP2C0         1 //P2C0 Falling
#define PCI7296_INT2_RP2C0_FP2C3   2 //P2C0 Rising or P2C3 Falling
#define PCI7296_INT2_TIMER_COUNTER 3 //Timer counter down to 0

/*Interrupt Source Control/Status for PCI7396*/
#define PCI7396_INT1_COS           0 //COS INT1
#define PCI7396_INT1_FP1C0         1 //P1C0 Falling
#define PCI7396_INT1_RP1C0_FP1C3   2 //P1C0 Rising or P1C3 Falling
#define PCI7396_INT1_EVENT_COUNTER 3 //Event counter down to 0
#define PCI7396_INT1_DISABLE       4 //INT1 Disable
#define PCI7396_INT2_COS           0 //COS INT2
#define PCI7396_INT2_FP2C0         1 //P2C0 Falling
#define PCI7396_INT2_RP2C0_FP2C3   2 //P2C0 Rising or P2C3 Falling
#define PCI7396_INT2_TIMER_COUNTER 3 //Timer counter down to 0
#define PCI7396_INT2_DISABLE       4 //INT2 Disable

/*PCI7300 DIO Config*/
/*Trigger Source*/
#define PCI7300_TRG_INT_PACER        DAS_TRG_INT_PACER
#define PCI7300_TRG_EXT_STROBE       DAS_TRG_EXT_STROBE
#define PCI7300_TRG_HANDSHAKE        DAS_TRG_HANDSHAKE
#define PCI7300_TRG_CLK_10MHZ        DAS_TRG_HANDSHAKE+1
#define PCI7300_TRG_CLK_20MHZ        DAS_TRG_HANDSHAKE+2
//hua+ For 7300b DO
#define PCI7300_TRG_CLK_ACK          DAS_TRG_HANDSHAKE
#define PCI7300_TRG_DO_CLK_TIMER_ACK DAS_TRG_HANDSHAKE+3
#define PCI7300_TRG_DO_CLK_10M_ACK   DAS_TRG_HANDSHAKE+4
#define PCI7300_TRG_DO_CLK_20M_ACK   DAS_TRG_HANDSHAKE+5
//hua-
/*StartMode*/
#define PCI7300_START_WAIT_NO        0
#define PCI7300_START_WAIT_TRG       1
#define PCI7300_START_WAIT_FIFO      2
#define PCI7300_START_WAIT_BOTH      3
/*Terminator control*/
#define PCI7300_TERM_OFF             0
#define PCI7300_TERM_ON              1
/*PCI7300 Channel Count*/
#define PCI7300_CHANNEL_DO           0
#define PCI7300_CHANNEL_EDO          1
#define PCI7300_CHANNEL_DI           0
#define PCI7300_CHANNEL_EDI          1
//hua+
//Control signals polarity
#define PCI7300_DIREQ_POS            0x00000000L
#define PCI7300_DIREQ_NEG            0x00000001L
#define PCI7300_DIACK_POS            0x00000000L
#define PCI7300_DIACK_NEG            0x00000002L
#define PCI7300_DITRIG_POS           0x00000000L
#define PCI7300_DITRIG_NEG           0x00000004L
#define PCI7300_DOREQ_POS            0x00000000L
#define PCI7300_DOREQ_NEG            0x00000008L
#define PCI7300_DOACK_POS            0x00000000L
#define PCI7300_DOACK_NEG            0x00000010L
#define PCI7300_DOTRIG_POS           0x00000000L
#define PCI7300_DOTRIG_NEG           0x00000020L
//hua-

/*Channel Defination for PCI7433, PCI7434*/
#define PCI7433_CHANNEL_DI_LOW  0
#define PCI7433_CHANNEL_DI_HIGH 1
#define PCI7434_CHANNEL_DO_LOW  0
#define PCI7434_CHANNEL_DO_HIGH 1
#define PCI7434R_CHANNEL_DO_LED 2

/*Interrupt Source Control/Status for PCI9111*/
#define PCI9111_INT1_EOC     0 //Ending of AD conversion
#define PCI9111_INT1_FIFO_HF 1 //FIFO Half Full
#define PCI9111_INT2_PACER   0 //Every Timer tick
#define PCI9111_INT2_EXT_TRG 1 //ExtTrig High->Low
/*Channel Count*/
#define PCI9111_CHANNEL_DO   0
#define PCI9111_CHANNEL_EDO  1
#define PCI9111_CHANNEL_DI   0
#define PCI9111_CHANNEL_EDI  1
/*EDO function*/
#define PCI9111_EDO_INPUT    1
#define PCI9111_EDO_OUT_EDO  2
#define PCI9111_EDO_OUT_CHN  3

/*Interrupt Source Control/Status for PCI9113*/
#define PCI9113_INT1_EOC     0 //Ending of AD conversion
#define PCI9113_INT1_FIFO_HF 1 //FIFO Half Full
#define PCI9113_INT2_PACER   0 //Every Timer tick
#define PCI9113_INT2_EXT_TRG 0 //ExtTrig High->Low

/*Interrupt Source Control/Status for PCI9114*/
#define PCI9114_INT1_EOC     0 //Ending of AD conversion
#define PCI9114_INT1_FIFO_HF 1 //FIFO Half Full
#define PCI9114_INT2_PACER   0 //Every Timer tick

/*9116 setting*/
/*-- Trigger Mode --*/
#define PCI9116_TRGMOD_SOFT  0
#define PCI9116_TRGMOD_POST  1
#define PCI9116_TRGMOD_DELAY 2
#define PCI9116_TRGMOD_PRE   3
#define PCI9116_TRGMOD_MIDL  4
#define PCI9116_TRGMOD_MAX   4

/*PCI9812 Constant define*/
/*-- Channel Gain - Channel --*/
#define PCI9812_CHANNEL_CNT1   1
#define PCI9812_CHANNEL_CNT2   2
#define PCI9812_CHANNEL_CNT4   4
/*-- Trigger Mode --*/
#define PCI9812_TRGMOD_SOFT    0
#define PCI9812_TRGMOD_POST    1
#define PCI9812_TRGMOD_PRE     2
#define PCI9812_TRGMOD_DELAY   3
#define PCI9812_TRGMOD_MIDL    4
#define PCI9812_TRGMOD_MAX     4
/*-- Trigger Source --*/
#define PCI9812_TRGSRC_CH0     0
#define PCI9812_TRGSRC_CH1     1
#define PCI9812_TRGSRC_CH2     2
#define PCI9812_TRGSRC_CH3     3
#define PCI9812_TRGSRC_EXT_DIG 4
#define PCI9812_TRGSLP_POS     0
#define PCI9812_TRGSLP_NEG     1
/*-- Trigger Level --*/
#define PCI9812_TRGLVL_MIN     -5.00
#define PCI9812_TRGLVL_STEP    0.04
/*-- Frequency Selection --*/
#define PCI9812_AD2_GT_PCI     1
#define PCI9812_AD2_LT_PCI     0
/*-- Clock Source --*/
#define PCI9812_CLKSRC_INT     0
#define PCI9812_CLKSRC_EXT_SIN 1
#define PCI9812_CLKSRC_EXT_DIG 2
#define PCI9812_CLKSRC_MAX     2
/**/
#define ExtTrigLine            100

#endif
