#include <linux/version.h>
#include <linux/module.h>   /* Specifically, a module */
#include <linux/kernel.h>   /* We're doing kernel work */
//#include <linux/config.h>   // for PCI, SMP configuration
#include <linux/mm.h>
#include <linux/delay.h>
#include <linux/pci.h>
#include <linux/fs.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <linux/interrupt.h>

#ifndef MODULE
#define MODULE
#endif

#ifdef CONFIG_SMP
#ifndef __SMP__
#define __SMP__
#endif
#endif

#if defined(__SMP__)
#include <linux/smp.h>
#endif

#if defined(MODVERSIONS)
#if defined(__SMP__)
#include <linux/modversions-smp.h>
#else
#include <linux/modversions.h>
#endif
#endif


#if ( LINUX_VERSION_CODE < KERNEL_VERSION(2,4,13) )
#include <linux/malloc.h>
#else
#include <linux/slab.h>
#endif

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
#include <unistd.h>
#endif
#endif

#include "../include/Adlsys.h"
#include "../include/Adldev.h"
#include "../include/Adllib.h"
#include "../include/Adlplx90.h"
#include "../include/AdlS593X.h"
#include "../include/Adli8254.h"

/* Our own ioctl numbers */
/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
/*---- The Channel Limit ----*/
#define ADL_BYTES_DIO    1
#define ADL_EXTBOARD_MAX 3
/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/
//  Board Port Relative Address - Port Offset
#define ADL_PT_BASE   0x00
#define ADL_PT_SIZE   2
#define ADL_PT_DO     0x00
#define ADL_PT_DI     0x01
#define ADL_PT_DIO    0x00
#define ADL_PT_CLRIRQ 0x08
/*>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>*/

/* The name for our device, as it will appear in
 * /proc/devices */
#define DEVICE_NAME   "p7250"
#define ADL_VENDOR_ID 0x144A
#define ADL_DEVICE_ID 0x7250

extern DAS_IOT_CARD_DESP  iniCardDesp;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
extern irqreturn_t p7250_interrupt_handler(int irq, void *dev_id, struct pt_regs *regs);
#else
extern irqreturn_t p7250_interrupt_handler(int irq, void *dev_id);
#endif

PCI_DEVEXT DevExtension[32]; // instance for every device_open
static PCI_DRVEXT drv_ext;
int device_number = 0; // return by AdlFindDevice(...) in init_module

static int MAJOR = 0;
int irq_result =0;
// static int READ_PORT,WRITE_PORT;

static int pci7250_init(struct pci_dev *, const struct pci_device_id *);
void pci7250_remove(struct pci_dev *);

int p7250_open(struct inode *inode, struct file *file)
{
    int dev_index;
    PPCI_DEVEXT dev_ext;
    U32  dwPort;
    U8* bpCtrl;
    U16  wIter, wValue;

    //printk ("DRIVE: device_open(%p)\n", file);
    dev_index = MINOR(inode->i_rdev);
    if( dev_index > device_number )
        return DAS_ERR_NO_DEVICE;
    if( !DevExtension[dev_index].openable )
        return DAS_ERR_NO_DEVICE;
    dev_ext = &(DevExtension[dev_index]);
    if( !AdlOpenDevice( dev_ext ) )
        return DAS_ERR_NO_DEVICE;
    dwPort = OP_Port(dev_ext);
    bpCtrl = (U8*)&(dev_ext->dwDevRA);
    *bpCtrl = 1;
    bpCtrl++;
    /*--------------------------------------*/
    for (wIter=1; wIter<= ADL_EXTBOARD_MAX; wIter++) {
        PLX90_SetUserIoPin(dwPort,wIter, 0);
        PLX90_GetUserIoPin(dwPort,wIter, &wValue);
        *bpCtrl = wValue ? 0 : 1;
        bpCtrl++;
    }
    /*--------------------------------------*/
    dwPort = PT_Port(dev_ext);
    PT_outb(dwPort, CLRIRQ, 0);
    /*--------------------------------------*/

    file->private_data = dev_ext;
    //printk("DRIVE: minor number is %d\n",INDEX);
    dev_ext->Reference ++;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
    MOD_INC_USE_COUNT;
#endif

    return DAS_ERR_NO;
}


int p7250_release(struct inode *inode,struct file *file)
{

    PPCI_DEVEXT pDevExt = file->private_data;

    pDevExt->Reference --;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
    MOD_DEC_USE_COUNT;
#endif

    return DAS_ERR_NO;
}

int AdlCheckPIO(PPCI_DEVEXT  pDevExt, DAS_IOT_PIO* pSbuf)
{
    unsigned char  bTmp;

    if ( pSbuf->wChannel > ADL_EXTBOARD_MAX ) {
        pSbuf->wErrorCode = DAS_ERR_CHANNEL_NUMBER;
        return 0;
    }

    bTmp = ((unsigned char*)&pDevExt->dwDevRA)[pSbuf->wChannel];
    if ( bTmp == 0 ) {
        pSbuf->wErrorCode = DAS_ERR_CHANNEL_LINE;
        return 0;
    }

    return 1;
}
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,36)
int p7250_ioctl( 
	struct inode *inode,  
	struct file *file,
    unsigned int ioctl_num,/* The number of the ioctl */
    unsigned long ioctl_param) /* The parameter to it */
{
#else
long p7250_ioctl( 
	struct file *file,
    unsigned int ioctl_num,/* The number of the ioctl */
    unsigned long ioctl_param) /* The parameter to it */
{
	struct inode  *inode = file->f_mapping->host;
#endif
    PPCI_DEVEXT  pDevExt = file->private_data;
    void* temp_param;
    short error_code;
    U32 dwPort,vi;

    dwPort = PT_Port(pDevExt) + ADL_PT_BASE;

    switch ( ioctl_num ){
        case DAS_IOC_GET_DESP :
            temp_param = (void *)kmalloc( sizeof( DAS_IOT_CARD_DESP ), GFP_ATOMIC );
            if (! AdlIocGetDesp(pDevExt, (DAS_IOT_CARD_DESP*)temp_param )){
                error_code = ((DAS_IOT_CARD_DESP*)temp_param)->wErrorCode;
                if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_CARD_DESP))){
                    kfree(temp_param);
                    return -EFAULT;
                }
                kfree( temp_param );
                return error_code;
            }
            if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_CARD_DESP))){
                kfree(temp_param);
                return -EFAULT;
            }
            kfree( temp_param );
            break;
        //yuan add 05/10/01
        case DAS_IOC_GET_RESRC :
            temp_param = (void *)kmalloc( sizeof( DAS_IOT_RESRC ), GFP_ATOMIC );
            for(vi=0; vi<4; vi++)
                ((DAS_IOT_RESRC*)temp_param)->PortAddr[vi] = (U16) (pDevExt->dwPortAddr[vi]);
            if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_RESRC))){
                kfree(temp_param);
                return -EFAULT;
            }
            kfree( temp_param );
            break;
        case DAS_IOC_DUALINT_SET :
            temp_param = (void *)kmalloc( sizeof(DAS_IOT_DUALINT), GFP_ATOMIC );
            if(copy_from_user ((void *)temp_param, (const void*)ioctl_param, sizeof(DAS_IOT_DUALINT))) {
                kfree(temp_param);
                return -EFAULT;
            }
            if ( ! pDevExt->pDin ) {
                ((DAS_IOT_DUALINT *) temp_param)->wErrorCode = DAS_ERR_NO_FUNCTION;
                if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_DUALINT))){
                    kfree(temp_param);
                    return -EFAULT;
                }
                kfree(temp_param);
                return -EFAULT;
            }
            // chained the Signal request to fasync_queue , before the AdlSetDualInt
            if( ((DAS_IOT_DUALINT *) temp_param)->bInt1Status ){ // change of SIGNAL async_queue is needed , fa->magic will be set to FASYNC_MAGIC
                if( fasync_helper( (int)inode, file, ((DAS_IOT_DUALINT *)temp_param)->bInt1Enable? 1:0, &(pDevExt->pDin->async_queue1) ) == -ENOMEM ){
                    // no enough memory for kmalloc or no matched
                    ((DAS_IOT_DUALINT *) temp_param)->wErrorCode = DAS_ERR_POOL_ALLOC;
                    if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_DUALINT))){
                        kfree(temp_param);
                        return -EFAULT;
                    }
                    kfree(temp_param);
                    return -1;
                }
            }

            if( ((DAS_IOT_DUALINT *) temp_param)->bInt2Status ){ // change of SIGNAL async_queue is needed
                if( (fasync_helper( (int)inode, file, ((DAS_IOT_DUALINT *) temp_param)->bInt2Enable? 1:0, \
                    &(pDevExt->pDin->async_queue2)) ) == -ENOMEM ){
                    // no enough memory for kmalloc or no matched
                    // clear the linked fasync_queue before
                    if( ((DAS_IOT_DUALINT *) temp_param)->bInt1Status && ((DAS_IOT_DUALINT *)temp_param)->bInt1Enable ){
                        fasync_helper( (int)inode, file, 0, &(pDevExt->pDin->async_queue1) );
                    }

                    ((DAS_IOT_DUALINT *) temp_param)->wErrorCode = DAS_ERR_POOL_ALLOC;
                    if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_DUALINT))){
                        kfree(temp_param);
                        return -EFAULT;
                    }
                    kfree(temp_param);
                    return -1;
                }
            }


            ((DAS_IOT_DUALINT*)temp_param)->bInt1Mode &= 1;
            ((DAS_IOT_DUALINT*)temp_param)->bInt2Mode &= 1;
            PLX90_IocSetDualInt(pDevExt, (DAS_IOT_DUALINT *) temp_param);
            PT_outb(dwPort, CLRIRQ, 0);
            //AdlSetDualInt(pDevExt, (DAS_IOT_DUALINT*)temp_param);
            kfree( temp_param );
            break;
        case DAS_IOC_DUALINT_GET :
            temp_param = (void *)kmalloc( sizeof(DAS_IOT_DUALINT), GFP_ATOMIC );
            if(copy_from_user ((void *)temp_param, (const void*)ioctl_param, sizeof(DAS_IOT_DUALINT))) {
                kfree(temp_param);
                return -EFAULT;
            }
            if ( ! pDevExt->pDin ) {
                ((DAS_IOT_DUALINT *) temp_param)->wErrorCode = DAS_ERR_NO_FUNCTION;
                if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_DUALINT))){
                    kfree(temp_param);
                    return -EFAULT;
                }
                kfree(temp_param);
                return -EFAULT;
            }
            PLX90_IocGetDualInt(pDevExt, (DAS_IOT_DUALINT*)temp_param);
            if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_DUALINT))){
                kfree(temp_param);
                return -EFAULT;
            }
            kfree( temp_param );
            break;
        /*-----------------------------------------------------------*/
        case DAS_IOC_DI_PIO_READ :
            temp_param = (void *)kmalloc( sizeof(DAS_IOT_PIO), GFP_ATOMIC );
            //copy_from_user ((void *)temp_param, (const void*)ioctl_param, sizeof(DAS_IOT_PIO));
            if(copy_from_user ((void *)temp_param, (const void*)ioctl_param, sizeof(DAS_IOT_PIO))) {
                kfree(temp_param);
                return -EFAULT;
            }
            if ( ! AdlCheckPIO(pDevExt, (DAS_IOT_PIO*)temp_param) ){
                error_code = ((DAS_IOT_PIO*)temp_param)->wErrorCode;
                if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_PIO))){
                    kfree(temp_param);
                    return -EFAULT;
                }
                kfree( temp_param );
                return error_code;
            }
            dwPort += ((DAS_IOT_PIO*)temp_param)->wChannel * ADL_PT_SIZE;

            ((DAS_IOT_PIO*)temp_param)->dwValue = PT_inpb(dwPort, DI);

            if ( ! AdlIocPioRead(1, (DAS_IOT_PIO*)temp_param) ){
                error_code = ((DAS_IOT_PIO*)temp_param)->wErrorCode;
                if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_PIO))){
                    kfree(temp_param);
                    return -EFAULT;
                }
                kfree( temp_param );
                return error_code;
            }
            if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_PIO))){
                kfree(temp_param);
                return -EFAULT;
            }
            kfree( temp_param );
            break;
        case DAS_IOC_DO_PIO_READ :
            temp_param = (void *)kmalloc( sizeof(DAS_IOT_PIO), GFP_ATOMIC );
            //copy_from_user ((void *)temp_param, (const void*)ioctl_param, sizeof(DAS_IOT_PIO));
            if(copy_from_user ((void *)temp_param, (const void*)ioctl_param, sizeof(DAS_IOT_PIO))) {
                kfree(temp_param);
                return -EFAULT;
            }
            if ( ! AdlCheckPIO(pDevExt, (DAS_IOT_PIO*)temp_param) ){
                error_code = ((DAS_IOT_PIO*)temp_param)->wErrorCode;
                if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_PIO))){
                    kfree(temp_param);
                    return -EFAULT;
                }
                kfree( temp_param );
                return error_code;
            }
            dwPort += ((DAS_IOT_PIO*)temp_param)->wChannel * ADL_PT_SIZE;
            ((DAS_IOT_PIO*)temp_param)->dwValue = PT_inpb(dwPort, DIO);

            if ( ! AdlIocPioRead(1, (DAS_IOT_PIO*)temp_param) ){
                error_code = ((DAS_IOT_PIO*)temp_param)->wErrorCode;
                if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_PIO))){
                    kfree(temp_param);
                    return -EFAULT;
                }
                kfree( temp_param );
                return error_code;
            }
            if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_PIO))){
                kfree(temp_param);
                return -EFAULT;
            }
            /*if(copy_from_user ((void *)temp_param, (const void*)ioctl_param, sizeof(DAS_IOT_PIO))) {
                 kfree(temp_param);
                 return -EFAULT;
            }*/
            //copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_PIO));
            kfree( temp_param );
            break;
        case DAS_IOC_DO_PIO_WRITE :
            temp_param = (void *)kmalloc( sizeof(DAS_IOT_PIO), GFP_ATOMIC );
            if(copy_from_user ((void *)temp_param, (const void*)ioctl_param, sizeof(DAS_IOT_PIO))) {
                kfree(temp_param);
                return -EFAULT;
            }
            if ( ! AdlCheckPIO(pDevExt, (DAS_IOT_PIO*)temp_param) ){
                error_code = ((DAS_IOT_PIO*)temp_param)->wErrorCode;
                if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_PIO))){
                    kfree(temp_param);
                    return -EFAULT;
                }
                kfree( temp_param );
                return error_code;
            }
            dwPort += ((DAS_IOT_PIO*)temp_param)->wChannel * ADL_PT_SIZE;
            if ( ! AdlIocPioWrite(1, (DAS_IOT_PIO*)temp_param, PT_inpb(dwPort, DIO)) ){
                error_code = ((DAS_IOT_PIO*)temp_param)->wErrorCode;
                if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_PIO))){
                    kfree(temp_param);
                    return -EFAULT;
                }
                kfree( temp_param );
                return error_code;
            }
            PT_outb(dwPort, DO, (U8)((DAS_IOT_PIO*)temp_param)->dwValue);
            kfree( temp_param );
            break;
        case DAS_IOC_PORT_READ :
            temp_param = (void *)kmalloc( sizeof(DAS_IOT_PORTIO), GFP_ATOMIC );
            if(copy_from_user ((void *)temp_param, (const void*)ioctl_param, sizeof(DAS_IOT_PORTIO))){
                kfree(temp_param);
                return -EFAULT;
            }
            AdlIocPortIoRead(pDevExt,1,(DAS_IOT_PORTIO*)temp_param);
            if(copy_to_user((void *)ioctl_param, (const void *)temp_param, sizeof(DAS_IOT_PORTIO))){
                kfree(temp_param);
                return -EFAULT;
            }
            kfree( temp_param );
            break;
        case DAS_IOC_PORT_WRITE :
            temp_param = (void *)kmalloc( sizeof(DAS_IOT_PORTIO), GFP_ATOMIC );
            if(copy_from_user ((void *)temp_param, (const void*)ioctl_param, sizeof(DAS_IOT_PORTIO))){
                kfree(temp_param);
                return -EFAULT;
            }
            AdlIocPortIoWrite(pDevExt,1,(DAS_IOT_PORTIO*)temp_param );
            kfree( temp_param );
            break;
        /*-----------------------------------------------------------*/
        default :
            return -EINVAL;

    };

    return DAS_ERR_NO;

}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
struct file_operations fops = {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,36)
    ioctl: p7250_ioctl,   /*ioctl */
#else
    unlocked_ioctl: p7250_ioctl,   /*ioctl */
#endif
    open: p7250_open,
    release: p7250_release  /* a.k.a. close */
};

#elif LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0)
struct file_operations fops = {
    NULL,   /* seek */
    NULL, // p7250_read,
    NULL, // p7250_write,
    NULL,   /* readdir */
    NULL,   /* select */
    p7250_ioctl,   /*ioctl */
    NULL,   /* mmap */
    p7250_open,
    NULL,  /* flush */
    p7250_release  /* a.k.a. close */
};

#endif

static struct pci_device_id pci7250_tbl[] =
{
    {0x144A, 0x7250, PCI_ANY_ID, PCI_ANY_ID, },
    {0, }
};

static struct pci_driver pci7250_driver =
{
    name    : "PCI-7250 Driver",
    id_table    : pci7250_tbl,
    probe   : pci7250_init,
    remove  : pci7250_remove,
};

MODULE_DEVICE_TABLE(pci, pci7250_tbl);
#define DUMMY_NAME "PCI7250W"
static struct class *p7250_class;

static int __devinit pci7250_init(struct pci_dev *pciptr, const struct pci_device_id *ent)
{
    int err=0;
    int vi;
    int irq_result = 0;
//    PPCI_DBX pDbx;

    memset(&(DevExtension[device_number]), 0, sizeof(PCI_DEVEXT));
    DevExtension[device_number].pDrvExt = &drv_ext;

    err = AdlInitDevice(0x144A, 0x7250, PCI_CTRL_PLX90, 2, \
        &(DevExtension[device_number]), pciptr, device_number
    );
    if(err<0)
    {
        printk(KERN_ALERT "[PCI-7250] pci7250_init: AdlInitDevice failed!\n");
        goto err_out;
    }

    for(vi=0; vi<PCI_PORT_MAX; vi++)
    {
        if((DevExtension[device_number].dwPortAddr[vi]>0)&&\
            (DevExtension[device_number].dwPortLen[vi]>0))
        {
            request_region(DevExtension[device_number].dwPortAddr[vi], \
                DevExtension[device_number].dwPortLen[vi], "pci7250");
        }
    }

    if(DevExtension[device_number].Irq_no!=0)
    {
        irq_result = request_irq(DevExtension[device_number].Irq_no, \
            (irq_handler_t)p7250_interrupt_handler, \
            /*SA_INTERRUPT|SA_SHIRQ*/IRQF_SHARED, \
           "pci7250", &DevExtension[device_number]);

    if(irq_result!=0){ // unsuccess irq request
            DevExtension[device_number].openable = FALSE;
            goto allocate_error;
        }
    }
    DevExtension[device_number].openable = TRUE;
                                                                                                                                                #if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
    // replace the code segment for __WAIT_QUEUE_HEAD_INITIALIZER
    init_waitqueue_head(&(DevExtension[device_number].sync_wq));
#endif

#ifdef CONFIG_DEVFS_FS
    sprintf(buf, "pci7250W%ld", device_number);
    #if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
        err = devfs_mk_cdev(MKDEV(MAJOR, device_number), \
            S_IFCHR|S_IRUGO|S_IWUGO, buf
        );
        //printk("devfs_mk_cdev error: %d\n", err);
    #else if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
        DevExtension[device_number].devfs_handle = \
            devfs_register(NULL, buf, DEVFS_FL_NONE, MAJOR, device_number, \
                S_IFCHR|S_IRUSR|S_IWUSR, &pci7250_fops, NULL
            );
        if(DevExtension[device_number].devfs_handle==NULL){
            printk("[PCI-7250] pci7250_init: devfs_register( pci7250W%ld ) failed!!\n", device_number);
        //goto iounmap_out;
    }
    #endif
#endif
    //pciptr->driver_data = &(DevExtension[device_number];
    pci_set_drvdata(pciptr, &(DevExtension[device_number]));
    printk(KERN_ALERT "[PCI-7250] pci7250_init: Inserting pci7250 device: %d \n", DevExtension[device_number].wCard);

    /* register your own device in sysfs, and this will cause udevd to create corresponding device node */
    device_create(p7250_class, NULL, MKDEV(MAJOR, device_number), NULL, \
        DUMMY_NAME "%d", device_number
    );
    device_number++;
    return 0;

allocate_error :
    // use the i in AdlDbxAllocate( &DevExtension[i] )
    if(DevExtension[device_number].pDin)
        AdlDropDsen(DevExtension[device_number].pDin);
    if(DevExtension[device_number].pTcn)
        AdlDropDsen(DevExtension[device_number].pTcn);
err_out:
    pci_disable_device(pciptr);
    return -ENODEV;
}

void pci7250_remove(struct pci_dev *pdev)
{
    int vi;
    PPCI_DEVEXT pDevExt;
    //PPCI_DBX pDbx;

    pDevExt = pci_get_drvdata(pdev);

    for(vi=0; vi<PCI_PORT_MAX; vi++)
    {
    if((pDevExt->dwPortAddr[vi]>0)&&(pDevExt->dwPortLen[vi]>0))
        release_region(pDevExt->dwPortAddr[vi], pDevExt->dwPortLen[vi]);
    }

    //free_irq
    if((pDevExt->Irq_no!=0)&&(pDevExt->openable==TRUE))
    free_irq(pDevExt->Irq_no, pDevExt);

    //release the buffer
    //for(vi=0; vi<DAS_FUNC_COUNT; vi++)
    //{
    //pDbx = pDevExt->pDbx[vi];
    //AdlDropDbx(pDbx);
    //}
    if(pDevExt->pDin)
    AdlDropDsen(pDevExt->pDin);
    if(pDevExt->pTcn)
    AdlDropDsen(pDevExt->pTcn);

    pci_set_drvdata(pdev, NULL);
#ifdef CONFIG_DEVFS_FS
#if :LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
    devfs_remove("pci7250W%d", pDevExt->wCard);
#else if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
    if(pDevExt->devfs_handle != NULL){
    devfs_unregister((devfs_handle_t)(pDevExt->devfs_handle));
    pDevExt->devfs_handle = NULL;
    }
#endif
#endif

    pci_disable_device(pdev);
    printk(KERN_ALERT "[PCI-7250] pci7250_remove: Removing pci7250 device: %d\n", pDevExt->wCard);
    device_destroy(p7250_class, MKDEV(MAJOR, pDevExt->wCard));
    device_number--;
}

int p7250_module_init(void)
{
    int rel_val;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
    EXPORT_NO_SYMBOLS;
#endif

    AdlInitModule(&drv_ext);

    rel_val = register_chrdev(MAJOR, DEVICE_NAME, &fops);
    if(rel_val<0)
    {
        printk("[PCI-7250] module_init: register_chrdev ");
        return rel_val;
    }
    MAJOR = rel_val;

    printk("[PCI-7250] module_init: register_chrdev successed\n");
    p7250_class = class_create(THIS_MODULE, "pci7250");
    if(IS_ERR(p7250_class))
    {
    printk("Error: class creation failed");
    #if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 23)
        rel_val = unregister_chrdev(MAJOR, DEVICE_NAME);
    if(rel_val<0)
        printk(KERN_ALERT "unregister_chrdev failed: %d\n", rel_val);
    else
    #else
    unregister_chrdev(MAJOR, DEVICE_NAME);
    #endif
    return -1;
    }
    return pci_module_init(&pci7250_driver);
}

/* Cleanup - unregister the appropriate file from /proc */
void p7250_module_exit(void)
{
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
    int ret;
#endif
    device_number = 0;

    pci_unregister_driver(&pci7250_driver);

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,23)
    ret = unregister_chrdev(MAJOR, DEVICE_NAME);
    if (ret <0)
    {
        printk(KERN_ALERT "[PCI-7250] Error in module_unregister_chrdev: %d\n", ret);
    }
    else
#else
    unregister_chrdev(MAJOR, DEVICE_NAME);
#endif
    printk ("[PCI-7250] module_exit: Unregistering the character device: %s \n", DEVICE_NAME );
    class_destroy(p7250_class);
}

module_init( p7250_module_init );
module_exit( p7250_module_exit );
