#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include "dask.h"
#include "Adlsys.h"
#include "dask_p.h"
#include "DAQHeader.h"


int ai_scale(U16 wCardNumber, U16 adRange, U16 reading, F64 *voltage);

int ai_scale32(U16 wCardNumber, U16 adRange, U32 reading, F64 *voltage);

int ai_scale32_9524(U16 wCardNumber, U16 adRange, U32 reading, F64 *voltage);

int ai_get_compensation_9524(U16 wCardNumber, U16 Group, U16 ADC_Gain,
    U16 ADC_Speed, F64 *res_offset, F64 *scaling);

int write_data_to_file(U16 *buf, FILE *fp, U16 byte_cnt, BOOLEAN exchange,
    U32 count);

void exch_data(U16 *tbuf, U32 cnt);


extern I16 write_header_to_file(U16 wCardNumber, FILE *fp, U16 data_width,
    U16 ad_range, U16 num_chan, U8 channel_no, U32 count,
    F64 fSampleRate, U16 num_of_channel_range);


extern REGISTER_CARD_TYPE RegisterCards[MAX_CARD];


I16 AI_9112_Config(
    U16 wCardNumber,
    U16 TrigSource
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    //Transfer Mode
    iotConf.mData.mAIO.wTrigSource = TrigSource;
    CurrentCard.AI_Status.CurrentConfig = iotConf;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }

    return NoError;
}


I16 AI_9111_Config(
    U16 wCardNumber,
    U16 TrigSource,
    U16 PreTrgEn,
    U16 TraceCnt
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber,0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    //Transfer Mode
    iotConf.mData.m9111AI.wTrigSource = TrigSource;
    iotConf.mData.m9111AI.wTraceEnable = PreTrgEn;
    iotConf.mData.m9111AI.wTraceCount = TraceCnt;
    //yuan modified 04/13/99
    //iotConf.mData.m9111AI.wEDO_Function = EDO_Fun;
    CurrentCard.AI_Status.CurrentConfig = iotConf;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }

    return NoError;
}


I16 AI_9113_Config(
    U16 wCardNumber,
    U16 TrigSource
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    iotConf.mData.m9113AI.wTrigSource = TrigSource;
    CurrentCard.AI_Status.CurrentConfig = iotConf;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }

    return NoError;
}


I16 AI_9114_Config(
    U16 wCardNumber,
    U16 TrigSource
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber,0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    iotConf.mData.m9114AI.wTrigSource = TrigSource;
    CurrentCard.AI_Status.CurrentConfig = iotConf;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }

    return NoError;
}


I16 AI_9114_PreTrigConfig(
    U16 wCardNumber,
    U16 PreTrgEn,
    U16 TraceCnt
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;
    U16 wTrigSource;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    //Transfer Mode
    //yuan 07/31/00
    wTrigSource = CurrentCard.AI_Status.CurrentConfig.mData.wValue[0];
    iotConf.mData.m9111AI.wTrigSource = 0x100;
    iotConf.mData.m9111AI.wTraceEnable = PreTrgEn;
    iotConf.mData.m9111AI.wTraceCount = TraceCnt;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }
    //yuan 07/31/00
    iotConf.mData.m9111AI.wTrigSource = wTrigSource;
    CurrentCard.AI_Status.CurrentConfig = iotConf;
    CurrentCard.AI_Status.CurrentConfig.wErrorCode = NoError;

    return NoError;
}


I16 AI_9116_Config(
    U16 wCardNumber,
    U16 ConfigCtrl,
    U16 TrigCtrl,
    U16 PostCnt,
    U16 MCnt,
    U16 ReTrgCnt
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    iotConf.mData.wValue[0] = ((ConfigCtrl|TrigCtrl)&0x3fff);
    iotConf.mData.m9116AI.wPostTriggerCounter = PostCnt;
    iotConf.mData.m9116AI.wMCounter = MCnt;
    iotConf.mData.m9116AI.wRetrgCnt = ReTrgCnt;
    //yuan 07/31/00
    CurrentCard.AI_Status.CurrentConfig = iotConf;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }

    return NoError;
}


I16 AI_9116_CounterInterval(
    U16 wCardNumber,
    U32 ScanIntrv,
    U32 SampIntrv
    )
{
    I16 err;
    DAS_IOT_SI iotSI;

    memset(&iotSI, '\0', sizeof(DAS_IOT_SI));

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    //if(ScanIntrv<240||ScanIntrv<240||)
    iotSI.dwScanInterval = ScanIntrv;
    iotSI.dwSampInterval = SampIntrv;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_9116SC_CONFIG, &(iotSI))
        !=DAS_ERR_NO)
    {
        return iotSI.wErrorCode;
    }
    //yuan add for scan rate
    CurrentCard.AI_Status.dwPacerValue = ScanIntrv;
    //yuan add 04/24/01 for channel count
    CurrentCard.AI_Status.wNumChans = (U16)(ScanIntrv/SampIntrv);

    return NoError;
}


I16 AI_9118_Config(
    U16 wCardNumber,
    U16 ModeCtrl,
    U16 FunCtrl,
    U16 BurstCnt,
    U16 PostCnt
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber,0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    iotConf.mData.wValue[0] = ((ModeCtrl|FunCtrl)&0x3ff);
    iotConf.mData.m9118AI.wBurstConversionCounter = BurstCnt;
    iotConf.mData.m9118AI.wPostTriggerCounter = PostCnt;
    //yuan add 06/12/03
    if(CurrentCard.wCardType==PCI_9118HR)
        iotConf.mData.m9118AI.fUnused = 1;
    CurrentCard.AI_Status.CurrentConfig = iotConf;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }

    return NoError;
}


I16 AI_9812_Config(
    U16 wCardNumber,
    U16 TrgMode,
    U16 TrgSrc,
    U16 TrgPol,
    U16 ClkSel,
    U16 TrgLevel,
    U16 PostCnt
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    iotConf.mData.wValue[0] = ((TrgMode|TrgSrc|TrgPol|ClkSel)&0x3ff);
    iotConf.mData.m9812AI.wTriggerLevel = TrgLevel;
    iotConf.mData.m9812AI.wPostTriggerCount = PostCnt;
    CurrentCard.AI_Status.CurrentConfig = iotConf;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }

    //01/10/08
    if(TrgSrc<P9812_TRGSRC_EXT_DIG)
        usleep(1);

    return NoError;
}


//yuan 09/13/00 for 9812
I16 AI_9812_SetDiv(
    U16 wCardNumber,
    U32 PacerVal
    )
{
    I16 err;

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }
    CurrentCard.AI_Status.dwPacerValue = PacerVal;

    return NoError;
}


I16 AI_AsyncCheck(
    U16     wCardNumber,
    BOOLEAN *bStopped,
    U32     *dwAccessCnt
    )
{
    I16 err;
    U32 tmpCount;
    U16 turn_flag;
    DAS_IOT_ASYNC iotAsync;

    memset(&iotAsync, '\0', sizeof(DAS_IOT_ASYNC));

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!= NoError)
    {
        return err;
    }

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_ASY_CHECK, &(iotAsync))
        !=DAS_ERR_NO)
    {
        return iotAsync.wErrorCode;
    }
    *bStopped = (BOOLEAN)iotAsync.wStopped;

    //01/25/08
    if((CurrentCard.wCardType==PCI_9221)||(CurrentCard.wCardType==PCI_9524)||
        (CurrentCard.wCardType==PCI_9222)||(CurrentCard.wCardType==PCI_9223)||
        (CurrentCard.wCardType==PCI_922A))
    {
        *dwAccessCnt = iotAsync.dwAccessCount;
    }
    else{
        tmpCount = CurrentCard_AIFunc.dwBufferSize/
            (U32)((CurrentCard_AIFunc.wResoluation+7)/8);
        tmpCount = tmpCount/2;
        turn_flag = (U16)((iotAsync.dwAccessCount&0x80000000)>>31);
        if(!CurrentCard.AI_Status.bDblBufferMode){
            if(turn_flag)
                *dwAccessCnt = CurrentCard.AI_Status.dwTransferCount;
            else
                *dwAccessCnt = iotAsync.dwAccessCount&0x0fffffff;
        }
        else{
            if((iotAsync.dwAccessCount&0x0fffffff)<tmpCount)
                *dwAccessCnt = iotAsync.dwAccessCount&0x0fffffff;
            else if((iotAsync.dwAccessCount&0x0fffffff)==tmpCount)
                *dwAccessCnt = CurrentCard.AI_Status.dwTransferCount;
            else
                *dwAccessCnt = (iotAsync.dwAccessCount&0x0fffffff)-tmpCount;
        }
    }

    return NoError;
}


I16 AI_AsyncClear(
    U16 wCardNumber,
    U32 *dwAccessCnt
    )
{
    I16 err;
    U32 tmpCount, cur_pos, ch_scan, data_res, BufId;
    U16 turn_flag;
    U8 *tBuf1, *tBuf2;
    U32 i, n;
    U16 *p, *q;
    BOOLEAN HalfReady, fstop;
    U16 RdyTrigCnt;
    FILE *fp;
    DAS_IOT_ASYNC iotAsync;
    DAS_IOT_CONFIG iotConf;

    memset(&iotAsync, '\0', sizeof(DAS_IOT_ASYNC));
    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_ASY_CLEAR, &(iotAsync))
        !=DAS_ERR_NO)
    {
        return iotAsync.wErrorCode;
    }

    if((CurrentCard.wCardType==PCI_9221)||(CurrentCard.wCardType==PCI_9524)||
        (CurrentCard.wCardType==PCI_9222)||(CurrentCard.wCardType==PCI_9223)||
        (CurrentCard.wCardType==PCI_922A))
    {
        iotConf = CurrentCard.AI_Status.CurrentConfig;
        BufId = CurrentCard.AI_Status.pDAQbufHandle;
        *dwAccessCnt = iotAsync.dwAccessCount;
        if(!CurrentCard.AI_Status.bDblBufferMode){
            /*SingleBuffer*/
            if(!CurrentCard.AI_Status.bToFile){
                /*Not to File*/
                //return NoError;
            }
            else{
                /*ToFile*/
                //07/29/08
                /*Infinite Trigger*/
                if(((CurrentCard.wCardType==PCI_9222)||
                    (CurrentCard.wCardType==PCI_9223)||
                    (CurrentCard.wCardType==PCI_922A))&&
                    (iotConf.mData.m922xAI.fEnReTrig)&&
                    (iotConf.mData.m922xAI.ReTriggerCnt==0)&&
                    (CurrentCard.AI_Status.pDAQbuf[1]))
                {
                    /*Get the index of working buffer of Double Buffer*/
                    //AI_AsyncDblBufferHalfReady(wCardNumber, &HalfReady, &fstop);
                    AI_AsyncReTrigNextReady(wCardNumber, &HalfReady, &fstop,
                        &RdyTrigCnt
                    );
                    if(HalfReady){
                        AI_AsyncDblBufferToFile(wCardNumber);  //transfer to file
                        CurrentCard.AI_Status.wDblBufferIndex = \
                            (CurrentCard.AI_Status.wDblBufferIndex==0)? 1:0;
                    }
                    if((*dwAccessCnt!=CurrentCard.AI_Status.dwTransferCount)&&
                        (*dwAccessCnt!=0))
                    {
                        write_data_to_file(CurrentCard.AI_Status.pDAQbuf[\
                            CurrentCard.AI_Status.wDblBufferIndex],
                            CurrentCard.AI_Status.pwBuffer,
                            (U16)((CurrentCard_AIFunc.wResoluation+7)/8),
                            FALSE, (*dwAccessCnt)
                        );
                    }
                }
                else{
                    write_data_to_file(
                        (U16 *)CurrentCard.AI_Status.pDAQbuf[0],
                        CurrentCard.AI_Status.pwBuffer,
                        (U16)((CurrentCard_AIFunc.wResoluation+7)/8), FALSE,
                        *dwAccessCnt
                    );
                }
                fp = (FILE *)(CurrentCard.AI_Status.pwBuffer);
                data_res = (U16)((CurrentCard_AIFunc.wResoluation+7)/8);
                cur_pos = ftell(fp);
                if(!CurrentCard.AI_Status.bChanRange)
                    ch_scan = (cur_pos-60)/data_res;
                else{
                    ch_scan = \
                        (cur_pos-60-CurrentCard.AI_Status.wNumChans*
                        sizeof(DAQChannelRange))/data_res;
                }
                //12/27/07
                if(CurrentCard.wCardType==PCI_9524){
                    /*add 2 double(4*32bit(Data Res)) for one channel*/
                    ch_scan -= 4*CurrentCard.AI_Status.wNumChans;
                }
                ch_scan = ch_scan/CurrentCard.AI_Status.wNumChans;
                fseek(fp, 15, SEEK_SET);
                fwrite(&ch_scan, sizeof(long), 1, fp);
                fseek(fp, 0, SEEK_END);

                fclose((FILE *)CurrentCard.AI_Status.pwBuffer);
            }

            //06/05/08
            if(CurrentCard.wCardType==PCI_9221){
                if(iotConf.mData.m9221AI.fAutoResetBuf){
                    CurrentCard.AI_Status.pDAQbuf[BufId] = NULL;
                    CurrentCard.AI_Status.dwBufCount[BufId] = 0;
                }
            }
            else if(CurrentCard.wCardType==PCI_9524){
                if((U32)(CurrentCard.AI_Status.p7300buf)==0){
                    /*Load cell group*/
                    if(iotConf.mData.m9524AI.fAutoResetBuf){
                        CurrentCard.AI_Status.pDAQbuf[BufId] = NULL;
                        CurrentCard.AI_Status.dwBufCount[BufId] = 0;
                    }
                }
                else if((U32)(CurrentCard.AI_Status.p7300buf)==1){
                    /*General purpose group*/
                    iotConf.mData.dwValue[0] = \
                        CurrentCard.AI_Status.dwPacerValue;
                    iotConf.mData.dwValue[1] = \
                        CurrentCard.AI_Status.dwSampleIntv;
                    if(iotConf.mData.m9524AI.fAutoResetBuf){
                        CurrentCard.AI_Status.pDAQbuf[BufId] = NULL;
                        CurrentCard.AI_Status.dwBufCount[BufId] = 0;
                    }
                }
            }
            else if((CurrentCard.wCardType==PCI_9222)||
                (CurrentCard.wCardType==PCI_9223)||
                (CurrentCard.wCardType==PCI_922A))
            {
                if(iotConf.mData.m922xAI.fAutoResetBuf){
                    CurrentCard.AI_Status.pDAQbuf[BufId] = NULL;
                    CurrentCard.AI_Status.dwBufCount[BufId] = 0;
                }
            }
        }
        else{
            /*DblBufferMode*/
            if(!CurrentCard.AI_Status.bToFile){
                /*Get the index of working buffer of Double Buffer*/
                AI_AsyncDblBufferHalfReady(wCardNumber, &HalfReady, &fstop);
                if(HalfReady){
                    *dwAccessCnt += CurrentCard.AI_Status.dwTransferCount;
                    //07/08/08 Removed, it is not necessary to mark Buffer Ready
                    //CurrentCard.AI_Status.wDblBufferIndex |= 0x02;
                }
                //return NoError;
            }
            else{
                /*write data to file*/
                /*Get the index of working buffer of Double Buffer*/
                AI_AsyncDblBufferHalfReady(wCardNumber, &HalfReady, &fstop);
                if(HalfReady){
                    AI_AsyncDblBufferToFile(wCardNumber);
                    CurrentCard.AI_Status.wDblBufferIndex = \
                        (CurrentCard.AI_Status.wDblBufferIndex==0)? 1:0;
                }

                if((*dwAccessCnt!=CurrentCard.AI_Status.dwTransferCount)&&
                    (*dwAccessCnt!=0))
                {
                    U16 BufIdx = CurrentCard.AI_Status.wDblBufferIndex;
                    write_data_to_file(
                        CurrentCard.AI_Status.pDAQbuf[BufIdx],
                        CurrentCard.AI_Status.pwBuffer,
                        (U16)((CurrentCard_AIFunc.wResoluation+7)/8),
                        FALSE, (*dwAccessCnt)
                    );
                }
                fp = (FILE *)CurrentCard.AI_Status.pwBuffer;
                data_res = (U16)((CurrentCard_AIFunc.wResoluation+7)/8);
                cur_pos = ftell(fp);
                if(!CurrentCard.AI_Status.bChanRange)
                    ch_scan = (cur_pos-60)/data_res;
                else{
                    ch_scan = \
                        (cur_pos-60-CurrentCard.AI_Status.wNumChans*
                        sizeof(DAQChannelRange))/data_res;
                }
                //12/27/07
                if(CurrentCard.wCardType==PCI_9524)
                    ch_scan -= 4*CurrentCard.AI_Status.wNumChans;
                ch_scan = ch_scan/CurrentCard.AI_Status.wNumChans;
                fseek(fp, 15, SEEK_SET);
                fwrite(&ch_scan, sizeof(long), 1, fp);
                fseek(fp, 0, SEEK_END);
                fclose((FILE *)CurrentCard.AI_Status.pwBuffer);
            }

            //06/05/08
            if(CurrentCard.wCardType==PCI_9221){
                if(iotConf.mData.m9221AI.fAutoResetBuf){
                    CurrentCard.AI_Status.pDAQbuf[0] = NULL;
                    CurrentCard.AI_Status.pDAQbuf[1] = NULL;
                    CurrentCard.AI_Status.dwBufCount[0] = 0;
                    CurrentCard.AI_Status.dwBufCount[1] = 0;
                }
            }
            else if(CurrentCard.wCardType==PCI_9524){
                if((U32)(CurrentCard.AI_Status.p7300buf)==0){
                    /*Load cell group*/
                    if(iotConf.mData.m9524AI.fAutoResetBuf){
                        CurrentCard.AI_Status.pDAQbuf[0] = NULL;
                        CurrentCard.AI_Status.pDAQbuf[1] = NULL;
                        CurrentCard.AI_Status.dwBufCount[0] = 0;
                        CurrentCard.AI_Status.dwBufCount[1] = 0;
                    }
                }
                else if((U32)(CurrentCard.AI_Status.p7300buf)==1){
                    /*General purpose group*/
                    iotConf.mData.dwValue[0] = \
                        CurrentCard.AI_Status.dwPacerValue;
                    iotConf.mData.dwValue[1] = \
                        CurrentCard.AI_Status.dwSampleIntv;
                    if(iotConf.mData.m9524AI.fAutoResetBuf){
                        CurrentCard.AI_Status.pDAQbuf[0] = NULL;
                        CurrentCard.AI_Status.pDAQbuf[1] = NULL;
                        CurrentCard.AI_Status.dwBufCount[0] = 0;
                        CurrentCard.AI_Status.dwBufCount[1] = 0;
                    }
                }
            }
            else if((CurrentCard.wCardType==PCI_9222)||
                (CurrentCard.wCardType==PCI_9223)||
                (CurrentCard.wCardType==PCI_922A))
            {
                if(iotConf.mData.m922xAI.fAutoResetBuf){
                    CurrentCard.AI_Status.pDAQbuf[0] = NULL;
                    CurrentCard.AI_Status.pDAQbuf[1] = NULL;
                    CurrentCard.AI_Status.dwBufCount[0] = 0;
                    CurrentCard.AI_Status.dwBufCount[1] = 0;
                }
            }
        }
    }
    else{
        tmpCount = \
            CurrentCard_AIFunc.dwBufferSize/
            (U32)((CurrentCard_AIFunc.wResoluation+7)/8);
        tmpCount = tmpCount/2;
        turn_flag = (U16)((iotAsync.dwAccessCount&0x80000000)>>31);
        tBuf1 = CurrentCard.AI_Status.pwBuffer;
        tBuf2 = CurrentCard_AIFunc.pView[0];
        if(!CurrentCard.AI_Status.bDblBufferMode){
            *dwAccessCnt = iotAsync.dwAccessCount&0x0fffffff;
            if(!CurrentCard.AI_Status.bToFile){
                /*Copy the AI data from internal buffer to user buffer*/
                if(!turn_flag){
                    if(CurrentCard.AI_Status.pwBuffer){
                        if((CurrentCard.wCardType!=PCI_9118DG)&&
                            (CurrentCard.wCardType!=PCI_9118HG)&&
                            (CurrentCard.wCardType!=PCI_9118HR))
                        {
                            memcpy(CurrentCard.AI_Status.pwBuffer,
                                CurrentCard_AIFunc.pView[0],
                                CurrentCard.AI_Status.dwTransferCount*
                                ((U32)((CurrentCard_AIFunc.wResoluation+7)/8))
                            );
                        }
                        else{
                            //9118
                            p = (U16 *)CurrentCard_AIFunc.pView[0];
                            q = (U16 *)CurrentCard.AI_Status.pwBuffer;
                            for(i=0; i<CurrentCard.AI_Status.dwTransferCount;
                                i++)
                            {
                                *(q+i) = ((i%2)? (*(p+i-1)):(*(p+i+1)));
                            }
                        }
                    }
                }
                else{
                    if(CurrentCard.AI_Status.pwBuffer){
                        if((CurrentCard.wCardType!=PCI_9118DG)&&
                            (CurrentCard.wCardType!=PCI_9118HG)&&
                            (CurrentCard.wCardType!=PCI_9118HR))
                        {
                            memcpy(tBuf1, tBuf2+(*dwAccessCnt)*
                                (U32)((CurrentCard_AIFunc.wResoluation+7)/8),
                                (CurrentCard.AI_Status.dwTransferCount-
                                (*dwAccessCnt))*
                                (U32)((CurrentCard_AIFunc.wResoluation+7)/8)
                            );
                            memcpy(tBuf1+(CurrentCard.AI_Status.dwTransferCount-
                                (*dwAccessCnt))*
                                ((U32)((CurrentCard_AIFunc.wResoluation+7)/8)),
                                tBuf2, (*dwAccessCnt)*
                                ((U32)((CurrentCard_AIFunc.wResoluation+7)/8))
                            );
                        }
                        else{
                            p = (U16 *)tBuf2+(*dwAccessCnt);
                            q = (U16 *)tBuf1;
                            n = CurrentCard.AI_Status.dwTransferCount-
                                (*dwAccessCnt);
                            for(i=0; i<n; i++){
                                *(q+i) = ((i%2)? (*(p+i-1)):(*(p+i+1)));
                            }
                            p = (U16 *)tBuf2;
                            for(i=n; i<CurrentCard.AI_Status.dwTransferCount;
                                i++)
                            {
                                *(q+i) = ((i%2)? (*(p+i-n-1)):(*(p+i-n+1)));
                            }
                        }
                    }
                    *dwAccessCnt = CurrentCard.AI_Status.dwTransferCount;
                }
            }
            else{
                /*write data to file*/
                if(!turn_flag){
                    if((CurrentCard.wCardType!=PCI_9118DG)&&
                        (CurrentCard.wCardType!=PCI_9118HG)&&
                        (CurrentCard.wCardType!=PCI_9118HR))
                    {
                        write_data_to_file(
                            (U16 *)CurrentCard_AIFunc.pView[0],
                            (FILE *)CurrentCard.AI_Status.pwBuffer,
                            (U16)((CurrentCard_AIFunc.wResoluation+7)/8),
                            FALSE,
                            CurrentCard.AI_Status.dwTransferCount
                        );
                    }
                    else{
                        //9118
                        write_data_to_file(
                            (U16 *)CurrentCard_AIFunc.pView[0],
                            (FILE *)CurrentCard.AI_Status.pwBuffer,
                            (U16)((CurrentCard_AIFunc.wResoluation+7)/8),
                            TRUE,
                            CurrentCard.AI_Status.dwTransferCount
                        );
                    }
                }
                else{
                    if((CurrentCard.wCardType!=PCI_9118DG)&&
                        (CurrentCard.wCardType!=PCI_9118HG)&&
                        (CurrentCard.wCardType!=PCI_9118HR))
                    {
                        write_data_to_file(
                            (U16 *)(tBuf2+(*dwAccessCnt)*
                            (U32)((CurrentCard_AIFunc.wResoluation+7)/8)),
                            (FILE *)CurrentCard.AI_Status.pwBuffer,
                            (U16)((CurrentCard_AIFunc.wResoluation+7)/8),
                            FALSE,
                            CurrentCard.AI_Status.dwTransferCount-(*dwAccessCnt)
                        );
                        write_data_to_file(
                            (U16 *)tBuf2,
                            (FILE *)CurrentCard.AI_Status.pwBuffer,
                            (U16)((CurrentCard_AIFunc.wResoluation+7)/8),
                            FALSE,
                            (*dwAccessCnt)
                        );
                    }
                    else{
                        //9118
                        write_data_to_file(
                            (U16 *)(tBuf2+(*dwAccessCnt)*
                            (U32)((CurrentCard_AIFunc.wResoluation+7)/8)),
                            (FILE *)CurrentCard.AI_Status.pwBuffer,
                            (U16)((CurrentCard_AIFunc.wResoluation+7)/8),
                            TRUE,
                            CurrentCard.AI_Status.dwTransferCount-(*dwAccessCnt)
                        );
                        write_data_to_file(
                            (U16 *)tBuf2,
                            (FILE *)CurrentCard.AI_Status.pwBuffer,
                            (U16)((CurrentCard_AIFunc.wResoluation+7)/8),
                            TRUE,
                            (*dwAccessCnt)
                        );
                    }
                    *dwAccessCnt = CurrentCard.AI_Status.dwTransferCount;
                }
                fclose((FILE *)CurrentCard.AI_Status.pwBuffer);
            }
        }
        else{
            /*double-buffer mode*/
            if((iotAsync.dwAccessCount&0x0fffffff)<tmpCount)
                *dwAccessCnt = iotAsync.dwAccessCount&0x0fffffff;
            else
                *dwAccessCnt = (iotAsync.dwAccessCount&0x0fffffff)-tmpCount;
            //yuan add 05/31/01
            if(!CurrentCard.AI_Status.bToFile){
                /*Get the index of working buffer of Double Buffer*/
                AI_AsyncDblBufferHalfReady(wCardNumber, &HalfReady, &fstop);
                if(HalfReady){
                    *dwAccessCnt += CurrentCard.AI_Status.dwTransferCount;
                    CurrentCard.AI_Status.wDblBufferIndex |= 0x02;
                }
            }
            else{
                /*Get the index of working buffer of Double Buffer*/
                AI_AsyncDblBufferHalfReady(wCardNumber, &HalfReady, &fstop);
                //AI_AsyncClear(card, &count);
                if(HalfReady){
                    /*transfer to file*/
                    AI_AsyncDblBufferTransfer(wCardNumber, NULL);
                    CurrentCard.AI_Status.wDblBufferIndex = \
                        (CurrentCard.AI_Status.wDblBufferIndex==0)? 1:0;
                }
                if((*dwAccessCnt!=CurrentCard.AI_Status.dwTransferCount)&&
                    (*dwAccessCnt!=0))
                {
                    U16 *ptmpAddr;
                    U16 wtmpBufIdx;

                    wtmpBufIdx = CurrentCard.AI_Status.wDblBufferIndex;
                    ptmpAddr = (U16 *)CurrentCard_AIFunc.pView[wtmpBufIdx];

                    if((CurrentCard.wCardType!=PCI_9118DG)&&
                        (CurrentCard.wCardType!=PCI_9118HG)&&
                        (CurrentCard.wCardType!=PCI_9118HR))
                    {
                        write_data_to_file(
                            ptmpAddr,
                            (FILE *)CurrentCard.AI_Status.pwBuffer,
                            (U16)((CurrentCard_AIFunc.wResoluation+7)/8),
                            FALSE,
                            (*dwAccessCnt)
                        );
                    }
                    else{
                        //9118
                        write_data_to_file(
                            ptmpAddr,
                            (FILE *)CurrentCard.AI_Status.pwBuffer,
                            (U16)((CurrentCard_AIFunc.wResoluation+7)/8),
                            TRUE,
                            (*dwAccessCnt)
                        );
                    }
                }
                fp = (FILE *)CurrentCard.AI_Status.pwBuffer;
                data_res = (U16)((CurrentCard_AIFunc.wResoluation+7)/8);
                cur_pos = ftell(fp);
                if(!CurrentCard.AI_Status.bChanRange)
                    ch_scan = (cur_pos-60)/data_res;
                else{
                    ch_scan = \
                        (cur_pos-60-CurrentCard.AI_Status.wNumChans*
                        sizeof(DAQChannelRange))/data_res;
                }
                ch_scan = ch_scan/CurrentCard.AI_Status.wNumChans;
                fseek(fp, 15, SEEK_SET);
                fwrite(&ch_scan, sizeof(long), 1, fp);
                fseek(fp, 0, SEEK_END);
                fclose((FILE *)CurrentCard.AI_Status.pwBuffer);
            }
        }
    }

    return NoError;
}


I16 AI_AsyncDblBufferHalfReady(
    U16     wCardNumber,
    BOOLEAN *bHalfReady,
    BOOLEAN *StopFlag
    )
{
    I16 err;
    U32 dwAccessCnt;
    BOOLEAN fstop;
    DAS_IOT_DBF iotDbf;

    memset(&iotDbf, '\0', sizeof(DAS_IOT_DBF));

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_DBF_HALF_READY, &(iotDbf))
        !=DAS_ERR_NO)
    {
        return iotDbf.wErrorCode;
    }
    *bHalfReady = (BOOLEAN)iotDbf.wReady;
    CurrentCard.AI_Status.wDblBufferIndex = iotDbf.wViewIndex;
    AI_AsyncCheck(wCardNumber, &fstop, &dwAccessCnt);
    *StopFlag = fstop;

    return NoError;
}


I16 AI_AsyncDblBufferHalfReadyIndex(
    U16     wCardNumber,
    BOOLEAN *bHalfReady,
    BOOLEAN *StopFlag,
    int     *pViewIndex
    )
{
    I16 err;
    U32 dwAccessCnt;
    BOOLEAN fstop;
    DAS_IOT_DBF iotDbf;

    memset(&iotDbf, '\0', sizeof(DAS_IOT_DBF));

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_DBF_HALF_READY, &(iotDbf))
        !=DAS_ERR_NO)
    {
        return iotDbf.wErrorCode;
    }

    *bHalfReady = (BOOLEAN)iotDbf.wReady;
    CurrentCard.AI_Status.wDblBufferIndex = iotDbf.wViewIndex;
    AI_AsyncCheck(wCardNumber, &fstop, &dwAccessCnt);
    *StopFlag = fstop;
    *pViewIndex = iotDbf.wViewIndex;

    return NoError;
}


I16 AI_AsyncDblBufferMode(
    U16     wCardNumber,
    BOOLEAN bEnable
    )
{
    I16 err;
    DAS_IOT_DBF iotDbf;

    memset(&iotDbf, '\0', sizeof(DAS_IOT_DBF));

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    iotDbf.wEnable = bEnable;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_DBF_MODE, &(iotDbf))
        !=DAS_ERR_NO)
    {
        return iotDbf.wErrorCode;
    }
    CurrentCard.AI_Status.bDblBufferMode = bEnable;

    return NoError;
}


I16 AI_AsyncDblBufferTransfer(
    U16 wCardNumber,
    U16 *pwBuffer
    )
{
    I16 err;
    U32 i;
    U16 *p;
    U16 DblBufId;
    U16 wBufId;
    volatile PPCI_DBXORUN pOvRun;

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }
    if(!CurrentCard.AI_Status.bDblBufferMode)
        return ErrorNotDoubleBufferMode;

    DblBufId = CurrentCard.AI_Status.wDblBufferIndex;
    CurrentCard.AI_Status.wDblBufferIndex &= 0x01;
    wBufId = CurrentCard.AI_Status.wDblBufferIndex;
    if(!CurrentCard.AI_Status.bToFile){
        /*to user buffer*/
        if((CurrentCard.wCardType!=PCI_9118DG)&&
            (CurrentCard.wCardType!=PCI_9118HG)&&
            (CurrentCard.wCardType!=PCI_9118HR))
        {
            memcpy(
                pwBuffer,
                CurrentCard_AIFunc.pView[wBufId],
                CurrentCard.AI_Status.dwTransferCount*
                (U16)((CurrentCard_AIFunc.wResoluation+7)/8)
            );
        }
        else{
            //9118
            p = (U16 *)CurrentCard_AIFunc.pView[wBufId];
            for(i=0; i<CurrentCard.AI_Status.dwTransferCount; i++){
                pwBuffer[i] = ((i%2)? (*(p+i-1)):(*(p+i+1)));
            }
        }
        //yuan add 07/25/01
        if((DblBufId>>1)&0x01){
            CurrentCard.AI_Status.wDblBufferIndex = \
                (CurrentCard.AI_Status.wDblBufferIndex==0)? 1:0;
            wBufId = CurrentCard.AI_Status.wDblBufferIndex;
        }
    }
    else{
        /*to file*/
        if((CurrentCard.wCardType!=PCI_9118DG)&&
            (CurrentCard.wCardType!=PCI_9118HG)&&
            (CurrentCard.wCardType!=PCI_9118HR))
        {
            write_data_to_file(
                (U16 *)CurrentCard_AIFunc.pView[wBufId],
                (FILE *)CurrentCard.AI_Status.pwBuffer,
                (U16)((CurrentCard_AIFunc.wResoluation+7)/8),
                FALSE,
                CurrentCard.AI_Status.dwTransferCount
            );
        }
        else{
            //9118
            write_data_to_file(
                (U16 *)CurrentCard_AIFunc.pView[wBufId],
                (FILE *)CurrentCard.AI_Status.pwBuffer,
                (U16)((CurrentCard_AIFunc.wResoluation+7)/8),
                TRUE,
                CurrentCard.AI_Status.dwTransferCount
            );
        }
    }
    //yuan add 09/11/02
    if(((CurrentCard.wCardType>=PCI_9111DG)&&
        (CurrentCard.wCardType<=PCI_9812))||
        (CurrentCard.wCardType==PCI_9116))
    {
        pOvRun = (volatile PPCI_DBXORUN)(CurrentCard_AOFunc.pView[0]);
        if(pOvRun){
            if(!pOvRun->xFlag.f.overrun){
                pOvRun->xFlag.f.lock = 1;
                if(pOvRun->dwCtr_isr[wBufId]>=1)
                    pOvRun->dwCtr_isr[wBufId]-=1;
                pOvRun->xFlag.f.lock = 0;
            }
        }
    }

    return NoError;
}


/*pwBuffer is of no use while double-buffer mode enabled*/
I16 AI_ContReadChannel(
    U16 wCardNumber,
    U16 wChannel,
    U16 wAdRange,
    U16 *pwBuffer,
    U32 dwReadCount,
    F64 fSampleRate,
    U16 wSyncMode
    )
{
    I16 err;
    U16 wGain;
    U32 i;
    U16 *p;
    DAS_IOT_CONFIG iotConf;
    DAS_IOT_CONT iotCont;
    DAS_IOT_SG_CONT iotSCont; //for PCI9221

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));
    memset(&iotCont, '\0', sizeof(DAS_IOT_CONT));
    memset(&iotSCont, '\0', sizeof(DAS_IOT_SG_CONT));

    //00/02/03 jeffrey, re-asign the iotConf value
    iotConf = CurrentCard.AI_Status.CurrentConfig;

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 1, &wChannel,
        dwReadCount*(U32)((CurrentCard_AIFunc.wResoluation+7)/8),
        TRUE))!=NoError)
    {
        return err;
    }

    //Check wAdRange and get Gain
    if(!validate_ad_range(wCardNumber, wAdRange, &wGain))
        return ErrorInvalidAdRange;

    /*Scatter-gather*/
    if((CurrentCard.wCardType==PCI_9221)||(CurrentCard.wCardType==PCI_9524)||
        (CurrentCard.wCardType==PCI_9222)||(CurrentCard.wCardType==PCI_9223)||
        (CurrentCard.wCardType==PCI_922A))
    {
        /*Set*/
        if(!CurrentCard.AI_Status.bDblBufferMode){
            if((*pwBuffer>1)||(!CurrentCard.AI_Status.pDAQbuf[*pwBuffer]))
                return ErrorInvalidBufferID;
            CurrentCard.AI_Status.pDAQbufHandle = *pwBuffer;
        }
        else{
            if((*pwBuffer>1)||(!CurrentCard.AI_Status.pDAQbuf[0])||
                (!CurrentCard.AI_Status.pDAQbuf[1]))
            {
                return ErrorInvalidBufferID;
            }
            CurrentCard.AI_Status.pDAQbufHandle = 0;
        }
        CurrentCard.AI_Status.bToFile = FALSE;
        CurrentCard.AI_Status.wNumChans = 1;
        CurrentCard.AI_Status.dwTransferCount = dwReadCount;
        /*Set*/
        switch(CurrentCard.wCardType){
            case PCI_9221:
                if((CurrentCard.AI_Status.dwPacerValue==0)||
                    (CurrentCard.AI_Status.dwSampleIntv==0))
                {
                    return ErrorInvalidCounterState;
                }
                else{
                    iotSCont.dwPacerValue[1] = \
                        CurrentCard.AI_Status.dwSampleIntv;
                    iotSCont.dwPacerValue[0] = \
                        CurrentCard.AI_Status.dwPacerValue;
                }
                break;
            case PCI_9222:
            case PCI_9223:
            case PCI_922A:
                if(((CurrentCard.AI_Status.bDblBufferMode)||
                    (wSyncMode==SYNCH_OP))&&
                    (iotConf.mData.m922xAI.fEnReTrig))
                {
                    return ErrorInvalidTriggerMode;
                }
                //06/18/08
                if(iotConf.mData.m922xAI.fConvSrc==0){
                    if((CurrentCard.AI_Status.dwPacerValue==0)||
                        (CurrentCard.AI_Status.dwSampleIntv==0))
                    {
                        return ErrorInvalidCounterState;
                    }
                    else{
                        iotSCont.dwPacerValue[1] = \
                            CurrentCard.AI_Status.dwSampleIntv;
                        iotSCont.dwPacerValue[0] = \
                            CurrentCard.AI_Status.dwPacerValue;
                    }
                }
                break;
            case PCI_9524:
                if(wChannel<4){ /*load cell group channel 0*/
                    iotConf.mData.m9524AI.fGroup = 0;
                    if(wGain)
                        return ErrorInvalidAdRange;
                    //06/05/08
                    CurrentCard.AI_Status.p7300buf = (void *)0;
                }
                else if(wChannel<8){ /*general purpose group channel 0*/
                    iotConf.mData.dwValue[0] = \
                        CurrentCard.AI_Status.dwPacerValue;
                    iotConf.mData.dwValue[1] = \
                        CurrentCard.AI_Status.dwSampleIntv;
                    iotConf.mData.m9524AI.fGroup = 1;
                    if((!wGain)||(wGain>4))
                        return ErrorInvalidAdRange;
                    //06/05/08
                    CurrentCard.AI_Status.p7300buf = (void *)1;
                }
                else
                    return ErrorInvalidIoChannel;

                if(iotConf.mData.m9524AI.fXMode!=1)
                    return ErrorConfigFailed;

                if(!get_pacervalue(wCardNumber, fSampleRate,
                    &iotSCont.dwPacerValue[0]))
                {
                    return ErrorInvalidSampleRate;
                }

                iotConf.mData.m9524AI.fAutoScan = 0;
                iotConf.mData.m9524AI.ADC_Channel = wChannel;
                iotConf.mData.m9524AI.ADC_Gain = wGain;
                iotConf.mData.m9524AI.ADC_Speed = \
                    (U8)(iotSCont.dwPacerValue[0]&0xff);
                iotConf.mData.m9524AI.Reserved1 = 1;

                if(iotConf.mData.m9524AI.fGroup==0){
                    CurrentCard.AI_Status.CurrentConfig = iotConf;
                }
                else{
                    /*use the two valuables to store the config of General Purpose group*/
                    CurrentCard.AI_Status.dwPacerValue = \
                        iotConf.mData.dwValue[0];
                    CurrentCard.AI_Status.dwSampleIntv = \
                        iotConf.mData.dwValue[1];
                }
                break;
        }
        iotSCont.wAutoScan = 0;
        iotSCont.wSyncMode = wSyncMode;
        iotSCont.wChannelCount = 1;
        iotSCont.dwAccessCount = dwReadCount;
        //06/12/08
        if((CurrentCard.wCardType==PCI_9222)&&(wChannel>=8))
            iotSCont.dwChannelGain[0].wChannel = wChannel+8;
        else
            iotSCont.dwChannelGain[0].wChannel = wChannel;
        iotSCont.dwChannelGain[0].wGain = wGain;
        iotSCont.dwBufferAddr = *pwBuffer;

        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CON_READ_SG, &(iotSCont))
            !=DAS_ERR_NO)
        {
            return iotSCont.wErrorCode;
        }
    }
    else{
        if(CurrentCard.AI_Status.bDblBufferMode)
            dwReadCount /= 2;
        CurrentCard.AI_Status.bToFile = FALSE;
        CurrentCard.AI_Status.dwTransferCount = dwReadCount;
        CurrentCard.AI_Status.pwBuffer = pwBuffer;

        //Set PacerValue and TimerCascade
        if(((CurrentCard.wCardType==PCI_9812)||(CurrentCard.wCardType==PCI_9810))
            &&((((iotConf.mData.wValue[0])&0x0300)==P9812_CLKSRC_EXT_SIN)||
            (((iotConf.mData.wValue[0])&0x0300)==P9812_CLKSRC_EXT_DIG)))
        {
            //iotCont.dwPacerValue = 2;
            iotCont.dwPacerValue = CurrentCard.AI_Status.dwPacerValue;
        }
        else if(CurrentCard.wCardType!=PCI_9116){
            //Set PacerValue and TimerCascade
            if(!get_pacervalue(wCardNumber, fSampleRate,
                &iotCont.dwPacerValue))
            {
                return ErrorInvalidSampleRate;
            }
        }

        if((CurrentCard.wCardType!=PCI_9116)&&
            (CurrentCard.wCardType!=PCI_9114DG)&&
            (CurrentCard.wCardType!=PCI_9114HG))
        {
            //Set Timer Cascade
            iotConf = CurrentCard.AI_Status.CurrentConfig;
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CONFIG, &(iotConf))
                !=DAS_ERR_NO)
            {
                return iotConf.wErrorCode;
            }
        }

        iotCont.wAutoScan = 0;
        iotCont.wSyncMode = wSyncMode;
        iotCont.wChannelCount = 1;
        if(CurrentCard.wCardType==PCI_9116){
            iotCont.dwAccessCount = dwReadCount*iotCont.wChannelCount;
        }
        else{
            iotCont.dwAccessCount = dwReadCount;
        }
        iotCont.dwChannelGain[0].wChannel = wChannel;
        iotCont.dwChannelGain[0].wGain = wGain;
        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CON_READ, &(iotCont))
            !=DAS_ERR_NO)
        {
            return iotCont.wErrorCode;
        }

        if((!CurrentCard.AI_Status.bDblBufferMode)&&(wSyncMode==SYNCH_OP)){
            if(CurrentCard.AI_Status.pwBuffer){
                if((CurrentCard.wCardType!=PCI_9118DG)&&
                    (CurrentCard.wCardType!=PCI_9118HG)&&
                    (CurrentCard.wCardType!=PCI_9118HR))
                {
                    memcpy(CurrentCard.AI_Status.pwBuffer,
                        CurrentCard_AIFunc.pView[0],
                        CurrentCard.AI_Status.dwTransferCount*
                        (U32)((CurrentCard_AIFunc.wResoluation+7)/8)
                    );
                }
                else{  //9118
                    p = (U16 *)CurrentCard_AIFunc.pView[0];
                    for(i=0; i<CurrentCard.AI_Status.dwTransferCount; i++){
                        pwBuffer[i] = ((i%2)? (*(p+i-1)):(*(p+i+1)));
                    }
                }
            }
        }
    }

    return NoError;
}


I16 AI_ContReadMultiChannels(
    U16 wCardNumber,
    U16 wNumChans,
    U16 *pwChans,
    U16 *pwAdRanges,
    U16 *pwBuffer,
    U32 dwReadCount,
    F64 fSampleRate,
    U16 wSyncMode
    )
{
    I16 err;
    U16 wRange, wGain;
    U16 wDiffGains[256];
    U16 wNum;
    U32 i;
    U16 *p, *q;
    DAS_IOT_CONFIG iotConf;
    DAS_IOT_CONT iotCont;
    DAS_IOT_SG_CONT iotSCont;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));
    memset(&iotCont, '\0', sizeof(DAS_IOT_CONT));
    memset(&iotSCont, '\0', sizeof(DAS_IOT_SG_CONT));

    //00/02/03 jeffrey, re-asign the iotConf value
    iotConf = CurrentCard.AI_Status.CurrentConfig;

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, wNumChans,
        pwChans, dwReadCount*(U32)((CurrentCard_AIFunc.wResoluation+7)/8),
        TRUE))!=NoError)
    {
        return err;
    }
    if(!CurrentCard_AIFunc.wFlag.bAutoScan)
        return ErrorFuncNotSupport;

    if((CurrentCard_AIFunc.wFlag.bScanEndingZero)&&
        ((pwChans[wNumChans-1]!=0)&&(pwChans[0]!=0)))
    {
        return ErrorLastChannelNotZero;
    }

    if(CurrentCard_AIFunc.wArraySize<=1){
        wRange = pwAdRanges[0];
        for(i=1; i<wNumChans; i++){
            if(pwAdRanges[i]!=wRange)
                return ErrorDiffRangeNotSupport;
        }
        if(!validate_ad_range(wCardNumber, pwAdRanges[0], &wGain))
            return ErrorInvalidAdRange;
        if(!CurrentCard_AIFunc.wFlag.bScanIncrease){ //Descending
            for(i=0; i<wNumChans-1; i++){
                if(pwChans[i]!=(pwChans[i+1]+1))
                    return ErrorChannelNotDescending;
            }
        }
        else{ //Ascending
            for(i=0; i<wNumChans-1; i++){
                if(pwChans[i]!=(pwChans[i+1]-1))
                    return ErrorChannelNotAscending;
            }
        }
        wNum = 1;
    }
    else{
        for(i=0; i<wNumChans; i++){
            if(!validate_ad_range(wCardNumber, pwAdRanges[i], &wDiffGains[i]))
                return ErrorInvalidAdRange;
        }
        wNum = wNumChans;
    }

    if((CurrentCard.wCardType==PCI_9221)||(CurrentCard.wCardType==PCI_9524)||
        (CurrentCard.wCardType==PCI_9222)||(CurrentCard.wCardType==PCI_9223)||
        (CurrentCard.wCardType==PCI_922A))
    {
        /*Set*/
        if(!CurrentCard.AI_Status.bDblBufferMode){
            if((*pwBuffer>1)||(!CurrentCard.AI_Status.pDAQbuf[*pwBuffer]))
                return ErrorInvalidBufferID;
            CurrentCard.AI_Status.pDAQbufHandle = *pwBuffer;
        }
        else{
            if((*pwBuffer>1)||(!CurrentCard.AI_Status.pDAQbuf[0])||
                (!CurrentCard.AI_Status.pDAQbuf[1]))
            {
                return ErrorInvalidBufferID;
            }
            CurrentCard.AI_Status.pDAQbufHandle = 0;
        }
        CurrentCard.AI_Status.bToFile = FALSE;
        CurrentCard.AI_Status.wNumChans = wNumChans;
        CurrentCard.AI_Status.dwTransferCount = dwReadCount;
        /*Set*/
        switch(CurrentCard.wCardType){
            case PCI_9221:
                if((CurrentCard.AI_Status.dwPacerValue==0)||
                    (CurrentCard.AI_Status.dwSampleIntv==0)||
                    (CurrentCard.AI_Status.dwPacerValue<
                    CurrentCard.AI_Status.dwSampleIntv*wNum))
                {
                    return ErrorInvalidCounterState;
                }
                else{
                    iotSCont.dwPacerValue[1] = \
                        CurrentCard.AI_Status.dwSampleIntv;
                    iotSCont.dwPacerValue[0] = \
                        CurrentCard.AI_Status.dwPacerValue;
                }
                break;
            case PCI_9222:
            case PCI_9223:
            case PCI_922A:
                if(((CurrentCard.AI_Status.bDblBufferMode)||
                    (wSyncMode==SYNCH_OP))&&
                    (iotConf.mData.m922xAI.fEnReTrig))
                {
                    return ErrorInvalidTriggerMode;
                }
                if(iotConf.mData.m922xAI.fConvSrc==0){
                    if((CurrentCard.AI_Status.dwPacerValue==0)||
                        (CurrentCard.AI_Status.dwSampleIntv==0)||
                        (CurrentCard.AI_Status.dwPacerValue<
                        CurrentCard.AI_Status.dwSampleIntv*wNum))
                    {
                        return ErrorInvalidCounterState;
                    }
                    else{
                        iotSCont.dwPacerValue[1] = \
                            CurrentCard.AI_Status.dwSampleIntv;
                        iotSCont.dwPacerValue[0] = \
                            CurrentCard.AI_Status.dwPacerValue;
                    }
                }
                break;
            case PCI_9524:
                if(*pwChans==0){ /*load cell group channel 0*/
                    iotConf.mData.m9524AI.fGroup = 0;
                    if(*wDiffGains)
                        return ErrorInvalidAdRange;
                    CurrentCard.AI_Status.p7300buf = (void *)0;
                }
                else if(*pwChans==4){ /*general purpose group channel 0*/
                    iotConf.mData.dwValue[0] = \
                        CurrentCard.AI_Status.dwPacerValue;
                    iotConf.mData.dwValue[1] = \
                        CurrentCard.AI_Status.dwSampleIntv;
                    iotConf.mData.m9524AI.fGroup = 1;
                    if((!(*wDiffGains))||(*wDiffGains>4))
                        return ErrorInvalidAdRange;
                    CurrentCard.AI_Status.p7300buf = (void *)1;
                }

                if(iotConf.mData.m9524AI.fXMode!=1)
                    return ErrorConfigFailed;

                if(!get_pacervalue(wCardNumber, fSampleRate,
                    &iotSCont.dwPacerValue[0]))
                {
                    return ErrorInvalidSampleRate;
                }

                iotConf.mData.m9524AI.fAutoScan = 1;
                iotConf.mData.m9524AI.ADC_Channel = \
                    wNumChans-1+iotConf.mData.m9524AI.fGroup*4;
                iotConf.mData.m9524AI.ADC_Gain = *wDiffGains;
                iotConf.mData.m9524AI.ADC_Speed = \
                    (U8)(iotSCont.dwPacerValue[0]&0xff);
                iotConf.mData.m9524AI.Reserved1 = 1;

                if(iotConf.mData.m9524AI.fGroup==0){
                    CurrentCard.AI_Status.CurrentConfig = iotConf;
                }
                else{
                    /*use the two valuables to store the config of General Purpose group*/
                    CurrentCard.AI_Status.dwPacerValue = \
                        iotConf.mData.dwValue[0];
                    CurrentCard.AI_Status.dwSampleIntv = \
                        iotConf.mData.dwValue[1];
                }
                break;
        }
        iotSCont.wAutoScan = 1;
        iotSCont.wSyncMode = wSyncMode;
        iotSCont.wChannelCount = wNum;
        iotSCont.dwAccessCount = dwReadCount;
        for(i=0; i<wNum; i++){
            if((CurrentCard.wCardType==PCI_9222)&&(pwChans[i]>=8)){
                //PCI-9222: CH0~7=>CH0~7; CH8~15=>CH16~CH23
                iotSCont.dwChannelGain[i].wChannel = pwChans[i]+8;
            }
            else{
                iotSCont.dwChannelGain[i].wChannel = pwChans[i];
            }
            iotSCont.dwChannelGain[i].wGain = wDiffGains[i];
        }
        iotSCont.dwBufferAddr = *pwBuffer;
        /*IoControl*/
        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CON_READ_SG, &(iotSCont))
            !=DAS_ERR_NO)
        {
            return iotSCont.wErrorCode;
        }
    }
    else{
        if(CurrentCard.AI_Status.bDblBufferMode)
            dwReadCount /= 2;

        CurrentCard.AI_Status.bToFile = FALSE;

        if(CurrentCard.wCardType!=PCI_9116)
            CurrentCard.AI_Status.dwTransferCount = dwReadCount;
        else
            CurrentCard.AI_Status.dwTransferCount = dwReadCount*wNum;
        //CurrentCard.AI_Status.dwTransferCount = dwReadCount;
        CurrentCard.AI_Status.pwBuffer = pwBuffer;

        if(((CurrentCard.wCardType==PCI_9812)||(CurrentCard.wCardType==PCI_9810))
            &&((((iotConf.mData.wValue[0])&0x0300)==P9812_CLKSRC_EXT_SIN)||
            (((iotConf.mData.wValue[0])&0x0300)==P9812_CLKSRC_EXT_DIG)))
        {
            //iotCont.dwPacerValue = 2;
            iotCont.dwPacerValue = CurrentCard.AI_Status.dwPacerValue;
        }
        else if(CurrentCard.wCardType!=PCI_9116){
            //Set PacerValue and TimerCascade
            if(!get_pacervalue(wCardNumber, fSampleRate,
                &iotCont.dwPacerValue))
            {
                return ErrorInvalidSampleRate;
            }
        }

        if((CurrentCard.wCardType!=PCI_9116)&&
            (CurrentCard.wCardType!=PCI_9114DG)&&
            (CurrentCard.wCardType!=PCI_9114HG))
        {
            //Set Timer Cascade
            iotConf = CurrentCard.AI_Status.CurrentConfig;
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CONFIG, &(iotConf))
                !=DAS_ERR_NO)
            {
                return iotConf.wErrorCode;
            }
        }

        iotCont.wChannelCount = wNum;
        if(CurrentCard.wCardType==PCI_9116){
            iotCont.dwAccessCount = dwReadCount*iotCont.wChannelCount;
        }
        else{
            iotCont.dwAccessCount = dwReadCount;
        }
        //iotCont.dwAccessCount = dwReadCount;
        iotCont.wAutoScan = 1;
        iotCont.wSyncMode = wSyncMode;
        //iotCont.wChannelCount = wNum;
        if((wNum==1)&&((CurrentCard.wCardType!=PCI_9118DG)&&
            (CurrentCard.wCardType!=PCI_9118HG)&&
            (CurrentCard.wCardType!=PCI_9118HR)&&
            (CurrentCard.wCardType!=PCI_9116)))
        {
            if(CurrentCard_AIFunc.wFlag.bScanIncrease)
                iotCont.dwChannelGain[0].wChannel = pwChans[wNumChans-1];
            else
                iotCont.dwChannelGain[0].wChannel = pwChans[0];
            iotCont.dwChannelGain[0].wGain = wGain;
        }
        else{
            for(i=0; i<wNum; i++){
                iotCont.dwChannelGain[i].wChannel = pwChans[i];
                iotCont.dwChannelGain[i].wGain = wDiffGains[i];
            }
        }

        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CON_READ, &(iotCont))
            !=DAS_ERR_NO)
        {
            return iotCont.wErrorCode;
        }

        if(!(CurrentCard.AI_Status.bDblBufferMode)&&(wSyncMode==SYNCH_OP)){
            if(CurrentCard.AI_Status.pwBuffer){
                if((CurrentCard.wCardType!=PCI_9118DG)&&
                    (CurrentCard.wCardType!=PCI_9118HG)&&
                    (CurrentCard.wCardType!=PCI_9118HR))
                {
                    memcpy(
                        CurrentCard.AI_Status.pwBuffer,
                        CurrentCard_AIFunc.pView[0],
                        CurrentCard.AI_Status.dwTransferCount*
                        (U32)((CurrentCard_AIFunc.wResoluation+7)/8)
                    );
                }
                else{  //9118
                    p = (U16 *)CurrentCard_AIFunc.pView[0];
                    q = (U16 *)pwBuffer;
                    for(i=0; i<CurrentCard.AI_Status.dwTransferCount; i++){
                        *(q+i) = ((i%2)? (*(p+i-1)):(*(p+i+1)));
                    }
                }
            }
        }
    }

    return NoError;
}


/*pwBuffer is of no use while double-buffer mode enabled*/
I16 AI_ContScanChannels(
    U16 wCardNumber,
    U16 wChannel,
    U16 wAdRange,
    U16 *pwBuffer,
    U32 dwReadCount,
    F64 fSampleRate,
    U16 wSyncMode
    )
{
    I16 err;
    U16 wGain;
    U16 Chans[256], Ranges[256];
    U32 i;
    U16 *p, *q;
    DAS_IOT_CONFIG iotConf;
    DAS_IOT_CONT   iotCont;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));
    memset(&iotCont, '\0', sizeof(DAS_IOT_CONT));

    //00/02/03 jeffrey, re-asign the iotConf value
    iotConf = CurrentCard.AI_Status.CurrentConfig;

    if((CurrentCard.wCardType==PCI_9118DG)||
        (CurrentCard.wCardType==PCI_9118HG)||
        (CurrentCard.wCardType==PCI_9118HR)||
        (CurrentCard.wCardType==PCI_9116)||
        (CurrentCard.wCardType==PCI_9221)||
        (CurrentCard.wCardType==PCI_9222)||
        (CurrentCard.wCardType==PCI_9223)||
        (CurrentCard.wCardType==PCI_9524)||
        (CurrentCard.wCardType==PCI_922A))
    {
        if(CurrentCard.wCardType==PCI_9524){
            if(wChannel<4){
                for(i=0; i<=wChannel; i++){
                    Chans[i] = (U16)i;
                    Ranges[i] = wAdRange;
                }
            }
            else if(wChannel<8){
                wChannel -= 4;
                for(i=0; i<=wChannel; i++){
                    Chans[i] = (U16)(i+4);
                    Ranges[i] = wAdRange;
                }
            }
            else{
                return ErrorInvalidIoChannel;
            }
        }
        else{
            for(i=0; i<=wChannel; i++){
                Chans[i] = (U16)i;
                Ranges[i] = wAdRange;
            }
        }
        return (AI_ContReadMultiChannels(wCardNumber, wChannel+1, Chans, \
            Ranges, pwBuffer, dwReadCount, fSampleRate, wSyncMode));
    }

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 1, &wChannel,
        dwReadCount*(U32)((CurrentCard_AIFunc.wResoluation+7)/8), TRUE))
        !=NoError)
    {
        return err;
    }
    if(!CurrentCard_AIFunc.wFlag.bAutoScan)
        return ErrorFuncNotSupport;

    /*Check wAdRange and get Gain*/
    if(!validate_ad_range(wCardNumber, wAdRange, &wGain))
        return ErrorInvalidAdRange;

    if(CurrentCard.AI_Status.bDblBufferMode)
        dwReadCount /= 2;

    CurrentCard.AI_Status.bToFile = FALSE;
    CurrentCard.AI_Status.dwTransferCount = dwReadCount;
    CurrentCard.AI_Status.pwBuffer = pwBuffer;

    if(((CurrentCard.wCardType==PCI_9812)||(CurrentCard.wCardType==PCI_9810))
        &&((((iotConf.mData.wValue[0])&0x0300)==P9812_CLKSRC_EXT_SIN)||
        (((iotConf.mData.wValue[0])&0x0300)==P9812_CLKSRC_EXT_DIG)))
    {
        //iotCont.dwPacerValue = 2;
        iotCont.dwPacerValue = CurrentCard.AI_Status.dwPacerValue;
    }
    else{
        //Set PacerValue and TimerCascade
        if(!get_pacervalue(wCardNumber, fSampleRate, &iotCont.dwPacerValue))
            return ErrorInvalidSampleRate;
    }

    if((CurrentCard.wCardType!=PCI_9116)&&
        (CurrentCard.wCardType!=PCI_9114DG)&&
        (CurrentCard.wCardType!=PCI_9114HG))
    {
        //Set Timer Cascade
        iotConf = CurrentCard.AI_Status.CurrentConfig;
        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CONFIG, &(iotConf))
            !=DAS_ERR_NO)
        {
            return iotConf.wErrorCode;
        }
    }
    //yuan 07/31/00
    iotCont.dwAccessCount = dwReadCount;
    iotCont.wAutoScan = 1;
    iotCont.wSyncMode = wSyncMode;
    iotCont.wChannelCount = 1;
    iotCont.dwChannelGain[0].wChannel = wChannel;
    iotCont.dwChannelGain[0].wGain = wGain;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CON_READ, &(iotCont))
        !=DAS_ERR_NO)
    {
        return iotCont.wErrorCode;
    }

    if(!(CurrentCard.AI_Status.bDblBufferMode)&&(wSyncMode==SYNCH_OP)){
        if(CurrentCard.AI_Status.pwBuffer){
            if((CurrentCard.wCardType!=PCI_9118DG)&&
                (CurrentCard.wCardType!=PCI_9118HG)&&
                (CurrentCard.wCardType!=PCI_9118HR))
            {
                memcpy(
                    CurrentCard.AI_Status.pwBuffer,
                    CurrentCard_AIFunc.pView[0],
                    CurrentCard.AI_Status.dwTransferCount*
                    (U32)((CurrentCard_AIFunc.wResoluation+7)/8)
                );
            }
            else{  //9118
                p = (U16 *)CurrentCard_AIFunc.pView[0];
                q = (U16 *)pwBuffer;
                for(i=0; i<CurrentCard.AI_Status.dwTransferCount; i++){
                    *(q+i) = ((i%2)? (*(p+i-1)):(*(p+i+1)));
                }
            }
        }
    }

    return NoError;
}


/*pwBuffer is of no use while double-buffer mode enabled*/
I16 AI_ContReadChannelToFile(
    U16 wCardNumber,
    U16 wChannel,
    U16 wAdRange,
    U8  *fileName,
    U32 dwReadCount,
    F64 fSampleRate,
    U16 wSyncMode
    )
{
    I16 err;
    U16 wGain;
    FILE *ad_file;
    char svFileName[128];
    U16 byte_width;
    U32 dwPacerValue;
    DAS_IOT_CONFIG iotConf;
    DAS_IOT_CONT iotCont;
    DAS_IOT_SG_CONT iotSCont;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));
    memset(&iotCont, '\0', sizeof(DAS_IOT_CONT));
    memset(&iotSCont, '\0', sizeof(DAS_IOT_SG_CONT));

    //00/02/03 jeffrey, re-asign the iotConf value
    iotConf = CurrentCard.AI_Status.CurrentConfig;

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 1, &wChannel,
        dwReadCount*(U32)((CurrentCard_AIFunc.wResoluation+7)/8), TRUE))
        !=NoError)
    {
        return err;
    }
    //Check wAdRange and get Gain
    if(!validate_ad_range(wCardNumber, wAdRange, &wGain))
        return ErrorInvalidAdRange;

    sprintf(svFileName, "%s.dat", fileName);
    if((ad_file = fopen(svFileName, "w+b"))==NULL)
        return ErrorOpenFile;

    /*Scatter-gather*/
    if((CurrentCard.wCardType==PCI_9221)||(CurrentCard.wCardType==PCI_9524)||
        (CurrentCard.wCardType==PCI_9222)||(CurrentCard.wCardType==PCI_9223)||
        (CurrentCard.wCardType==PCI_922A))
    {
        /*Set*/
        /*always started from first buffer*/
        if(!CurrentCard.AI_Status.bDblBufferMode){
            if(!CurrentCard.AI_Status.pDAQbuf[0])
                return ErrorInvalidBufferID;
            CurrentCard.AI_Status.pDAQbufHandle = 0;
        }
        else{
            if((!CurrentCard.AI_Status.pDAQbuf[0])||
                (!CurrentCard.AI_Status.pDAQbuf[1]))
            {
                return ErrorInvalidBufferID;
            }
            CurrentCard.AI_Status.pDAQbufHandle = 0;
        }
        CurrentCard.AI_Status.bToFile = TRUE;
        CurrentCard.AI_Status.bChanRange = 0;
        CurrentCard.AI_Status.dwTransferCount = dwReadCount;
        CurrentCard.AI_Status.wNumChans = 1;
        CurrentCard.AI_Status.pwBuffer = ad_file;
        byte_width = (CurrentCard_AIFunc.wResoluation+7)/8;

        /*Set*/
        switch(CurrentCard.wCardType){
            case PCI_9221:
                if((CurrentCard.AI_Status.dwPacerValue==0)||
                    (CurrentCard.AI_Status.dwSampleIntv==0))
                {
                    return ErrorInvalidCounterState;
                }
                else{
                    dwPacerValue = CurrentCard.AI_Status.dwPacerValue;
                    fSampleRate = ((F64)CurrentCardInfo.dwCtrKHz*1000)/
                        ((F64)dwPacerValue);
                    iotSCont.dwPacerValue[1] = \
                        CurrentCard.AI_Status.dwSampleIntv;
                    iotSCont.dwPacerValue[0] = \
                        CurrentCard.AI_Status.dwPacerValue;
                }
                break;
            case PCI_9222:
            case PCI_9223:
            case PCI_922A:
                if(((CurrentCard.AI_Status.bDblBufferMode)||
                    (wSyncMode==SYNCH_OP))&&
                    (iotConf.mData.m922xAI.fEnReTrig))
                {
                    return ErrorInvalidTriggerMode;
                }
                if(iotConf.mData.m922xAI.fConvSrc==0){
                    if((CurrentCard.AI_Status.dwPacerValue==0)||
                        (CurrentCard.AI_Status.dwSampleIntv==0))
                    {
                        return ErrorInvalidCounterState;
                    }
                    else{
                        dwPacerValue = CurrentCard.AI_Status.dwPacerValue;
                        fSampleRate = ((F64)CurrentCardInfo.dwCtrKHz*1000)/
                            ((F64)dwPacerValue);
                        iotSCont.dwPacerValue[1] = \
                            CurrentCard.AI_Status.dwSampleIntv;
                        iotSCont.dwPacerValue[0] = \
                            CurrentCard.AI_Status.dwPacerValue;
                    }
                }
                break;
            case PCI_9524:
                if(wChannel<4){ /*load cell group channel 0*/
                    F64 actRate[2];
                    iotConf.mData.m9524AI.fGroup = 0;
                    if(wGain)
                        return ErrorInvalidAdRange;

                    if(!get_pacervalue(wCardNumber, fSampleRate,
                        &iotSCont.dwPacerValue[0]))
                    {
                        return ErrorInvalidSampleRate;
                    }
                    get_actualrate_9524(iotSCont.dwPacerValue[0], actRate);

                    if(iotConf.mData.m9524AI.fAZEn){
                        fSampleRate = (*(actRate+1))/2;
                    }
                    else{
                        fSampleRate = *actRate;
                    }
                    CurrentCard.AI_Status.p7300buf = (void *)0;
                }
                else if(wChannel<8){ /*general purpose group channel 0*/
                    F64 actRate[2];
                    iotConf.mData.dwValue[0] = \
                        CurrentCard.AI_Status.dwPacerValue;
                    iotConf.mData.dwValue[1] = \
                        CurrentCard.AI_Status.dwSampleIntv;
                    iotConf.mData.m9524AI.fGroup = 1;
                    if((!wGain)||(wGain>4))
                        return ErrorInvalidAdRange;

                    if(!get_pacervalue(wCardNumber, fSampleRate,
                        &iotSCont.dwPacerValue[0]))
                    {
                        return ErrorInvalidSampleRate;
                    }
                    get_actualrate_9524(iotSCont.dwPacerValue[0], actRate);

                    fSampleRate = *actRate;
                    CurrentCard.AI_Status.p7300buf = (void *)1;
                }
                else
                    return ErrorInvalidIoChannel;

                if(iotConf.mData.m9524AI.fXMode!=1)
                    return ErrorConfigFailed;

                iotConf.mData.m9524AI.fAutoScan = 0;
                iotConf.mData.m9524AI.ADC_Channel = wChannel;
                iotConf.mData.m9524AI.ADC_Gain = wGain;
                iotConf.mData.m9524AI.ADC_Speed = \
                    (U8)(iotSCont.dwPacerValue[0]&0xff);
                iotConf.mData.m9524AI.Reserved1 = 1;

                if(iotConf.mData.m9524AI.fGroup==0){
                    CurrentCard.AI_Status.CurrentConfig = iotConf;
                }
                else{
                    /*use the two valuables to store the config of General Purpose group*/
                    CurrentCard.AI_Status.dwPacerValue = \
                        iotConf.mData.dwValue[0];
                    CurrentCard.AI_Status.dwSampleIntv = \
                        iotConf.mData.dwValue[1];
                }
                break;
        }

        iotSCont.wAutoScan = 0;
        iotSCont.wSyncMode = wSyncMode;
        iotSCont.wChannelCount = 1;
        iotSCont.dwAccessCount = dwReadCount;
        if((CurrentCard.wCardType==PCI_9222)&&(wChannel>=8))
            iotSCont.dwChannelGain[0].wChannel = wChannel+8;
        else
            iotSCont.dwChannelGain[0].wChannel = wChannel;
        iotSCont.dwChannelGain[0].wGain = wGain;
        iotSCont.dwBufferAddr = 0;

        write_header_to_file(wCardNumber, ad_file, byte_width, wAdRange, 1,
            wChannel, dwReadCount, fSampleRate, 0
        );

        if(CurrentCard.wCardType==PCI_9524){
            double *Compensation;
            double res_offset, scaling;
            int ret;

            //hMem = GlobalAlloc(GMEM_FIXED, 2*sizeof(double));
            //Compensation = (double *)GlobalLock(hMem);
            Compensation = (double *)malloc(2*sizeof(double));
            if(Compensation==NULL)
                return ErrorAllocateMemory;

            if(wChannel<4){ //Load Cell Group
                if((ret = ai_get_compensation_9524(
                    wCardNumber, 0, wGain, (U16)iotSCont.dwPacerValue[0],
                    &res_offset, &scaling))<0)
                {
                    return ret;
                }
                *(Compensation) = res_offset;
                *(Compensation+1) = scaling;
            }
            else if(wChannel<8){ //General Purpose Group
                if((ret = ai_get_compensation_9524(
                    wCardNumber, 1, wGain, (U16)iotSCont.dwPacerValue[0],
                    &res_offset, &scaling))<0)
                {
                    return ret;
                }
                *(Compensation) = res_offset;
                *(Compensation+1) = scaling;
            }

            fwrite(Compensation, sizeof(double)*2, 1, ad_file);
            //GlobalUnfix(hMem);
            //GlobalUnlock(hMem);
            //GlobalFree(hMem);
            free(Compensation);
        }

        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CON_READ_SG, &(iotSCont))
            !=DAS_ERR_NO)
        {
            return iotSCont.wErrorCode;
        }

        if((!CurrentCard.AI_Status.bDblBufferMode)&&(wSyncMode==SYNCH_OP)){
            write_data_to_file(
                (U16 *)CurrentCard.AI_Status.pDAQbuf[0],
                ad_file,
                (U16)((CurrentCard_AIFunc.wResoluation+7)/8),
                FALSE,/*1,*/
                CurrentCard.AI_Status.dwTransferCount
            );
            fclose(ad_file);
        }
    }
    else{
        if(CurrentCard.AI_Status.bDblBufferMode)
            dwReadCount /= 2;

        CurrentCard.AI_Status.bToFile = TRUE;
        CurrentCard.AI_Status.bChanRange = 0;
        CurrentCard.AI_Status.dwTransferCount = dwReadCount;
        CurrentCard.AI_Status.wNumChans = 1;
        CurrentCard.AI_Status.pwBuffer = ad_file;
        byte_width = (CurrentCard_AIFunc.wResoluation+7)/8;

        if(((CurrentCard.wCardType==PCI_9812)||
            (CurrentCard.wCardType==PCI_9810))&&
            ((((iotConf.mData.wValue[0])&0x0300)==P9812_CLKSRC_EXT_SIN)||
            (((iotConf.mData.wValue[0])&0x0300)==P9812_CLKSRC_EXT_DIG)))
        {
            //iotCont.dwPacerValue = 2;
            iotCont.dwPacerValue = CurrentCard.AI_Status.dwPacerValue;
        }
        else if(CurrentCard.wCardType!=PCI_9116){
            //Set PacerValue and TimerCascade
            if(!get_pacervalue(wCardNumber, fSampleRate, &iotCont.dwPacerValue))
                return ErrorInvalidSampleRate;
        }

        if((CurrentCard.wCardType!=PCI_9116)&&
            (CurrentCard.wCardType!=PCI_9114DG)&&
            (CurrentCard.wCardType!=PCI_9114HG))
        {
            //Set Timer Cascade
            iotConf = CurrentCard.AI_Status.CurrentConfig;
            //yuan 07/31/00
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CONFIG, &(iotConf))
                !=DAS_ERR_NO)
            {
                return iotConf.wErrorCode;
            }
        }

        iotCont.wAutoScan = 0;
        iotCont.wSyncMode = wSyncMode;
        iotCont.wChannelCount = 1;
        if(CurrentCard.wCardType==PCI_9116){
            iotCont.dwAccessCount = dwReadCount*iotCont.wChannelCount;
            if(iotConf.mData.wValue[0]&0x100)
                fSampleRate = (F64)0;
            else{
                dwPacerValue = CurrentCard.AI_Status.dwPacerValue;
                fSampleRate = \
                    ((F64)CurrentCardInfo.dwCtrKHz*1000)/((F64)dwPacerValue);
            }
        }
        else{
            iotCont.dwAccessCount = dwReadCount;
        }
        iotCont.dwChannelGain[0].wChannel = wChannel;
        iotCont.dwChannelGain[0].wGain = wGain;

        write_header_to_file(wCardNumber, ad_file, byte_width, wAdRange, 1,
            wChannel, dwReadCount, fSampleRate, 0
        );

        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CON_READ, &(iotCont))
            !=DAS_ERR_NO)
        {
            return iotCont.wErrorCode;
        }

        if((!CurrentCard.AI_Status.bDblBufferMode)&&(wSyncMode==SYNCH_OP)){
            if((CurrentCard.wCardType!=PCI_9118DG)&&
                (CurrentCard.wCardType!=PCI_9118HG)&&
                (CurrentCard.wCardType!=PCI_9118HR))
            {
                write_data_to_file(
                    (U16 *)CurrentCard_AIFunc.pView[0],
                    ad_file,
                    (U16)((CurrentCard_AIFunc.wResoluation+7)/8),
                    FALSE,
                    CurrentCard.AI_Status.dwTransferCount
                );
            }
            else{
                //9118
                write_data_to_file(
                    (U16 *)CurrentCard_AIFunc.pView[0],
                    ad_file,
                    (U16)((CurrentCard_AIFunc.wResoluation+7)/8),
                    TRUE,
                    CurrentCard.AI_Status.dwTransferCount
                );
            }
            fclose(ad_file);
        }
    }

    return NoError;
}


I16 AI_ContReadMultiChannelsToFile(
    U16 wCardNumber,
    U16 wNumChans,
    U16 *pwChans,
    U16 *pwAdRanges,
    U8  *fileName,
    U32 dwReadCount,
    F64 fSampleRate,
    U16 wSyncMode
    )
{
    I16 err;
    U16 wRange, wGain;
    U16 wDiffGains[256];
    U16 wNum;
    char svFileName[128];
    U16 i;
    U16 byte_width;
    FILE *ad_file;
    DAQChannelRange *chRange;
    U32 dwPacerValue;
    DAS_IOT_CONFIG iotConf;
    DAS_IOT_CONT iotCont;
    DAS_IOT_SG_CONT iotSCont;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));
    memset(&iotCont, '\0', sizeof(DAS_IOT_CONT));
    memset(&iotSCont, '\0', sizeof(DAS_IOT_SG_CONT));

    //00/02/03 jeffrey, re-asign the iotConf value
    iotConf = CurrentCard.AI_Status.CurrentConfig;

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, wNumChans,
        pwChans, dwReadCount*(U32)((CurrentCard_AIFunc.wResoluation+7)/8),
        TRUE))!=NoError)
    {
        return err;
    }

    if(!CurrentCard_AIFunc.wFlag.bAutoScan)
        return ErrorFuncNotSupport;

    if((CurrentCard_AIFunc.wFlag.bScanEndingZero)&&
        ((pwChans[wNumChans-1]!=0)&&(pwChans[0]!=0)))
    {
        return ErrorLastChannelNotZero;
    }

    if(CurrentCard_AIFunc.wArraySize<=1){
        wRange = pwAdRanges[0];
        for(i=1; i<wNumChans; i++){
            if(pwAdRanges[i]!=wRange)
                return ErrorDiffRangeNotSupport;
        }
        if(!validate_ad_range(wCardNumber, pwAdRanges[0], &wGain))
            return ErrorInvalidAdRange;
        if(!CurrentCard_AIFunc.wFlag.bScanIncrease){ //Descending
            for(i=0; i<wNumChans-1; i++){
                if(pwChans[i]!=(pwChans[i+1]+1))
                    return ErrorChannelNotDescending;
            }
        }
        else{ //Ascending
            for(i=0; i<wNumChans-1; i++){
                if(pwChans[i]!=(pwChans[i+1]-1))
                    return ErrorChannelNotAscending;
            }
        }
        wNum = 1;
    }
    else{
        for(i=0; i<wNumChans; i++){
            if(!validate_ad_range(wCardNumber, pwAdRanges[i], &wDiffGains[i]))
                return ErrorInvalidAdRange;
        }
        wNum = wNumChans;
    }

    sprintf(svFileName, "%s.dat", fileName);
    if((ad_file = fopen(svFileName, "w+b"))==NULL)
        return ErrorOpenFile;

    if((CurrentCard.wCardType==PCI_9221)||(CurrentCard.wCardType==PCI_9524)||
        (CurrentCard.wCardType==PCI_9222)||(CurrentCard.wCardType==PCI_9223)||
        (CurrentCard.wCardType==PCI_922A))
    {
        /*Set*/
        /*always started from first buffer*/
        if(!CurrentCard.AI_Status.bDblBufferMode){
            if(!CurrentCard.AI_Status.pDAQbuf[0])
                return ErrorInvalidBufferID;
            CurrentCard.AI_Status.pDAQbufHandle = 0;
        }
        else{
            if((!CurrentCard.AI_Status.pDAQbuf[0])||
                (!CurrentCard.AI_Status.pDAQbuf[1]))
            {
                return ErrorInvalidBufferID;
            }
            CurrentCard.AI_Status.pDAQbufHandle = 0;
        }
        CurrentCard.AI_Status.bToFile = TRUE;
        CurrentCard.AI_Status.wNumChans = wNumChans;
        CurrentCard.AI_Status.pwBuffer = ad_file;
        CurrentCard.AI_Status.dwTransferCount = dwReadCount;
        byte_width = (CurrentCard_AIFunc.wResoluation+7)/8;
        /*Set*/
        switch(CurrentCard.wCardType){
            case PCI_9221:
                if((CurrentCard.AI_Status.dwPacerValue==0)||
                    (CurrentCard.AI_Status.dwSampleIntv==0)||
                    (CurrentCard.AI_Status.dwPacerValue<
                    CurrentCard.AI_Status.dwSampleIntv*wNum))
                {
                    return ErrorInvalidCounterState;
                }
                else{
                    dwPacerValue = CurrentCard.AI_Status.dwPacerValue;
                    fSampleRate = ((F64)CurrentCardInfo.dwCtrKHz*1000)/
                        ((F64)dwPacerValue);
                    iotSCont.dwPacerValue[1] = \
                        CurrentCard.AI_Status.dwSampleIntv;
                    iotSCont.dwPacerValue[0] = \
                        CurrentCard.AI_Status.dwPacerValue;
                }
                break;
            case PCI_9222:
            case PCI_9223:
            case PCI_922A:
                if(((CurrentCard.AI_Status.bDblBufferMode)||
                    (wSyncMode==SYNCH_OP))&&
                    (iotConf.mData.m922xAI.fEnReTrig))
                {
                    return ErrorInvalidTriggerMode;
                }
                if(iotConf.mData.m922xAI.fConvSrc==0){
                    if((CurrentCard.AI_Status.dwPacerValue==0)||
                        (CurrentCard.AI_Status.dwSampleIntv==0)||
                        (CurrentCard.AI_Status.dwPacerValue<
                        CurrentCard.AI_Status.dwSampleIntv*wNum))
                    {
                        return ErrorInvalidCounterState;
                    }
                    else{
                        dwPacerValue = CurrentCard.AI_Status.dwPacerValue;
                        fSampleRate = ((F64)CurrentCardInfo.dwCtrKHz*1000)/
                            ((F64)dwPacerValue);
                        iotSCont.dwPacerValue[1] = \
                            CurrentCard.AI_Status.dwSampleIntv;
                        iotSCont.dwPacerValue[0] = \
                            CurrentCard.AI_Status.dwPacerValue;
                    }
                }
                break;
            case PCI_9524:
                if(*pwChans==0){ /*load cell group channel 0*/
                    F64 actRate[2];

                    iotConf.mData.m9524AI.fGroup = 0;

                    if(*wDiffGains)
                        return ErrorInvalidAdRange;

                    if(!get_pacervalue(wCardNumber, fSampleRate,
                        &iotSCont.dwPacerValue[0]))
                    {
                        return ErrorInvalidSampleRate;
                    }

                    get_actualrate_9524(iotSCont.dwPacerValue[0], actRate);

                    if(iotConf.mData.m9524AI.fAZEn){
                        fSampleRate = (*(actRate+1))/2;
                    }
                    else{
                        fSampleRate = *(actRate+1);
                    }

                    CurrentCard.AI_Status.p7300buf = (void *)0;
                }
                else if(*pwChans==4){ /*general purpose group channel 0*/
                    F64 actRate[2];

                    iotConf.mData.dwValue[0] = \
                        CurrentCard.AI_Status.dwPacerValue;
                    iotConf.mData.dwValue[1] = \
                        CurrentCard.AI_Status.dwSampleIntv;
                    iotConf.mData.m9524AI.fGroup = 1;

                    if((!(*wDiffGains))||(*wDiffGains>4))
                        return ErrorInvalidAdRange;

                    if(!get_pacervalue(wCardNumber, fSampleRate,
                        &iotSCont.dwPacerValue[0]))
                    {
                        return ErrorInvalidSampleRate;
                    }

                    get_actualrate_9524(iotSCont.dwPacerValue[0], actRate);

                    fSampleRate = *(actRate+1);

                    CurrentCard.AI_Status.p7300buf = (void *)1;
                }
                else
                    return ErrorInvalidIoChannel;

                if(iotConf.mData.m9524AI.fXMode!=1)
                    return ErrorConfigFailed;

                iotConf.mData.m9524AI.fAutoScan = 1;
                iotConf.mData.m9524AI.ADC_Channel = \
                    wNumChans-1+iotConf.mData.m9524AI.fGroup*4;
                iotConf.mData.m9524AI.ADC_Gain = *wDiffGains;
                iotConf.mData.m9524AI.ADC_Speed = \
                    (U8)(iotSCont.dwPacerValue[0]&0xff);
                iotConf.mData.m9524AI.Reserved1 = 1;

                if(iotConf.mData.m9524AI.fGroup==0){
                    CurrentCard.AI_Status.CurrentConfig = iotConf;
                }
                else{
                    /*use the two valuables to store the config of General Purpose group*/
                    CurrentCard.AI_Status.dwPacerValue = \
                        iotConf.mData.dwValue[0];
                    CurrentCard.AI_Status.dwSampleIntv = \
                        iotConf.mData.dwValue[1];
                }
                break;
        }
        iotSCont.wAutoScan = 1;
        iotSCont.wSyncMode = wSyncMode;
        iotSCont.wChannelCount = wNum;
        iotSCont.dwAccessCount = dwReadCount;
        for(i=0; i<wNum; i++){
            if((CurrentCard.wCardType==PCI_9222)&&(pwChans[i]>=8)){
                //PCI-9222: CH0~7=>CH0~7; CH8~15=>CH16~CH23
                iotSCont.dwChannelGain[i].wChannel = pwChans[i]+8;
            }
            else{
                iotSCont.dwChannelGain[i].wChannel = pwChans[i];
            }
            iotSCont.dwChannelGain[i].wGain = wDiffGains[i];
        }
        iotSCont.dwBufferAddr = 0;

        write_header_to_file(wCardNumber, ad_file, byte_width, pwAdRanges[0],
            wNum, pwChans[0], dwReadCount, fSampleRate, wNum
        );

        CurrentCard.AI_Status.bChanRange = 1;

        chRange = (DAQChannelRange *)malloc(wNum * sizeof(DAQChannelRange));
        if(chRange==NULL)
            return ErrorAllocateMemory;
        for(i=0; i<wNum; i++){
            chRange[i].channel = (char)pwChans[i];
            chRange[i].range = (char)pwAdRanges[i];
        }
        fwrite(chRange, sizeof(DAQChannelRange), wNum, ad_file);
        free(chRange);

        if(CurrentCard.wCardType==PCI_9524){
            double *Compensation;
            double res_offset, scaling;
            int ret;
            int getComploop;

            //hMem = GlobalAlloc(GMEM_FIXED, wNum*2*sizeof(double));
            //Compensation = (double *)GlobalLock(hMem);
            Compensation = (double *)malloc(wNum*2*sizeof(double));
            if(Compensation==NULL)
                return ErrorAllocateMemory;

            if(*pwChans==0){
                /*Load Cell Group*/
                for(getComploop=0; getComploop<wNum; getComploop++){
                    if((ret=ai_get_compensation_9524(wCardNumber, 0,
                        *wDiffGains, (U16)iotSCont.dwPacerValue[0],
                        &res_offset, &scaling))<0)
                    {
                        return ret;
                    }
                    *(Compensation+getComploop*2) = res_offset;
                    *(Compensation+getComploop*2+1) = scaling;
                }
            }
            else if(*pwChans==4){
                /*General Purpose Group*/
                for(getComploop=0; getComploop<wNum; getComploop++){
                    if((ret=ai_get_compensation_9524(wCardNumber, 1,
                        *wDiffGains, (U16)iotSCont.dwPacerValue[0],
                        &res_offset, &scaling))<0)
                    {
                        return ret;
                    }
                    *(Compensation+getComploop*2) = res_offset;
                    *(Compensation+getComploop*2+1) = scaling;
                }
            }

            fwrite(Compensation, sizeof(double)*2, wNum, ad_file);
            //GlobalUnfix(hMem);
            //GlobalUnlock(hMem);
            //GlobalFree(hMem);
            free(Compensation);
        }

        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CON_READ_SG, &(iotSCont))
            !=DAS_ERR_NO)
        {
            return iotSCont.wErrorCode;
        }

        if((!CurrentCard.AI_Status.bDblBufferMode)&&(wSyncMode==SYNCH_OP)){
            write_data_to_file(
                (U16 *)CurrentCard.AI_Status.pDAQbuf[0],
                ad_file,
                (U16)((CurrentCard_AIFunc.wResoluation+7)/8),
                FALSE, /*1,*/
                CurrentCard.AI_Status.dwTransferCount
            );
            fclose(ad_file);
        }
    }
    else{
        if(CurrentCard.AI_Status.bDblBufferMode)
            dwReadCount /= 2;

        CurrentCard.AI_Status.bToFile = TRUE;
        //CurrentCard.AI_Status.dwTransferCount = dwReadCount;
        CurrentCard.AI_Status.wNumChans = wNumChans;
        CurrentCard.AI_Status.pwBuffer = ad_file;
        if(CurrentCard.wCardType!=PCI_9116)
            CurrentCard.AI_Status.dwTransferCount = dwReadCount;
        else
            CurrentCard.AI_Status.dwTransferCount = dwReadCount*wNum;

        byte_width = (CurrentCard_AIFunc.wResoluation+7)/8;

        if(((CurrentCard.wCardType==PCI_9812)||
            (CurrentCard.wCardType==PCI_9810))&&
            ((((iotConf.mData.wValue[0])&0x0300)==P9812_CLKSRC_EXT_SIN)||
            (((iotConf.mData.wValue[0])&0x0300)==P9812_CLKSRC_EXT_DIG)))
        {
            //iotCont.dwPacerValue = 2;
            iotCont.dwPacerValue = CurrentCard.AI_Status.dwPacerValue;
        }
        else if(CurrentCard.wCardType!=PCI_9116){
            //Set PacerValue and TimerCascade
            if(!get_pacervalue(wCardNumber, fSampleRate,&iotCont.dwPacerValue))
                return ErrorInvalidSampleRate;
        }
        //Set PacerValue and TimerCascade
        if((CurrentCard.wCardType!=PCI_9116)&&
            (CurrentCard.wCardType!=PCI_9114DG)&&
            (CurrentCard.wCardType!=PCI_9114HG))
        {
            //Set Timer Cascade
            iotConf = CurrentCard.AI_Status.CurrentConfig;
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CONFIG, &(iotConf))
                !=DAS_ERR_NO)
            {
                return iotConf.wErrorCode;
            }
        }

        //iotCont.dwAccessCount = dwReadCount;
        iotCont.wAutoScan = 1;
        iotCont.wSyncMode = wSyncMode;
        iotCont.wChannelCount = wNum;
        if(CurrentCard.wCardType==PCI_9116){
            iotCont.dwAccessCount = dwReadCount*iotCont.wChannelCount;
            if(iotConf.mData.wValue[0]&0x100)
                fSampleRate = (F64)0;
            else{
                dwPacerValue = CurrentCard.AI_Status.dwPacerValue;
                fSampleRate = \
                    ((F64)CurrentCardInfo.dwCtrKHz*1000)/((F64)dwPacerValue);
            }
        }
        else{
            iotCont.dwAccessCount = dwReadCount;
        }
        if((wNum==1)&&((CurrentCard.wCardType!=PCI_9118DG)&&
            (CurrentCard.wCardType!=PCI_9118HG)&&
            (CurrentCard.wCardType!=PCI_9118HR)&&
            (CurrentCard.wCardType!=PCI_9116)))
        {
            if(CurrentCard_AIFunc.wFlag.bScanIncrease)
                iotCont.dwChannelGain[0].wChannel = pwChans[wNumChans-1];
            else
                iotCont.dwChannelGain[0].wChannel = pwChans[0];
            iotCont.dwChannelGain[0].wGain = wGain;
            CurrentCard.AI_Status.bChanRange = 0;
            write_header_to_file(wCardNumber, ad_file, byte_width,
                pwAdRanges[0], wNum, pwChans[0], dwReadCount, fSampleRate, 0
            );
        }
        else{
            for(i=0; i<wNum; i++){
                iotCont.dwChannelGain[i].wChannel = pwChans[i];
                iotCont.dwChannelGain[i].wGain = wDiffGains[i];
            }
            CurrentCard.AI_Status.bChanRange = 1;
            write_header_to_file(wCardNumber, ad_file, byte_width,
                pwAdRanges[0], wNum, pwChans[0], dwReadCount, fSampleRate, wNum
            );
            chRange = (DAQChannelRange *)malloc(wNum*sizeof(DAQChannelRange));
            if(chRange==NULL)
                return ErrorAllocateMemory;
            for(i=0; i<wNum; i++){
                chRange[i].channel = (char)pwChans[i];
                chRange[i].range = (char)pwAdRanges[i];
            }
            fwrite(chRange, sizeof(DAQChannelRange), wNum, ad_file);
            free(chRange);
        }
        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CON_READ, &(iotCont))
            !=DAS_ERR_NO)
        {
            return iotCont.wErrorCode;
        }

        if((!CurrentCard.AI_Status.bDblBufferMode)&&(wSyncMode==SYNCH_OP)){
            if((CurrentCard.wCardType!=PCI_9118DG)&&
                (CurrentCard.wCardType!=PCI_9118HG)&&
                (CurrentCard.wCardType!=PCI_9118HR))
            {
                write_data_to_file(
                    (U16 *)CurrentCard_AIFunc.pView[0],
                    ad_file,
                    (U16)((CurrentCard_AIFunc.wResoluation+7)/8),
                    FALSE,
                    CurrentCard.AI_Status.dwTransferCount
                );
            }
            else{  //9118
                write_data_to_file(
                    (U16 *)CurrentCard_AIFunc.pView[0],
                    ad_file,
                    (U16)((CurrentCard_AIFunc.wResoluation+7)/8),
                    TRUE,
                    CurrentCard.AI_Status.dwTransferCount
                );
            }
            fclose(ad_file);
        }
    }
    return NoError;
}


/*pwBuffer is of no use while double-buffer mode enabled*/
I16 AI_ContScanChannelsToFile(
    U16 wCardNumber,
    U16 wChannel,
    U16 wAdRange,
    U8  *fileName,
    U32 dwReadCount,
    F64 fSampleRate,
    U16 wSyncMode
    )
{
    I16 err;
    U16 wGain;
    U16 Chans[256], Ranges[256];
    U32 i;
    U16 byte_width;
    char svFileName[128];
    FILE *ad_file;
    DAS_IOT_CONFIG iotConf;
    DAS_IOT_CONT   iotCont;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));
    memset(&iotCont, '\0', sizeof(DAS_IOT_CONT));

    //00/02/03 jeffrey, re-asign the iotConf value
    iotConf = CurrentCard.AI_Status.CurrentConfig;

    if((CurrentCard.wCardType==PCI_9118DG)||
        (CurrentCard.wCardType==PCI_9118HG)||
        (CurrentCard.wCardType==PCI_9118HR)||
        (CurrentCard.wCardType==PCI_9116)||
        (CurrentCard.wCardType==PCI_9221)||
        (CurrentCard.wCardType==PCI_9222)||
        (CurrentCard.wCardType==PCI_9223)||
        (CurrentCard.wCardType==PCI_9524)||
        (CurrentCard.wCardType==PCI_922A))
    {
        if(CurrentCard.wCardType==PCI_9524){
            if(wChannel<4){
                for(i=0; i<=wChannel; i++){
                    Chans[i] = (U16)i;
                    Ranges[i] = wAdRange;
                }
            }
            else if(wChannel<8){
                wChannel -= 4;
                for(i=0; i<=wChannel; i++){
                    Chans[i] = (U16)(i+4);
                    Ranges[i] = wAdRange;
                }
            }
            else{
                return ErrorInvalidIoChannel;
            }
        }
        else{
            for(i=0; i<=wChannel; i++){
                Chans[i] = (U16)i;
                Ranges[i] = wAdRange;
            }
        }

        return (AI_ContReadMultiChannelsToFile(wCardNumber, wChannel+1, Chans, \
            Ranges, fileName, dwReadCount, fSampleRate, wSyncMode));
    }

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 1, &wChannel,
        dwReadCount*(U32)((CurrentCard_AIFunc.wResoluation+7)/8), TRUE))
        !=NoError)
    {
        return err;
    }
    if(!CurrentCard_AIFunc.wFlag.bAutoScan)
        return ErrorFuncNotSupport;

    //Check wAdRange and get Gain
    if(!validate_ad_range(wCardNumber, wAdRange, &wGain))
        return ErrorInvalidAdRange;

    if(CurrentCard.AI_Status.bDblBufferMode)
        dwReadCount /= 2;

    sprintf(svFileName, "%s.dat", fileName);
    if((ad_file = fopen(svFileName, "w+b"))==NULL)
        return ErrorOpenFile;

    CurrentCard.AI_Status.bToFile = TRUE;
    CurrentCard.AI_Status.bChanRange = 0;
    CurrentCard.AI_Status.dwTransferCount = dwReadCount;
    CurrentCard.AI_Status.wNumChans = wChannel+1;
    CurrentCard.AI_Status.pwBuffer = ad_file;
    byte_width = (CurrentCard_AIFunc.wResoluation+7)/8;

    if(((CurrentCard.wCardType==PCI_9812)||(CurrentCard.wCardType==PCI_9810))&&
        ((((iotConf.mData.wValue[0])&0x0300)==P9812_CLKSRC_EXT_SIN)||
        (((iotConf.mData.wValue[0])&0x0300)==P9812_CLKSRC_EXT_DIG)))
    {
        //iotCont.dwPacerValue = 2;
        iotCont.dwPacerValue = CurrentCard.AI_Status.dwPacerValue;
    }
    else{
        //Set PacerValue and TimerCascade
        if(!get_pacervalue(wCardNumber, fSampleRate, &iotCont.dwPacerValue))
            return ErrorInvalidSampleRate;
    }

    if((CurrentCard.wCardType!=PCI_9116)&&
        (CurrentCard.wCardType!=PCI_9114DG)&&
        (CurrentCard.wCardType!=PCI_9114HG))
    {
        //Set Timer Cascade
        iotConf = CurrentCard.AI_Status.CurrentConfig;
        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CONFIG, &(iotConf))
            !=DAS_ERR_NO)
        {
            return iotConf.wErrorCode;
        }
    }

    iotCont.dwAccessCount = dwReadCount;
    iotCont.wAutoScan = 1;
    iotCont.wSyncMode = wSyncMode;
    iotCont.wChannelCount = 1;
    iotCont.dwChannelGain[0].wChannel = wChannel;
    iotCont.dwChannelGain[0].wGain = wGain;

    write_header_to_file(wCardNumber, ad_file, byte_width, wAdRange,
        wChannel+1, wChannel, dwReadCount, fSampleRate, 0
    );

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CON_READ, &(iotCont))
        !=DAS_ERR_NO)
    {
        return iotCont.wErrorCode;
    }

    if(!(CurrentCard.AI_Status.bDblBufferMode)&&(wSyncMode==SYNCH_OP)){
        if((CurrentCard.wCardType!=PCI_9118DG)&&
            (CurrentCard.wCardType!=PCI_9118HG)&&
            (CurrentCard.wCardType!=PCI_9118HR))
        {
            write_data_to_file(
                (U16 *)CurrentCard_AIFunc.pView[0],
                ad_file,
                (U16)((CurrentCard_AIFunc.wResoluation+7)/8),
                FALSE, /*wChannel+1,*/
                CurrentCard.AI_Status.dwTransferCount
            );
        }
        else{
            /*9118 exchange data*/
            write_data_to_file(
                (U16 *)CurrentCard_AIFunc.pView[0],
                ad_file,
                (U16)((CurrentCard_AIFunc.wResoluation+7)/8),
                TRUE, /*wChannel+1,*/
                CurrentCard.AI_Status.dwTransferCount
            );
        }

        fclose(ad_file);
    }

    return NoError;
}


I16 AI_ReadChannel(
    U16 wCardNumber,
    U16 wChannel,
    U16 wAdRange,
    U16 *pwValue
    )
{
    I16 err;
    U16 wGain;
    DAS_IOT_PIO iotPio;

    memset(&iotPio, '\0', sizeof(DAS_IOT_PIO));

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 1, &wChannel,
        0L, FALSE))!=NoError)
    {
        return err;
    }

    //Check wAdRange and get Gain
    if(!validate_ad_range(wCardNumber, wAdRange, &wGain))
        return ErrorInvalidAdRange;

    if((CurrentCard.wCardType==PCI_9222)&&(wChannel>=8)){
        //PCI-9222: CH0~7=>CH0~7; CH8~15=>CH16~CH23
        iotPio.wChannel = wChannel+8;
    }
    else
        iotPio.wChannel = wChannel;
    iotPio.wLine_Gain = wGain;
    iotPio.wByteCount = (U16)((CurrentCard_AIFunc.wResoluation+7)/8);
    if((CurrentCard.wCardType==PCI_9113)||(CurrentCard.wCardType==PCI_9114DG)||
        (CurrentCard.wCardType==PCI_9114HG))
    {
        iotPio.wByteCount /= 2;
    }
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_PIO_READ, &(iotPio))
        !=DAS_ERR_NO)
    {
        return iotPio.wErrorCode;
    }
    *pwValue = iotPio.dwValue;

    return NoError;
}


I16 AI_InitialMemoryAllocated(
    U16 wCardNumber,
    U32 *MemSize
    )
{
    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    *MemSize = CurrentCard_AIFunc.dwBufferSize/1024; //in KB
    if((CurrentCard.wCardType==PCI_9221)||
        (CurrentCard.wCardType==PCI_9524)||
        (CurrentCard.wCardType==PCI_9222)||
        (CurrentCard.wCardType==PCI_9223)||
        (CurrentCard.wCardType==PCI_922A))
    {
        if(*MemSize){
            *MemSize = (*MemSize-4)*256;
        }
    }

    return NoError;
}


I16 AI_ContStatus(
    U16 wCardNumber,
    U16 *Status
    )
{
    DAS_IOT_STATUS iotStat;

    memset(&iotStat, '\0', sizeof(DAS_IOT_STATUS));

    if(wCardNumber >= MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CON_STATUS, &(iotStat))
        !=DAS_ERR_NO)
    {
        return iotStat.wErrorCode;
    }
    *Status = iotStat.mStat.wValue;

    return NoError;
}


I16 AI_VReadChannel(
    U16 wCardNumber,
    U16 wChannel,
    U16 wAdRange,
    F64 *voltage
    )
{
    I16 err;
    U16 wValue;
    U32 dwValue;

    switch(CurrentCard.wCardType)
    {
        case PCI_9524:
            if((err = AI_ReadChannel32(wCardNumber, wChannel, wAdRange,
                &dwValue))!=NoError)
            {
                return err;
            }
            ai_scale32_9524(wCardNumber, wAdRange, dwValue, voltage);
            break;
        default:
            if((err = AI_ReadChannel(wCardNumber, wChannel, wAdRange,
                &wValue))!=NoError)
            {
                return err;
            }
            ai_scale(wCardNumber, wAdRange, wValue, voltage);
            break;
    }
    return NoError;
}


I16 AI_VoltScale(
    U16 wCardNumber,
    U16 adRange,
    I16 reading,
    F64 *voltage
    )
{
    I16 err;
    U16 wGain;

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }
    //Check wAdRange and get Gain
    if(!validate_ad_range(wCardNumber, adRange, &wGain))
        return ErrorInvalidAdRange;

    ai_scale(wCardNumber, adRange, reading, voltage);

    return NoError;
}


I16 AI_ContVScale(
    U16  wCardNumber,
    U16  adRange,
    void *readingArray,
    F64  *voltageArray,
    I32  count
    )
{
    I16 err;
    U16 wGain;
    I32 i;
    U16 *p16;
    U32 *p32;

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }
    //Check wAdRange and get Gain
    if(!validate_ad_range(wCardNumber, adRange, &wGain))
        return ErrorInvalidAdRange;

    if((CurrentCard.wCardType==PCI_9113)||(CurrentCard.wCardType==PCI_9114DG)||
        (CurrentCard.wCardType==PCI_9114HG))
    {
        p32 = (U32 *)readingArray;
        for(i=0; i<count; i++)
            ai_scale(wCardNumber, adRange, (U16)p32[i], &voltageArray[i]);
    }
    else if(CurrentCard.wCardType==PCI_9524){
        p32 = (U32 *)readingArray;
        for(i=0; i<count; i++){
            ai_scale32_9524(wCardNumber, adRange, p32[i], &voltageArray[i]);
        }
    }
    else if((CurrentCard.wCardType==PCI_9222)||
        (CurrentCard.wCardType==PCI_9223)||
        (CurrentCard.wCardType==PCI_922A))
    {
        if(CurrentCard_AIFunc.wResoluation==32){
            p32 = (U32 *)readingArray;
            for(i=0; i<count; i++){
                ai_scale32(wCardNumber, adRange, p32[i], &voltageArray[i]);
            }
        }
        else{
            p16 = (U16 *)readingArray;
            for(i=0; i<count; i++){
                ai_scale(wCardNumber, adRange, p16[i], &voltageArray[i]);
            }
        }
    }
    else{
        p16 = (U16 *)readingArray;
        for(i=0; i<count; i++)
            ai_scale(wCardNumber, adRange, p16[i], &voltageArray[i]);
    }

    return(NoError);
}


int ai_scale(
    U16 wCardNumber,
    U16 adRange,
    U16 reading,
    F64 *voltage
    )
{
    I16 value;

    switch(CurrentCard.wCardType){
        case PCI_9111DG:
            value = ((I16)reading)>>4;
            switch(adRange){
                case AD_B_10_V:
                    *voltage = (F64)value/2048*10;
                    break;
                case AD_B_5_V:
                    *voltage = (F64)value/2048*5.0;
                    break;
                case AD_B_2_5_V:
                    *voltage = (F64)value/2048*2.5;
                    break;
                case AD_B_1_25_V:
                    *voltage = (F64)value/2048*1.25;
                    break;
                case AD_B_0_625_V:
                    *voltage = (F64)value/2048*0.625;
                    break;
            }
            break;
        case PCI_9111HR:
            value = (I16)reading;
            switch(adRange){
                case AD_B_10_V:
                    *voltage = (F64)value/32768*10;
                    break;
                case AD_B_5_V:
                    *voltage = (F64)value/32768*5.0;
                    break;
                case AD_B_2_5_V:
                    *voltage = (F64)value/32768*2.5;
                    break;
                case AD_B_1_25_V:
                    *voltage = (F64)value/32768*1.25;
                    break;
                case AD_B_0_625_V:
                    *voltage = (F64)value/32768*0.625;
                    break;
            }
            break;
        case PCI_9112:
            value = (reading>>4)&0x0fff;
            switch(adRange){
                case AD_B_10_V:
                    *voltage = (F64)(value-2048)/2048*10;
                    break;
                case AD_B_5_V:
                    *voltage = (F64)(value-2048)/2048.0*5.0;
                    break;
                case AD_B_2_5_V:
                    *voltage = (F64)(value-2048)/2048*2.5;
                    break;
                case AD_B_1_25_V:
                    *voltage = (F64)(value-2048)/2048*1.25;
                    break;
                case AD_B_0_625_V:
                    *voltage = (F64)(value-2048)/2048*0.625;
                    break;
                case AD_U_10_V:
                    *voltage = (F64)value/4095*10;
                    break;
                case AD_U_5_V:
                    *voltage = (F64)value/4095*5;
                    break;
                case AD_U_2_5_V:
                    *voltage = (F64)value/4095*2.5;
                    break;
                case AD_U_1_25_V:
                    *voltage = (F64)value/4095*1.25;
                    break;
            }
            break;
        case PCI_9113:
            value = (reading&0x0fff);
            switch(adRange){
                case AD_B_10_V:
                    *voltage = (F64)(value-2048)/2048*10;
                    break;
                case AD_B_1_V:
                    *voltage = (F64)(value-2048)/2048*1;
                    break;
                case AD_B_0_1_V:
                    *voltage = (F64)(value-2048)/2048*0.1;
                    break;
                case AD_B_5_V:
                    *voltage = (F64)(value-2048)/2048.0*5.0;
                    break;
                case AD_B_0_5_V:
                    *voltage = (F64)(value-2048)/2048*0.5;
                    break;
                case AD_B_0_05_V:
                    *voltage = (F64)(value-2048)/2048*0.05;
                    break;
                case AD_U_10_V:
                    *voltage = (F64)value/4095*10;
                    break;
                case AD_U_1_V:
                    *voltage = (F64)value/4095*1;
                    break;
                case AD_U_0_1_V:
                    *voltage = (F64)value/4095*0.1;
                    break;
            }
            break;
        case PCI_9114DG:
            value = (I16)reading;
            switch(adRange){
                case AD_B_10_V:
                    *voltage = (F64)value/32768*10;
                    break;
                case AD_B_5_V:
                    *voltage = (F64)value/32768*5.0;
                    break;
                case AD_B_2_5_V:
                    *voltage = (F64)value/32768*2.5;
                    break;
                case AD_B_1_25_V:
                    *voltage = (F64)value/32768*1.25;
                    break;
            }
            break;
        case PCI_9114HG:
            value = (I16)reading;
            switch(adRange){
                case AD_B_10_V:
                    *voltage = (F64)value/32768*10;
                    break;
                case AD_B_1_V:
                    *voltage = (F64)value/32768*1.0;
                    break;
                case AD_B_0_1_V:
                    *voltage = (F64)value/32768*0.1;
                    break;
                case AD_B_0_01_V:
                    *voltage = (F64)value/32768*0.01;
                    break;
            }
            break;
        case PCI_9116:
            value = (I16)reading;
            switch(adRange) {
                case AD_B_5_V:
                    *voltage = (F64)value/32768*5.0;
                    break;
                case AD_B_2_5_V:
                    *voltage = (F64)value/32768*2.5;
                    break;
                case AD_B_1_25_V:
                    *voltage = (F64)value/32768*1.25;
                    break;
                case AD_B_0_625_V:
                    *voltage = (F64)value/32768*0.625;
                    break;
                case AD_U_10_V:
                    *voltage = ((F64)value+32768)/65535*10;
                    break;
                case AD_U_5_V:
                    *voltage = ((F64)value+32768)/65535*5;
                    break;
                case AD_U_2_5_V:
                    *voltage = ((F64)value+32768)/65535*2.5;
                    break;
                case AD_U_1_25_V:
                    *voltage = ((F64)value+32768)/65535*1.25;
                    break;
            }
            break;
        case PCI_9118DG:
            value = (reading>>4)&0x0fff;
            switch(adRange){
                case AD_B_5_V:
                    *voltage = (F64)(value-2048)/2048*5.0;
                    break;
                case AD_B_2_5_V:
                    *voltage = (F64)(value-2048)/2048*2.5;
                    break;
                case AD_B_1_25_V:
                    *voltage = (F64)(value-2048)/2048*1.25;
                    break;
                case AD_B_0_625_V:
                    *voltage = (F64)(value-2048)/2048*0.625;
                    break;
                case AD_U_10_V:
                    *voltage = (F64)value/4095*10;
                    break;
                case AD_U_5_V:
                    *voltage = (F64)value/4095*5;
                    break;
                case AD_U_2_5_V:
                    *voltage = (F64)value/4095*2.5;
                    break;
                case AD_U_1_25_V:
                    *voltage = (F64)value/4095*1.25;
                    break;
            }
            break;
        case PCI_9118HG:
            value = (reading>>4)&0x0fff;
            switch(adRange){
                case AD_B_5_V:
                    *voltage = (F64)(value-2048)/2048*5.0;
                    break;
                case AD_B_0_5_V:
                    *voltage = (F64)(value-2048)/2048*0.5;
                    break;
                case AD_B_0_05_V:
                    *voltage = (F64)(value-2048)/2048*0.05;
                    break;
                case AD_B_0_005_V:
                    *voltage = (F64)(value-2048)/2048*0.005;
                    break;
                case AD_U_10_V:
                    *voltage = (F64)value/4095*10;
                    break;
                case AD_U_1_V:
                    *voltage = (F64)value/4095*1;
                    break;
                case AD_U_0_1_V:
                    *voltage = (F64)value/4095*0.1;
                    break;
                case AD_U_0_01_V:
                    *voltage = (F64)value/4095*0.01;
                    break;
            }
            break;
        case PCI_9118HR:
            value = (I16)reading;
            switch(adRange){
                case AD_B_5_V:
                    *voltage = (F64)value/32768*5.0;
                    break;
                case AD_B_2_5_V:
                    *voltage = (F64)value/32768*2.5;
                    break;
                case AD_B_1_25_V:
                    *voltage = (F64)value/32768*1.25;
                    break;
                case AD_B_0_625_V:
                    *voltage = (F64)value/32768*0.625;
                    break;
                case AD_U_10_V:
                    *voltage = ((F64)value+32768)/65535*10;
                    break;
                case AD_U_5_V:
                    *voltage = ((F64)value+32768)/65535*5;
                    break;
                case AD_U_2_5_V:
                    *voltage = ((F64)value+32768)/65535*2.5;
                    break;
                case AD_U_1_25_V:
                    *voltage = ((F64)value+32768)/65535*1.25;
                    break;
            }
            break;
        case PCI_9812:
            value = ((I16)reading)>>4;
            switch(adRange){
                case AD_B_1_V:
                    *voltage = (F64)value/2048*1.0;
                    break;
                case AD_B_5_V:
                    *voltage = (F64)value/2048*5.0;
                    break;
            }
            break;
        case PCI_9810:
            value = ((I16)reading)>>6;
            switch(adRange){
                case AD_B_1_V:
                    *voltage = (F64)value/512*1.0;
                    break;
                case AD_B_5_V:
                    *voltage = (F64)value/512*5.0;
                    break;
            }
            break;
        case PCI_9221:
            value = (I16)reading;
            switch(adRange){
                case AD_B_5_V:
                    *voltage = (F64)value/32768*5.0;
                    break;
                case AD_B_1_V:
                    *voltage = (F64)value/32768*1.0;
                    break;
                case AD_B_0_5_V:
                    *voltage = (F64)value/32768*0.5;
                    break;
                case AD_B_0_2_V:
                    *voltage = (F64)value/32768*0.2;
                    break;
            }
            break;
        case PCI_9222:
        case PCI_9223:
            value = (I16)reading;
            switch(adRange){
                case AD_B_10_V:
                    *voltage = (F64)value/32768*10.0;
                    break;
                case AD_B_5_V:
                    *voltage = (F64)value/32768*5.0;
                    break;
                case AD_B_2_5_V:
                    *voltage = (F64)value/32768*2.5;
                    break;
                case AD_B_2_V:
                    *voltage = (F64)value/32768*2.0;
                    break;
                case AD_B_1_25_V:
                    *voltage = (F64)value/32768*1.25;
                    break;
                case AD_B_1_V:
                    *voltage = (F64)value/32768*1.0;
                    break;
                case AD_B_0_5_V:
                    *voltage = (F64)value/32768*0.5;
                    break;
                case AD_B_0_25_V:
                    *voltage = (F64)value/32768*0.25;
                    break;
            }
            break;
        case PCI_922A:
            value = (I16)reading;
            switch(adRange){
                case AD_B_10_V:
                    *voltage = (F64)value/32768*10.0;
                    break;
                case AD_B_5_V:
                    *voltage = (F64)value/32768*5.0;
                    break;
            }
            break;
    }
    return NoError;
}


int ai_get_compensation_9524(
    U16 wCardNumber,
    U16 Group,
    U16 ADC_Gain,
    U16 ADC_Speed,
    F64 *res_offset,
    F64 *scaling
    )
{
    U16 LU_Gain = 0;
    U16 LU_Speed = 0;
    int i;
    DAS_IOT_CALCONST_9524 *p9524_Cal;
    ROM_SysAI0IntCalConst *SysAI0IntCalConst;
    ROM_SysAI1IntCalConst *SysAI1IntCalConst;

    U16 LU_Gain_TB[4] = {1, 2, 3, 4};
    U16 LU_Speed_TB[16] = {
        0x03, 0x13, 0x23, 0x33, 0x43, 0x53, 0x63, 0x72,
        0x82, 0x92, 0xa1, 0xb0, 0xc0, 0xd0, 0xe0, 0xf0
    };

    p9524_Cal = (DAS_IOT_CALCONST_9524 *)CurrentCard.pCAL_Const;
    SysAI0IntCalConst = &(p9524_Cal->SysAI0IntCalConst);
    SysAI1IntCalConst = &(p9524_Cal->SysAI1IntCalConst);

    if(Group==0){ //Load Cell Group
        if(ADC_Gain!=0)
            return ErrorInvalidAdRange;
        LU_Gain = 0; //Load Cell always 0
        for(i=0; i<16; i++){
            if(ADC_Speed==LU_Speed_TB[i]){
                LU_Speed = i;
                break;
            }
            if(i==15)
                return ErrorInvalidSampleRate;
        }
        *res_offset = SysAI0IntCalConst->EE_RESIDUAL_OFFSET[LU_Gain][LU_Speed];
        *scaling = SysAI0IntCalConst->EE_SCALING[LU_Gain][LU_Speed];
    }
    else if(Group==1){ //General Purpose Group
        for(i=0; i<4; i++){
            if(ADC_Gain==LU_Gain_TB[i]){
                LU_Gain = i;
                break;
            }
            if(i==3)
                return ErrorInvalidAdRange;
        }
        for(i=0; i<16; i++){
            if(ADC_Speed==LU_Speed_TB[i]){
                LU_Speed = i;
                break;
            }
            if(i==15)
                return ErrorInvalidSampleRate;
        }
        *res_offset = SysAI1IntCalConst->EE_RESIDUAL_OFFSET[LU_Gain][LU_Speed];
        *scaling = SysAI1IntCalConst->EE_SCALING[LU_Gain][LU_Speed];
    }
    else
        return ErrorFuncNotSupport;

    return NoError;
}


int ai_scale32_9524(
    U16 wCardNumber,
    U16 wAdRange,
    U32 reading,
    F64 *voltage
    )
{
    DAS_IOT_CONFIG iotConf;
    U16 wGain;
    U16 wSpeed;
    I32 value;
    F64 res_offset;
    F64 scaling;
    I16 err;
    /*
    U16 LU_Gain_TB[4] = {1, 2, 3, 4};
    U16 LU_Speed_TB[16] = {
        0x03, 0x13, 0x23, 0x33, 0x43, 0x53, 0x63, 0x72,
        0x82, 0x92, 0xa1, 0xb0, 0xc0, 0xd0, 0xe0, 0xf0
    };
    U16 factor[5] = {1, 1, 2, 4, 8};
    */

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));
    value = ((I32)reading)>>8;

    if(!validate_ad_range(wCardNumber, wAdRange, &wGain))
        return ErrorInvalidAdRange;

    if(wGain==0){ //Load Cell Group
        iotConf = CurrentCard.AI_Status.CurrentConfig;
        wSpeed = iotConf.mData.m9524AI.ADC_Speed;

        if((err = (I16)ai_get_compensation_9524(wCardNumber, 0/*group*/,
            wGain, wSpeed, &res_offset, &scaling))<0)
        {
            return err;
        }
    }
    else{ //General Purpose Group
        iotConf.mData.m9524AI = *((DAS_9524AI_CONFIG *)(\
            &CurrentCard.AI_Status.dwPacerValue)
        );
        wSpeed = iotConf.mData.m9524AI.ADC_Speed;

        if((err = (I16)ai_get_compensation_9524(wCardNumber, 1/*group*/,
            wGain, wSpeed, &res_offset, &scaling))<0)
        {
            return err;
        }
    }
    *voltage = (((F64)value-res_offset)*scaling)/*/factor[wGain]*/;

    return NoError;
}


int ai_scale32(
    U16 wCardNumber,
    U16 adRange,
    U32 reading,
    F64 *voltage
    )
{
    //I32 dwvalue = 0;
    I16 wvalue = 0;

    switch(CurrentCard.wCardType){
        case PCI_9222:
        case PCI_9223:
            wvalue = (I16)(reading&0xffff);
            switch(adRange){
                case AD_B_10_V:
                    *voltage = (double)wvalue/32768*10.0;
                    break;
                case AD_B_5_V:
                    *voltage = (double)wvalue/32768*5.0;
                    break;
                case AD_B_2_5_V:
                    *voltage = (double)wvalue/32768*2.5;
                    break;
                case AD_B_2_V:
                    *voltage = (double)wvalue/32768*2.0;
                    break;
                case AD_B_1_25_V:
                    *voltage = (double)wvalue/32768*1.25;
                    break;
                case AD_B_1_V:
                    *voltage = (double)wvalue/32768*1.0;
                    break;
                case AD_B_0_5_V:
                    *voltage = (double)wvalue/32768*0.5;
                    break;
                case AD_B_0_25_V:
                    *voltage = (double)wvalue/32768*0.25;
                    break;
            }
            break;
        case PCI_922A:
            wvalue = (I16)(reading&0xffff);
            switch(adRange){
                case AD_B_10_V:
                    *voltage = (double)wvalue/32768*10.0;
                    break;
                case AD_B_5_V:
                    *voltage = (double)wvalue/32768*5.0;
                    break;
            }
            break;
        default:
            return ErrorFuncNotSupport;
    }

    return NoError;
}


int write_data_to_file(
    U16     *buf,
    FILE    *fp,
    U16     byte_cnt,
    BOOLEAN exchange,
    U32     count
    )
{
    U16 *p16;

    if(exchange){
        if(byte_cnt==2){ //16-bit
            p16 = buf;
            exch_data(p16, count);
            fwrite(buf, byte_cnt, count, fp);
        }
    }
    else{
        fwrite(buf, byte_cnt, count, fp);
    }

    return 0;
}


void exch_data(
    U16 *tbuf,
    U32 cnt
    )
{
    U16 temp;
    int i;

    for(i=0; i<cnt; i++){
        if(!(i%2)){
            temp = tbuf[i];
            tbuf[i] = tbuf[i+1];
            tbuf[i+1] = temp;
        }
    }

    return;
}


I16 AI_AsyncDblBufferOverrun(
    U16 wCardNumber,
    U16 op,
    U16 *overrunFlag
    )
{
    I16 err;
    volatile PPCI_DBXORUN pOvRun;
    //DAS_IOT_CALCONST_9221 *calPtr;

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    if(CurrentCard.wCardType==PCI_9221){
        //calPtr = (DAS_IOT_CALCONST_9221 *)CurrentCard.pCAL_Const;
        //pOvRun = &(calPtr->OvRun);
        pOvRun = &(((DAS_IOT_CALCONST_9221 *)CurrentCard.pCAL_Const)->OvRun);
    }
    else if(CurrentCard.wCardType==PCI_9524){
        pOvRun = &(((DAS_IOT_CALCONST_9524 *)CurrentCard.pCAL_Const)->OvRun);
    }
    else if((CurrentCard.wCardType==PCI_9222)||
        (CurrentCard.wCardType==PCI_9223)||
        (CurrentCard.wCardType==PCI_922A))
    {
        //01/26/08
        pOvRun = \
            &(((DAS_IOT_CALCONST_922X *)CurrentCard.pCAL_Const)->OvRun[0]);
    }
    //yuan add 09/11/02
    else if(((CurrentCard.wCardType>=PCI_9111DG)&&
        (CurrentCard.wCardType<=PCI_9812))||
        (CurrentCard.wCardType==PCI_9116))
    {
        pOvRun = (volatile PPCI_DBXORUN)(CurrentCard_AOFunc.pView[0]);
    }
    else{
        return ErrorFuncNotSupport;
    }

    switch(op){
        case 0: //check
            *overrunFlag = pOvRun->xFlag.f.overrun;
            break;
        case 1: //clear
            memset(pOvRun, '\0', sizeof(PCI_DBXORUN));
            break;
    }

    return NoError;
}


I16 AI_9221_Config(
    U16     wCardNumber,
    U16     ConfigCtrl,
    U16     TrigCtrl,
    BOOLEAN AutoResetBuf
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    iotConf.mData.m9221AI.fInputType = ConfigCtrl&0x03;
    iotConf.mData.m9221AI.fTimeBaseMode = (ConfigCtrl>>7)&0x01;
    iotConf.mData.m9221AI.fTimeBaseSrc = (ConfigCtrl>>4)&0x07;
    iotConf.mData.m9221AI.fTriggerMode = (TrigCtrl>>3)&0x01;
    iotConf.mData.m9221AI.fTrigSrc = TrigCtrl&0x07;
    iotConf.mData.m9221AI.fTrigPol = (TrigCtrl>>4)&0x1;
    iotConf.mData.m9221AI.fAutoResetBuf = AutoResetBuf;
    CurrentCard.AI_Status.CurrentConfig = iotConf;
    if(iotConf.mData.m9221AI.fInputType==P9221_AI_Differential)
        CurrentCard_AIFunc.wChannelCount = 8;
    else
        CurrentCard_AIFunc.wChannelCount = 16;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }

    //CurrentCard.AI_Status.CurrentConfig.wErrorCode = NoError;
    return NoError;
}


I16 AI_9221_CounterInterval(
    U16 wCardNumber,
    U32 ScanIntrv,
    U32 SampIntrv
    )
{
    I16 err;
    DAS_IOT_SI iotSI;

    memset(&iotSI, '\0', sizeof(DAS_IOT_SI));

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    if((ScanIntrv==0)||(SampIntrv==0)||(ScanIntrv<SampIntrv))
        return ErrorInvalidCounterValue;

    iotSI.dwScanInterval = ScanIntrv;
    iotSI.dwSampInterval = SampIntrv;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_9116SC_CONFIG, &(iotSI))
        !=DAS_ERR_NO)
    {
        return iotSI.wErrorCode;
    }
    CurrentCard.AI_Status.dwPacerValue = ScanIntrv;
    CurrentCard.AI_Status.dwSampleIntv = SampIntrv;

    return NoError;
}


I16 AI_ScanReadChannels(
    U16 wCardNumber,
    U16 wChannel,
    U16 wAdRange,
    U16 *pwBuffer
    )
{
    int i;
    U16 pwChans[256];
    U16 pwAdRanges[256];

    for(i=0; i<=wChannel; i++){
        *(pwChans+i) = (U16)i;
        *(pwAdRanges+i) = wAdRange;
    }

    return (AI_ReadMultiChannels(wCardNumber, wChannel+1, pwChans,
        pwAdRanges, pwBuffer));
}


I16 AI_ReadMultiChannels(
    U16 wCardNumber,
    U16 wNumChans,
    U16 *pwChans,
    U16 *pwAdRanges,
    U16 *pwBuffer
    )
{
    U16 wGain;
    U16 chNum;
    DAS_IOT_SIMU_PIO iotSPio;
    int i;
    I16 err;

    chNum = wNumChans;
    memset(&iotSPio, '\0', sizeof(DAS_IOT_SIMU_PIO));

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, wNumChans,
        pwChans, 0, FALSE))!=NoError)
    {
        return err;
    }

    iotSPio.wChannelCount = chNum;
    iotSPio.wDirectory = 0;
    for(i=0; i<chNum; i++){
        if((CurrentCard.wCardType==PCI_9222)&&(pwChans[i]>=8)){
            //PCI-9222: CH0~7=>CH0~7; CH8~15=>CH16~CH23
            iotSPio.dwChannelGain[i].wChannel = pwChans[i]+8;
        }
        else
            iotSPio.dwChannelGain[i].wChannel = pwChans[i];
        if(!validate_ad_range(wCardNumber, pwAdRanges[i], &wGain))
            return ErrorInvalidAdRange;
        iotSPio.dwChannelGain[i].wGain = wGain;
    }

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_SIMU_AI_PIO, &(iotSPio))
        !=DAS_ERR_NO)
    {
        return iotSPio.wErrorCode;
    }

    memcpy(
        pwBuffer,
        (U16 *)(iotSPio.dwChannelGain),
        ((CurrentCard_AIFunc.wResoluation+7)/8)*chNum
    );

    return NoError;
}


I16 AI_ContBufferSetup(
    U16  wCardNumber,
    void *pwBuffer,
    U32  dwReadCount,
    U16  *BufferId
    )
{
    I16 err;
    DAS_IOT_MBUF2 iotMbuf;

    memset(&iotMbuf, '\0', sizeof(DAS_IOT_MBUF2));

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL,
        dwReadCount, TRUE))!=NoError)
    {
        return err;
    }

    iotMbuf.dwBufferAddr = (U32)pwBuffer;
    iotMbuf.dwAccessCount = dwReadCount;
    iotMbuf.wDir = 0;
    iotMbuf.wBufferId =1;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_MBF_SET, &(iotMbuf))
        !=DAS_ERR_NO)
    {
        return iotMbuf.wErrorCode;
    }
    *BufferId = iotMbuf.wBufferId;
    CurrentCard.AI_Status.pDAQbuf[*BufferId] = pwBuffer;
    CurrentCard.AI_Status.dwBufCount[*BufferId] = dwReadCount;

    return NoError;
}


I16 AI_ContBufferReset(
    U16 wCardNumber
    )
{
    I16 err;
    DAS_IOT_MBUF2 iotMbuf;

    memset(&iotMbuf, '\0', sizeof(DAS_IOT_MBUF2));

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_MBF_RESET, &(iotMbuf))
        !=DAS_ERR_NO)
    {
        return iotMbuf.wErrorCode;
    }
    CurrentCard.AI_Status.pDAQbuf[0] = NULL;
    CurrentCard.AI_Status.pDAQbuf[1] = NULL;
    CurrentCard.AI_Status.dwBufCount[0] = 0;
    CurrentCard.AI_Status.dwBufCount[1] = 0;

    return NoError;
}


I16 AI_SetTimeOut(
    U16 wCardNumber,
    U32 dwTimeOut
    )
{
    I16 err;
    DAS_IOT_SI iotSI;

    memset(&iotSI, '\0', sizeof(DAS_IOT_SI));

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    iotSI.dwScanInterval = dwTimeOut;
    iotSI.dwSampInterval = 0;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_TIMEOUT_SET, &(iotSI))
        !=DAS_ERR_NO)
    {
        return iotSI.wErrorCode;
    }

    return NoError;
}


I16 AI_AsyncDblBufferToFile(
    U16 wCardNumber
    )
{
    I16 err;
    volatile PPCI_DBXORUN pOvRun;
    U16 BufIdx;

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    //if(!CurrentCard.AI_Status.bDblBufferMode)
    //    return ErrorNotDoubleBufferMode;

    CurrentCard.AI_Status.wDblBufferIndex &= 0x01;
    BufIdx = CurrentCard.AI_Status.wDblBufferIndex;
    write_data_to_file(
        (U16 *)CurrentCard.AI_Status.pDAQbuf[BufIdx],
        (FILE *)CurrentCard.AI_Status.pwBuffer,
        (U16)((CurrentCard_AIFunc.wResoluation+7)/8),
        FALSE,
        CurrentCard.AI_Status.dwTransferCount
    );

    switch(CurrentCard.wCardType){
        case PCI_9221:
            pOvRun = \
                &(((DAS_IOT_CALCONST_9221 *)CurrentCard.pCAL_Const)->OvRun);
            break;
        case PCI_9222:
        case PCI_9223:
        case PCI_922A:
            pOvRun = \
                &(((DAS_IOT_CALCONST_922X *)CurrentCard.pCAL_Const)->OvRun[0]);
            break;
        case PCI_9524:
            pOvRun = \
                &(((DAS_IOT_CALCONST_9524 *)CurrentCard.pCAL_Const)->OvRun);
            break;
        default:
            return ErrorFuncNotSupport;
    }

    if(pOvRun){
        if(!pOvRun->xFlag.f.overrun){
            pOvRun->xFlag.f.lock = 1;
            if(pOvRun->dwCtr_isr[CurrentCard.AI_Status.wDblBufferIndex]>=1)
                pOvRun->dwCtr_isr[CurrentCard.AI_Status.wDblBufferIndex] -= 1;
            pOvRun->xFlag.f.lock = 0;
        }
    }

    return NoError;
}


I16 AI_AsyncDblBufferHandled(
    U16 wCardNumber
    )
{
    I16 err;
    volatile PPCI_DBXORUN pOvRun;

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    if(!CurrentCard.AI_Status.bDblBufferMode)
        return ErrorNotDoubleBufferMode;

    switch(CurrentCard.wCardType){
        case PCI_9221:
            pOvRun = \
                &(((DAS_IOT_CALCONST_9221 *)CurrentCard.pCAL_Const)->OvRun);
            break;
        case PCI_9524:
            pOvRun = \
                &(((DAS_IOT_CALCONST_9524 *)CurrentCard.pCAL_Const)->OvRun);
            break;
        case PCI_9222:
        case PCI_9223:
        case PCI_922A:
            pOvRun = \
                &(((DAS_IOT_CALCONST_922X *)CurrentCard.pCAL_Const)->OvRun[0]);
            break;
        default:
            return ErrorFuncNotSupport;
    }

    if(pOvRun){
        if(!pOvRun->xFlag.f.overrun){
            pOvRun->xFlag.f.lock = 1;
            if(pOvRun->dwCtr_isr[CurrentCard.AI_Status.wDblBufferIndex]>=1)
                pOvRun->dwCtr_isr[CurrentCard.AI_Status.wDblBufferIndex] -= 1;
            pOvRun->xFlag.f.lock = 0;
        }
    }

    return NoError;
}


I16 AI_9524_SetDSP(
    U16 wCardNumber,
    U16 wChannel,         /*load cell channel number*/
    U16 wMode,            /*enable or disable spk_rej*/
    U16 wDFStage,         /*dsp stage*/
    U32 dwSPKRejThreshold /*reject threshold*/
    )
{
    DAS_IOT_PIO iotPio;

    memset(&iotPio, '\0', sizeof(DAS_IOT_PIO));

    if(CurrentCard.wCardType!=PCI_9524)
        return ErrorFuncNotSupport;
    /*SHOULD ADD ERROR CHECK IN HERE*/
    if(wChannel>3)
        return ErrorInvalidIoChannel;
    if(wMode>1)
        return ErrorUndefinedParameter;
    if(wDFStage>10)
        return ErrorUndefinedParameter;
    if((!dwSPKRejThreshold)||(dwSPKRejThreshold>0xffffff))
        return ErrorUndefinedParameter;

    //iotPio.wErrorCode = ErrorConfigIoctl;
    /*0x8000, let driver know it's a DSP setting*/
    iotPio.wChannel = (wChannel|0x8000);
    /*enable or disable SPK_REJ*/
    iotPio.wLine_Gain = wMode;
    /*DSP TAP*/
    iotPio.wByteCount = wDFStage;
    /*REJ Threshold*/
    iotPio.dwValue = dwSPKRejThreshold;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_PIO_READ, &(iotPio))
        !=DAS_ERR_NO)
    {
        return iotPio.wErrorCode;
    }

    return NoError;
}


I16 AI_9524_Config(
    U16 wCardNumber,
    U16 wGroup,      /*Load Cell group or General Purpose group*/
    U16 wXMode,      /*transfer mode, polling or dma mode*/
    U16 wConfigCtrl, /*Config*/
    U16 wTrigCtrl,   /*Trigger*/
    U32 dwTrigValue  /*Trigger Value*/
    )
{
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if(CurrentCard.wCardType!=PCI_9524)
        return ErrorFuncNotSupport;
    if(wGroup>1)
        return ErrorUndefinedParameter;
    if(wXMode>1)
        return ErrorUndefinedParameter;

    /*Group0: Load Cell or Group1: General Purpose*/
    iotConf.mData.m9524AI.fGroup = wGroup&0x1;
    /*AI transfer mode - 0: Poll, 1: DMA*/
    iotConf.mData.m9524AI.fXMode = wXMode&0x1;
    /*the folling three parameters are only valid in LC Group*/
    if(wGroup==0){
        /*0: Exec. voltage 5V, 1: Exec. voltage 10V*/
        iotConf.mData.m9524AI.fExecV = (wConfigCtrl)&0x1;
        /*0: Local sense, 1: remove sense*/
        iotConf.mData.m9524AI.fRMTen = (wConfigCtrl>>1)&0x1;
        /*enable AZ*/
        iotConf.mData.m9524AI.fAZEn = (wConfigCtrl>>2)&0x1;
    }
    if(wXMode==1){
        /*the folling parameter are only valid in DMA mode*/
        /*enable buffer autoreset*/
        iotConf.mData.m9524AI.fAutoResetBuf = (wConfigCtrl>>3)&0x1;
    }
    else{
        //06/09/08 add for AI EOC
        iotConf.mData.m9524AI.fEOCInt = (wConfigCtrl>>4)&0x1;
    }
    /*Trigger ctrl*/
    iotConf.mData.m9524AI.fTrigSrc = wTrigCtrl&0x7;
    iotConf.mData.m9524AI.fTrigPol = (wTrigCtrl>>3)&0x1;
    /*does not support re-trigger now*/
    iotConf.mData.m9524AI.fReTrig = (wTrigCtrl>>4)&0x1;
    /*Trigger Count*/
    //While trigger source is set to PG0 or QD0, this parameter means
    //the number of pulse generated or the number of motion encoded
    iotConf.mData.m9524AI.dwTrigCount = dwTrigValue;
    //iotConf.wErrorCode = ErrorConfigIoctl;
    iotConf.mData.m9524AI.Reserved1 = 0;
    if(wGroup==0){
        CurrentCard.AI_Status.CurrentConfig = iotConf;
    }
    else{
        /*use the two valuables to store the config of General Purpose group*/
        CurrentCard.AI_Status.dwPacerValue = iotConf.mData.dwValue[0];
        CurrentCard.AI_Status.dwSampleIntv = iotConf.mData.dwValue[1];
    }
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }

    //CurrentCard.AI_Status.CurrentConfig.wErrorCode = NoError;
    return NoError;
}


I16 AI_9524_PollConfig(
    U16 wCardNumber,
    U16 wGroup,
    U16 wPollChannel,
    U16 wPollRange,
    U16 wPollSpeed
    )
{
    DAS_IOT_CONFIG iotConf;
    U32 dwSampleRate;
    U16 wGain;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if(CurrentCard.wCardType!=PCI_9524)
        return ErrorFuncNotSupport;
    if(!validate_ad_range(wCardNumber, wPollRange, &wGain))
        return ErrorInvalidAdRange;
    /*ADC Speed supports 16 speed scaler*/
    if(wPollSpeed>15)
        return ErrorInvalidSampleRate;
    /*Load Cell Channel 0 ~ 3; General Purpose Channel 0 ~ 3; ScanAllChannels*/
    //10/19/09 add 9, 10, 11
    /*
     * 0~3 : for one channel 0~3 of load cell group
     * 4~7 : for one channel 0~3 of general purpose group
     * 8   : for scanning channel 0 to 3
     * 9   : for scanning channel 0 to 2
     * 10  : for scanning channel 0 to 1
     * 11  : for scanning channel 0 to 0
     */
    if(wPollChannel>11)
        return ErrorInvalidIoChannel;

    /*Load Cell group only supports Gain x25(0)*/
    /*General Purpose group supports Gain x1(1), x2(2), x4(3), x8(4)*/
    if(wGroup==0){
        iotConf = CurrentCard.AI_Status.CurrentConfig;
        if(wGain)
            return ErrorInvalidAdRange;
        if((wPollChannel>3)&&(wPollChannel<8))
            return ErrorInvalidIoChannel;
    }
    else if(wGroup==1){
        iotConf.mData.dwValue[0] = CurrentCard.AI_Status.dwPacerValue;
        iotConf.mData.dwValue[1] = CurrentCard.AI_Status.dwSampleIntv;
        if((!wGain)||(wGain>4))
            return ErrorInvalidAdRange;
        if(wPollChannel<4)
            return ErrorInvalidIoChannel;
    }
    else{
        return ErrorUndefinedParameter;
    }

    if(iotConf.mData.m9524AI.fXMode!=0)
        return ErrorConfigFailed;

    /*Group0: Load Cell or Group1: General Purpose*/
    iotConf.mData.m9524AI.fGroup = wGroup&0x1;
    /*AutoScan*/
    //02/19/08
    if(wPollChannel==8){
        iotConf.mData.m9524AI.fAutoScan = 1;
        if(iotConf.mData.m9524AI.fGroup==0){
            /*autoscan biggest channel number*/
            iotConf.mData.m9524AI.ADC_Channel = 3;
        }
        else{
            /*autoscan biggest channel number*/
            iotConf.mData.m9524AI.ADC_Channel = 7;
        }
    }
    //10/19/09
    else if(wPollChannel==9){
        iotConf.mData.m9524AI.fAutoScan = 1;
        if(iotConf.mData.m9524AI.fGroup==0){
            /*autoscan biggest channel number*/
            iotConf.mData.m9524AI.ADC_Channel = 2;
        }
        else{
            /*autoscan biggest channel number*/
            iotConf.mData.m9524AI.ADC_Channel = 6;
        }
    }
    //10/19/09
    else if(wPollChannel==10){
        iotConf.mData.m9524AI.fAutoScan = 1;
        if(iotConf.mData.m9524AI.fGroup==0){
            /*autoscan biggest channel number*/
            iotConf.mData.m9524AI.ADC_Channel = 1;
        }
        else{
            /*autoscan biggest channel number*/
            iotConf.mData.m9524AI.ADC_Channel = 5;
        }
    }
    //10/19/09
    else if(wPollChannel==11){
        iotConf.mData.m9524AI.fAutoScan = 1;
        if(iotConf.mData.m9524AI.fGroup==0){
            /*autoscan biggest channel number*/
            iotConf.mData.m9524AI.ADC_Channel = 0;
        }
        else{
            /*autoscan biggest channel number*/
            iotConf.mData.m9524AI.ADC_Channel = 4;
        }
    }
    else{
        iotConf.mData.m9524AI.fAutoScan = 0;
        iotConf.mData.m9524AI.ADC_Channel = wPollChannel&0x7;
    }
    /*ADC Config*/
    iotConf.mData.m9524AI.ADC_Gain = wGain;
    /*ADC Speed*/
    get_pacervalue(wCardNumber, (F64)wPollSpeed, &dwSampleRate);
    iotConf.mData.m9524AI.ADC_Speed = (U8)(dwSampleRate&0xff);
    //iotConf.wErrorCode = ErrorConfigIoctl;
    iotConf.mData.m9524AI.Reserved1 = 1;
    if(wGroup==0){
        CurrentCard.AI_Status.CurrentConfig = iotConf;
    }
    else{
        /*use the two valuables to store the config of General Purpose group*/
        CurrentCard.AI_Status.dwPacerValue = iotConf.mData.dwValue[0];
        CurrentCard.AI_Status.dwSampleIntv = iotConf.mData.dwValue[1];
    }

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }

    //CurrentCard.AI_Status.CurrentConfig.wErrorCode = NoError;
    return NoError;
}


I16 AI_ReadChannel32(
    U16 wCardNumber,
    U16 wChannel,
    U16 wAdRange,
    U32 *pdwValue
    )
{
    I16 err;
    U16 wGain;
    DAS_IOT_PIO iotPio;

    memset(&iotPio, '\0', sizeof(DAS_IOT_PIO));

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 1, &wChannel,
        0L, FALSE))!= NoError)
    {
        return err;
    }

    switch(CurrentCard.wCardType)
    {
        case PCI_9524:
            if(!validate_ad_range(wCardNumber, wAdRange, &wGain))
                return ErrorInvalidAdRange;
            if(wChannel<4){ /*load cell group channel 0*/
                if(wGain)
                    return ErrorInvalidAdRange;
            }
            else if(wChannel<8){ /*general purpose group channel 0*/
                if((!wGain)||(wGain>4))
                    return ErrorInvalidAdRange;
            }
            iotPio.wChannel = wChannel;
            iotPio.wLine_Gain = wGain;
            iotPio.wByteCount = (U16)((CurrentCard_AIFunc.wResoluation+7)/8);
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_PIO_READ, &(iotPio))
                !=DAS_ERR_NO)
            {
                return iotPio.wErrorCode;
            }
            *pdwValue = iotPio.dwValue;
            break;
        default:
            return ErrorFuncNotSupport;
    }

    return NoError;
}


I16 AI_VoltScale32(
    U16 wCardNumber,
    U16 adRange,
    I32 reading,
    F64 *voltage
    )
{
    I16 err;
    U16 wGain;
    I16 wReading;

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    //Check wAdRange and get Gain
    if(!validate_ad_range(wCardNumber, adRange, &wGain))
        return ErrorInvalidAdRange;

    switch(CurrentCard.wCardType){
        case PCI_9524:
            ai_scale32_9524(wCardNumber, adRange, reading, voltage);
            break;
        case PCI_9222:
        case PCI_9223:
        case PCI_922A:
            ai_scale32(wCardNumber, adRange, reading, voltage);
            break;
        default:
            wReading = (I16)reading;
            ai_scale(wCardNumber, adRange, wReading, voltage);
            break;
    }

    return NoError;
}



/*Does not support at this moment*/
I16 AI_ReadMultiChannels32(
    U16 wCardNumber,
    U16 wNumChans,
    U16 *pwChans,
    U16 *pwAdRanges,
    U32 *pdwBuffer
    )
{
    I16 err;
    U16 wGain;
    U16 chNum;
    DAS_IOT_SIMU_PIO iotSPio;
    int i;

    chNum = wNumChans;
    memset(&iotSPio, '\0', sizeof(DAS_IOT_SIMU_PIO));

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, wNumChans,
        pwChans, 0, FALSE))!=NoError)
    {
        return err;
    }

    iotSPio.wChannelCount = chNum;
    iotSPio.wDirectory = 0;
    for(i=0; i<chNum; i++){
        iotSPio.dwChannelGain[i].wChannel = pwChans[i];
        if(!validate_ad_range(wCardNumber, pwAdRanges[i], &wGain))
            return ErrorInvalidAdRange;
        iotSPio.dwChannelGain[i].wGain = wGain;
    }

    if(CurrentCard.wCardType==PCI_9524){
        if(iotSPio.dwChannelGain[0].wChannel==0){
            /*load cell group channel 0*/
            if(iotSPio.dwChannelGain[0].wGain)
                return ErrorInvalidAdRange;
        }
        else if(iotSPio.dwChannelGain[0].wChannel==4){
            /*general purpose group channel 0*/
            if((!iotSPio.dwChannelGain[0].wGain)||\
                (iotSPio.dwChannelGain[0].wGain>4))
            {
                return ErrorInvalidAdRange;
            }
        }
    }

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_SIMU_AI_PIO, &(iotSPio))
        !=DAS_ERR_NO)
    {
        return iotSPio.wErrorCode;
    }

    memcpy(
        pdwBuffer,
        (U32 *)(iotSPio.dwChannelGain),
        ((CurrentCard_AIFunc.wResoluation+7)/8)*chNum
    );

    return NoError;
}


I16 AI_ScanReadChannels32(
    U16 wCardNumber,
    U16 wChannel,
    U16 wAdRange,
    U32 *pdwBuffer
    )
{
    I16 err;
    int i;
    U16 pwChans[256];
    U16 pwAdRanges[256];
    //10/19/09
    DAS_IOT_CONFIG iotConf = {0};
    U16 MaxScanCh;

    switch(CurrentCard.wCardType){
        case PCI_9524:
            if(!wChannel){
                //10/19/09
                iotConf = CurrentCard.AI_Status.CurrentConfig;
                MaxScanCh = iotConf.mData.m9524AI.ADC_Channel;
                for(i=0; i<=MaxScanCh; i++){
                    *(pwChans+i) = (U16)i;
                    *(pwAdRanges+i) = wAdRange;
                }
            }
            else if(wChannel==1){
                //10/19/09
                iotConf.mData.dwValue[0] = CurrentCard.AI_Status.dwPacerValue;
                iotConf.mData.dwValue[1] = CurrentCard.AI_Status.dwSampleIntv;
                MaxScanCh = iotConf.mData.m9524AI.ADC_Channel-4;
                for(i=0; i<=MaxScanCh; i++){
                    *(pwChans+i) = (U16)(i+4);
                    *(pwAdRanges+i) = wAdRange;
                }
            }
            else{
                return ErrorInvalidIoChannel;
            }
            //10/19/09
            if((err = AI_ReadMultiChannels32(wCardNumber, (U16)(MaxScanCh+1)/*4*/,
                pwChans, pwAdRanges, pdwBuffer))!=NoError)
            {
                return err;
            }
            break;
        default:
            return ErrorFuncNotSupport;
    }

    return NoError;
}


I16 ai_922x_config(
    U16     wCardNumber,
    U16     ConfigCtrl,
    U16     TrigCtrl,
    U32     ReTriggerCnt,
    BOOLEAN AutoResetBuf
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }
    if(CurrentCard.wCardType==PCI_922A){
        if(ConfigCtrl>P922x_AI_Differential)
            return ErrorUndefinedParameter;
        if(TrigCtrl)
            return ErrorInvalidTriggerMode;
        ReTriggerCnt = 0;
    }
    if((TrigCtrl&0x1)&&(((TrigCtrl>>4)&0xf)==0)){
        //Gated Trigger & Soft Trigger Source
        return ErrorInvalidTriggerMode;
    }
    if((TrigCtrl&0x1)&&((TrigCtrl>>9)&0x1)){
        //Gated Trigger & Retrigger
        return ErrorInvalidTriggerMode;
    }
    if((((TrigCtrl>>4)&0xf)==0)&&((TrigCtrl>>9)&0x1)){
        //Soft Trigger Source & ReTrigger
        return ErrorInvalidTriggerMode;
    }

    /*Single-Endded or Differential*/
    iotConf.mData.m922xAI.fInputType = ConfigCtrl&0x3;
    /*Conversion source*/
    iotConf.mData.m922xAI.fConvSrc = (ConfigCtrl>>4)&0xf;
    /*Trigger Mode*/
    iotConf.mData.m922xAI.fTriggerMode = TrigCtrl&0x1;
    if(iotConf.mData.m922xAI.fTriggerMode){
        //Gated Trigger 4 bytes per sample
        CurrentCard_AIFunc.wResoluation = 32;
    }
    else{
        CurrentCard_AIFunc.wResoluation = 16;
    }
    /*Trigger source*/
    iotConf.mData.m922xAI.fTrigSrc = (TrigCtrl>>4)&0xf;
    /*Trigger polarity*/
    iotConf.mData.m922xAI.fTrigPol = (TrigCtrl>>8)&0x1;
    /*Enable ReTrigger*/
    iotConf.mData.m922xAI.fEnReTrig = (TrigCtrl>>9)&0x1;
    /*Auto reset buffer?*/
    iotConf.mData.m922xAI.fAutoResetBuf = AutoResetBuf;
    if(iotConf.mData.m922xAI.fEnReTrig)
        iotConf.mData.m922xAI.ReTriggerCnt = ReTriggerCnt;
    //iotConf.wErrorCode = ErrorConfigIoctl;
    CurrentCard.AI_Status.CurrentConfig = iotConf;
    //9222 has 8/16 AI channels while 9223 has 16/32 AI channels
    //in Single-Ended/Differential Mode
    if(CurrentCard.wCardType==PCI_9222){
        if(iotConf.mData.m922xAI.fInputType==P922x_AI_Differential){
            CurrentCard_AIFunc.wChannelCount = 8;
        }
        else{
            CurrentCard_AIFunc.wChannelCount = 16;
        }
    }
    else if((CurrentCard.wCardType==PCI_9223)||
        (CurrentCard.wCardType==PCI_922A))
    {
        if(iotConf.mData.m922xAI.fInputType==P922x_AI_Differential){
            CurrentCard_AIFunc.wChannelCount = 16;
        }
        else{
            CurrentCard_AIFunc.wChannelCount = 32;
        }
    }
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }

    //CurrentCard.AI_Status.CurrentConfig.wErrorCode = NoError;
    return NoError;
}


I16 AI_9222_Config(
    U16     wCardNumber,
    U16     ConfigCtrl,
    U16     TrigCtrl,
    U32     ReTriggerCnt,
    BOOLEAN AutoResetBuf
    )
{
    return ai_922x_config(wCardNumber, ConfigCtrl, TrigCtrl, ReTriggerCnt, \
        AutoResetBuf);
}


I16 AI_9222_CounterInterval(
    U16 wCardNumber,
    U32 ScanIntrv,
    U32 SampIntrv
    )
{
    return AI_9221_CounterInterval(wCardNumber, ScanIntrv, SampIntrv);
}


I16 AI_9223_Config(
    U16     wCardNumber,
    U16     ConfigCtrl,
    U16     TrigCtrl,
    U32     ReTriggerCnt,
    BOOLEAN AutoResetBuf
    )
{
    return ai_922x_config(wCardNumber, ConfigCtrl, TrigCtrl, ReTriggerCnt, \
        AutoResetBuf);
}


I16 AI_9223_CounterInterval(
    U16 wCardNumber,
    U32 ScanIntrv,
    U32 SampIntrv
    )
{
    return AI_9221_CounterInterval(wCardNumber, ScanIntrv, SampIntrv);
}


I16 AI_AsyncReTrigNextReady(
    U16     wCardNumber,
    BOOLEAN *bReady,
    BOOLEAN *StopFlag,
    U16     *RdyTrigCnt
    )
{
    I16 err;
    DAS_IOT_DBF iotDbf = {0};
    DAS_IOT_CONFIG iotConf;
    U8 fInfiniteTrigger = 0;

    iotConf = CurrentCard.AI_Status.CurrentConfig;

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    if((CurrentCard.wCardType==PCI_9222)||(CurrentCard.wCardType==PCI_9223)||
        (CurrentCard.wCardType==PCI_922A))
    {
        if(!iotConf.mData.m922xAI.fEnReTrig)
            return ErrorInvalidTriggerMode;
        if(iotConf.mData.m922xAI.ReTriggerCnt==0)
            fInfiniteTrigger = 1;
    }

    iotDbf.wEnable = 1;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_DBF_HALF_READY, &(iotDbf))
        !=DAS_ERR_NO)
    {
        return iotDbf.wErrorCode;
    }
    *bReady = (BOOLEAN)(iotDbf.wReady&0x1);
    *StopFlag = (BOOLEAN)((iotDbf.wReady>>1)&0x1);
    if(fInfiniteTrigger){
        /*Infinit Trigger - Return Half Ready Buffer Index*/
        *RdyTrigCnt = iotDbf.wViewIndex;
        CurrentCard.AI_Status.wDblBufferIndex = iotDbf.wViewIndex;
    }
    else{
        /*Finite Trigger - Return Comming Trigger Count*/
        *RdyTrigCnt = iotDbf.wViewIndex|(iotDbf.wEnable<<16);
    }

    return NoError;
}


I16 AI_922A_Config(
    U16     wCardNumber,
    U16     ConfigCtrl,
    U16     TrigCtrl,
    U32     ReTriggerCnt,
    BOOLEAN AutoResetBuf
    )
{
    return ai_922x_config(wCardNumber, ConfigCtrl, TrigCtrl, ReTriggerCnt, \
        AutoResetBuf);
}


I16 AI_922A_CounterInterval(
    U16 wCardNumber,
    U32 ScanIntrv,
    U32 SampIntrv
    )
{
    return AI_9221_CounterInterval(wCardNumber, ScanIntrv, SampIntrv);
}
