#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include "dask.h"
#include "Adlsys.h"
#include "dask_p.h"


int ao_scale(U16 wCardNumber, U16 wChannel, F64 Voltage, I16 *binValue);


extern REGISTER_CARD_TYPE RegisterCards[MAX_CARD];


I16 AO_6208A_Config(
    U16 wCardNumber,
    U16 V2AMode
    )
{
    I16 err;
    //00/02/03 jeffrey, localize the global vars.
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if((err = fundamental_check(&CurrentCard_AOFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    iotConf.mData.m6208AO.wCurrentRange = V2AMode;
    CurrentCard.AO_Status.CurrentConfig = iotConf;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AO_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }

    return NoError;
}


I16 AO_6308A_Config(
    U16 wCardNumber,
    U16 V2AMode
    )
{
    I16 err;
    //00/02/03 jeffrey, localize the global vars.
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if((err = fundamental_check(&CurrentCard_AOFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    iotConf.mData.m6208AO.wCurrentRange = V2AMode;
    CurrentCard.AO_Status.CurrentConfig = iotConf;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AO_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }

    return NoError;
}


I16 AO_WriteChannel(
    U16 wCardNumber,
    U16 wChannel,
    I16 wValue
    )
{
    I16 err;
    DAS_IOT_PIO iotPio;

    memset(&iotPio, '\0', sizeof(DAS_IOT_PIO));

    if((err = fundamental_check(&CurrentCard_AOFunc, wCardNumber, 1,
        &wChannel, 0L, FALSE))!= NoError)
    {
        return err;
    }

    iotPio.wChannel = wChannel;
    iotPio.dwValue = (U16)wValue;
    iotPio.wByteCount = (U16)((CurrentCard_AOFunc.wResoluation+7)/8);
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AO_PIO_WRITE, &(iotPio))
        !=DAS_ERR_NO)
    {
        return iotPio.wErrorCode;
    }

    return NoError;
}


I16 AO_SimuWriteChannel(
    U16 wCardNumber,
    U16 wGroup,
    I16 *pwBuffer
    )
{
    I16 i, wNum, err;
    DAS_IOT_CONT iotCont;

    memset(&iotCont, '\0', sizeof(DAS_IOT_CONT));

    //08/27/09
    /*
     * Since 6202 does not support simultaneous AO polling
     * use DMA mode to simulate it
     */
    if(CurrentCard.wCardType==PCI_6202){
        U16 wChans[4] = {0, 1, 2, 3};
        U32 pdwBuffer[4];
        U16 BufferId;
        U32 dwWriteCnt;
        BOOLEAN bStopFlag;

        for(i=0; i<4; i++){
            pdwBuffer[i] = pwBuffer[i];
        }

        err = AO_6202_Config(wCardNumber, 0, 0, 1, 0, 0, 0);
        if(err<0){
            return err;
        }

        err = AO_ContBufferSetup(wCardNumber, pdwBuffer, 4, &BufferId);
        if(err<0){
            return err;
        }

        err = AO_AsyncDblBufferMode(wCardNumber , 0);
        if(err<0){
            return err;
        }

        err = AO_ContWriteMultiChannels(wCardNumber, 4, wChans, BufferId,
            4, 1, 8000, 1, ASYNCH_OP
        );
        if(err<0){
            return err;
        }
        i = 0;
        do{
            AO_AsyncCheck(wCardNumber, &bStopFlag, &dwWriteCnt);
            i++;
        }while((!bStopFlag)&&(i<10000));
        AO_AsyncClear(wCardNumber, &dwWriteCnt, 0);

        err = AO_ContBufferReset(wCardNumber);
        if(err<0){
            return err;
        }
    }
    else{
        if(CurrentCard.wCardType==PCI_9524){
            iotCont.wChannelCount = wNum = 2;
        }
        else{
            /*Original setting (for 6308)*/
            //yuan 07/31/00
            iotCont.wSyncMode = wGroup;
            wNum = 4;
        }
        for(i=0; i<wNum; i++){
            iotCont.dwChannelGain[i].wGain = (U16)pwBuffer[i];
        }
        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AO_CON_WRITE, &(iotCont))
            !=DAS_ERR_NO)
        {
            return iotCont.wErrorCode;
        }
    }

    return NoError;
}


I16 AO_SimuVWriteChannel(
    U16 wCardNumber,
    U16 wGroup,
    F64 *pfBuffer
    )
{
    I16 i, wNum, err ;
    I16 binValue[4];
    //00/02/03 jeffrey, localize the global vars.
    DAS_IOT_CONT iotCont;

    memset(&iotCont, '\0', sizeof(DAS_IOT_CONT));

    //08/27/09
    /*
     * Since 6202 does not support simultaneous AO polling
     * use DMA mode to simulate it
     */
    if(CurrentCard.wCardType==PCI_6202){
        U16 wChans[4] = {0, 1, 2, 3};
        U16 pwBuffer[4];
        U32 pdwBuffer[4];
        U16 BufferId;
        U32 dwWriteCnt;
        BOOLEAN bStopFlag;

        for(i=0; i<4; i++){
            err = AO_VoltScale(wCardNumber, i, pfBuffer[i],
                (I16 *)(pwBuffer+i)
            );
            if(err<0)
                return err;
            pdwBuffer[i] = pwBuffer[i];
        }

        err = AO_6202_Config(wCardNumber, 0, 0, 1, 0, 0, 0);
        if(err<0){
            return err;
        }

        err = AO_ContBufferSetup(wCardNumber, pdwBuffer, 4, &BufferId);
        if(err<0){
            return err;
        }

        err = AO_AsyncDblBufferMode(wCardNumber , 0);
        if(err<0){
            return err;
        }

        err = AO_ContWriteMultiChannels(wCardNumber, 4, wChans, BufferId,
            4, 1, 8000, 1, ASYNCH_OP
        );
        if(err<0){
            return err;
        }
        i = 0;
        do{
            AO_AsyncCheck(wCardNumber, &bStopFlag, &dwWriteCnt);
            i++;
        }while((!bStopFlag)&&(i<1000));
        AO_AsyncClear(wCardNumber, &dwWriteCnt, 0);

        err = AO_ContBufferReset(wCardNumber);
        if(err<0){
            return err;
        }
    }
    else{
        if(CurrentCard.wCardType==PCI_9524){
            iotCont.wChannelCount = wNum = 2;
        }
        else{
            /*Original setting (for 6308)*/
            //yuan 07/31/00
            iotCont.wSyncMode = wGroup;
            wNum = 4;
        }
        for(i=0; i<wNum; i++){
            if((err = AO_VoltScale(wCardNumber, i, pfBuffer[i], &(binValue[i])))
                !=NoError)
            {
                return err;
            }
            iotCont.dwChannelGain[i].wGain = binValue[i];
        }
        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AO_CON_WRITE, &(iotCont))
            !=DAS_ERR_NO)
        {
            return iotCont.wErrorCode;
        }
    }

    return NoError;
}


I16 AO_VoltScale(
    U16 wCardNumber,
    U16 Channel,
    F64 Voltage,
    I16 *binValue
    )
{
    I16 err;

    if((err = fundamental_check(&CurrentCard_AOFunc, wCardNumber, 1, &Channel,
        0L, FALSE))!=NoError)
    {
        return err;
    }
    ao_scale(wCardNumber, Channel, Voltage, binValue);

    return NoError;
}


I16 AO_VWriteChannel(
    U16 wCardNumber,
    U16 wChannel,
    F64 Voltage
    )
{
    I16 err;
    I16 binValue;

    if((err = AO_VoltScale(wCardNumber, wChannel, Voltage, &binValue))!=NoError)
        return err;
    if((err = AO_WriteChannel(wCardNumber, wChannel, binValue))!=NoError)
        return err;

    return NoError;
}


I16 AO_6308V_Config(
    U16 wCardNumber,
    U16 wChannel,
    U16 wOutputPolarity,
    F64 refVoltage
    )
{
    I16 err;

    if((err = fundamental_check(&CurrentCard_AOFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    if(wChannel==P6308V_AO_CH0_3){//channel 0-3
        CurrentCard.AO_Status.CurrentConfig.mData.m6208AO.wCurrentRange = \
            wOutputPolarity;
        CurrentCard.RefVoltage[0] = refVoltage;
    }
    else{//channel 4-7
        CurrentCard.AO_Status.CurrentConfig.mData.m6208AO.wReserve1 = \
            wOutputPolarity;
        CurrentCard.RefVoltage[1] = refVoltage;
    }

    return NoError;
}


I16 AO_9111_Config(
    U16 wCardNumber,
    U16 wOutputPolarity
    )
{
    I16 err;

    if((err = fundamental_check(&CurrentCard_AOFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    CurrentCard.AO_Status.CurrentConfig.mData.m9111AO.wOutputPolarity = \
        wOutputPolarity;

    return NoError;
}


I16 AO_9112_Config(
    U16 wCardNumber,
    U16 wChannel,
    F64 refVoltage
    )
{
    I16 err;

    if((err = fundamental_check(&CurrentCard_AOFunc, wCardNumber, 1, &wChannel,
        0L, FALSE))!=NoError)
    {
        return err;
    }
    if((refVoltage>10)||(refVoltage<-10))
        return ErrorInvalidDaRefVoltage;

    CurrentCard.RefVoltage[wChannel] = refVoltage;

    return NoError;
}


int ao_scale(
    U16 wCardNumber,
    U16 wChannel,
    F64 Voltage,
    I16 *binValue
    )
{
    switch(CurrentCard.wCardType){
        case PCI_6208V:
            *binValue = (I16)(Voltage*32767/10);
            break;
        case PCI_6208A:
            switch(CurrentCard_AOCurConfig.mData.m6208AO.wCurrentRange){
                case P6208_CURRENT_0_20MA:
                    *binValue = (I16)(Voltage*32767/20);
                    break;
                case P6208_CURRENT_5_25MA:
                    *binValue = (I16)((Voltage-5)*32767/20);
                    break;
                case P6208_CURRENT_4_20MA:
                    *binValue = (I16)((Voltage-4)*32767/16);
                    break;
            }
            break;
        case PCI_6308V:
            switch(wChannel){
                case 0:
                case 1:
                case 2:
                case 3:
                    if(CurrentCard_AOCurConfig.mData.m6208AO.wCurrentRange==\
                        P6308V_AO_UNIPOLAR)
                    {
                        *binValue = \
                           (U16)(Voltage*4095/CurrentCard.RefVoltage[0])&0x0fff;
                    }
                    else //Bipolar +/-10V
                        *binValue = (U16)((Voltage+10)*4095/20)&0x0fff;
                    break;
                case 4:
                case 5:
                case 6:
                case 7:
                    if(CurrentCard_AOCurConfig.mData.m6208AO.wReserve1==\
                        P6308V_AO_UNIPOLAR)
                    {
                        *binValue = \
                           (U16)(Voltage*4095/CurrentCard.RefVoltage[1])&0x0fff;
                    }
                    else //Bipolar +/-10V
                        *binValue = (U16)((Voltage+10)*4095/20)&0x0fff;
                    break;
            }
            break;
        case PCI_6308A:
            switch(CurrentCard_AOCurConfig.mData.m6208AO.wCurrentRange){
                case P6308_CURRENT_0_20MA:
                    *binValue = (U16)(Voltage*4095/20)&0x0fff;
                    break;
                case P6308_CURRENT_5_25MA:
                    *binValue = (U16)((Voltage-5)*4095/20)&0x0fff;
                    break;
                case P6308_CURRENT_4_20MA:
                    *binValue = (U16)((Voltage-4)*4095/16)&0x0fff;
                    break;
            }
            break;
        case PCI_9111DG:
        case PCI_9111HR:
            if(CurrentCard_AOCurConfig.mData.m9111AO.wOutputPolarity==\
                P9111_AO_UNIPOLAR)
            {
                *binValue = (I16)(Voltage*4095/10);
            }
            else //Bipolar +/-10V
                *binValue = (I16)((Voltage+10)*4095/20);
            break;
        case PCI_9112:
            if(Voltage*CurrentCard.RefVoltage[wChannel]>0)
                return ErrorDaVoltageOutOfRange;
            *binValue = (I16)(Voltage*4095/(-CurrentCard.RefVoltage[wChannel]));
            break;
        case PCI_9118DG:
        case PCI_9118HG:
        case PCI_9118HR:
            *binValue = (I16)((Voltage+10)*4095/20);
            break;
        case PCI_9221:
            if(Voltage>=0.0)
                *binValue = (I16)(Voltage*32767/5);
            else
                *binValue = (I16)(Voltage*32768/5);
            break;
        case PCI_9524:
            *binValue = (I16)((unsigned)((Voltage+10)*65535/20));
            break;
        case PCI_6202:
        case PCI_9222:
        case PCI_9223:
            *binValue = (I16)((unsigned)((Voltage+10)*65535/20));
            break;
    }

    return NoError;
}


I16 AO_6202_Config(
    U16 wCardNumber,
    U16 ConfigCtrl,
    U16 TrigCtrl,
    U32 ReTrgCnt,
    U32 DLY1Cnt,
    U32 DLY2Cnt,
    BOOLEAN AutoResetBuf
    )
{
    I16 err;
    DAS_EAO_CONFIG extend_6202AO;

    memset(&extend_6202AO, '\0', sizeof(DAS_EAO_CONFIG));

    if((err = fundamental_check(&CurrentCard_AOFunc, wCardNumber,
        0, NULL, 0L, FALSE))!=NoError)
    {
        return err;
    }

    extend_6202AO.iotConf.mData.m6202AO.fDaWRSrc = ConfigCtrl&0x03;
    extend_6202AO.iotConf.mData.m6202AO.fTrigSrc = TrigCtrl&0x03;
    if(TrigCtrl&P6202_DA_TRGMOD_DELAY){
        extend_6202AO.iotConf.mData.m6202AO.fTriggerMode = 1;
    }
    if(TrigCtrl&P6202_DA_ReTrigEn){
        extend_6202AO.iotConf.mData.m6202AO.fEnableReTrig = 1;
    }
    if(AutoResetBuf){ //david 04/22/08
        extend_6202AO.iotConf.mData.m6202AO.fAutoResetBuf = 1;
    }
    extend_6202AO.Counter_Value[0] = ReTrgCnt;
    extend_6202AO.Counter_Value[1] = DLY1Cnt;
    //Delay 2 Trigger by Tim 20091021
    //extend_6202AO.Counter_Value[2] = DLY2Cnt;
    extend_6202AO.iotConf.mData.m6202AO.fEnableDLY2 = (TrigCtrl>>8)&0x1;
    if(extend_6202AO.iotConf.mData.m6202AO.fEnableDLY2)
    {
        extend_6202AO.Counter_Value[2] = DLY2Cnt;
    }
    CurrentCard.AO_Status.CurrentConfig = extend_6202AO.iotConf;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_EAO_CONFIG, &(extend_6202AO))
        !=DAS_ERR_NO)
    {
        return extend_6202AO.iotConf.wErrorCode;
    }

    //CurrentCard.AO_Status.CurrentConfig.wErrorCode = NoError;
    return NoError;
}


I16 AO_ContWriteChannel(
    U16 wCardNumber,
    U16 wChannel,
    U16 BufId,
    U32 dwWriteCount,
    U32 wIterations,
    U32 dwCHUI,
    U16 definite,
    U16 wSyncMode
    )
{
    I16 err;
    DAS_IOT_SG_CONT iotCont;
    DAS_IOT_CONFIG iotConf;
    U8 fAutoReset = 0;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));
    memset(&iotCont, '\0', sizeof(DAS_IOT_SG_CONT));

    iotConf = CurrentCard.AO_Status.CurrentConfig;

    if((err = fundamental_check(&CurrentCard_AOFunc, wCardNumber, 1, &wChannel,
        dwWriteCount*4, TRUE))!= NoError)
    {
        return err;
    }

    /*SyncMode must be 1 iteration*/
    //06/27/08 modified
    if(((wIterations!=1)||(!definite))&&(wSyncMode==SYNCH_OP))
        return ErrorUndefinedParameter;

    if((BufId>1)||(!CurrentCard.AO_Status.pDAQbuf[BufId]))
        return ErrorInvalidBufferID;

    if(CurrentCard.wCardType==PCI_6202){
        //Infinite repeat dose not work with retrigger mode
        if(((!wIterations)||(!definite))&&(iotConf.mData.m6202AO.fEnableReTrig))
            return ErrorInvalidTriggerMode;
        //06/19/08 modified
        //Check conversion rate if internal conversion source is selected
        if(iotConf.mData.m6202AO.fDaWRSrc==0){
            if(dwCHUI<80)
                return ErrorInvalidSampleRate;
        }
        else{
            //09/26/08
            //External Conversion Source only valid for 1 or infinite interations
            if((definite)&&(wIterations>1))
                return ErrorUndefinedParameter;
            //External Conversion Source only valid for ASYNCMODE
            if(wSyncMode==SYNCH_OP)
                return ErrorUndefinedParameter;
        }
        fAutoReset = (U8)iotConf.mData.m6202AO.fAutoResetBuf;
    }
    if((CurrentCard.wCardType==PCI_9222)||(CurrentCard.wCardType==PCI_9223)){
        //Infinite repeat dose not work with retrigger mode
        if(((!wIterations)||(!definite))&&(iotConf.mData.m922xAO.fEnReTrig))
            return ErrorInvalidTriggerMode;
        //Check conversion rate if internal conversion source is selected
        if(iotConf.mData.m922xAO.fConvSrc==0){
            if(dwCHUI<80)
                return ErrorInvalidSampleRate;
        }
        else{
            //09/25/08
            //External Conversion Source only valid for 1 or infinite interations
            if((definite)&&(wIterations>1))
                return ErrorUndefinedParameter;
            //External Conversion Source only valid for ASYNCMODE
            if(wSyncMode==SYNCH_OP)
                return ErrorUndefinedParameter;
        }
        fAutoReset = (U8)iotConf.mData.m922xAO.fAutoResetBuf;
    }

    CurrentCard.AO_Status.dwTransferCount = dwWriteCount;

    if(!CurrentCard.AO_Status.bDblBufferMode)
        CurrentCard.AO_Status.pDAQbufHandle = (U32)BufId;
    else
        CurrentCard.AO_Status.pDAQbufHandle = 0;

    if(CurrentCard.AO_Status.dwTransferCount>
        CurrentCard.AO_Status.dwBufCount[\
        ((U16)CurrentCard.AO_Status.pDAQbufHandle)])
    {
        return ErrorTransferCountTooLarge;
    }

    iotCont.dwPacerValue[0] = dwCHUI;
    iotCont.dwPacerValue[1] = /*dwCHUI*/definite;
    iotCont.dwAccessCount = dwWriteCount;
    iotCont.wAutoScan = 0;
    iotCont.wIterations = wIterations;
    iotCont.wSyncMode = wSyncMode;
    iotCont.wChannelCount = 1;
    iotCont.dwChannelGain[0].wChannel = wChannel;
    //...iotCont.dwBufferAddr = BufId;
    iotCont.dwBufferAddr = (U32)CurrentCard.AO_Status.pDAQbufHandle;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AO_CON_WRITE_SG, &(iotCont))
        !=DAS_ERR_NO)
    {
        return iotCont.wErrorCode;
    }

    //06/27/08 Add for auto reset buffer
    if((wSyncMode==SYNCH_OP)&&(fAutoReset)){
        CurrentCard.AO_Status.pDAQbuf[BufId] = NULL;
        CurrentCard.AO_Status.dwBufCount[BufId] = 0;
    }

    return NoError;
}


I16 AO_ContWriteMultiChannels(
    U16 wCardNumber,
    U16 wNnmChans,
    U16 *wChans,
    U16 BufId,
    U32 dwWriteCount,
    U32 wIterations,
    U32 dwCHUI,
    U16 definite,
    U16 wSyncMode
    )
{
    int i;
    I16 err;
    DAS_IOT_SG_CONT iotCont;
    DAS_IOT_CONFIG iotConf;
    U8 fAutoReset = 0;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));
    memset(&iotCont, '\0', sizeof(DAS_IOT_SG_CONT));

    iotConf = CurrentCard.AO_Status.CurrentConfig;

    if((err = fundamental_check(&CurrentCard_AOFunc, wCardNumber, wNnmChans, wChans,
        dwWriteCount*4, TRUE))
        != NoError)
    {
        return err;
    }

    /*SyncMode must be 1 iteration*/
    //06/27/08 modified
    if(((wIterations!=1)||(!definite))&&(wSyncMode==SYNCH_OP))
        return ErrorUndefinedParameter;

    if((BufId>1)||(!CurrentCard.AO_Status.pDAQbuf[BufId]))
        return ErrorInvalidBufferID;

    if(CurrentCard.wCardType==PCI_6202){
        //Infinite repeat dose not work with retrigger mode
        if(((!wIterations)||(!definite))&&(iotConf.mData.m6202AO.fEnableReTrig))
            return ErrorInvalidTriggerMode;
        //06/27/08 modified
        //Check conversion rate if internal conversion source is selected
        if(iotConf.mData.m6202AO.fDaWRSrc==0){
            if(dwCHUI<80)
                return ErrorInvalidSampleRate;
        }
        else
        {
            //09/26/08
            //External Conversion Source only valid for 1 or infinite interations
            if((definite)&&(wIterations>1))
                return ErrorUndefinedParameter;
            //External Conversion Source only valid for ASYNCMODE
            if(wSyncMode==SYNCH_OP)
                return ErrorUndefinedParameter;
        }
        fAutoReset = (U8)iotConf.mData.m6202AO.fAutoResetBuf;
    }
    if((CurrentCard.wCardType==PCI_9222)||(CurrentCard.wCardType==PCI_9223)){
        //Infinite repeat dose not work with retrigger mode
        if(((!wIterations)||(!definite))&&(iotConf.mData.m922xAO.fEnReTrig))
            return ErrorInvalidTriggerMode;
        //Check conversion rate if internal conversion source is selected
        if(iotConf.mData.m922xAO.fConvSrc==0){
            if(dwCHUI<80)
                return ErrorInvalidSampleRate;
        }
        else{
            //09/25/08
            //External Conversion Source only valid for 1 or infinite interations
            if((definite)&&(wIterations>1))
                return ErrorUndefinedParameter;
            //External Conversion Source only valid for ASYNCMODE
            if(wSyncMode==SYNCH_OP)
                return ErrorUndefinedParameter;
        }
        fAutoReset = (U8)iotConf.mData.m922xAO.fAutoResetBuf;
    }

    CurrentCard.AO_Status.dwTransferCount = dwWriteCount;

    if(!CurrentCard.AO_Status.bDblBufferMode)
        CurrentCard.AO_Status.pDAQbufHandle = (U32)BufId;
    else
        CurrentCard.AO_Status.pDAQbufHandle = 0;

    if(CurrentCard.AO_Status.dwTransferCount>
        CurrentCard.AO_Status.dwBufCount[\
        ((U16)CurrentCard.AO_Status.pDAQbufHandle)])
    {
        return ErrorTransferCountTooLarge;
    }

    iotCont.dwPacerValue[0] = dwCHUI;
    iotCont.dwPacerValue[1] = /*dwCHUI*/definite;
    iotCont.dwAccessCount = dwWriteCount;
    iotCont.wAutoScan = 1;
    iotCont.wIterations = wIterations;
    iotCont.wSyncMode = wSyncMode;
    iotCont.wChannelCount = wNnmChans;
    for(i=0; i<wNnmChans; i++)
        iotCont.dwChannelGain[i].wChannel = wChans[i];
    //...iotCont.dwBufferAddr = BufId;
    iotCont.dwBufferAddr = (U32)CurrentCard.AO_Status.pDAQbufHandle;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AO_CON_WRITE_SG, &(iotCont))
        !=DAS_ERR_NO)
    {
        return iotCont.wErrorCode;
    }

    //06/27/08 Add for auto reset buffer
    if((wSyncMode==SYNCH_OP)&&(fAutoReset)){
        CurrentCard.AO_Status.pDAQbuf[BufId] = NULL;
        CurrentCard.AO_Status.dwBufCount[BufId] = 0;
    }

    return NoError;
}


I16 AO_ContBufferSetup(
    U16  wCardNumber,
    void *pwBuffer,
    U32  dwWriteCount,
    U16  *BufferId
    )
{
    DAS_IOT_MBUF2 iotMbuf;

    memset(&iotMbuf, '\0', sizeof(DAS_IOT_MBUF2));

    iotMbuf.dwBufferAddr = (U32)pwBuffer;
    iotMbuf.dwAccessCount = dwWriteCount;
    iotMbuf.wDir = 1; /*output*/
    iotMbuf.wBufferId = 1; //??
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_MBF_SET, &(iotMbuf))
        !=DAS_ERR_NO)
    {
        return iotMbuf.wErrorCode;
    }
    *BufferId = iotMbuf.wBufferId&0xffff;
    CurrentCard.AO_Status.pDAQbuf[*BufferId] = pwBuffer;
    CurrentCard.AO_Status.dwBufCount[*BufferId] = dwWriteCount;

    return NoError;
}


I16 AO_ContBufferReset(
    U16 wCardNumber
    )
{
    DAS_IOT_MBUF2 iotMbuf;

    memset(&iotMbuf, '\0', sizeof(DAS_IOT_MBUF2));

    iotMbuf.wDir = 1;
    iotMbuf.dwAccessCount = 0;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_MBF_RESET, &(iotMbuf))
        !=DAS_ERR_NO)
    {
        return iotMbuf.wErrorCode;
    }
    CurrentCard.AO_Status.pDAQbuf[0] = NULL;
    CurrentCard.AO_Status.pDAQbuf[1] = NULL;
    CurrentCard.AO_Status.dwBufCount[0] = 0;
    CurrentCard.AO_Status.dwBufCount[1] = 0;

    return NoError;
}


I16 AO_AsyncCheck(
    U16     wCardNumber,
    BOOLEAN *bStopped,
    U32     *dwAccessCnt
    )
{
    I16 err;
    DAS_IOT_ASYNC iotAsync;

    memset(&iotAsync, '\0', sizeof(DAS_IOT_ASYNC));

    if((err = fundamental_check(&CurrentCard_AOFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AO_ASY_CHECK, &(iotAsync))
        !=DAS_ERR_NO)
    {
        return iotAsync.wErrorCode;
    }
    *bStopped = (BOOLEAN)(iotAsync.wStopped);
    *dwAccessCnt = iotAsync.dwAccessCount;

    return NoError;
}


I16 AO_AsyncClear(
    U16 wCardNumber,
    U32 *dwAccessCnt,
    U16 stop_mode
    )
{
    I16 err;
    DAS_IOT_ASYNC iotAsync;
    DAS_IOT_CONFIG iotConf;
    U8 fAutoReset;
    U32 BufId;

    memset(&iotAsync, '\0', sizeof(DAS_IOT_ASYNC));
    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    iotConf = CurrentCard.AO_Status.CurrentConfig;

    if((err = fundamental_check(&CurrentCard_AOFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    iotAsync.wStopped = stop_mode;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AO_ASY_CLEAR, &(iotAsync))
        !=DAS_ERR_NO)
    {
        return iotAsync.wErrorCode;
    }
    *dwAccessCnt = iotAsync.dwAccessCount;

    //06/27/08 Add for clear buffer
    switch(CurrentCard.wCardType){
        case PCI_9222:
        case PCI_9223:
            fAutoReset = (U8)iotConf.mData.m922xAO.fAutoResetBuf;
            break;
        case PCI_6202:
            fAutoReset = (U8)iotConf.mData.m6202AO.fAutoResetBuf;
            break;
        default:
            fAutoReset = 0;
            break;
    }
    if(fAutoReset){
        if(!CurrentCard.AO_Status.bDblBufferMode){
            BufId = (U32)CurrentCard.AO_Status.pDAQbufHandle;
            CurrentCard.AO_Status.pDAQbuf[BufId] = NULL;
            CurrentCard.AO_Status.dwBufCount[BufId] = 0;
        }
        else{
            CurrentCard.AO_Status.pDAQbuf[0] = NULL;
            CurrentCard.AO_Status.pDAQbuf[1] = NULL;
            CurrentCard.AO_Status.dwBufCount[0] = 0;
            CurrentCard.AO_Status.dwBufCount[1] = 0;
        }
    }

    return NoError;
}


I16 AO_AsyncDblBufferMode(
    U16     wCardNumber,
    BOOLEAN bEnable
    )
{
    I16 err;
    DAS_IOT_DBF iotDbf;
    DAS_IOT_CONFIG iotConf;

    memset(&iotDbf, '\0', sizeof(DAS_IOT_DBF));
    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    iotConf = CurrentCard.AO_Status.CurrentConfig;

    //08/04/08 add error check
    switch(CurrentCard.wCardType){
        case PCI_6202:
            if(bEnable&&(iotConf.mData.m6202AO.fEnableReTrig))
                return ErrorDblBufModeNotAllowed;
            break;
        case PCI_9222:
        case PCI_9223:
            if(bEnable&&(iotConf.mData.m922xAO.fEnReTrig))
                return ErrorDblBufModeNotAllowed;
            break;
        default:
            return ErrorFuncNotSupport;
    }

    if((err = fundamental_check(&CurrentCard_AOFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    iotDbf.wEnable = bEnable;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AO_DBF_MODE, &(iotDbf))
        !=DAS_ERR_NO)
    {
        return iotDbf.wErrorCode;
    }
    CurrentCard.AO_Status.bDblBufferMode = bEnable;

    return NoError;
}


I16 AO_AsyncDblBufferHalfReady(
    U16     wCardNumber,
    BOOLEAN *bHalfReady
    )
{
    I16 err;
    DAS_IOT_DBF iotDbf;

    memset(&iotDbf, '\0', sizeof(DAS_IOT_DBF));

    if((err = fundamental_check(&CurrentCard_AOFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AO_DBF_HALF_READY, &(iotDbf))
        !=DAS_ERR_NO)
    {
        return iotDbf.wErrorCode;
    }
    *bHalfReady = iotDbf.wReady;
    CurrentCard.AO_Status.wDblBufferIndex = iotDbf.wViewIndex;

    return NoError;
}


I16 AO_9222_Config(
    U16     wCardNumber,
    U16     ConfigCtrl,
    U16     TrigCtrl,
    U32     ReTrgCnt,
    U32     DLY1Cnt,
    U32     DLY2Cnt,
    BOOLEAN AutoResetBuf
    )
{
    I16 err;
    DAS_EAO_CONFIG m9222ao;

    memset(&m9222ao, '\0', sizeof(DAS_EAO_CONFIG));

    if((err = fundamental_check(&CurrentCard_AOFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    /*Conversion Source*/
    m9222ao.iotConf.mData.m922xAO.fConvSrc = ConfigCtrl&0xf;
    /*Auto Reset Buffer?*/
    m9222ao.iotConf.mData.m922xAO.fAutoResetBuf = AutoResetBuf&0x1;
    /*Trigger Mode*/
    m9222ao.iotConf.mData.m922xAO.fTriggerMode = TrigCtrl&0x1;
    if(m9222ao.iotConf.mData.m922xAO.fTriggerMode){
        //Delay 1 Trigger Coumt
        m9222ao.Counter_Value[1] = DLY1Cnt;
    }
    /*Trigger Source*/
    m9222ao.iotConf.mData.m922xAO.fTrigSrc = (TrigCtrl>>4)&0xf;
    /*Trigger Polarity*/
    m9222ao.iotConf.mData.m922xAO.fTrigPol = (TrigCtrl>>8)&0x1;
    /*ReTrigger*/
    m9222ao.iotConf.mData.m922xAO.fEnReTrig = (TrigCtrl>>9)&0x1;
    if(m9222ao.iotConf.mData.m922xAO.fEnReTrig){
        //ReTriggerCount
        m9222ao.Counter_Value[0] = ReTrgCnt;
    }
    /*Delay 2 Trigger*/
    m9222ao.iotConf.mData.m922xAO.fEnDly2 = (TrigCtrl>>10)&0x1;
    if(m9222ao.iotConf.mData.m922xAO.fEnDly2){
        //Delay 2 Trigger Count
        m9222ao.Counter_Value[2] = DLY2Cnt;
    }
    CurrentCard.AO_Status.CurrentConfig = m9222ao.iotConf;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_EAO_CONFIG, &m9222ao)
        !=DAS_ERR_NO)
    {
        return m9222ao.iotConf.wErrorCode;
    }

    return NoError;
}


I16 AO_9223_Config(
    U16     wCardNumber,
    U16     ConfigCtrl,
    U16     TrigCtrl,
    U32     ReTrgCnt,
    U32     DLY1Cnt,
    U32     DLY2Cnt,
    BOOLEAN AutoResetBuf
    )
{
    return AO_9222_Config(wCardNumber, ConfigCtrl, TrigCtrl, ReTrgCnt,
        DLY1Cnt, DLY2Cnt, AutoResetBuf);
}


I16 AO_SetTimeOut(
    U16 wCardNumber,
    U32 dwTimeOut
    )
{
    I16 err;
    DAS_IOT_SI iotSI;

    memset(&iotSI, '\0', sizeof(DAS_IOT_SI));

    if((err = fundamental_check(&CurrentCard_AOFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    switch(CurrentCard.wCardType){
        case PCI_9222:
        case PCI_9223:
        case PCI_6202:
            iotSI.dwScanInterval = dwTimeOut;
            iotSI.dwSampInterval = 0;
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AO_TIMEOUT_SET, &(iotSI))
                !=DAS_ERR_NO)
            {
                return iotSI.wErrorCode;
            }
            break;
        default:
            return ErrorFuncNotSupport;
    }

    return NoError;
}


I16 AO_InitialMemoryAllocated(
    U16 wCardNumber,
    U32 *MemSize
    )
{
    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    switch(CurrentCard.wCardType){
        case PCI_6202:
        case PCI_9222:
        case PCI_9223:
            *MemSize = CurrentCard_AOFunc.dwBufferSize/1024;
            if(*MemSize)
                *MemSize = (*MemSize-4)*256;
            break;
        default:
            return ErrorFuncNotSupport;
    }

    return NoError;
}


I16 AO_ContStatus(
    U16 wCardNumber,
    U16 *Status
    )
{
    DAS_IOT_STATUS iotStat;

    memset(&iotStat, '\0', sizeof(DAS_IOT_STATUS));

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AO_CON_STATUS, &(iotStat))
        !=DAS_ERR_NO)
    {
        return iotStat.wErrorCode;
    }
    *Status = iotStat.mStat.wValue; //AO Status from Kernel

    return NoError;
}


I16 AO_ContBufferCompose(
    U16  wCardNumber,
    U16  TotalChnCount,
    U16  ChnNum,
    U32  UpdateCount,
    void *ConBuffer,
    void *Buffer
    )
{
    U32 vi;
    U32 *TargetBuf;
    U32 *SrcBuf;

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    switch(CurrentCard.wCardType){
        case PCI_6202:
            if((!TotalChnCount)||(TotalChnCount>4))
                return ErrorUndefinedParameter;
            if(ChnNum>3)
                return ErrorUndefinedParameter;
            TargetBuf = (U32 *)ConBuffer;
            SrcBuf = (U32 *)Buffer;
            break;
        case PCI_9222:
        case PCI_9223:
            if((!TotalChnCount)||(TotalChnCount>2))
                return ErrorUndefinedParameter;
            if(ChnNum>1)
                return ErrorUndefinedParameter;
            TargetBuf = (U32 *)ConBuffer;
            SrcBuf = (U32 *)Buffer;
            break;
        default:
            return ErrorFuncNotSupport;
    }

    for(vi=0; vi<UpdateCount; vi++){
        *(TargetBuf+vi*TotalChnCount+ChnNum) = *(SrcBuf+vi);
    }

    return NoError;
}
