#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <sys/ioctl.h>
#include "dask.h"
#include "Adlsys.h"
#include "dask_p.h"
#include "DAQHeader.h"

void W_9221_Read_EEPROM(U32 *lcr_ba, U8 address, U16 *data);

void W_922X_Read_EEPROM(U32 *lcr_ba, U8 address, U16 *data);

void W_922X_TrimDAC_WR(U32 LcrAddr, U8 channel, U8 func, U8 data);

I16 W_922X_AI_CalReadChannel(U16 CardNumber, U16 AdRange, U16 Cal,
    U16 Cal2, U16 CalCMRR, U16 CalSrc, U16 *Value);

void get_local_time(U16 *YEAR, U16 *MONTH, U16 *DAY, U16 *WEEKDAY,
    U16 *HOUR, U16 *MINUTE, U16 *SECOND);

extern REGISTER_CARD_TYPE RegisterCards[MAX_CARD];


I16 W_9221_AD_offset_Average(
    U16 wCardNumber,
    U16 gain,
    U16 *offset_err
    )
{
    DAS_IOT_CTR iotCtr;
    U16 BufferID;
    U16 buffer[256];

    U16 fifo_slice = 0, offset_tmp = 0, offset = 0;
    U32 summy = 0;
    int i;

    memset(&iotCtr, '\0', sizeof(DAS_IOT_CTR));
    memset(buffer, '\0', 256*2);

    AI_ContBufferSetup(wCardNumber, buffer, 256, &BufferID);

    iotCtr.wCounterIndex = gain;
    iotCtr.wBinBcd = BufferID;
    iotCtr.wMode = 0;   //offset

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_CAL_ACT, &(iotCtr))
        !=DAS_ERR_NO)
    {
          return iotCtr.wErrorCode;
    }

    for(i=0;i<256;i++){
        fifo_slice = *(buffer+i);

        if((fifo_slice>>15)>0){
            fifo_slice = fifo_slice&0x7FFF;
        }
        else{
            fifo_slice = fifo_slice+0x8000;
        }
        summy+=fifo_slice;
    }

    offset_tmp = summy/0x100;
    if((offset_tmp-0x8000)>=0){
        offset = 0xFFFF-offset_tmp+0x8001 ;
    }
    else{
        offset = 0xFFFF-(offset_tmp|0x8000)+1;
    }

    AI_ContBufferReset(wCardNumber);

    *offset_err = offset;

    return NoError;
}


I16 W_9221_AD_gain_Average(
    U16 wCardNumber,
    U16 gain,
    U16 *gain_err
    )
{
    DAS_IOT_CTR iotCtr;
    U16 BufferID;
    U16 buffer[256];

    U16 fifo_slice, offset_tmp, offset;
    U32 summy = 0, gain_cal=0;
    int i;

    memset(&iotCtr, '\0', sizeof(DAS_IOT_CTR));
    memset(buffer, '\0', 256*2);

    AI_ContBufferSetup(wCardNumber, buffer, 256, &BufferID);

    iotCtr.wCounterIndex = gain;
    iotCtr.wBinBcd = BufferID;
    iotCtr.wMode = 1;   //gain

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_CAL_ACT, &(iotCtr))
        !=DAS_ERR_NO)
    {
          return iotCtr.wErrorCode;
    }

    for(i=0;i<256;i++){
        fifo_slice = *(buffer+i);
        if((fifo_slice>>15)>0){
            fifo_slice = fifo_slice&0x7FFF;
        }
        else{
            fifo_slice = fifo_slice+0x8000;
        }
        summy+=fifo_slice;
    }

    offset_tmp = summy/0x100;
    if((offset_tmp-0x8000)>=0){
        offset = offset_tmp-0x8000;
    }
    else{
        offset = offset_tmp|0x8000;
    }

    gain_cal = ((gain_cal|(0x7fff-offset))<<16)/offset;
    offset = gain_cal&0xFFFF;

    AI_ContBufferReset(wCardNumber);

    *gain_err = offset;

    return NoError;
}


I16 W_9221_AD_Read_gain_Average(
    U16 wCardNumber,
    U16 gain,
    U16 *gain_err
    )
{
    DAS_IOT_CTR iotCtr;
    U16 BufferID;
    U16 buffer[256];
    U32 dwReadCount = 256;

    U16 fifo_slice, offset_tmp, offset;
    U32 summy = 0;
    int i;

    memset(&iotCtr, '\0', sizeof(DAS_IOT_CTR));
    memset(buffer, '\0', 256*2);

    AI_ContBufferSetup(wCardNumber, buffer, dwReadCount, &BufferID);

    iotCtr.wCounterIndex = gain;
    iotCtr.wBinBcd = BufferID;
    iotCtr.wMode = 2;   //gain

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_CAL_ACT, &(iotCtr))
        !=DAS_ERR_NO)
    {
          return iotCtr.wErrorCode;
    }

    for(i=0;i<256;i++){
        fifo_slice = *(buffer+i);
        if((fifo_slice>>15)>0){
            fifo_slice = fifo_slice&0x7FFF;
        }
        else{
            fifo_slice = fifo_slice+0x8000;
        }
        summy+=fifo_slice;
    }

    offset_tmp = summy/0x100;

    if((offset_tmp-0x8000)>=0){
        offset = offset_tmp-0x8000;
    }
    else{
        offset = offset_tmp|0x8000;
    }

    AI_ContBufferReset(wCardNumber);

    *gain_err = offset;

    return NoError;
}


I16 W_9221_AD_Read_Average(
    U16 wCardNumber,
    U16 gain,
    U16 *gain_err
    )
{
    DAS_IOT_CTR iotCtr;
    U16 BufferID;
    U16 buffer[256];

    U16 fifo_slice, offset_tmp;
    U32 summy = 0;
    int i;

    memset(&iotCtr, '\0', sizeof(DAS_IOT_CTR));
    memset(buffer, '\0', 256*2);

    AI_ContBufferSetup(wCardNumber, buffer, 256, &BufferID);

    iotCtr.wCounterIndex = gain;
    iotCtr.wBinBcd = BufferID;
    iotCtr.wMode = 1;   //gain

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_CAL_ACT, &(iotCtr))
        !=DAS_ERR_NO)
    {
          return iotCtr.wErrorCode;
    }

    for(i=0;i<256;i++){
        fifo_slice = *(buffer+i);
        if((fifo_slice>>15)>0){
            fifo_slice = fifo_slice&0x7FFF;
        }
        else{
            fifo_slice = fifo_slice+0x8000;
        }
        summy+=fifo_slice;
    }

    offset_tmp = summy/0x100;

    AI_ContBufferReset(wCardNumber);

    *gain_err = offset_tmp;

    return NoError;
}


void da_code_transfer(
    U32 dwPort,
    U32 da_out
    )
{
    U32 da_real_out = 0x00000000;
    U32 da_low;
    U32 da_high;

    da_low = da_out&0x1FF;
    da_high = (da_out>>8)&0x1FE;

    da_real_out= (da_high<<9)|da_low;

    *((volatile U32 *)dwPort) = da_real_out;

    return;
}


I16 W_9221_DA_CAL(
    U16 wCardNumber,
    U16 DA_Ch,
    U32 MEMADDR
    )
{
    U32 DA_5v=0x17FFF, DA_0v=0xFFFF, DA_N_5v=0x7FFF;
    U32 da_out;
    U16 ad_read;
    U16 set_gain;
    I16 deviation;
    U16 counter=0;
    I16 err;
    U32 dwPort;
    DAS_IOT_CALCONST_9221 *calPtr;

    //01/11/08
    //calPtr = CurrentCard.CAL_Const;
    calPtr = (DAS_IOT_CALCONST_9221 *)CurrentCard.pCAL_Const;

    if(DA_Ch==0){
        dwPort = MEMADDR+0x94;
        set_gain = 0x6000;
    }
    else{
        dwPort = MEMADDR+0x98;
        set_gain = 0x6800;
    }

    //+5V
    da_out = DA_5v;
    da_code_transfer(dwPort, da_out);
    usleep(5000); //Sleep(5);
    err = W_9221_AD_Read_gain_Average(wCardNumber, set_gain, &ad_read);
    if(err<0)
        return err;

    deviation=0x7fff-ad_read;
    do{
        if(deviation==0x0){
            if(DA_Ch==0)
                da_out = da_out-38;
            else
                da_out = da_out /*-1*/-38;
        }
        else
            da_out = da_out+1;

        da_code_transfer(dwPort, da_out);
        usleep(5000); //Sleep(5);
        err = W_9221_AD_Read_gain_Average(wCardNumber, set_gain, &ad_read);
        if(err<0)
            return err;

        deviation=0x7fff-ad_read;
        counter++;
    }while((deviation!=0x1)&&(counter!=1000));

    if(counter==1000)
        return ErrorCalibrationTimeOut;
    else{
        da_out++;
        if(DA_Ch==0)
            calPtr->constant_DA0[1] = da_out/*(((da_out>>8)&0x1fe)<<9)|(da_out&0x1ff)*/;
        else
            calPtr->constant_DA1[1] = da_out/*(((da_out>>8)&0x1fe)<<9)|(da_out&0x1ff)*/;
    }
    //printf("calPtr->constant_DA%d[1]: %x\n", DA_Ch, da_out);
    counter=0;

    //-5V
    da_out=DA_N_5v;
    da_code_transfer(dwPort, da_out);
    usleep(5000); //Sleep(5);
    err = W_9221_AD_Read_gain_Average(wCardNumber, set_gain, &ad_read);
    if(err<0)
        return err;

    deviation=ad_read-0x8000;
    do{
        if(deviation==0x0){
            if(DA_Ch==0)
                da_out = da_out+38;
            else
                da_out = da_out /*+1*/+38;
        }
        else
            da_out = da_out-1;

        da_code_transfer(dwPort, da_out);
        usleep(5000); //Sleep(5);
        err = W_9221_AD_Read_gain_Average(wCardNumber, set_gain, &ad_read);
        if(err<0)
            return err;

        deviation=ad_read-0x8000;
        counter = counter+1;
    }while((deviation!=0x1)&&(counter!=1000));

    if(counter==1000)
        return ErrorCalibrationTimeOut;
    else{
        da_out--;
        if(DA_Ch==0)
            calPtr->constant_DA0[2] = da_out/*(((da_out>>8)&0x1fe)<<9)|(da_out&0x1ff)*/;
        else
            calPtr->constant_DA1[2] = da_out/*(((da_out>>8)&0x1fe)<<9)|(da_out&0x1ff)*/;
    }
    //printf("calPtr->constant_DA%d[2]: %x\n", DA_Ch, da_out);
    counter=0;

    //0V
    da_out=DA_0v;
    da_code_transfer(dwPort,da_out);
    usleep(5000); //Sleep(5);
    err = W_9221_AD_Read_Average(wCardNumber, set_gain, &ad_read);
    if(err<0)
        return err;

    deviation = ad_read-0x8000;
    do{
        if(ad_read>0x8000){
            if(fabs(deviation)>32)
                da_out = da_out-32;
            else
                da_out = da_out-1;
        }
        else{
            if(fabs(deviation)>32)
                da_out=da_out+32;
            else
                da_out=da_out+1;
        }

        da_code_transfer(dwPort,da_out);
        usleep(5000); //Sleep(5);
        err = W_9221_AD_Read_Average(wCardNumber, set_gain, &ad_read);
        if(err<0)
            return err;

        deviation = ad_read-0x8000;
        counter++;
    }while((fabs(deviation)>1)&&(counter!=1000));

    if(counter==1000)
        return ErrorCalibrationTimeOut;
    else{
        if(ad_read>0x8000)
            da_out--;
        else if(ad_read<0x8000)
            da_out++;

        if(DA_Ch==0)
            calPtr->constant_DA0[0] = da_out/*(((da_out>>8)&0x1fe)<<9)|(da_out&0x1ff)*/;
        else
            calPtr->constant_DA1[0] = da_out/*(((da_out>>8)&0x1fe)<<9)|(da_out&0x1ff)*/;
    }
    //printf("calPtr->constant_DA%d[0]: %x\n", DA_Ch, da_out);

    return NoError;
}


I16 W_9221_DB_Auto_Calibration_ALL(
    U16 wCardNumber
    )
{
    U32 dwPort;
    DAS_IOT_CALCONST_9221 *calPtr;
    U16 offset_err, gain_err;
    int i;
    U32 DA_1v = 0x23331, DA_0_5v = 0x21998, DA_0_2v = 0x206EA;
    U32 adref_read = 0x00000000;
    U16 adread, data;
    I16 deviation;
    I16 err;
    F32 f, LSB_real;
    I16 inaccuracy, DA_5v;

    //dwPort = (CurrentCard.CAL_Const->IOP_mem);
    dwPort = (U32)(CurrentCard_AIFunc.pView[1]);
    //01/14/08
    //calPtr = CurrentCard.CAL_Const;
    calPtr = (DAS_IOT_CALCONST_9221 *)CurrentCard.pCAL_Const;

    *((volatile U32 *)(dwPort+0x9C)) = 0x1ffff;
    *((volatile U32 *)(dwPort+0x98)) = 0x1f000;
    *((volatile U32 *)(dwPort+0x94)) = 0x1f000;

    for(i=0; i<4; i++){
        calPtr->constant_offset_AD[i] = 0;
        calPtr->constant_gain_AD[i] = 0;
    }

    AI_SetTimeOut(wCardNumber, 60000);

    //Offset0 ~ Offset3
    for(i=0; i<4; i++){
        err = W_9221_AD_offset_Average(wCardNumber, i, &offset_err);
        if(err<0)
            return err;
        calPtr->constant_offset_AD[i] = offset_err;
        /*printf("constant_offset_AD[%d]: %x\n", i, calPtr->constant_offset_AD[i]);*/
    }

    //Read 5V constant from board EEPROM
    W_9221_Read_EEPROM((U32 *)dwPort, 32+21, &(data));
    DA_5v = (I16)data;
    W_9221_Read_EEPROM((U32 *)dwPort, 32+22, &(data));

    f = ((float)DA_5v/10)+(float)data/100000;
    LSB_real = f/65535;
    inaccuracy = abs(((f-5)*65535/f)*2.2/4.99);

    //Gain0
    err = W_9221_AD_Read_gain_Average(wCardNumber, 0x4800, &adread);
    if(err<0)
        return err;

    if(f>5)
       adread = adread-inaccuracy;
    else
       adread = adread+inaccuracy;

    adref_read = ((adref_read|(0x7fff-adread))<<16)/adread;

    calPtr->constant_gain_AD[0] = adref_read&0xffff;
    //printf("constant_gain_AD[0]: %x\n",calPtr->constant_gain_AD[0]);

    //Gain1
    *((volatile U32 *)(dwPort+0x94)) = DA_1v;
    usleep(5000); //Sleep(5); 5ms
    err = W_9221_AD_Read_gain_Average(wCardNumber, 0x6000, &adread);
    if(err<0)
        return err;

    deviation = adread-0x1999;
    i = 0;
    do{
        if(adread>0x1999){
            if(fabs(deviation)>10){
                DA_1v = DA_1v-10;
            }
            else{
                DA_1v = DA_1v-1;
            }
        }
        else{
            if(fabs(deviation)>10){
                DA_1v = DA_1v+10;
            }
            else{
                DA_1v = DA_1v+1;
            }
        }

        *((volatile U32 *)(dwPort+0x94)) = DA_1v;
        usleep(5000); //Sleep(5);
        err = W_9221_AD_Read_gain_Average(wCardNumber, 0x6000, &adread);
        if(err<0)
            return err;
        deviation = adread-0x1999;

        if((++i)>10000)
            return ErrorCalibrationTimeOut;
    }while(fabs(deviation)>1);

    err = W_9221_AD_gain_Average(wCardNumber, 0x6100, &gain_err);
    if(err<0)
        return err;
    calPtr->constant_gain_AD[1] = gain_err;
    //printf("constant_gain_AD[1]: %x\n",calPtr->constant_gain_AD[1]);

    //Gain2
    *((volatile U32 *)(dwPort+0x94)) = DA_0_5v;
    usleep(5000); //Sleep(5);
    err = W_9221_AD_Read_gain_Average(wCardNumber, 0x6100, &adread);
    if(err<0)
        return err;

    deviation = adread-0x3FFF;
    i = 0;
    do{
        if(adread>0x3FFF){
            if(fabs(deviation)>3){
                DA_0_5v = DA_0_5v-3;
            }
            else{
                DA_0_5v = DA_0_5v-1;
            }
        }
        else{
            if(fabs(deviation)>3){
                DA_0_5v = DA_0_5v+3;
            }
            else{
                DA_0_5v = DA_0_5v+1;
            }
        }

        *((volatile U32 *)(dwPort+0x94)) = DA_0_5v;
        usleep(5000); //Sleep(5);
        err = W_9221_AD_Read_gain_Average(wCardNumber, 0x6100, &adread);
        if(err<0)
            return err;
        deviation = adread-0x3FFF;

        if((++i)>10000)
            return ErrorCalibrationTimeOut;
    }while(fabs(deviation)>3);

    err = W_9221_AD_gain_Average(wCardNumber, 0x6200, &gain_err);
    if(err<0)
        return err;
    calPtr->constant_gain_AD[2] = gain_err;
    //printf("constant_gain_AD[2]: %x\n",calPtr->constant_gain_AD[2]);

    //Gain3
    *((volatile U32 *)(dwPort+0x94)) = DA_0_2v;
    usleep(5000); //Sleep(5);
    err = W_9221_AD_Read_gain_Average(wCardNumber, 0x6200, &adread);
    if(err<0)
        return err;

    deviation = adread-0x3332;
    i = 0;
    do{
        if(adread>0x3332)
            DA_0_2v = DA_0_2v-1;
        else
            DA_0_2v = DA_0_2v+1;

        *((volatile U32 *)(dwPort+0x94)) = DA_0_2v;
        usleep(5000); //Sleep(5);
        err = W_9221_AD_Read_gain_Average(wCardNumber, 0x6200, &adread);
        if(err<0)
            return err;

        deviation = adread-0x3332;

        if((++i)>10000)
            return ErrorCalibrationTimeOut;
    }while(fabs(deviation)>4);

    err = W_9221_AD_gain_Average(wCardNumber, 0x6300, &gain_err);
    if(err<0)
        return err;
    calPtr->constant_gain_AD[3] = gain_err;
    //printf("constant_gain_AD[3]: %x\n",calPtr->constant_gain_AD[3]);

    //Write to register
    *((volatile U32 *)(dwPort+0xc0)) = \
        ((calPtr->constant_offset_AD[1])<<16)|(calPtr->constant_offset_AD[0]);
    *((volatile U32 *)(dwPort+0xc4)) = \
        ((calPtr->constant_offset_AD[3])<<16)|(calPtr->constant_offset_AD[2]);
    *((volatile U32 *)(dwPort+0xc8)) = \
        ((calPtr->constant_gain_AD[1])<<16)|(calPtr->constant_gain_AD[0]);
    *((volatile U32 *)(dwPort+0xcc)) = \
        ((calPtr->constant_gain_AD[3])<<16)|(calPtr->constant_gain_AD[2]);

    err = W_9221_DA_CAL(wCardNumber, 0, dwPort);
    if(err<0)
        return err;
    err = W_9221_DA_CAL(wCardNumber, 1, dwPort);
    if(err<0)
        return err;

    return NoError;
}


I16 W_9524_DB_Auto_Calibration_ALL(
    U16 wCardNumber
    )
{
    DAS_IOT_CTR iotCtr;
    DAS_IOT_CALCONST_9524 *calPtr;
    U16 YEAR, MONTH, DAY;

    memset(&iotCtr, '\0', sizeof(DAS_IOT_CTR));

    iotCtr.wMode = 0;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_CAL_ACT, &(iotCtr))
        !=DAS_ERR_NO)
    {
          return iotCtr.wErrorCode;
    }

    calPtr = (DAS_IOT_CALCONST_9524 *)CurrentCard.pCAL_Const;
    get_local_time(
        &YEAR, &MONTH, &DAY, (U16 *)NULL, (U16 *)NULL, (U16 *)NULL, (U16 *)NULL
    );
    calPtr->SysExtCalConst.EE_YEAR = YEAR;
    calPtr->SysExtCalConst.EE_MMDD = MONTH*100+DAY;

    return NoError;
}


U8 optimize_val_code(
    U16 array_num,
    F64 *array_data,
    F64 *min_data
    )
{
    U16 i;
    U8 min_index;

    *min_data = array_data[0];
    min_index = 0;
    for(i=1; i<=array_num; i++){
        if(fabs(*min_data)>fabs(array_data[i])){
            *min_data = array_data[i];
            min_index = (U8)i;
        }
    }

    return min_index;
}


I16 W_922X_AO_SetRef(
    U32 LcrAddr,
    U16 wChannel,
    I16 wEnable
    )
{
    U32 tmp_value;

    tmp_value = (*((volatile U32 *)(LcrAddr+0x498)))|((U32)wChannel&0x07);
    if(wEnable){
        tmp_value |= 0x10;
    }
    else{
        tmp_value &= 0xffffffef ;
    }
    *((volatile U32 *)(LcrAddr+0x498)) = tmp_value;

    return NoError;
}


I16 W_922X_AO_RegWriteChannel(
    U32 LcrAddr,
    U16 wChannel,
    I16 wValue
    )
{
    *((volatile U32 *)(LcrAddr+0x10)) = 0x80;
    *((volatile U32 *)(LcrAddr+0x2c)) = (U32)wValue;
    *((volatile U32 *)(LcrAddr+0x498)) = \
        ((*((volatile U32 *)(LcrAddr+0x498)))&0x10)|((U32)wChannel&0x3)|0x40;

    return NoError;
}


I16 W_922X_SetCalConfig(
    U32 LcrAddr,
    U16 wCal,
    U16 wCal2,
    U16 wCalCMRR,
    U16 wCalSrc
    )
{
    U32 AIConfigRB;

    if(wCal>1)
        return -1;
    if(wCal2>1)
        return -1;
    if(wCalCMRR>1)
        return -1;
    if(wCalSrc>7)
        return -2;

    AIConfigRB = *((volatile U32 *)(LcrAddr+0x40C));
    AIConfigRB &= (~(0x3f<<11));
    AIConfigRB |= wCalSrc<<11;
    AIConfigRB |= wCal<<14;
    AIConfigRB |= wCalCMRR<<15;
    AIConfigRB |= wCal2<<17;
    *((volatile U32 *)(LcrAddr+0x40C)) = AIConfigRB;

    return NoError;
}


I16 W_922X_AO_CAL(
    U16 wCardNumber
    )
{
    I16 err;
    U16 i;
    U8 j, k;
    U8 n_channel;
    U16 AO_cal_sel[2] = {2, 3};
    F64 optimize_val[256];
    F64 GAN_POS;
    F64 GAN_NEG;
    F64 min_value_LTY;
    F64 min_value_OS;
    F64 min_value_GAIN;
    F64 OFT_ZERO;
    DAS_IOT_CALCONST_922X *calPtr;
    U32 *p;
    U32 LcrAddr;
    U16 CalValue;
    F64 CalVoltage[11];

    calPtr = (DAS_IOT_CALCONST_922X *)CurrentCard.pCAL_Const;
    p = (U32 *)(CurrentCard_AIFunc.pView[1]);
    LcrAddr = (U32)(CurrentCard_AIFunc.pView[1]);

    for(n_channel=0; n_channel<2; n_channel++){
        W_922X_SetCalConfig(LcrAddr, 1, 0, 0, AO_cal_sel[n_channel]);

        /*-----------------------*/
        /* Linearity calibration */
        /*-----------------------*/
        W_922X_AO_SetRef(LcrAddr, n_channel, 0);
        for(i=0; i<256; i++){
            //printf("Channel: %d, Linear: %d\n", n_channel, i);
            GAN_POS = 0;
            GAN_NEG = 0;

            W_922X_TrimDAC_WR(LcrAddr, n_channel, 0, (U8)i);
            usleep(5000);//Sleep(5);

            W_922X_AO_RegWriteChannel(LcrAddr, n_channel, (I16)0xffff);
            usleep(5000);//Sleep(5);
            for(j=0; j<8; j++){
                usleep(1000);//Sleep(1);
                CalVoltage[10] = 0.0;
                for(k=0; k<10; k++){
                    err = W_922X_AI_CalReadChannel(wCardNumber, AD_B_10_V, 1,
                        0, 0, AO_cal_sel[n_channel], &CalValue
                    );
                    if(err!=NoError)
                        return err;
                    AI_VoltScale(wCardNumber, AD_B_10_V, CalValue,
                        &CalVoltage[k]
                    );
                    CalVoltage[10] += CalVoltage[k];
                }
                CalVoltage[10] /= 10;
                GAN_POS += CalVoltage[10];
            }

            W_922X_AO_RegWriteChannel(LcrAddr, n_channel, 0);
            usleep(5000);//Sleep(5);
            for(j=0; j<8; j++){
                usleep(1000);//Sleep(1);
                CalVoltage[10] = 0.0;
                for(k=0; k<10; k++){
                    err = W_922X_AI_CalReadChannel(wCardNumber, AD_B_10_V, 1,
                        0, 0, AO_cal_sel[n_channel], &CalValue
                    );
                    if(err!=NoError)
                        return err;
                    AI_VoltScale(wCardNumber, AD_B_10_V, CalValue,
                        &CalVoltage[k]
                    );
                    CalVoltage[10] += CalVoltage[k];
                }
                CalVoltage[10] /= 10;
                GAN_NEG += CalVoltage[10];
            }

            optimize_val[i] = fabs(GAN_POS-GAN_NEG)/8;
            //printf("    Optimize Value: %lf\n", optimize_val[i]);
        }
        W_922X_TrimDAC_WR(LcrAddr, n_channel, 0,
            optimize_val_code(255, optimize_val, &min_value_LTY)
        );
        calPtr->CALConst_DA_Linear[n_channel] = \
            optimize_val_code(255, optimize_val, &min_value_LTY);

        /*----------------------------------*/
        /* Bipolar Offset Error calibration */
        /*----------------------------------*/
        AO_WriteChannel(wCardNumber, n_channel, (I16)0x8000);
        for(i=0; i<256; i++){
            //printf("Channel: %d, Offset: %d\n", n_channel, i);
            OFT_ZERO = 0;

            W_922X_TrimDAC_WR(LcrAddr, n_channel, 1, (U8)i);
            usleep(5000);//Sleep(5);
            for(j=0; j<8; j++){
                usleep(1000);//Sleep(1);
                CalVoltage[10] = 0.0;
                for(k=0; k<10; k++){
                    err = W_922X_AI_CalReadChannel(wCardNumber, AD_B_10_V, 1,
                        0, 0, AO_cal_sel[n_channel], &CalValue
                    );
                    if(err!=NoError)
                        return err;
                    AI_VoltScale(wCardNumber, AD_B_10_V, CalValue,
                        &CalVoltage[k]
                    );
                    CalVoltage[10] += CalVoltage[k];
                }
                CalVoltage[10] /= 10;
                OFT_ZERO += CalVoltage[10];
            }

            optimize_val[i] = fabs(OFT_ZERO)/8;
            //printf("    Optimize Value: %lf\n", optimize_val[i]);
        }
        W_922X_TrimDAC_WR(LcrAddr, n_channel, 1,
            optimize_val_code(255, optimize_val, &min_value_OS)
        );
        calPtr->CALConst_DA_Offset[n_channel] = \
            optimize_val_code(255, optimize_val, &min_value_OS);

        /*--------------------------------*/
        /* Bipolar Gain Error calibration */
        /*--------------------------------*/
        for(i=0; i<256; i++){
            //printf("Channel: %d, Gain: %d\n", n_channel, i);
            GAN_POS = 0;
            GAN_NEG = 0;

            W_922X_TrimDAC_WR(LcrAddr, n_channel, 2, (U8)i);
            usleep(5000);//Sleep(5);

            AO_WriteChannel(wCardNumber, n_channel, (I16)(0xffff));
            usleep(5000);//Sleep(5);
            for(j=0; j<8; j++){
                usleep(1000);//Sleep(1);
                CalVoltage[10] = 0.0;
                for(k=0; k<10; k++){
                    err = W_922X_AI_CalReadChannel(wCardNumber, AD_B_10_V, 1,
                        0, 0, AO_cal_sel[n_channel], &CalValue
                    );
                    if(err!=NoError)
                        return err;
                    AI_VoltScale(wCardNumber, AD_B_10_V, CalValue,
                        &CalVoltage[k]
                    );
                    CalVoltage[10] += CalVoltage[k];
                }
                CalVoltage[10] /= 10;
                GAN_POS += CalVoltage[10];
            }

            AO_WriteChannel(wCardNumber, n_channel, 0);
            usleep(5000);//Sleep(5);
            for(j=0; j<8; j++){
                usleep(1000);//Sleep(1);
                CalVoltage[10] = 0.0;
                for(k=0; k<10; k++){
                    err = W_922X_AI_CalReadChannel(wCardNumber, AD_B_10_V, 1,
                        0, 0, AO_cal_sel[n_channel], &CalValue
                    );
                    if(err!=NoError)
                        return err;
                    AI_VoltScale(wCardNumber, AD_B_10_V, CalValue,
                        &CalVoltage[k]
                    );
                    CalVoltage[10] += CalVoltage[k];
                }
                CalVoltage[10] /= 10;
                GAN_NEG += CalVoltage[10];
            }

            optimize_val[i] = (GAN_POS-GAN_NEG)/8;
            optimize_val[i] = fabs(optimize_val[i]-19.9996);
            //printf("    Optimize Value: %lf\n", optimize_val[i]);
        }

        W_922X_TrimDAC_WR(LcrAddr, n_channel, 2,
            optimize_val_code(255, optimize_val, &min_value_GAIN)
        );
        calPtr->CALConst_DA_Gain[n_channel] = \
            optimize_val_code(255, optimize_val, &min_value_GAIN);
    }

    return NoError;
}


void W_922X_AO_CalPWM(
    U32 LcrAddr,
    U32 WriteData
    )
{
    *((volatile U32 *)(LcrAddr+0x420)) = (1<<31)|(WriteData&0x3ffff);
}


I16 W_922X_AI_CalReadChannel(
    U16 wCardNumber,
    U16 wAdRange,
    U16 wCal,
    U16 wCal2,
    U16 wCalCMRR,
    U16 wCalSrc,
    U16 *pwValue
    )
{
    U16 wGain, calconfig;
    DAS_IOT_PIO iotPio = {0};

    if(!validate_ad_range(wCardNumber, wAdRange, &wGain))
        return ErrorInvalidAdRange;

    if(wCal>1)
        return -1;
    if(wCal2>1)
        return -1;
    if(wCalCMRR>1)
        return -1;
    if(wCalSrc>7)
        return -2;
    calconfig = (wCalSrc&0x7);
    calconfig |= ((wCal&0x1)<<3);
    calconfig |= ((wCalCMRR&0x1)<<4);
    calconfig |= ((wCal2&0x1)<<5);

    iotPio.wChannel = calconfig|0x8000;
    iotPio.wLine_Gain = wGain;
    iotPio.wByteCount = 2;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_AI_PIO_READ, &(iotPio))
        !=DAS_ERR_NO)
    {
        return iotPio.wErrorCode;
    }
    *pwValue = (U16)iotPio.dwValue;

    return NoError;
}


I16 W_922X_AI_CAL(
    U16 wCardNumber
    )
{
    I16 err;
    U16 read_val_code = 0;
    I32 total_code = 0;
    U16 Gain_array[8];
    U16 i, j;
    I32 offset_gain_code[8];
    U32 gain_error[8];
    DAS_IOT_CALCONST_922X *calPtr;
    U32 *p;
    U32 LcrAddr;
    U32 code_volt[8];
    U16 EE_Read, EE_Read2;
    U16 CapPoint = 100;

    calPtr = (DAS_IOT_CALCONST_922X *)CurrentCard.pCAL_Const;
    p = (U32 *)(CurrentCard_AIFunc.pView[1]);
    LcrAddr = (U32)(CurrentCard_AIFunc.pView[1]);
    Gain_array[0] = AD_B_10_V;
    Gain_array[1] = AD_B_5_V;
    Gain_array[2] = AD_B_2_5_V;
    Gain_array[3] = AD_B_2_V;
    Gain_array[4] = AD_B_1_25_V;
    Gain_array[5] = AD_B_1_V;
    Gain_array[6] = AD_B_0_5_V;
    Gain_array[7] = AD_B_0_25_V;

    /*Write AI Offset/Gain Register to Zero*/
    *((volatile U32 *)(LcrAddr+0x588)) = 0;
    *((volatile U32 *)(LcrAddr+0x58C)) = 0;
    *((volatile U32 *)(LcrAddr+0x590)) = 0;
    *((volatile U32 *)(LcrAddr+0x594)) = 0;
    *((volatile U32 *)(LcrAddr+0x598)) = 0;
    *((volatile U32 *)(LcrAddr+0x59C)) = 0;
    *((volatile U32 *)(LcrAddr+0x5A0)) = 0;
    *((volatile U32 *)(LcrAddr+0x5A4)) = 0;

    /*-------------------------------*/
    /* AI Bipolar offset calibration */
    /*-------------------------------*/
    for(i=0; i<8; i++){
        total_code = 0;
        for(j=0; j<CapPoint; j++){
            if((err = W_922X_AI_CalReadChannel(wCardNumber, Gain_array[i], 1, 0,
                0, 1, &read_val_code))!=NoError)
            {
                return err;
            }
            usleep(10000);//Sleep(10);
            total_code += (I16)read_val_code;
        }
        offset_gain_code[i] = total_code/CapPoint;

        if(offset_gain_code[i]<=0){
            offset_gain_code[i] += 65536;
        }
        calPtr->CALConst_AD_Offset[i] = (U16)(65536-offset_gain_code[i]);
    }

    /*Set Gain to 1 for AO_CalPWM*/
    if((err = W_922X_AI_CalReadChannel(wCardNumber, AD_B_10_V, 1, 0, 0, 1,
        &read_val_code))!=NoError)
    {
        return err;
    }

    /*Write To AI Offset Register*/
    *((volatile U32 *)(LcrAddr+0x588)) = \
        calPtr->CALConst_AD_Offset[0]|(calPtr->CALConst_AD_Offset[1]<<16);
    *((volatile U32 *)(LcrAddr+0x58C)) = \
        calPtr->CALConst_AD_Offset[2]|(calPtr->CALConst_AD_Offset[3]<<16);
    *((volatile U32 *)(LcrAddr+0x590)) = \
        calPtr->CALConst_AD_Offset[4]|(calPtr->CALConst_AD_Offset[5]<<16);
    *((volatile U32 *)(LcrAddr+0x594)) = \
        calPtr->CALConst_AD_Offset[6]|(calPtr->CALConst_AD_Offset[7]<<16);


    /*-----------------------------*/
    /* AI Bipolar Gain calibration */
    /*-----------------------------*/
    for(i=0; i<8; i++){
        W_922X_Read_EEPROM(p, (U8)(15+i*2), &EE_Read);
        W_922X_Read_EEPROM(p, (U8)(16+i*2), &EE_Read2);
        code_volt[i] = (EE_Read2<<16)+EE_Read;
    }

    for(i=0; i<8; i++){
        total_code = 0;
        W_922X_AO_CalPWM(LcrAddr, code_volt[i]);
        sleep(2);//Sleep(1500);
        for(j=0; j<CapPoint; j++){
            if((err = W_922X_AI_CalReadChannel(wCardNumber, Gain_array[i], 1, 0,
                0, 6, &read_val_code))!=NoError)
            {
                return err;
            }
            usleep(10000);//Sleep(10);
            total_code += (I16)read_val_code;
        }
        gain_error[i] = \
            ((0X7fff-(total_code/CapPoint))*0x10000)/(total_code/CapPoint);
        calPtr->CALConst_AD_Gain[i] = (U16)gain_error[i];
    }

    /*Set Gain to 1 for AO_CalPWM*/
    if((err = W_922X_AI_CalReadChannel(wCardNumber, AD_B_10_V, 1, 0, 0, 1,
        &read_val_code))!=NoError)
    {
        return err;
    }

    /*Write To AI Gain Register*/
    *((volatile U32 *)(LcrAddr+0x598)) =
        calPtr->CALConst_AD_Gain[0]|(calPtr->CALConst_AD_Gain[1]<<16);
    *((volatile U32 *)(LcrAddr+0x59C)) =
        calPtr->CALConst_AD_Gain[2]|(calPtr->CALConst_AD_Gain[3]<<16);
    *((volatile U32 *)(LcrAddr+0x5A0)) =
        calPtr->CALConst_AD_Gain[4]|(calPtr->CALConst_AD_Gain[5]<<16);
    *((volatile U32 *)(LcrAddr+0x594)) =
        calPtr->CALConst_AD_Gain[6]|(calPtr->CALConst_AD_Gain[7]<<16);

    return NoError;
}


I16 W_922X_AI_CALEx(
    U16 wCardNumber
    )
{
    I16 err;
    U16 read_val_code = 0;
    I32 total_code = 0;
    U16 Gain_array[2];
    U16 i, j;
    I32 offset_gain_code[2];
    U32 gain_error[2];
    DAS_IOT_CALCONST_922X *calPtr;
    U32 *p;
    U32 LcrAddr;
    //U32 code_volt[2];
    //U16 EE_Read, EE_Read2;
    U16 CapPoint = 5000;
    U16 DIF10V = 0;
    U16 gain_set[2] = {2, 1};

    calPtr = (DAS_IOT_CALCONST_922X *)CurrentCard.pCAL_Const;
    p = (U32 *)(CurrentCard_AIFunc.pView[1]);
    LcrAddr = (U32)(CurrentCard_AIFunc.pView[1]);
    Gain_array[0] = AD_B_10_V;
    Gain_array[1] = AD_B_5_V;

    /*Write AI Offset/Gain Register to Zero*/
    *((volatile U32 *)(LcrAddr+0x588)) = 0;
    *((volatile U32 *)(LcrAddr+0x598)) = 0;

    /*-------------------------------*/
    /* AI Bipolar offset calibration */
    /*-------------------------------*/
    for(i=0; i<2; i++){
        total_code = 0;
        for(j=0; j<CapPoint; j++){
            if((err = W_922X_AI_CalReadChannel(wCardNumber, Gain_array[i], 1, 0,
                0, 0, &read_val_code))!=NoError)
            {
                return err;
            }
            usleep(10000);//Sleep(10);
            total_code += (I16)read_val_code;
        }
        offset_gain_code[i] = total_code/CapPoint;

        if(offset_gain_code[i]<=0){
            offset_gain_code[i] += 65536;
        }
        calPtr->CALConst_AD_Offset[i] = (U16)(65536-offset_gain_code[i]);
    }

    /*Set Gain to 1 to avoid circuit clamping*/
    if((err = W_922X_AI_CalReadChannel(wCardNumber, AD_B_10_V, 1, 0, 0, 0,
        &read_val_code))!=NoError)
    {
        return err;
    }

    /*Write To AI Offset Register*/
    *((volatile U32 *)(LcrAddr+0x588)) = \
        calPtr->CALConst_AD_Offset[0]|(calPtr->CALConst_AD_Offset[1]<<16);

    /*-----------------------------*/
    /* AI Bipolar Gain calibration */
    /*-----------------------------*/

    W_922X_Read_EEPROM(p, 39, &DIF10V);
    //DIF10V = 0;

    for(i=0; i<2; i++){
        total_code = 0;
        //W_922X_AO_CalPWM(LcrAddr, code_volt[i]);
        //Sleep(1500);
        usleep(10000);//Sleep(10);
        for(j=0; j<CapPoint; j++){
            if((err = W_922X_AI_CalReadChannel(wCardNumber, Gain_array[i], 1, 0,
                0, gain_set[i], &read_val_code))!=NoError)
            {
                return err;
            }
            usleep(10000);//Sleep(10);
            total_code += (I16)read_val_code;
        }
        if(i==0){
            gain_error[i] = ((0X7fff-DIF10V-(total_code/CapPoint))*0x10000)/
                (total_code/CapPoint);
        }
        else{
            gain_error[i] = ((0X7fff-(total_code/CapPoint))*0x10000)/
                (total_code/CapPoint);
        }
        calPtr->CALConst_AD_Gain[i] = (U16)gain_error[i];
    }

    /*Write To AI Gain Register*/
    *((volatile U32 *)(LcrAddr+0x598)) = \
        calPtr->CALConst_AD_Gain[0]|(calPtr->CALConst_AD_Gain[1]<<16);

    /*Set Gain to 1 to avoid circuit clamping*/
    if((err = W_922X_AI_CalReadChannel(wCardNumber, AD_B_10_V, 1, 0, 0, 0,
        &read_val_code))!=NoError)
    {
        return err;
    }

    return NoError;
}


I16 W_922X_DB_Auto_Calibration_ALL(
    U16 wCardNumber
    )
{
    I16 err;

    if(CurrentCard.wCardType==PCI_922A){
        if((err = W_922X_AI_CALEx(wCardNumber))!=NoError)
            return err;
    }
    else{
        if((err = W_922X_AI_CAL(wCardNumber))!=NoError)
            return err;
        if((err = W_922X_AO_CAL(wCardNumber))!=NoError)
            return err;
    }

    return NoError;
}


I16 PCI_DB_Auto_Calibration_ALL(
    U16 wCardNumber
    )
{
    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    switch(CurrentCard.wCardType){
        case PCI_9221:
            return W_9221_DB_Auto_Calibration_ALL(wCardNumber);
        case PCI_9222:
        case PCI_9223:
        case PCI_922A:
            return W_922X_DB_Auto_Calibration_ALL(wCardNumber);
        case PCI_9524:
            return W_9524_DB_Auto_Calibration_ALL(wCardNumber);
        default:
            return ErrorFuncNotSupport;
    }
}


void W_9221_Read_EEPROM(
    U32 *lcr_ba,
    U8  address,
    U16 *data
    )
{
    int i;
    U8 addr_temp;

    *data = 0;

    //chip select enable
    *((volatile U8 *)lcr_ba+3) = 0x00;
    *((volatile U8 *)lcr_ba+3) = 0x02;
    *((volatile U8 *)lcr_ba+3) = 0x03;
    *((volatile U8 *)lcr_ba+3) = 0x02;

    //start bit
    *((volatile U8 *)lcr_ba+3) = 0x06;
    *((volatile U8 *)lcr_ba+3) = 0x07;
    *((volatile U8 *)lcr_ba+3) = 0x06;

    //read command 2'b10
    *((volatile U8 *)lcr_ba+3) = 0x07;
    *((volatile U8 *)lcr_ba+3) = 0x06;
    *((volatile U8 *)lcr_ba+3) = 0x02;
    *((volatile U8 *)lcr_ba+3) = 0x03;
    *((volatile U8 *)lcr_ba+3) = 0x02;

    //8 bits address
    for(i=0; i<=7; i++){
        addr_temp = (address&(0x80>>i));
        if(i<5)
            addr_temp = addr_temp>>(5-i);
        else if(i>5)
            addr_temp = addr_temp<<(i-5);
        *((volatile U8 *)lcr_ba+3) = (0x02|addr_temp);
        *((volatile U8 *)lcr_ba+3) = (0x03|addr_temp);
        *((volatile U8 *)lcr_ba+3) = (0x02|addr_temp);
    }

    //read 16 bis data
    for(i=0; i<=15; i++){
        *((volatile U8 *)lcr_ba+3) = 0x22;
        *((volatile U8 *)lcr_ba+3) = 0x33;

        if(i<=4){
            *data = (*data)|
                (((*((volatile U16 *)((volatile U8 *)lcr_ba+2)))&0x0800)<<(4-i));
        }
        else{
            *data = (*data)|
                (((*((volatile U16 *)((volatile U8 *)lcr_ba+2)))&0x0800)>>(i-4));
        }
    }

    *((volatile U8 *)lcr_ba+3) = 0x22;
    *((volatile U8 *)lcr_ba+3) = 0x00;

    return;
}


void W_9221_Load_CAL_Constant(
    I16 wCardNumber,
    U16 bank
    )
{
    U8 bank_start = 0;
    U16 val, val2;
    U32 *p;
    U32 Lcr_Addr; //07/23/08
    DAS_IOT_CALCONST_9221 *calPtr;

    switch(bank){
        case EEPROM_DEFAULT_BANK:
            bank_start = 0;
            break;
        case EEPROM_USER_BANK1:
            bank_start = 32;
            break;
    }

    //01/14/08
    //calPtr = CurrentCard.CAL_Const;
    calPtr = (DAS_IOT_CALCONST_9221 *)CurrentCard.pCAL_Const;
    //p = (U32 *)(CurrentCard.CAL_Const->IOP_mem);
    p = (U32 *)(CurrentCard_AIFunc.pView[1]);
    Lcr_Addr = (U32)(CurrentCard_AIFunc.pView[1]);

    /*offset*/
    W_9221_Read_EEPROM(p, bank_start+1, &val);
    calPtr->constant_offset_AD[0] = val;
    W_9221_Read_EEPROM(p, bank_start+2, &val);
    calPtr->constant_offset_AD[1] = val;
    W_9221_Read_EEPROM(p, bank_start+3, &val);
    calPtr->constant_offset_AD[2] = val;
    W_9221_Read_EEPROM(p, bank_start+4, &val);
    calPtr->constant_offset_AD[3] = val;

    /*gain*/
    W_9221_Read_EEPROM(p, bank_start+5, &val);
    calPtr->constant_gain_AD[0] = val;
    W_9221_Read_EEPROM(p, bank_start+6, &val);
    calPtr->constant_gain_AD[1] = val;
    W_9221_Read_EEPROM(p, bank_start+7, &val);
    calPtr->constant_gain_AD[2] = val;
    W_9221_Read_EEPROM(p, bank_start+8, &val);
    calPtr->constant_gain_AD[3] = val;

    //Write to register
    //07/23/08 modified
    *((volatile U32 *)(Lcr_Addr+0xc0)) = \
        ((calPtr->constant_offset_AD[1])<<16)|(calPtr->constant_offset_AD[0]);
    *((volatile U32 *)(Lcr_Addr+0xc4)) = \
        ((calPtr->constant_offset_AD[3])<<16)|(calPtr->constant_offset_AD[2]);
    *((volatile U32 *)(Lcr_Addr+0xc8)) = \
        ((calPtr->constant_gain_AD[1])<<16)|(calPtr->constant_gain_AD[0]);
    *((volatile U32 *)(Lcr_Addr+0xcc)) = \
        ((calPtr->constant_gain_AD[3])<<16)|(calPtr->constant_gain_AD[2]);
    /*
    *(p+0xc0) = ((calPtr->constant_offset_AD[1])<<16)|\
    (calPtr->constant_offset_AD[0]);
    *(p+0xc4) = ((calPtr->constant_offset_AD[3])<<16)|\
    (calPtr->constant_offset_AD[2]);
    *(p+0xc8) = ((calPtr->constant_gain_AD[1])<<16)|\
    (calPtr->constant_gain_AD[0]);
    *(p+0xcc) = ((calPtr->constant_gain_AD[3])<<16)|\
    (calPtr->constant_gain_AD[2]);
    */

    /*0V*/
    W_9221_Read_EEPROM(p, bank_start+9, &val);
    W_9221_Read_EEPROM(p, bank_start+10, &val2);
    calPtr->constant_DA0[0] = ((val2&0x1fe)<<8)|(val&0x1ff);

    /*+5V*/
    W_9221_Read_EEPROM(p, bank_start+11, &val);
    W_9221_Read_EEPROM(p, bank_start+12, &val2);
    calPtr->constant_DA0[1] = ((val2&0x1fe)<<8)|(val&0x1ff);

    /*-5V*/
    W_9221_Read_EEPROM(p, bank_start+13, &val);
    W_9221_Read_EEPROM(p, bank_start+14, &val2);
    calPtr->constant_DA0[2] = ((val2&0x1fe)<<8)|(val&0x1ff);

    /*0V*/
    W_9221_Read_EEPROM(p, bank_start+15, &val);
    W_9221_Read_EEPROM(p, bank_start+16, &val2);
    calPtr->constant_DA1[0] = ((val2&0x1fe)<<8)|(val&0x1ff);

    /*+5V*/
    W_9221_Read_EEPROM(p, bank_start+17, &val);
    W_9221_Read_EEPROM(p, bank_start+18, &val2);
    calPtr->constant_DA1[1] = ((val2&0x1fe)<<8)|(val&0x1ff);

    /*-5V*/
    W_9221_Read_EEPROM(p, bank_start+19, &val);
    W_9221_Read_EEPROM(p, bank_start+20, &val2);
    calPtr->constant_DA1[2] = ((val2&0x1fe)<<8)|(val&0x1ff);
/*
    printf("constant_offset_AD[0]: %x\n", calPtr->constant_offset_AD[0]);
    printf("constant_offset_AD[1]: %x\n", calPtr->constant_offset_AD[1]);
    printf("constant_offset_AD[2]: %x\n", calPtr->constant_offset_AD[2]);
    printf("constant_offset_AD[3]: %x\n", calPtr->constant_offset_AD[3]);

    printf("constant_gain_AD[0]: %x\n", calPtr->constant_gain_AD[0]);
    printf("constant_gain_AD[1]: %x\n", calPtr->constant_gain_AD[1]);
    printf("constant_gain_AD[2]: %x\n", calPtr->constant_gain_AD[2]);
    printf("constant_gain_AD[3]: %x\n", calPtr->constant_gain_AD[3]);

    printf("constant_DA0[0]: %x\n", calPtr->constant_DA0[0]);
    printf("constant_DA0[1]: %x\n", calPtr->constant_DA0[1]);
    printf("constant_DA0[2]: %x\n", calPtr->constant_DA0[2]);

    printf("constant_DA1[0]: %x\n", calPtr->constant_DA1[0]);
    printf("constant_DA1[1]: %x\n", calPtr->constant_DA1[1]);
    printf("constant_DA1[2]: %x\n", calPtr->constant_DA1[2]);

    W_9221_Read_EEPROM(p, bank_start+21, &val);
    W_9221_Read_EEPROM(p, bank_start+22, &val2);
    printf("5V: %f\n", (float)val/10+(float)val2/100000);
*/
    return;
}


I16 W_9524_Load_CAL_Constant(
    U16 wCardNumber,
    U16 bank
    )
{
    DAS_IOT_CTR iotCtr;

    memset(&iotCtr, '\0', sizeof(DAS_IOT_CTR));

    iotCtr.wMode = 2;
    iotCtr.wState = bank;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_CAL_ACT, &(iotCtr))
        !=DAS_ERR_NO)
    {
          return iotCtr.wErrorCode;
    }

    return NoError;
}


void W_922X_Read_EEPROM(
    U32 *lcr_ba,
    U8  address,
    U16 *data
    )
{
    int i;
    U8 addr_temp;

    *data = 0;

    //chip select enable
    *((volatile U8 *)lcr_ba+9) = 0x00;
    *((volatile U8 *)lcr_ba+9) = 0x02;
    *((volatile U8 *)lcr_ba+9) = 0x03;
    *((volatile U8 *)lcr_ba+9) = 0x02;

    //start bit
    *((volatile U8 *)lcr_ba+9) = 0x06;
    *((volatile U8 *)lcr_ba+9) = 0x07;
    *((volatile U8 *)lcr_ba+9) = 0x06;

    //read command 2'b10
    *((volatile U8 *)lcr_ba+9) = 0x07;
    *((volatile U8 *)lcr_ba+9) = 0x06;
    *((volatile U8 *)lcr_ba+9) = 0x02;
    *((volatile U8 *)lcr_ba+9) = 0x03;
    *((volatile U8 *)lcr_ba+9) = 0x02;

    //8 bits address
    for(i=0; i<=7; i++){
        addr_temp = (address&(0x80>>i));
        if(i<5)
            addr_temp = addr_temp>>(5-i);
        else if(i>5)
            addr_temp = addr_temp<<(i-5);
        *((volatile U8 *)lcr_ba+9) = (0x02|addr_temp);
        *((volatile U8 *)lcr_ba+9) = (0x03|addr_temp);
        *((volatile U8 *)lcr_ba+9) = (0x02|addr_temp);
    }

    *((volatile U8 *)lcr_ba+9) = 0x22;
    *((volatile U8 *)lcr_ba+9) = 0x33;

    //read 16 bis data
    for(i=0; i<=15; i++){
        *((volatile U8 *)lcr_ba+9) = 0x22;
        *((volatile U8 *)lcr_ba+9) = 0x33;
        usleep(1000);//Sleep(1);

        if(i>=12){
            *data = (*data)|
                (((*((volatile U16 *)((volatile U8 *)lcr_ba+9)))&0x08)>>(i-12));
        }
        else{
            *data = (*data)|
                (((*((volatile U16 *)((volatile U8 *)lcr_ba+9)))&0x08)<<(12-i));
        }
    }

    *((volatile U8 *)lcr_ba+9) = 0x22;
    *((volatile U8 *)lcr_ba+9) = 0x00;

    return;
}


void W_922X_TrimDAC_WR(
    U32 LcrAddr,
    U8  channel,
    U8  func,
    U8  data
    )
{
    U32 TDAC_Value;

    TDAC_Value = *((volatile U32 *)(LcrAddr+0x4dc));
    TDAC_Value &= 0x8000ffff;
    TDAC_Value |= (data<<20); //Set TrimDAC_Data
    switch(func){
        case 0:
            TDAC_Value |= 0x10000000; //Set TrimDAC_CS
            TDAC_Value |= (((channel*2+1)&0xf)<<16); //Set TrimDAC_Addr
            break;
        case 1:
            TDAC_Value |= 0x10000000; //Set TrimDAC_CS
            TDAC_Value |= (((channel*2+2)&0xf)<<16); //Set TrimDAC_Addr
            break;
        case 2:
            TDAC_Value |= 0x10000000; //Set TrimDAC_CS
            TDAC_Value |= (((channel+5)&0xf)<<16); //Set TrimDAC_Addr
            break;
    }
    *((volatile U32 *)(LcrAddr+0x4dc)) = TDAC_Value;
}


void W_922X_VREDP_WR(
    U32 LcrAddr,
    U8  dp_data
    )
{
    U32 TDAC_Value;

    TDAC_Value = *((volatile U32 *)(LcrAddr+0x4dc));
    TDAC_Value &= 0xffffff00;
    TDAC_Value |= 0x100; //set DP_CS
    TDAC_Value |= dp_data; //set DP_Data

    *((volatile U32 *)(LcrAddr+0x4dc)) = TDAC_Value;
}


void W_922X_Load_CAL_Constant(
    U16 wCardNumber,
    U16 bank
    )
{
    U8 bank_start = 0;
    U16 val;
    U32 *p;
    U32 LcrAddr;
    DAS_IOT_CALCONST_922X *calPtr;

    switch(bank){
        case EEPROM_DEFAULT_BANK:
            bank_start = 0;
            break;
        case EEPROM_USER_BANK1:
            bank_start = 64;
            break;
    }

    calPtr = (DAS_IOT_CALCONST_922X *)CurrentCard.pCAL_Const;
    LcrAddr = (U32)(CurrentCard_AIFunc.pView[1]);
    p = (U32 *)(CurrentCard_AIFunc.pView[1]);

    /*DA 5V*/
    W_922X_Read_EEPROM(p, (U8)(bank_start+0), &val);
    calPtr->CALConst_DA_5V = val;
    if(CurrentCard.wCardType!=PCI_922A){
        /*DA CH0 Linear*/
        W_922X_Read_EEPROM(p, (U8)(bank_start+1), &val);
        calPtr->CALConst_DA_Linear[0] = val;
        /*DA CH0 Offset*/
        W_922X_Read_EEPROM(p, (U8)(bank_start+2), &val);
        calPtr->CALConst_DA_Offset[0] = val;
        /*DA CH0 Gain*/
        W_922X_Read_EEPROM(p, (U8)(bank_start+3), &val);
        calPtr->CALConst_DA_Gain[0] = val;
        /*DA CH1 Linear*/
        W_922X_Read_EEPROM(p, (U8)(bank_start+4), &val);
        calPtr->CALConst_DA_Linear[1] = val;
        /*DA CH1 Offset*/
        W_922X_Read_EEPROM(p, (U8)(bank_start+5), &val);
        calPtr->CALConst_DA_Offset[1] = val;
        /*DA CH1 Gain*/
        W_922X_Read_EEPROM(p, (U8)(bank_start+6), &val);
        calPtr->CALConst_DA_Gain[1] = val;
    }
    /*AI Offset*/
    W_922X_Read_EEPROM(p, (U8)(bank_start+7), &val);
    calPtr->CALConst_AD_Offset[0] = val;
    W_922X_Read_EEPROM(p, (U8)(bank_start+8), &val);
    calPtr->CALConst_AD_Offset[1] = val;
    if(CurrentCard.wCardType!=PCI_922A){
        W_922X_Read_EEPROM(p, (U8)(bank_start+9), &val);
        calPtr->CALConst_AD_Offset[2] = val;
        W_922X_Read_EEPROM(p, (U8)(bank_start+10), &val);
        calPtr->CALConst_AD_Offset[3] = val;
        W_922X_Read_EEPROM(p, (U8)(bank_start+11), &val);
        calPtr->CALConst_AD_Offset[4] = val;
        W_922X_Read_EEPROM(p, (U8)(bank_start+12), &val);
        calPtr->CALConst_AD_Offset[5] = val;
        W_922X_Read_EEPROM(p, (U8)(bank_start+13), &val);
        calPtr->CALConst_AD_Offset[6] = val;
        W_922X_Read_EEPROM(p, (U8)(bank_start+14), &val);
        calPtr->CALConst_AD_Offset[7] = val;
    }
    /*AI Gain*/
    W_922X_Read_EEPROM(p, (U8)(bank_start+31), &val);
    calPtr->CALConst_AD_Gain[0] = val;
    W_922X_Read_EEPROM(p, (U8)(bank_start+32), &val);
    calPtr->CALConst_AD_Gain[1] = val;
    if(CurrentCard.wCardType!=PCI_922A){
        W_922X_Read_EEPROM(p, (U8)(bank_start+33), &val);
        calPtr->CALConst_AD_Gain[2] = val;
        W_922X_Read_EEPROM(p, (U8)(bank_start+34), &val);
        calPtr->CALConst_AD_Gain[3] = val;
        W_922X_Read_EEPROM(p, (U8)(bank_start+35), &val);
        calPtr->CALConst_AD_Gain[4] = val;
        W_922X_Read_EEPROM(p, (U8)(bank_start+36), &val);
        calPtr->CALConst_AD_Gain[5] = val;
        W_922X_Read_EEPROM(p, (U8)(bank_start+37), &val);
        calPtr->CALConst_AD_Gain[6] = val;
        W_922X_Read_EEPROM(p, (U8)(bank_start+38), &val);
        calPtr->CALConst_AD_Gain[7] = val;
    }
/*
{
    int i;
    printf("[DLL] CALConst_DA_5V: %#x\n", calPtr->CALConst_DA_5V);
    printf("[DLL] CALConst_DA_Linear[0]: %#x\n", calPtr->CALConst_DA_Linear[0]);
    printf("[DLL] CALConst_DA_Offset[0]: %#x\n", calPtr->CALConst_DA_Offset[0]);
    printf("[DLL] CALConst_DA_Gain[0]: %#x\n", calPtr->CALConst_DA_Gain[0]);
    printf("[DLL] CALConst_DA_Linear[1]: %#x\n", calPtr->CALConst_DA_Linear[1]);
    printf("[DLL] CALConst_DA_Offset[1]: %#x\n", calPtr->CALConst_DA_Offset[1]);
    printf("[DLL] CALConst_DA_Gain[1]: %#x\n", calPtr->CALConst_DA_Gain[1]);

    for(i=0; i<8; i++){
        printf("[DLL] CALConst_AD_Offset[%d]: %#x\n", i, calPtr->CALConst_AD_Offset[i]);
        printf("[DLL] CALConst_AD_Gain[%d]: %#x\n", i, calPtr->CALConst_AD_Gain[i]);
    }
}
*/
    /*Write To AI Offset Register*/
    *((volatile U32 *)(LcrAddr+0x588)) = \
        calPtr->CALConst_AD_Offset[0]|(calPtr->CALConst_AD_Offset[1]<<16);
    if(CurrentCard.wCardType!=PCI_922A){
        *((volatile U32 *)(LcrAddr+0x58C)) = \
            calPtr->CALConst_AD_Offset[2]|(calPtr->CALConst_AD_Offset[3]<<16);
        *((volatile U32 *)(LcrAddr+0x590)) = \
            calPtr->CALConst_AD_Offset[4]|(calPtr->CALConst_AD_Offset[5]<<16);
        *((volatile U32 *)(LcrAddr+0x594)) = \
            calPtr->CALConst_AD_Offset[6]|(calPtr->CALConst_AD_Offset[7]<<16);
    }
    /*Write To AI Gain Register*/
    *((volatile U32 *)(LcrAddr+0x598)) = \
        calPtr->CALConst_AD_Gain[0]|(calPtr->CALConst_AD_Gain[1]<<16);
    if(CurrentCard.wCardType!=PCI_922A){
        *((volatile U32 *)(LcrAddr+0x59C)) = \
            calPtr->CALConst_AD_Gain[2]|(calPtr->CALConst_AD_Gain[3]<<16);
        *((volatile U32 *)(LcrAddr+0x5A0)) = \
            calPtr->CALConst_AD_Gain[4]|(calPtr->CALConst_AD_Gain[5]<<16);
        *((volatile U32 *)(LcrAddr+0x5A4)) = \
            calPtr->CALConst_AD_Gain[6]|(calPtr->CALConst_AD_Gain[7]<<16);
    }
    /*AO TrimDAC*/
    W_922X_VREDP_WR(LcrAddr, (U8)(calPtr->CALConst_DA_5V));
    if(CurrentCard.wCardType!=PCI_922A){
        W_922X_TrimDAC_WR(LcrAddr, 0, 0, (U8)(calPtr->CALConst_DA_Linear[0]));
        W_922X_TrimDAC_WR(LcrAddr, 0, 1, (U8)(calPtr->CALConst_DA_Offset[0]));
        W_922X_TrimDAC_WR(LcrAddr, 0, 2, (U8)(calPtr->CALConst_DA_Gain[0]));
        W_922X_TrimDAC_WR(LcrAddr, 1, 0, (U8)(calPtr->CALConst_DA_Linear[1]));
        W_922X_TrimDAC_WR(LcrAddr, 1, 1, (U8)(calPtr->CALConst_DA_Offset[1]));
        W_922X_TrimDAC_WR(LcrAddr, 1, 2, (U8)(calPtr->CALConst_DA_Gain[1]));
    }
}


I16 PCI_Load_CAL_Data(
    U16 wCardNumber,
    U16 bank
    )
{
    I16 err;

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    switch(CurrentCard.wCardType){
        case PCI_9221:
            if((bank!=EEPROM_DEFAULT_BANK)&&(bank!=EEPROM_USER_BANK1))
                return ErrorUndefinedParameter;
            W_9221_Load_CAL_Constant(wCardNumber, bank);
            break;
        case PCI_9222:
        case PCI_9223:
        case PCI_922A:
            if((bank!=EEPROM_DEFAULT_BANK)&&(bank!=EEPROM_USER_BANK1))
                return ErrorUndefinedParameter;
            W_922X_Load_CAL_Constant(wCardNumber, bank);
            break;
        case PCI_9524:
            if(bank>3) //Bank 0 ~ Bank 3
                return ErrorUndefinedParameter;
            if((err = W_9524_Load_CAL_Constant(wCardNumber, bank))!=NoError)
                return err;
            break;
        default:
            return ErrorFuncNotSupport;
    }

    return NoError;
}


void W_9221_Write_EEPROM(
    U32 *lcr_ba,
    U8  address,
    U16 data
    )
{
    int i;
    U8 addr_temp;
    U16 data_temp;

    //chip select enable
    *((volatile U8 *)lcr_ba+3) = 0x00;
    *((volatile U8 *)lcr_ba+3) = 0x02;
    *((volatile U8 *)lcr_ba+3) = 0x03;
    *((volatile U8 *)lcr_ba+3) = 0x02;

    //start bit
    *((volatile U8 *)lcr_ba+3) = 0x06;
    *((volatile U8 *)lcr_ba+3) = 0x07;
    *((volatile U8 *)lcr_ba+3) = 0x06;

    //write enable command 4'b0011
    *((volatile U8 *)lcr_ba+3) = 0x02;
    *((volatile U8 *)lcr_ba+3) = 0x03;
    *((volatile U8 *)lcr_ba+3) = 0x02;
    *((volatile U8 *)lcr_ba+3) = 0x03;
    *((volatile U8 *)lcr_ba+3) = 0x02;
    *((volatile U8 *)lcr_ba+3) = 0x06;
    *((volatile U8 *)lcr_ba+3) = 0x07;
    *((volatile U8 *)lcr_ba+3) = 0x06;
    *((volatile U8 *)lcr_ba+3) = 0x07;
    *((volatile U8 *)lcr_ba+3) = 0x06;

    //don't care 7 bits
    *((volatile U8 *)lcr_ba+3) = 0x07;
    *((volatile U8 *)lcr_ba+3) = 0x06;
    *((volatile U8 *)lcr_ba+3) = 0x07;
    *((volatile U8 *)lcr_ba+3) = 0x06;
    *((volatile U8 *)lcr_ba+3) = 0x07;
    *((volatile U8 *)lcr_ba+3) = 0x06;
    *((volatile U8 *)lcr_ba+3) = 0x07;
    *((volatile U8 *)lcr_ba+3) = 0x06;
    *((volatile U8 *)lcr_ba+3) = 0x07;
    *((volatile U8 *)lcr_ba+3) = 0x06;
    *((volatile U8 *)lcr_ba+3) = 0x07;
    *((volatile U8 *)lcr_ba+3) = 0x06;
    *((volatile U8 *)lcr_ba+3) = 0x07;
    *((volatile U8 *)lcr_ba+3) = 0x06;

    //chip select disable
    *((volatile U8 *)lcr_ba+3) = 0x00;

    /****** write data ******/
    //chip select enable
    *((volatile U8 *)lcr_ba+3) = 0x00;
    *((volatile U8 *)lcr_ba+3) = 0x02;
    *((volatile U8 *)lcr_ba+3) = 0x03;
    *((volatile U8 *)lcr_ba+3) = 0x02;

    //start bit
    *((volatile U8 *)lcr_ba+3) = 0x06;
    *((volatile U8 *)lcr_ba+3) = 0x07;
    *((volatile U8 *)lcr_ba+3) = 0x06;

    //read command 2'b01
    *((volatile U8 *)lcr_ba+3) = 0x02;
    *((volatile U8 *)lcr_ba+3) = 0x03;
    *((volatile U8 *)lcr_ba+3) = 0x02;
    *((volatile U8 *)lcr_ba+3) = 0x06;
    *((volatile U8 *)lcr_ba+3) = 0x07;
    *((volatile U8 *)lcr_ba+3) = 0x06;


    //8 bits address
    for(i=0; i<=7; i++){
        addr_temp = (address&(0x80>>i));
        if(i<5)
            addr_temp = addr_temp>>(5-i);
        else if(i>5)
            addr_temp = addr_temp<<(i-5);
        *((volatile U8 *)lcr_ba+3) = (0x02|addr_temp);
        *((volatile U8 *)lcr_ba+3) = (0x03|addr_temp);
        *((volatile U8 *)lcr_ba+3) = (0x02|addr_temp);
    }

    //16 bits data
    for(i=0; i<=15; i++){
        data_temp = (data&(0x8000>>i));
        if(i<5)
            data_temp = data_temp>>(5-i);
        else if(i>5)
            data_temp = data_temp<<(i-5);
        data_temp = data_temp>>8;
        *((volatile U8 *)lcr_ba+3) = (0x02|data_temp);
        *((volatile U8 *)lcr_ba+3) = (0x03|data_temp);
        *((volatile U8 *)lcr_ba+3) = (0x02|data_temp);
    }

    //chip select disable
    *((volatile U8 *)lcr_ba+3) = 0x00;

    //chip select enable and check Din
    *((volatile U8 *)lcr_ba+3) = 0x22;
    *((volatile U8 *)lcr_ba+3) = 0x23;
    while(((*((volatile U32 *)lcr_ba))&0x08000000)==0){
        *((volatile U8 *)lcr_ba+3) = 0x22;
        *((volatile U8 *)lcr_ba+3) = 0x23;
    }

    //chip select disable
    *((volatile U8 *)lcr_ba+3) = 0x20;

    return;
}


void W_9221_Calibration_Date_Update(
    I16 wCardNumber,
    U16 bank
    )
{
    U32 *p;
    DAS_IOT_CALCONST_9221 *calPtr;

    //01/11/08
    //calPtr = CurrentCard.CAL_Const;
    calPtr = (DAS_IOT_CALCONST_9221 *)CurrentCard.pCAL_Const;
    //p = (U32 *)(CurrentCard.CAL_Const->IOP_mem);
    p = (U32 *)(CurrentCard_AIFunc.pView[1]);

    /*01~32: EEPROM_DEFAULT; 33~64: EEPROM_USER*/
    W_9221_Write_EEPROM(p, 33, calPtr->constant_offset_AD[0]);
    W_9221_Write_EEPROM(p, 34, calPtr->constant_offset_AD[1]);
    W_9221_Write_EEPROM(p, 35, calPtr->constant_offset_AD[2]);
    W_9221_Write_EEPROM(p, 36, calPtr->constant_offset_AD[3]);
    W_9221_Write_EEPROM(p, 37, calPtr->constant_gain_AD[0]);
    W_9221_Write_EEPROM(p, 38, calPtr->constant_gain_AD[1]);
    W_9221_Write_EEPROM(p, 39, calPtr->constant_gain_AD[2]);
    W_9221_Write_EEPROM(p, 40, calPtr->constant_gain_AD[3]);

    W_9221_Write_EEPROM(p, 41, (calPtr->constant_DA0[0])&0x1ff);
    W_9221_Write_EEPROM(p, 42, ((calPtr->constant_DA0[0])>>8)&0x1fe);

    W_9221_Write_EEPROM(p, 43, (calPtr->constant_DA0[1])&0x1ff);
    W_9221_Write_EEPROM(p, 44, ((calPtr->constant_DA0[1])>>8)&0x1fe);

    W_9221_Write_EEPROM(p, 45, (calPtr->constant_DA0[2])&0x1ff);
    W_9221_Write_EEPROM(p, 46, ((calPtr->constant_DA0[2])>>8)&0x1fe);

    W_9221_Write_EEPROM(p, 47, (calPtr->constant_DA1[0])&0x1ff);
    W_9221_Write_EEPROM(p, 48, ((calPtr->constant_DA1[0])>>8)&0x1fe);

    W_9221_Write_EEPROM(p, 49, (calPtr->constant_DA1[1])&0x1ff);
    W_9221_Write_EEPROM(p, 50, ((calPtr->constant_DA1[1])>>8)&0x1fe);

    W_9221_Write_EEPROM(p, 51, (calPtr->constant_DA1[2])&0x1ff);
    W_9221_Write_EEPROM(p, 52, ((calPtr->constant_DA1[2])>>8)&0x1fe);

    return;
}


I16 W_9524_Calibration_Date_Update(
    U16 wCardNumber,
    U16 bank
    )
{
    DAS_IOT_CTR iotCtr;

    memset(&iotCtr, '\0', sizeof(DAS_IOT_CTR));

    iotCtr.wMode = 1;
    iotCtr.wState = bank;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_CAL_ACT, &(iotCtr))
        !=DAS_ERR_NO)
    {
        return iotCtr.wErrorCode;
    }

    return NoError;
}


void W_922X_Write_EEPROM(
    U32 *lcr_ba,
    U8  address,
    U16 data
    )
{
    int i;
    U8 addr_temp;
    U16 data_temp;

    //chip select enable
    *((volatile U8 *)lcr_ba+9) = 0x00;
    *((volatile U8 *)lcr_ba+9) = 0x02;
    *((volatile U8 *)lcr_ba+9) = 0x03;
    *((volatile U8 *)lcr_ba+9) = 0x02;

    //start bit
    *((volatile U8 *)lcr_ba+9) = 0x06;
    *((volatile U8 *)lcr_ba+9) = 0x07;
    *((volatile U8 *)lcr_ba+9) = 0x06;

    //write enable command 4'b0011
    *((volatile U8 *)lcr_ba+9) = 0x02;
    *((volatile U8 *)lcr_ba+9) = 0x03;
    *((volatile U8 *)lcr_ba+9) = 0x02;
    *((volatile U8 *)lcr_ba+9) = 0x03;
    *((volatile U8 *)lcr_ba+9) = 0x02;
    *((volatile U8 *)lcr_ba+9) = 0x06;
    *((volatile U8 *)lcr_ba+9) = 0x07;
    *((volatile U8 *)lcr_ba+9) = 0x06;
    *((volatile U8 *)lcr_ba+9) = 0x07;
    *((volatile U8 *)lcr_ba+9) = 0x06;

    //don't care 7 bits
    *((volatile U8 *)lcr_ba+9) = 0x07;
    *((volatile U8 *)lcr_ba+9) = 0x06;
    *((volatile U8 *)lcr_ba+9) = 0x07;
    *((volatile U8 *)lcr_ba+9) = 0x06;
    *((volatile U8 *)lcr_ba+9) = 0x07;
    *((volatile U8 *)lcr_ba+9) = 0x06;
    *((volatile U8 *)lcr_ba+9) = 0x07;
    *((volatile U8 *)lcr_ba+9) = 0x06;
    *((volatile U8 *)lcr_ba+9) = 0x07;
    *((volatile U8 *)lcr_ba+9) = 0x06;
    *((volatile U8 *)lcr_ba+9) = 0x07;
    *((volatile U8 *)lcr_ba+9) = 0x06;
    *((volatile U8 *)lcr_ba+9) = 0x07;
    *((volatile U8 *)lcr_ba+9) = 0x06;

    //chip select disable
    *((volatile U8 *)lcr_ba+9) = 0x00;

    /****** write data ******/
    //chip select enable
    *((volatile U8 *)lcr_ba+9) = 0x00;
    *((volatile U8 *)lcr_ba+9) = 0x02;
    *((volatile U8 *)lcr_ba+9) = 0x03;
    *((volatile U8 *)lcr_ba+9) = 0x02;

    //start bit
    *((volatile U8 *)lcr_ba+9) = 0x06;
    *((volatile U8 *)lcr_ba+9) = 0x07;
    *((volatile U8 *)lcr_ba+9) = 0x06;

    //read command 2'b01
    *((volatile U8 *)lcr_ba+9) = 0x02;
    *((volatile U8 *)lcr_ba+9) = 0x03;
    *((volatile U8 *)lcr_ba+9) = 0x02;
    *((volatile U8 *)lcr_ba+9) = 0x06;
    *((volatile U8 *)lcr_ba+9) = 0x07;
    *((volatile U8 *)lcr_ba+9) = 0x06;


    //8 bits address
    for(i=0; i<=7; i++){
        addr_temp = (address&(0x80>>i));
        if(i<5)
            addr_temp = addr_temp>>(5-i);
        else if(i>5)
            addr_temp = addr_temp<<(i-5);
        *((volatile U8 *)lcr_ba+9) = (0x02|addr_temp);
        *((volatile U8 *)lcr_ba+9) = (0x03|addr_temp);
        *((volatile U8 *)lcr_ba+9) = (0x02|addr_temp);
    }

    //16 bits data
    for(i=0; i<=15; i++){
        data_temp = (data&(0x8000>>i));
        if(i<5)
            data_temp = data_temp>>(5-i);
        else if(i>5)
            data_temp = data_temp<<(i-5);
        data_temp = data_temp>>8;
        *((volatile U8 *)lcr_ba+9) = (0x02|data_temp);
        *((volatile U8 *)lcr_ba+9) = (0x03|data_temp);
        *((volatile U8 *)lcr_ba+9) = (0x02|data_temp);
    }

    //chip select disable
    *((volatile U8 *)lcr_ba+9) = 0x00;

    //chip select enable and check Din
    *((volatile U8 *)lcr_ba+9) = 0x22;
    *((volatile U8 *)lcr_ba+9) = 0x23;
    while(((*((volatile U8 *)lcr_ba+9))&0x08)==0){
        *((volatile U8 *)lcr_ba+9) = 0x22;
        *((volatile U8 *)lcr_ba+9) = 0x23;
    }

    //chip select disable
    *((volatile U8 *)lcr_ba+9) = 0x20;
}


void W_922X_Calibration_Date_Update(
    U16 wCardNumber,
    U16 bank
    )
{
    U32 *p;
    DAS_IOT_CALCONST_922X *calPtr;

    calPtr = (DAS_IOT_CALCONST_922X *)CurrentCard.pCAL_Const;
    p = (U32 *)(CurrentCard_AIFunc.pView[1]);

    /*AO Const*/
    W_922X_Write_EEPROM(p, 64, calPtr->CALConst_DA_5V);
    if(CurrentCard.wCardType!=PCI_922A){
        W_922X_Write_EEPROM(p, 65, calPtr->CALConst_DA_Linear[0]);
        W_922X_Write_EEPROM(p, 66, calPtr->CALConst_DA_Offset[0]);
        W_922X_Write_EEPROM(p, 67, calPtr->CALConst_DA_Gain[0]);
        W_922X_Write_EEPROM(p, 68, calPtr->CALConst_DA_Linear[1]);
        W_922X_Write_EEPROM(p, 69, calPtr->CALConst_DA_Offset[1]);
        W_922X_Write_EEPROM(p, 70, calPtr->CALConst_DA_Gain[1]);
    }
    /*AI Const*/
    W_922X_Write_EEPROM(p, 71, calPtr->CALConst_AD_Offset[0]);
    W_922X_Write_EEPROM(p, 72, calPtr->CALConst_AD_Offset[1]);
    if(CurrentCard.wCardType!=PCI_922A){
        W_922X_Write_EEPROM(p, 73, calPtr->CALConst_AD_Offset[2]);
        W_922X_Write_EEPROM(p, 74, calPtr->CALConst_AD_Offset[3]);
        W_922X_Write_EEPROM(p, 75, calPtr->CALConst_AD_Offset[4]);
        W_922X_Write_EEPROM(p, 76, calPtr->CALConst_AD_Offset[5]);
        W_922X_Write_EEPROM(p, 77, calPtr->CALConst_AD_Offset[6]);
        W_922X_Write_EEPROM(p, 78, calPtr->CALConst_AD_Offset[7]);
    }
    W_922X_Write_EEPROM(p, 95, calPtr->CALConst_AD_Gain[0]);
    W_922X_Write_EEPROM(p, 96, calPtr->CALConst_AD_Gain[1]);
    if(CurrentCard.wCardType!=PCI_922A){
        W_922X_Write_EEPROM(p, 97, calPtr->CALConst_AD_Gain[2]);
        W_922X_Write_EEPROM(p, 98, calPtr->CALConst_AD_Gain[3]);
        W_922X_Write_EEPROM(p, 99, calPtr->CALConst_AD_Gain[4]);
        W_922X_Write_EEPROM(p, 100, calPtr->CALConst_AD_Gain[5]);
        W_922X_Write_EEPROM(p, 101, calPtr->CALConst_AD_Gain[6]);
        W_922X_Write_EEPROM(p, 102, calPtr->CALConst_AD_Gain[7]);
    }

    return;
}


I16 PCI_EEPROM_CAL_Constant_Update(
    U16 wCardNumber,
    U16 bank
    )
{
    I16 err;

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    switch(CurrentCard.wCardType){
        case PCI_9221:
            /*only user defined eeprom can be written*/
            if(bank!=EEPROM_USER_BANK1)
                return ErrorUndefinedParameter;
            W_9221_Calibration_Date_Update(wCardNumber, bank);
            break;
        case PCI_9222:
        case PCI_9223:
        case PCI_922A:
            if(bank!=EEPROM_USER_BANK1)
                return ErrorUndefinedParameter;
            W_922X_Calibration_Date_Update(wCardNumber, bank);
            break;
        case PCI_9524:
            if((bank==0)||(bank>3)) //only write bank 1 ~ bank 3
                return ErrorUndefinedParameter;
            if((err = W_9524_Calibration_Date_Update(wCardNumber,bank))
                !=NoError)
            {
                return err;
            }
            break;
        default:
            return ErrorFuncNotSupport;
    }
    return NoError;
}


void get_local_time(
    U16 *YEAR,
    U16 *MONTH,
    U16 *DAY,
    U16 *WEEKDAY,
    U16 *HOUR,
    U16 *MINUTE,
    U16 *SECOND
    )
{
    time_t ltime;
    //char *strtime = (char *)malloc(25);

    time(&ltime);
    //strtime = ctime(&ltime);

    //www mnm dd hh:mm:ss yyyy\n
    if(YEAR){
        char year[5];
        //strncpy(year, strtime+20, 4);
        strncpy(year, ctime(&ltime)+20, 4);
        year[4] = '\0';
        *YEAR = atoi(year);
    }
    if(MONTH){
        char month[4];
        //strncpy(month, strtime+4, 3);
        strncpy(month, ctime(&ltime)+4, 3);
        month[3] = '\0';
        switch(month[0]){
            case 'J':
                if(month[1]=='a')
                    *MONTH = 1;
                else if(month[1]=='u'){
                    if(month[2]=='n')
                        *MONTH = 6;
                    else if(month[2]=='l')
                        *MONTH = 7;
                }
                break;
            case 'F':
                *MONTH = 2;
                break;
            case 'M':
                if(month[2]=='r')
                    *MONTH = 3;
                else if(month[2]=='y')
                    *MONTH = 5;
                break;
            case 'A':
                if(month[1]=='p')
                    *MONTH = 4;
                else if(month[1]=='u')
                    *MONTH = 8;
                break;
            case 'S':
                *MONTH = 9;
                break;
            case 'O':
                *MONTH = 10;
                break;
            case 'N':
                *MONTH = 11;
                break;
            case 'D':
                *MONTH = 12;
                break;
        }
    }
    if(DAY){
        char day[3];
        //strncpy(day, strtime+8, 2);
        strncpy(day, ctime(&ltime)+8, 2);
        day[2] = '\0';
        *DAY = atoi(day);
    }
    if(WEEKDAY){
        char wek[4];
        //strncpy(wek, strtime, 3);
        strncpy(wek, ctime(&ltime), 3);
        wek[3] = '\0';
        switch(wek[0]){
            case 'S':
                if(wek[1]=='u')
                    *WEEKDAY = 0;
                else if(wek[1]=='a')
                    *WEEKDAY = 6;
                break;
            case 'M':
                *WEEKDAY = 1;
                break;
            case 'T':
                if(wek[1]=='u')
                    *WEEKDAY = 2;
                else if(wek[1]=='h')
                    *WEEKDAY = 4;
                break;
            case 'W':
                *WEEKDAY = 3;
                break;
            case 'F':
                *WEEKDAY = 5;
                break;
        }
    }
    if(HOUR){
        char hour[3];
        //strncpy(hour, strtime+11, 2);
        strncpy(hour, ctime(&ltime)+11, 2);
        hour[2] = '\0';
        *HOUR = atoi(hour);
    }
    if(MINUTE){
        char minute[3];
        //strncpy(minute, strtime+14, 2);
        strncpy(minute, ctime(&ltime)+14, 2);
        minute[2] = '\0';
        *MINUTE = atoi(minute);
    }
    if(SECOND){
        char second[3];
        //strncpy(second, strtime+17, 2);
        strncpy(second, ctime(&ltime)+17, 2);
        second[2] = '\0';
        *SECOND = atoi(second);
    }

    //free(strtime);
}


I16 PCI9524_Acquire_AD_CalConst(
    U16 wCardNumber,
    U16 wGroup,
    U16 wADC_Range,
    U16 wADC_Speed,
    U32 *dwCalDate,
    F32 *fCalTemp,
    U32 *dwADC_offset,
    U32 *dwADC_gain,
    F64 *Residual_offset,
    F64 *Residual_scaling
    )
{
    DAS_IOT_CALCONST_9524 *calPtr;
    U32 Gain_index;
    U32 Speed_index;

    if(CurrentCard.wCardType!=PCI_9524)
        return ErrorFuncNotSupport;

    if(wGroup==0){
        if(wADC_Range)
             return ErrorUndefinedParameter;
    }
    else if(wGroup==1){
        if((wADC_Range==0)||(wADC_Range>4))
            return ErrorUndefinedParameter;
    }
    else{
        return ErrorUndefinedParameter;
    }

    if(wADC_Speed>15)
        return ErrorUndefinedParameter;


    calPtr = (DAS_IOT_CALCONST_9524 *)CurrentCard.pCAL_Const;
    Speed_index = 15-wADC_Speed;
    *dwCalDate = \
        calPtr->SysExtCalConst.EE_YEAR*10000+calPtr->SysExtCalConst.EE_MMDD;
    *fCalTemp = calPtr->SysExtCalConst.EE_CALTEMP;
    if(wGroup==0){
        Gain_index = 0;
        *dwADC_offset = \
            calPtr->SysAI0IntCalConst.EE_ADC_OFC[Gain_index][Speed_index];
        *dwADC_gain = \
            calPtr->SysAI0IntCalConst.EE_ADC_FSC[Gain_index][Speed_index];
        *Residual_offset = \
            calPtr->SysAI0IntCalConst.EE_RESIDUAL_OFFSET[Gain_index][Speed_index];
        *Residual_scaling = \
            calPtr->SysAI0IntCalConst.EE_SCALING[Gain_index][Speed_index];
    }
    else{
        Gain_index = wADC_Range-1;
        *dwADC_offset = \
            calPtr->SysAI1IntCalConst.EE_ADC_OFC[Gain_index][Speed_index];
        *dwADC_gain = \
            calPtr->SysAI1IntCalConst.EE_ADC_FSC[Gain_index][Speed_index];
        *Residual_offset = \
            calPtr->SysAI1IntCalConst.EE_RESIDUAL_OFFSET[Gain_index][Speed_index];
        *Residual_scaling = \
            calPtr->SysAI1IntCalConst.EE_SCALING[Gain_index][Speed_index];
    }

    return NoError;
}


I16 PCI9524_Acquire_DA_CalConst(
    U16 wCardNumber,
    U16 wChannel,
    U32 *dwCalDate,
    F32 *fCalTemp,
    U8  *bDAC_offset,
    U8  *bDAC_linearity,
    F32 *fGain_factor
    )
{
    DAS_IOT_CALCONST_9524 *calPtr;

    if(CurrentCard.wCardType!=PCI_9524)
        return ErrorFuncNotSupport;

    if(wChannel>1)
        return ErrorUndefinedParameter;

    calPtr = (DAS_IOT_CALCONST_9524 *)CurrentCard.pCAL_Const;

    *dwCalDate = \
        calPtr->SysExtCalConst.EE_YEAR*10000+calPtr->SysExtCalConst.EE_MMDD;
    *fCalTemp = calPtr->SysExtCalConst.EE_CALTEMP;
    *bDAC_offset = calPtr->SysAOIntCalConst.EE_DAC_OFFSET[wChannel];
    *bDAC_linearity = calPtr->SysAOIntCalConst.EE_DAC_LINEARITY[wChannel];
    *fGain_factor = calPtr->SysAOIntCalConst.EE_GAIN_FACTOR[wChannel];

    return NoError;
}


void W_6202_Read_EEPROM(
    U32 lcr_ba,
    U8  address,
    U16 *data
    )
{
    int i;
    volatile U8 *p_eeprom;
    U8 addr_temp;
    U16 data_temp;

    p_eeprom = (volatile U8 *)(lcr_ba+0x09);
//chip select enable
    *p_eeprom = 0x00;
    *p_eeprom = 0x02;
    *p_eeprom = 0x03;
    *p_eeprom = 0x02;
//start bit
    *p_eeprom = 0x06;
    *p_eeprom = 0x07;
    *p_eeprom = 0x06;
//read command 2'b10
    *p_eeprom = 0x07;
    *p_eeprom = 0x06;
    *p_eeprom = 0x02;
    *p_eeprom = 0x03;
    *p_eeprom = 0x02;
//8 bits address
    for(i=0; i<=7; i++){
        addr_temp = (address&(0x80>>i));
        if(i<5){
            addr_temp = addr_temp>>(5-i);
        }
        else if(i>5){
            addr_temp = addr_temp<<(i-5);
        }
        *p_eeprom = (0x02|addr_temp);
        *p_eeprom = (0x03|addr_temp);
        *p_eeprom = (0x02|addr_temp);
    }
    *p_eeprom = 0x22;  //delay one sck
    *p_eeprom = 0x23;
//read 16 bis data
    data_temp = 0;
    for(i=0; i<=15; i++){
        *p_eeprom = 0x22;
        *p_eeprom = 0x23;
        //WriteByte(Mem_Ptr+Gen_Csr+1,0x22);
        usleep(1000);//Sleep(1);
        if(i>=12)
            data_temp = data_temp|(((*p_eeprom)&0x08)>>(i-12));
        else
            data_temp = data_temp|(((*p_eeprom)&0x08)<<(12-i));
    }
    *p_eeprom = 0x22;
    *p_eeprom = 0x00;
    *data = data_temp;
}


void W_6202_Write_EEPROM(
    U32 lcr_ba,
    U8  address,
    U16 data
    )
{
    int i;
    volatile U8 *p_eeprom;
    U8 addr_temp;
    U16 data_temp;

    p_eeprom = (volatile U8 *)(lcr_ba+0x09);
//chip select enable
    *p_eeprom = 0x00;
    *p_eeprom = 0x02;
    *p_eeprom = 0x03;
    *p_eeprom = 0x02;
//start bit
    *p_eeprom = 0x06;
    *p_eeprom = 0x07;
    *p_eeprom = 0x06;
//write enable command 4'b0011
    *p_eeprom = 0x02;
    *p_eeprom = 0x03;
    *p_eeprom = 0x02;
    *p_eeprom = 0x03;
    *p_eeprom = 0x02;
    *p_eeprom = 0x06;
    *p_eeprom = 0x07;
    *p_eeprom = 0x06;
    *p_eeprom = 0x07;
    *p_eeprom = 0x06;
//don't care 7 bits
    *p_eeprom = 0x07;
    *p_eeprom = 0x06;
    *p_eeprom = 0x07;
    *p_eeprom = 0x06;
    *p_eeprom = 0x07;
    *p_eeprom = 0x06;
    *p_eeprom = 0x07;
    *p_eeprom = 0x06;
    *p_eeprom = 0x07;
    *p_eeprom = 0x06;
    *p_eeprom = 0x07;
    *p_eeprom = 0x06;
    *p_eeprom = 0x07;
    *p_eeprom = 0x06;
//chip select disable
    *p_eeprom = 0x00;
/****** write data ******/
//chip select enable
    *p_eeprom = 0x00;
    *p_eeprom = 0x02;
    *p_eeprom = 0x03;
    *p_eeprom = 0x02;
//start bit
    *p_eeprom = 0x06;
    *p_eeprom = 0x07;
    *p_eeprom = 0x06;
//write command 2'b10
    *p_eeprom = 0x02;
    *p_eeprom = 0x03;
    *p_eeprom = 0x02;
    *p_eeprom = 0x06;
    *p_eeprom = 0x07;
    *p_eeprom = 0x06;

    for(i=0; i<=7; i++){
        addr_temp = (address&(0x80>>i));
        if(i<5){
            addr_temp = addr_temp>>(5-i);
        }
        else if(i>5){
            addr_temp = addr_temp<<(i-5);
        }
        *p_eeprom = (0x02|addr_temp);
        *p_eeprom = (0x03|addr_temp);
        *p_eeprom = (0x02|addr_temp);
    }
//16 bis data
    for(i=0; i<=15; i++){
        data_temp = (data&(0x8000>>i));
        if(i<5)
            data_temp = data_temp>>(5-i);
        else if(i>5)
            data_temp = data_temp<<(i-5);
        data_temp = data_temp>>8;
        *p_eeprom = (0x02|data_temp);
        *p_eeprom = (0x03|data_temp);
        *p_eeprom = (0x02|data_temp);
    }
//chip select disable
    *p_eeprom = 0x00;
//chip select enable and check Din
    *p_eeprom = 0x22;
    *p_eeprom = 0x23;
    while(((*p_eeprom)&0x08)==0){
        *p_eeprom = 0x22;
        *p_eeprom = 0x23;
    }
//chip select disable
    *p_eeprom = 0x20;
}


/*
 * Add for customer to read/write remaining EEPROM space
 */
I16 PCI9524_Write_EEProm(
    U16 wCardNumber,
    U16 wWriteAddr,
    U8  *pbWriteData
    )
{
    DAS_IOT_CTR iotCtr;

    memset(&iotCtr, '\0', sizeof(DAS_IOT_CTR));

    /*Error Check*/
    if(CurrentCard.wCardType!=PCI_9524)
        return ErrorFuncNotSupport;
    if(wWriteAddr>999)
        return ErrorUndefinedParameter;
    if(pbWriteData==NULL)
        return ErrorUndefinedParameter;

    iotCtr.wMode = 4; //Let Kernel Know PCI9524_Write_EEProm will be performed
    iotCtr.wCounterIndex = wWriteAddr;
    iotCtr.wState = ((*(pbWriteData+1))<<8)|(*pbWriteData);
    iotCtr.wBinBcd = ((*(pbWriteData+3))<<8)|(*(pbWriteData+2));
    iotCtr.dwCounterValue = ((*(pbWriteData+7))<<24)|((*(pbWriteData+6))<<16)|
        ((*(pbWriteData+5))<<8)|(*(pbWriteData+4));
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_CAL_ACT, &(iotCtr))
        !=DAS_ERR_NO)
    {
        return iotCtr.wErrorCode;
    }

    return NoError;
}


I16 PCI9524_Read_EEProm(
    U16 wCardNumber,
    U16 wReadAddr,
    U8  *pbReadData
    )
{
    DAS_IOT_CTR iotCtr;

    memset(&iotCtr, '\0', sizeof(DAS_IOT_CTR));

    /*Error Check*/
    if(CurrentCard.wCardType!=PCI_9524)
        return ErrorFuncNotSupport;
    if(wReadAddr>999)
        return ErrorUndefinedParameter;
    if(pbReadData==NULL)
        return ErrorUndefinedParameter;

    iotCtr.wMode = 5; //Let Kernel Know PCI9524_Read_EEProm will be performed
    iotCtr.wCounterIndex = wReadAddr;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_CAL_ACT, &(iotCtr))
        !=DAS_ERR_NO)
    {
        return iotCtr.wErrorCode;
    }

    *pbReadData = (U8)(iotCtr.wState&0xff);
    *(pbReadData+1) = (U8)((iotCtr.wState>>8)&0xff);
    *(pbReadData+2) = (U8)(iotCtr.wBinBcd&0xff);
    *(pbReadData+3) = (U8)((iotCtr.wBinBcd>>8)&0xff);
    *(pbReadData+4) = (U8)(iotCtr.dwCounterValue&0xff);
    *(pbReadData+5) = (U8)((iotCtr.dwCounterValue>>8)&0xff);
    *(pbReadData+6) = (U8)((iotCtr.dwCounterValue>>16)&0xff);
    *(pbReadData+7) = (U8)((iotCtr.dwCounterValue>>24)&0xff);

    return NoError;
}
