#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <errno.h>
#include "dask.h"
#include "Adlsys.h"
#include "dask_p.h"


extern REGISTER_CARD_TYPE RegisterCards[MAX_CARD];


I16 DO_7200_Config(
    U16 wCardNumber,
    U16 wTrigSource,
    U16 wOutReqEn,
    U16 wOutTrigSig
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    iotConf.mData.m7200DIO.wTrigSource = wTrigSource;
    //iotConf.mData.m7200DIO.fOutRequestEnable=wOutReqEn;
    //iotConf.mData.m7200DIO.fOutTriggerHigh=wOutTrigSig;
    iotConf.mData.wValue[1] = wOutReqEn|wOutTrigSig;
    iotConf.mData.m7200DIO.fTimerCascade = 0;
    CurrentCard.DO_Status.CurrentConfig = iotConf;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }

    return NoError;
}


I16 DO_7300A_Config(
    U16 wCardNumber,
    U16 wPortWidth,
    U16 wTrigSource,
    U16 wWaitStatus,
    U16 wTerminator,
    U16 wO_REQ_Pol
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    switch(wPortWidth){
        case 0:
            wPortWidth = DAS_PORT_WIDTH_0;
            break;
        case 8:
            wPortWidth = DAS_PORT_WIDTH_8;
            break;
        case 16:
            wPortWidth = DAS_PORT_WIDTH_16;
            break;
        case 32:
            wPortWidth = DAS_PORT_WIDTH_32;
            break;
        default:
            return ErrorInvalidPortWidth;
    }

    iotConf.mData.m7300DIO.fPortWidth = wPortWidth;
    iotConf.mData.m7300DIO.wTrigSource = wTrigSource;
    iotConf.mData.m7300DIO.fStartWait = wWaitStatus;
    iotConf.mData.m7300DIO.fTerminator = wTerminator;
    iotConf.mData.m7300DIO.fRequestRising = wO_REQ_Pol;
    //yuan add 11/30/99
    iotConf.mData.m7300DIO.fConfig = 1;
    CurrentCard.DO_Status.CurrentConfig = iotConf;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }

    return NoError;
}


I16 DO_7300B_Config(
    U16 wCardNumber,
    U16 wPortWidth,
    U16 wTrigSource,
    U16 wWaitStatus,
    U16 wTerminator,
    U16 wO_Cntrl_Pol,
    U32 dwFifoThreshold
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    switch(wPortWidth){
        case 0:
            wPortWidth = DAS_PORT_WIDTH_0;
            break;
        case 8:
            wPortWidth = DAS_PORT_WIDTH_8;
            break;
        case 16:
            wPortWidth = DAS_PORT_WIDTH_16;
            break;
        case 32:
            wPortWidth = DAS_PORT_WIDTH_32;
            break;
        default:
            return ErrorInvalidPortWidth;
    }

    iotConf.mData.m7300bDIO.fPortWidth = wPortWidth;
    iotConf.mData.m7300bDIO.wTrigSource = wTrigSource;
    iotConf.mData.m7300bDIO.fStartWait = wWaitStatus;
    iotConf.mData.m7300bDIO.fTerminator = wTerminator;
    iotConf.mData.m7300bDIO.fRequestRising = wO_Cntrl_Pol;
    iotConf.mData.m7300bDIO.fifo_threshold = dwFifoThreshold;
    //yuan add 11/30/99
    iotConf.mData.m7300bDIO.fConfig = 1;
    CurrentCard.DO_Status.CurrentConfig = iotConf;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }

    return NoError;
}


I16 DO_AsyncCheck(
    U16     wCardNumber,
    BOOLEAN *bStopped,
    U32     *dwAccessCnt
    )
{
    I16 err;
    DAS_IOT_ASYNC iotAsync;

    memset(&iotAsync, '\0', sizeof(DAS_IOT_ASYNC));

    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_ASY_CHECK, &(iotAsync))
        !=DAS_ERR_NO)
    {
        return iotAsync.wErrorCode;
    }
    *bStopped = (BOOLEAN)iotAsync.wStopped;
    *dwAccessCnt = iotAsync.dwAccessCount;

    return NoError;
}


I16 DO_AsyncClear(
    U16 wCardNumber,
    U32 *dwAccessCnt
    )
{
    I16 err, i;
    U32 byte_width;
    I16 pwidth[6] = {0, 4, 8, 16, 24, 32};
    DAS_IOT_ASYNC iotAsync;
    DAS_IOT_CONFIG iotConf;
    U32 BufId;

    memset(&iotAsync, '\0', sizeof(DAS_IOT_ASYNC));
    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    iotConf = CurrentCard.DO_Status.CurrentConfig;

    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber, 0, NULL,
        0L, TRUE))!=NoError)
    {
        return err;
    }
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_ASY_CLEAR, &(iotAsync))
        !=DAS_ERR_NO)
    {
        return iotAsync.wErrorCode;
    }
    *dwAccessCnt = iotAsync.dwAccessCount;

    //08/04/08
    if((CurrentCard.wCardType==PCI_9222)||
        (CurrentCard.wCardType==PCI_9223))
    {
        //Reset Buffer
        if(iotConf.mData.m922xAO.fAutoResetBuf){
            if(!CurrentCard.DO_Status.bDblBufferMode){
                BufId = (U32)CurrentCard.DO_Status.pDAQbufHandle;
                CurrentCard.DO_Status.pDAQbuf[BufId] = NULL;
                CurrentCard.DO_Status.dwBufCount[BufId] = 0;
            }
            else{
                CurrentCard.DO_Status.pDAQbuf[0] = NULL;
                CurrentCard.DO_Status.pDAQbuf[1] = NULL;
                CurrentCard.DO_Status.dwBufCount[0] = 0;
                CurrentCard.DO_Status.dwBufCount[1] = 0;
            }
        }
    }
    else{
        if((CurrentCard.wCardType==PCI_7300A_RevA)||
            (CurrentCard.wCardType==PCI_7300A_RevB)||
            (CurrentCard.wCardType==PCI_7350))
        {
            if(CurrentCard.wCardType==PCI_7350)
                byte_width = iotConf.mData.m7350DIO.PortCount+1;
            else if(CurrentCard.wCardType==PCI_7300A_RevA)
                byte_width = pwidth[iotConf.mData.m7300DIO.fPortWidth]/8;
            else
                byte_width = pwidth[iotConf.mData.m7300bDIO.fPortWidth]/8;

            munlock(
                (const void *)CurrentCard.DO_Status.p7300buf,
                CurrentCard.DO_Status.dwTransferCount*byte_width
            );
            CurrentCard.DO_Status.p7300buf = NULL;

            //02/01/26 jeffrey added for MultiBuffer DO
            for(i=0; i<CurrentCard.DO_Status.p7300mbufCnt; i++){
                if(CurrentCard.wCardType==PCI_7350){
                    if(CurrentCard.DO_Status.p7350mbuf[i].dwBufferAddr){
                        munlock(
                            (const void *)
                            CurrentCard.DO_Status.p7350mbuf[i].dwBufferAddr,
                            CurrentCard.DO_Status.p7350mbuf[i].dwAccessCount*
                            byte_width
                        );
                    }
                    CurrentCard.DO_Status.p7350mbuf[i].dwBufferAddr = \
                        (U32)NULL;
                    CurrentCard.DO_Status.p7350mbuf[i].dwAccessCount = 0;
                    CurrentCard.DO_Status.p7350mbuf[i].wBufferId = -1;
                }
                else{
                    if(CurrentCard.DO_Status.p7300mbuf[i].dwBufferAddr){
                        munlock(
                            (const void *)
                            CurrentCard.DO_Status.p7300mbuf[i].dwBufferAddr,
                            CurrentCard.DO_Status.p7300mbuf[i].dwAccessCount*
                            byte_width
                        );
                    }
                    CurrentCard.DO_Status.p7300mbuf[i].dwBufferAddr = \
                        (U32)NULL;
                    CurrentCard.DO_Status.p7300mbuf[i].dwAccessCount = 0;
                    CurrentCard.DO_Status.p7300mbuf[i].wBufferId = -1;
                }
            }
            CurrentCard.DO_Status.p7300mbufCnt = 0;
        }
    }

    return NoError;
}


I16 DO_AsyncDblBufferHalfReady(
    U16     wCardNumber,
    BOOLEAN *bHalfReady
    )
{
    I16 err;
    DAS_IOT_DBF  iotDbf;

    memset(&iotDbf, '\0', sizeof(DAS_IOT_DBF));

    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_DBF_HALF_READY, &(iotDbf))
        !=DAS_ERR_NO)
    {
        return iotDbf.wErrorCode;
    }
    *bHalfReady = (BOOLEAN)iotDbf.wReady;
    CurrentCard.DO_Status.wDblBufferIndex = iotDbf.wViewIndex;

    return NoError;
}


I16 DO_AsyncDblBufferMode(
    U16     wCardNumber,
    BOOLEAN bEnable
    )
{
    I16 err;
    DAS_IOT_DBF  iotDbf;

    memset(&iotDbf, '\0', sizeof(DAS_IOT_DBF));

    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    iotDbf.wEnable = bEnable;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_DBF_MODE, &(iotDbf))
        !=DAS_ERR_NO)
    {
        return iotDbf.wErrorCode;
    }
    CurrentCard.DO_Status.bDblBufferMode = bEnable;

    return NoError;
}


I16 DO_AsyncDblBufferTransfer(
    U16  wCardNumber,
    void *pwBuffer
    )
{
    I16 err;

    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }
    if(!CurrentCard.DO_Status.bDblBufferMode)
        return ErrorNotDoubleBufferMode;

    memcpy(
        CurrentCard_DOFunc.pView[CurrentCard.DO_Status.wDblBufferIndex],
        pwBuffer,
        CurrentCard.DO_Status.dwTransferCount*
        (U16)((CurrentCard_DOFunc.wResoluation+7)/8)
    );

    return NoError;
}


I16 DO_ContWritePort(
    U16  wCardNumber,
    U16  wPort,
    void *pwBuffer,
    U32  dwWriteCount,
    U16  wIterations,
    F64  fSampleRate,
    U16  wSyncMode
    )
{
    I16 err;
    I16 pwidth[6] = {0, 4, 8, 16, 24, 32};
    DAS_IOT_CONFIG iotConf;
    DAS_IOT_CONT iotCont;
    DAS_IOT_SG_CONT iotSCont;
    //00/06/03 jeffrey added
    U32 byte_width = 0;
    U8 needlock = 0;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));
    memset(&iotCont, '\0', sizeof(DAS_IOT_CONT));

    iotConf = CurrentCard.DO_Status.CurrentConfig;

    if((CurrentCard.wCardType==PCI_9222)||(CurrentCard.wCardType==PCI_9223)){
        U16 BufferID = *((U16 *)pwBuffer);
        U32 SampIntrv = 0;
        U8 fAutoReset = (U8)iotConf.mData.m922xAO.fAutoResetBuf;

        if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber, 1,
            &wPort, dwWriteCount*(U32)((CurrentCard_DOFunc.wResoluation+7)/8),
            TRUE))!=NoError)
        {
            return err;
        }

        /*SyncMode must be 1 iteration*/
        //if((wIterations!=1)&&(wSyncMode==SYNCH_OP))
        //    return ErrorUndefinedParameter;

        //Infinite repeat dose not work with retrigger mode
        //if((!wIterations)&&(iotConf.mData.m922xAO.fEnReTrig))
        //    return ErrorInvalidTriggerMode;

        if(((CurrentCard.DO_Status.bDblBufferMode)||(wSyncMode==SYNCH_OP))&&
            (iotConf.mData.m922xAO.fEnReTrig))
        {
            return ErrorInvalidTriggerMode;
        }

        if(iotConf.mData.m922xAO.fConvSrc==0){
            if((fSampleRate==0)||(fSampleRate>2000000))
                return ErrorInvalidSampleRate;
            else
                SampIntrv = (U32)(80000000/fSampleRate);
        }
        else{
            //09/25/08
            //External Conversion Source only valid for 1 or infinite interations
            //if((definite)&&(wIterations>1))
                //return ErrorUndefinedParameter;
            //External Conversion Source only valid for ASYNCMODE
            if(wSyncMode==SYNCH_OP)
                return ErrorUndefinedParameter;
        }

        if(!CurrentCard.DO_Status.bDblBufferMode){
            if((BufferID>1)||(!CurrentCard.DO_Status.pDAQbuf[BufferID]))
                return ErrorInvalidBufferID;
            CurrentCard.DO_Status.pDAQbufHandle = (U32)(BufferID);
        }
        else{
            if((BufferID>1)||(!CurrentCard.DO_Status.pDAQbuf[0])||
                (!CurrentCard.DO_Status.pDAQbuf[1]))
            {
                return ErrorInvalidBufferID;
            }
            CurrentCard.DO_Status.pDAQbufHandle = 0;
        }
        CurrentCard.DO_Status.bToFile = FALSE;
        CurrentCard.DO_Status.wNumChans = 1;
        CurrentCard.DO_Status.dwTransferCount = dwWriteCount;

        //iotSCont.wErrorCode = ErrorContOPIoctl;
        iotSCont.wAutoScan = 0;
        iotSCont.wSyncMode = wSyncMode;
        iotSCont.wChannelCount = 1;
        iotSCont.dwAccessCount = dwWriteCount;
        iotSCont.dwChannelGain[0].wChannel = wPort;
        iotSCont.dwBufferAddr = (U32)CurrentCard.DO_Status.pDAQbufHandle;
        iotSCont.dwPacerValue[0] = SampIntrv;
        iotSCont.wIterations = wIterations;
        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_CON_WRITE_SG, &(iotSCont))
            !=DAS_ERR_NO)
        {
            return iotSCont.wErrorCode;
        }

        if((wSyncMode==SYNCH_OP)&&(fAutoReset)){
            CurrentCard.DO_Status.pDAQbuf[BufferID] = NULL;
            CurrentCard.DO_Status.dwBufCount[BufferID] = 0;
        }
    }
    else{
        if(CurrentCard.wCardType==PCI_7200){
            if((err = fundamental_check(&CurrentCard_DOFunc,
                wCardNumber, 1, &wPort, dwWriteCount*
                (U32)((CurrentCard_DOFunc.wResoluation+7)/8),
                TRUE))!=NoError)
            {
                return err;
            }
            if(pwBuffer){
                memcpy(
                    CurrentCard_DOFunc.pView[0],
                    pwBuffer,
                    dwWriteCount*(U32)((CurrentCard_DOFunc.wResoluation+7)/8)
                );
            }
        }
        if(CurrentCard.wCardType==PCI_7300A_RevA){
            if((err = fundamental_check(&CurrentCard_DOFunc,
                wCardNumber, 1, &wPort, dwWriteCount*
                (U32)pwidth[iotConf.mData.m7300DIO.fPortWidth]/8,
                TRUE))!=NoError)
            {
                return err;
            }
            //Set Buffer Address
            iotConf.mData.m7300DIO.dwBufferAddr = (U32)pwBuffer;
            //yuan add 11/30/99
            iotConf.mData.m7300DIO.fConfig = 0;
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_CONFIG, &(iotConf))
                !=DAS_ERR_NO)
            {
                return iotConf.wErrorCode;
            }
        }
        if(CurrentCard.wCardType==PCI_7300A_RevB){
            if((err = fundamental_check(&CurrentCard_DOFunc,
                wCardNumber, 1, &wPort, dwWriteCount*
                (U32)pwidth[iotConf.mData.m7300bDIO.fPortWidth]/8,
                TRUE))!=NoError)
            {
                return err;
            }
            //Set Buffer Address
            iotConf.mData.m7300bDIO.dwBufferAddr = (U32)pwBuffer;
            //yuan add 11/30/99
            iotConf.mData.m7300bDIO.fConfig = 0;
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_CONFIG, &(iotConf))
                !=DAS_ERR_NO)
            {
                return iotConf.wErrorCode;
            }
        }
        if(CurrentCard.wCardType==PCI_7350){
            if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber, 1,
                &wPort, dwWriteCount*(iotConf.mData.m7350DIO.PortCount+1),
                TRUE))!=NoError)
            {
                return err;
            }
        }

        //00/06/02 jeffrey , lock memory for the user-level buffer
        if((CurrentCard.wCardType==PCI_7300A_RevA)||
            (CurrentCard.wCardType==PCI_7300A_RevB)||
            (CurrentCard.wCardType==PCI_7350))
        {
            if(CurrentCard.wCardType==PCI_7350)
                byte_width = iotConf.mData.m7350DIO.PortCount+1;
            else if(CurrentCard.wCardType==PCI_7300A_RevA)
                byte_width = pwidth[iotConf.mData.m7300DIO.fPortWidth]/8;
            else
                byte_width = pwidth[iotConf.mData.m7300bDIO.fPortWidth]/8;

            err = mlock((const void *)pwBuffer, dwWriteCount*byte_width);
            if(err!=0){ //errno defined in errno.h
                needlock = 1;
                goto lockio;
                if(errno==ENOMEM)
                    printf("memory lock failed : errno = ENOMEM , please check the memory size\n");
                else if(errno==EPERM)
                    printf("memory lock failed : error =  EPERM, mlock system call is only for ROOT\n");
                else if(errno==EINVAL)
                    printf("memory lock failed : error = EINVAL, Invalate value\n");
                else
                    printf("memory lock failed : errno=%d\n", errno);
                return ErrorLockMemory;
            }
lockio:
            //record the user-level buffer for unlock in DO_AsyncClear
            CurrentCard.DO_Status.p7300buf = pwBuffer;
        }

        if(CurrentCard.DO_Status.bDblBufferMode)
            dwWriteCount /= 2;

        CurrentCard.DO_Status.dwTransferCount = dwWriteCount;

        //06/10/09
        if((CurrentCard.wCardType!=PCI_7350)||
            ((CurrentCard.wCardType==PCI_7350)&&
            (iotConf.mData.m7350DIO.OPMode!=1)&&
            (!iotConf.mData.m7350DIO.ExtClk)))
        {
            //Set PacerValue and TimerCascade
            if(!get_pacervalue(wCardNumber, fSampleRate, &iotCont.dwPacerValue))
                return ErrorInvalidSampleRate;
        }
        else if((CurrentCard.wCardType==PCI_7350)&&
            (iotConf.mData.m7350DIO.OPMode==1))
        {
            iotCont.dwPacerValue = 2;
        }

        if(CurrentCard.wCardType==PCI_7200){
            //Set Timer Cascade
            if(iotCont.dwPacerValue>65535)
                iotConf.mData.m7200DIO.fTimerCascade = 1;
            else
                iotConf.mData.m7200DIO.fTimerCascade = 0;

            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_CONFIG, &(iotConf))
                !=DAS_ERR_NO)
            {
                return iotConf.wErrorCode;
            }
        }
        if(CurrentCard.wCardType==PCI_7350){
            if(CurrentCard.DO_Status.bChanRange==0){
                /*Non-Multi-Buffer*/
                DAS_IOT_MBUF iotMbuf = {0};
                iotMbuf.dwBufferAddr = (U32)pwBuffer;
                if((iotConf.mData.m7350DIO.WaitTrig==1)&&
                    (iotConf.mData.m7350DIO.WaitSTrig==0))
                {
                    /*External Trigger*/
                    if(!CurrentCard.AO_Status.CurrentConfig.mData.dwValue[1]){
                        /*Infinite Trigger*/
                        iotMbuf.dwAccessCount = dwWriteCount*2;
                        CurrentCard.DO_Status.dwTransferCount *= 2;
                    }
                    else{
                        iotMbuf.dwAccessCount = dwWriteCount*
                           CurrentCard.AO_Status.CurrentConfig.mData.dwValue[1];
                        CurrentCard.DO_Status.dwTransferCount *=
                           CurrentCard.AO_Status.CurrentConfig.mData.dwValue[1];
                    }
                }
                else
                    iotMbuf.dwAccessCount = dwWriteCount;
                iotMbuf.wBufferId = 1|0x8000; //0: DI; 1:DO; //0x8000 single buffer
                //iotMbuf.wErrorCode = ErrorConfigIoctl;
                if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_MBF_SET, &(iotMbuf))
                    !=DAS_ERR_NO)
                {
                    munlock((const void *)pwBuffer, dwWriteCount*byte_width);
                    //yuan add 04/16/01
                    CurrentCard.DO_Status.p7300buf = NULL;
                    return iotMbuf.wErrorCode;
                }
            }
        }

        iotCont.dwAccessCount = dwWriteCount;
        iotCont.wAutoScan = 0;
        iotCont.wIterations = wIterations;
        iotCont.wSyncMode = wSyncMode;
        iotCont.wChannelCount = 1;
        iotCont.dwChannelGain[0].wChannel = wPort;
        iotCont.dwChannelGain[0].wGain = needlock;//0;
        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_CON_WRITE, &(iotCont))
            !=DAS_ERR_NO)
        {
            if((CurrentCard.wCardType==PCI_7300A_RevA)||
                (CurrentCard.wCardType==PCI_7300A_RevB)||
                (CurrentCard.wCardType==PCI_7350))
            {
                munlock((const void *)pwBuffer, dwWriteCount*byte_width);
                CurrentCard.DO_Status.p7300buf = NULL;
            }
            return iotCont.wErrorCode;
        }
        //00/06/02 jeffrey, for SyncMode , Unlock the user-level memory
        if(((CurrentCard.wCardType==PCI_7300A_RevA)||
            (CurrentCard.wCardType==PCI_7300A_RevB)||
            (CurrentCard.wCardType==PCI_7350))&&
            (wSyncMode==SYNCH_OP))
        {
            munlock((const void *)pwBuffer, dwWriteCount*byte_width);
            CurrentCard.DO_Status.p7300buf = NULL;
        }
    }

    return NoError;
}


I16 DO_PGStart(
    U16  wCardNumber,
    void *pwBuffer,
    U32  dwWriteCount,
    F64  fSampleRate
    )
{
    //00/06/02 jeffrey , lock memory for data_pattern
    I16 pwidth[6] = {0, 4, 8, 16, 24, 32};
    U32 byte_width = 0;
    U8 needlock = 0;
    DAS_IOT_CONFIG iotConf;
    DAS_IOT_PGEN iotPgen;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));
    memset(&iotPgen, '\0', sizeof(DAS_IOT_PGEN));

    iotConf = CurrentCard.DO_Status.CurrentConfig;
    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    if(CurrentCard.wCardType==PCI_7300A_RevA){
        //Set Buffer Address
        iotConf.mData.m7300DIO.dwBufferAddr = (U32)pwBuffer;
        //yuan add 11/30/99
        iotConf.mData.m7300DIO.fConfig = 0;
        byte_width = pwidth[iotConf.mData.m7300DIO.fPortWidth]/8;
        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_CONFIG, &(iotConf))
            !=DAS_ERR_NO)
        {
            return iotConf.wErrorCode;
        }
    }
    if(CurrentCard.wCardType==PCI_7300A_RevB){
        //Set Buffer Address
        iotConf.mData.m7300bDIO.dwBufferAddr = (U32)pwBuffer;
        //yuan add 11/30/99
        iotConf.mData.m7300bDIO.fConfig = 0;
        byte_width = pwidth[iotConf.mData.m7300bDIO.fPortWidth]/8;
        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_CONFIG, &(iotConf))
            !=DAS_ERR_NO)
        {
            return iotConf.wErrorCode;
        }
    }

    iotPgen.dwAccessCount = dwWriteCount;
    //Set PacerValue and TimerCascade
    if(!get_pacervalue(wCardNumber, fSampleRate, &iotPgen.dwPacerValue))
        return ErrorInvalidSampleRate;

    //00/06/02 jeffrey , lock memory for the user-level buffer
    if((CurrentCard.wCardType==PCI_7300A_RevA)||
        (CurrentCard.wCardType==PCI_7300A_RevB))
    {
        if(mlock((const void *)pwBuffer, dwWriteCount*byte_width)!=0){
            needlock = 1;
            goto lockio;
            if(errno==ENOMEM) //errno defined in errno.h
                printf("memory lock failed : errno = ENOMEM, please check the memory size\n");
            else if(errno==EPERM)
                printf("memory lock failed : error =  EPERM, mlock system call is only for ROOT\n");
            else if(errno==EINVAL)
                printf("memory lock failed : error = EINVAL, Invalate value\n");
            else
                printf("memory lock failed : errno=%d\n", errno);
            return ErrorLockMemory;
        }
lockio:
        CurrentCard.DO_Status.p7300buf = pwBuffer;
        CurrentCard.DO_Status.dwTransferCount = dwWriteCount;
    }
    iotPgen.wErrorCode = needlock;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_PG_START, &(iotPgen))
        !=DAS_ERR_NO)
    {
        return iotPgen.wErrorCode;
    }

    return NoError;
}


I16 DO_PGStop(
    U16 wCardNumber
    )
{
    //00/06/02 jeffrey , lock memory for data_pattern
    I16 pwidth[6] = {0, 4, 8, 16, 24, 32};
    U32 byte_width;
    DAS_IOT_PGEN iotPgen;

    memset(&iotPgen, '\0', sizeof(DAS_IOT_PGEN));

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_PG_STOP, &(iotPgen))
        !=DAS_ERR_NO)
    {
        return iotPgen.wErrorCode;
    }

    //00/06/02 jeffrey , unlock memory for data_pattern
    if(CurrentCard.wCardType==PCI_7300A_RevA||
        CurrentCard.wCardType==PCI_7300A_RevB)
    {
        if(CurrentCard.wCardType==PCI_7300A_RevA){
            byte_width = \
                pwidth[CurrentCard_DOCurConfig.mData.m7300DIO.fPortWidth]/8;
        }
        else{
            byte_width = \
                pwidth[CurrentCard_DOCurConfig.mData.m7300bDIO.fPortWidth]/8;
        }
        munlock(
            (const void *)CurrentCard.DO_Status.p7300buf,
            CurrentCard.DO_Status.dwTransferCount*byte_width
        );
    }

    return NoError;
}


#ifdef WHOLE
I16 DO_ContWriteMultiPorts(
    U16  wCardNumber,
    U16  wNumPorts,
    U16  *pwPorts,
    void *pwBuffer,
    U32  dwWriteCount,
    U16  wIterations,
    F64  fSampleRate,
    U16  wSyncMode
    )
{
/*
    I16 err;

    if((err = fundamental_check(&CurrentCardInfo.DO_Func, wCardNumber, wNumPorts,
        pwPorts, dwWriteCount*(U32)((CurrentCard_DOFunc.wResoluation+7)/8),
        TRUE))!=NoError)
    {
        return err;
    }

    memcpy(CurrentCardInfo.DO_Func.pView[0], pwBuffer, dwWriteCount);
    check pwPorts to determine if AutoScan set on
    {
        if(!CurrentCardInfo.DO_Func.fAutoScan) then no AutoScan
        if(ports do not start/end port-0 and fEndingZero) then no AutoScan
        if(ports are continuous){
            check bScanDirection to see if port ascending/decending is allowed
            if yes then AutoScan else no AutoScan
        }
        else no AutoScan
    }
    ioc=DAS_IOC_DO_CON_READ perform DO
    in :
    {   dwAccessCnt = dwWriteCount;
        wSampCntr = calculate_SampCntr(fSampleRate);
        wXferMode = CurrentCard.DO_Status.wXferMode;
        wTrigMode = CurrentCard.DO_Status.wTrigMode;
        wAutoScan = AutoScan;
        wSyncMode = wSyncMode;
        wChannelCnt = wNumPorts;
        for(i=0; i<wNumPorts; i++)
            wChannelGain[i].wChannel = pwPorts[i];
    }
*/
}
#endif


I16 DO_WritePort(
    U16 wCardNumber,
    U16 wPort,
    U32 Value
    )
{
    I16 err;
    DAS_IOT_PIO iotPio;

    memset(&iotPio, '\0', sizeof(DAS_IOT_PIO));

    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    iotPio.wChannel = wPort;
    //"Port Size" is different
    //09/22/08 modified
    switch(CurrentCard.wCardType){
        case PCI_6202:
            if(wPort==0){
                //ISO DO - 16 bits
                iotPio.wByteCount = 2;
            }
            else if(wPort==1){
                //TTL - 8 bits
                iotPio.wByteCount = 1;
            }
            break;
        case PCI_7350:
            iotPio.wByteCount = 1;
            break;
        case PCI_7396:
            if(wPort<Channel_P1)
                iotPio.wByteCount = 1;
            else
                iotPio.wByteCount = (CurrentCard_DOFunc.wResoluation+7)/8;
            break;
        case PCI_7442:
            if(wPort<2){
                //DO Channel - 32 bits Port
                iotPio.wByteCount = 4;
            }
            else if(wPort<4){
                //TTL port - 16 bits Port
                iotPio.wByteCount = 2;
            }
            break;
        case PCI_7443:
            iotPio.wByteCount = 2;
            break;
        case PCI_7444:
            if(wPort<4){
                //DO0 ~ DO3 channel - 32 bits
                iotPio.wByteCount = 4;
            }
            else if(wPort<6){
                //TTL port - 16 bits
                iotPio.wByteCount = 2;
            }
            break;
        default:
            iotPio.wByteCount = (CurrentCard_DOFunc.wResoluation+7)/8;
    }
    iotPio.dwValue = Value;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_PIO_WRITE, &(iotPio))
        !=DAS_ERR_NO)
    {
        return iotPio.wErrorCode;
    }

    return NoError;
}


I16 DO_ReadPort(
    U16 wCardNumber,
    U16 wPort,
    U32 *pwValue
    )
{
    I16 err;
    DAS_IOT_PIO iotPio;

    memset(&iotPio, '\0', sizeof(DAS_IOT_PIO));

    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber,/*1*/0,
        &wPort, 0L, FALSE))!=NoError)
    {
        return err;
    }

    iotPio.wChannel = wPort;
    iotPio.wByteCount = (CurrentCard_DOFunc.wResoluation+7)/8;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_PIO_READ, &(iotPio))
        !=DAS_ERR_NO)
    {
        return iotPio.wErrorCode;
    }
    *pwValue = iotPio.dwValue;

    return NoError;
}


#ifdef WHOLE
I16 DO_WriteDPort(
    U16 wCardNumber,
    U16 wPort,
    U16 wValue
    )
{
    DAS_IOT_PIO iotPio;
    U16 err, wDPort;

    memset(&iotPio, '\0', sizeof(DAS_IOT_PIO));
    wDPort = wPort*2;

    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber, 1, &wDPort,
        0L, FALSE))!=NoError)
    {
        return err;
    }

    iotPio.wChannel = wPort;
    iotPio.wByteCount = 2;
    iotPio.dwValue = wValue;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_PIO_WRITE, &(iotPio))
        !=DAS_ERR_NO)
    {
        return iotPio.wErrorCode;
    }

    return NoError;
}


I16 DO_WriteQPort(
    U16 wCardNumber,
    U16 wPort,
    U32 dwValue
    )
{
    U16 err, wQPort;
    DAS_IOT_PIO iotPio;

    memset(&iotPio, '\0', sizeof(DAS_IOT_PIO));

    wQPort = wPort*4;
    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber,1, &wQPort,
        0L, FALSE))!=NoError)
    {
        return err;
    }

    iotPio.wChannel = wPort;
    iotPio.wByteCount = 4;
    iotPio.dwValue = dwValue;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_PIO_WRITE, &(iotPio))
        !=DAS_ERR_NO)
    {
        return iotPio.wErrorCode;
    }

    return NoError;
}
#endif


I16 DO_WriteLine(
    U16 wCardNumber,
    U16 wPort,
    U16 wLine,
    U16 wValue
    )
{
    I16 err;
    DAS_IOT_PIO iotPio;

    memset(&iotPio, '\0', sizeof(DAS_IOT_PIO));

    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber,/*1*/0,
        &wPort, 0L, FALSE))!=NoError)
    {
        return err;
    }

    iotPio.wChannel = wPort;
    iotPio.wLine_Gain = wLine;
    iotPio.wByteCount = 0;
    iotPio.dwValue = wValue;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_PIO_WRITE, &(iotPio))
        !=DAS_ERR_NO)
    {
        return iotPio.wErrorCode;
    }

    return NoError;
}


I16 DO_WriteExtTrigLine(
    U16 wCardNumber,
    U16 wValue
    )
{
    //I16 err;
    DAS_IOT_PIO iotPio;

    memset(&iotPio, '\0', sizeof(DAS_IOT_PIO));

    iotPio.wChannel = 0;
    iotPio.wLine_Gain = ExtTrigLine;
    iotPio.wByteCount = 0;
    iotPio.dwValue = wValue;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_PIO_WRITE, &(iotPio))
        !=DAS_ERR_NO)
    {
        return iotPio.wErrorCode;
    }

    return NoError;
}


I16 DO_ReadLine(
    U16 wCardNumber,
    U16 wPort,
    U16 wLine,
    U16 *pwValue
    )
{
    I16 err;
    DAS_IOT_PIO iotPio;

    memset(&iotPio, '\0', sizeof(DAS_IOT_PIO));

    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber,/*1*/0,
        &wPort, 0L, FALSE))!=NoError)
    {
        return err;
    }

    iotPio.wChannel = wPort;
    iotPio.wLine_Gain = wLine;
    iotPio.wByteCount = 0;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_PIO_READ, &(iotPio))
        !=DAS_ERR_NO)
    {
        return iotPio.wErrorCode;
    }
    *pwValue = (U16)iotPio.dwValue;

    return NoError;
}


I16 DO_InitialMemoryAllocated(
    U16 wCardNumber,
    U32 *MemSize
    )
{
    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    //06/02/09 add for 7350
    if(CurrentCard.wCardType==PCI_7350){
        if(CurrentCard_DOFunc.dwBufferSize)
            *MemSize = (CurrentCard_DOFunc.dwBufferSize-128*20)/5;
        else
            *MemSize = 0;
    }
    else{
        *MemSize = CurrentCard_DOFunc.dwBufferSize/1024;  //in KB
        if((CurrentCard.wCardType==PCI_7300A_RevA)||
            (CurrentCard.wCardType == PCI_7300A_RevB)||
            (CurrentCard.wCardType==PCI_9222)||
            (CurrentCard.wCardType==PCI_9223))
        {
            //07/07/08 modified
            if(*MemSize)
                *MemSize = (*MemSize-4)*256;
        }
    }

    return NoError;
}


//yuan modified 04/13/99
//for 9111 EDO config
I16 EDO_9111_Config(
    U16 wCardNumber,
    U16 EDO_Fun
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    if((err = fundamental_check(&CurrentCard_AIFunc, wCardNumber,0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    iotConf.mData.m9111AI.wEDO_Function = EDO_Fun;
    CurrentCard.AI_Status.CurrentConfig = iotConf;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }

    return NoError;
}


I16 DO_ContStatus(
    U16 wCardNumber,
    U16 *Status
    )
{
    DAS_IOT_STATUS iotStat;

    memset(&iotStat, '\0', sizeof(DAS_IOT_STATUS));

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_CON_STATUS, &(iotStat))
        !=DAS_ERR_NO)
    {
        return iotStat.wErrorCode;
    }
    *Status = iotStat.mStat.wValue;

    return NoError;
}


/* 01/28/02 added for MultiBuffer DMA DO */
I16 DO_ContMultiBufferSetup(
    U16  wCardNumber,
    void *pwBuffer,
    U32  dwWriteCount,
    U16  *BufferId
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;
    DAS_IOT_MBUF iotMbuf;
    //00/06/02 jeffrey added
    I16 pwidth[6] = {0, 4, 8, 16, 24, 32};
    U32 byte_width;
    U8 needlock = 0;

    memset(&iotMbuf, '\0', sizeof(DAS_IOT_MBUF));
    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    iotConf = CurrentCard.DO_Status.CurrentConfig;

    if((CurrentCard.wCardType==PCI_7300A_RevB)||
        (CurrentCard.wCardType==PCI_7300A_RevA)||
        (CurrentCard.wCardType==PCI_7350))
    {
        if(CurrentCard.wCardType==PCI_7350)
            byte_width = iotConf.mData.m7350DIO.PortCount+1;
        else if(CurrentCard.wCardType==PCI_7300A_RevA)
            byte_width = pwidth[iotConf.mData.m7300DIO.fPortWidth]/8;
        else
            byte_width = pwidth[iotConf.mData.m7300bDIO.fPortWidth]/8;

        err = mlock((const void *)pwBuffer, dwWriteCount*byte_width);
        if(err!=0){ //errno defined in errno.h
            needlock = 1;
            goto lockio;
            if(errno==ENOMEM)
                printf("memory lock failed : errno = ENOMEM , please check the memory size\n");
            else if(errno==EPERM)
                printf("memory lock failed : error =  EPERM, mlock system call is only for ROOT\n");
            else if(errno==EINVAL)
                printf("memory lock failed : error = EINVAL, Invalate value\n");
            else
                printf("memory lock failed : errno=%d\n", errno);
            return ErrorLockMemory;
        }
lockio:
        //record the user-level buffer for unlock in DO_AsyncClear
        iotMbuf.dwBufferAddr = (U32)pwBuffer;
        iotMbuf.dwAccessCount = dwWriteCount;
        //yuan add 11/05/01
        iotMbuf.wBufferId = 1;
        iotMbuf.wErrorCode = needlock;
        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_MBF_SET, &(iotMbuf))
            !=DAS_ERR_NO)
        {
            munlock((const void *)pwBuffer, dwWriteCount*byte_width);
            return iotMbuf.wErrorCode;
        }
        *BufferId = iotMbuf.wBufferId;
        CurrentCard.DO_Status.p7300mbuf[*BufferId].dwBufferAddr =(U32)pwBuffer;
        CurrentCard.DO_Status.p7300mbuf[*BufferId].dwAccessCount =dwWriteCount;
        CurrentCard.DO_Status.p7300mbuf[*BufferId].wBufferId = *BufferId;
        CurrentCard.DO_Status.p7300mbufCnt += 1;
    }
    else
        return ErrorFuncNotSupport;

    if(CurrentCard.wCardType==PCI_7350){
        /*Multi-Buffer Mode*/
        CurrentCard.DO_Status.bChanRange = 1;
    }
    else{
        //yuan add 10/23/02
        CurrentCard.DO_Status.bDblBufferMode = 1;
    }

    return NoError;
}


I16 DO_ContMultiBufferReset(
    U16 wCardNumber
    )
{
    DAS_IOT_MBUF iotMbuf;
    I16 pwidth[6] = {0, 4, 8, 16, 24, 32};
    U32 byte_width;
    int i;

    memset(&iotMbuf, '\0', sizeof(DAS_IOT_MBUF));

    if((CurrentCard.wCardType==PCI_7300A_RevB)||
        (CurrentCard.wCardType==PCI_7300A_RevA))
    {
        iotMbuf.dwAccessCount = 0;
        //yuan add 11/05/01
        iotMbuf.wBufferId = 1;
        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_MBF_SET, &(iotMbuf))
            !=DAS_ERR_NO)
        {
            return iotMbuf.wErrorCode;
        }
        if(CurrentCard.wCardType==PCI_7300A_RevA){
            byte_width = \
                pwidth[CurrentCard_DOCurConfig.mData.m7300DIO.fPortWidth]/8;
        }
        else{
            byte_width = \
                pwidth[CurrentCard_DOCurConfig.mData.m7300bDIO.fPortWidth]/8;
        }

        //yuan add 04/11/01 for multibuffer unlock
        for(i=0; i<CurrentCard.DO_Status.p7300mbufCnt; i++){
            if(CurrentCard.DO_Status.p7300mbuf[i].dwBufferAddr){
                munlock(
                    (const void *)
                    CurrentCard.DO_Status.p7300mbuf[i].dwBufferAddr,
                    CurrentCard.DO_Status.p7300mbuf[i].dwAccessCount*
                    byte_width
                );
            }
            CurrentCard.DO_Status.p7300mbuf[i].dwBufferAddr = (U32)NULL;
            CurrentCard.DO_Status.p7300mbuf[i].dwAccessCount = 0;
            CurrentCard.DO_Status.p7300mbuf[i].wBufferId = -1;
        }
        CurrentCard.DO_Status.p7300mbufCnt = 0;
    }
    else
        return ErrorFuncNotSupport;

    //yuan add 10/23/02
    CurrentCard.DO_Status.bDblBufferMode = 0;

    return NoError;
}


I16 DO_ContMultiBufferStart(
    U16 wCardNumber,
    U16 wPort,
    F64 fSampleRate
    )
{
    //00/02/03 jeffrey, localize the global vars.
    DAS_IOT_CONT iotCont;
    DAS_IOT_CONFIG iotConf;

    memset(&iotCont, '\0', sizeof(DAS_IOT_CONT));
    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));

    iotConf = CurrentCard.DO_Status.CurrentConfig;

    //06/10/09
    if((CurrentCard.wCardType!=PCI_7350)||
        ((CurrentCard.wCardType==PCI_7350)&&
        (iotConf.mData.m7350DIO.OPMode!=1)&&
        (!iotConf.mData.m7350DIO.ExtClk)))
    {
        //Set PacerValue and TimerCascade
        if(!get_pacervalue(wCardNumber, fSampleRate, &iotCont.dwPacerValue))
            return ErrorInvalidSampleRate;
    }
    else if((CurrentCard.wCardType==PCI_7350)&&
        (iotConf.mData.m7350DIO.OPMode==1))
    {
        iotCont.dwPacerValue = 2;
    }

    //yuan 07/31/00
    iotCont.wAutoScan = 0;
    iotCont.wSyncMode = 0;
    iotCont.wChannelCount = 1;
    iotCont.dwChannelGain[0].wChannel = wPort;
    iotCont.dwChannelGain[0].wGain = 0;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_CON_WRITE, &(iotCont))
        !=DAS_ERR_NO)
    {
        return iotCont.wErrorCode;
    }

    return NoError;
}


I16 DO_AsyncMultiBufferNextReady(
    U16     wCardNumber,
    BOOLEAN *bNextReady,
    U16     *wBufferId
    )
{
    I16 err;
    DAS_IOT_DBF iotDbf;

    memset(&iotDbf, '\0', sizeof(DAS_IOT_DBF));

    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_DBF_HALF_READY, &(iotDbf))
        !=DAS_ERR_NO)
    {
        return iotDbf.wErrorCode;
    }
    *bNextReady = (BOOLEAN)iotDbf.wReady;
    *wBufferId = CurrentCard.DO_Status.wDblBufferIndex = iotDbf.wViewIndex;

    return NoError;
}


I16 DO_SimuWritePort(
    U16 wCardNumber,
    U16 wNumChans,
    U32 *pdwBuffer
    )
{
    I16 err;
    int i;
    DAS_IOT_SIMU_PIO iotSPio;

    memset(&iotSPio, '\0', sizeof(DAS_IOT_SIMU_PIO));

    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    switch(CurrentCard.wCardType){
        case PCI_7442:
            iotSPio.wChannelCount = wNumChans;
            if((wNumChans!=1)&&(wNumChans!=2))
                return ErrorUndefinedParameter;//wNumChans = 2;
            for(i=0;i<wNumChans;i++){
                iotSPio.dwChannelGain[i].wChannel = (*(pdwBuffer+i))&0xffff;
                iotSPio.dwChannelGain[i].wGain = ((*(pdwBuffer+i))>>16)&0xffff;
            }
            break;
        case PCI_7444:
            iotSPio.wChannelCount = wNumChans;
            if((wNumChans!=1)&&(wNumChans!=2)&&(wNumChans!=4))
                return ErrorUndefinedParameter;
            for(i=0;i<wNumChans;i++){
                iotSPio.dwChannelGain[i].wChannel = (*(pdwBuffer+i))&0xffff;
                iotSPio.dwChannelGain[i].wGain = ((*(pdwBuffer+i))>>16)&0xffff;
            }
            break;
        default:
            return ErrorFuncNotSupport;
    }

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_SIMU_AI_PIO, &(iotSPio))
        !=DAS_ERR_NO)
    {
        return iotSPio.wErrorCode;
    }

    return NoError;
}


I16 DO_SetTimeOut(
    U16 wCardNumber,
    U32 dwTimeOut
    )
{
    I16 err;
    DAS_IOT_SI iotSI;

    memset(&iotSI, '\0', sizeof(DAS_IOT_SI));

    /*
     * Error Check
     */
    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    /*
     * Setup Structure to be Passed to Kernel
     */
    //TimeOut Period(ms) for DO Sync. Mode
    iotSI.dwScanInterval = dwTimeOut;
    iotSI.dwSampInterval = 0;
    //iotSI.wErrorCode = ErrorConfigIoctl;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_TIMEOUT_SET, &(iotSI))
        !=DAS_ERR_NO)
    {
        return iotSI.wErrorCode;
    }

    return NoError;
}


I16 DO_ContBufferSetup(
    U16  wCardNumber,
    void *pwBuffer,
    U32  dwWriteCount,
    U16  *BufferId
    )
{
    DAS_IOT_MBUF2 iotMbuf = {0};
    I16 err;

    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber, 0, NULL,
        dwWriteCount, TRUE))!=NoError)
    {
        return err;
    }

    iotMbuf.dwBufferAddr = (U32)pwBuffer; //Buffer Address to be Setup
    iotMbuf.dwAccessCount = dwWriteCount; //Buffer Size to be Setup
    iotMbuf.wDir = 3; //3: Let Kernel Know that DO Buffer to be Setup
    iotMbuf.wBufferId = 1; //1: Lock Memory - Compatiable for Linux Buffer Set
    //iotMbuf.wErrorCode = ErrorConfigIoctl;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_MBF_SET, &(iotMbuf))
        !=DAS_ERR_NO)
    {
        return iotMbuf.wErrorCode;
    }
    *BufferId = iotMbuf.wBufferId; //Buffer ID Got from Kernal Mode
    //Set Buffer Structure
    CurrentCard.DO_Status.pDAQbuf[*BufferId] = pwBuffer;
    CurrentCard.DO_Status.dwBufCount[*BufferId] = dwWriteCount;

    return NoError;
}


I16 DO_ContBufferReset(
    U16 wCardNumber
    )
{
    DAS_IOT_MBUF2 iotMbuf = {0};
    I16 err;

    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber, 0, NULL, 0L,
        TRUE))!=NoError)
    {
        return err;
    }

    iotMbuf.wDir = 3; //3: Let Kernel Know that DO Buffer to be Reset
    //iotMbuf.wErrorCode = ErrorConfigIoctl;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_MBF_RESET, &(iotMbuf))
        !=DAS_ERR_NO)
    {
        return iotMbuf.wErrorCode;
    }

    //Reset DO Buffer Structure to Default
    CurrentCard.DO_Status.pDAQbuf[0] = NULL;
    CurrentCard.DO_Status.pDAQbuf[1] = NULL;
    CurrentCard.DO_Status.dwBufCount[0] = 0;
    CurrentCard.DO_Status.dwBufCount[1] = 0;

    return NoError;
}


I16 DO_9222_Config(
    U16     wCardNumber,
    U16     ConfigCtrl,
    U16     TrigCtrl,
    U32     ReTrgCnt,
    U32     DLY1Cnt,
    U32     DLY2Cnt,
    BOOLEAN AutoResetBuf
    )
{
    I16 err;
    DAS_EAO_CONFIG m9222ao;

    memset(&m9222ao, '\0', sizeof(DAS_EAO_CONFIG));

    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    /*Conversion Source*/
    m9222ao.iotConf.mData.m922xAO.fConvSrc = ConfigCtrl&0xf;
    /*Auto Reset Buffer?*/
    m9222ao.iotConf.mData.m922xAO.fAutoResetBuf = AutoResetBuf&0x1;
    /*Trigger Mode*/
    m9222ao.iotConf.mData.m922xAO.fTriggerMode = TrigCtrl&0x1;
    if(m9222ao.iotConf.mData.m922xAO.fTriggerMode){
        //Delay 1 Trigger Coumt
        m9222ao.Counter_Value[1] = DLY1Cnt;
    }
    /*Trigger Source*/
    m9222ao.iotConf.mData.m922xAO.fTrigSrc = (TrigCtrl>>4)&0xf;
    /*Trigger Polarity*/
    m9222ao.iotConf.mData.m922xAO.fTrigPol = (TrigCtrl>>8)&0x1;
    /*ReTrigger*/
    m9222ao.iotConf.mData.m922xAO.fEnReTrig = (TrigCtrl>>9)&0x1;
    if(m9222ao.iotConf.mData.m922xAO.fEnReTrig){
        //ReTriggerCount
        m9222ao.Counter_Value[0] = ReTrgCnt;
    }
    //Not Support Now
    /*Delay 2 Trigger*/
    m9222ao.iotConf.mData.m922xAO.fEnDly2 = (TrigCtrl>>10)&0x1;
    if(m9222ao.iotConf.mData.m922xAO.fEnDly2){
        //Delay 2 Trigger Count
        m9222ao.Counter_Value[2] = DLY2Cnt;
    }
    //m9222ao.iotConf.wErrorCode = ErrorConfigIoctl;
    CurrentCard.DO_Status.CurrentConfig = m9222ao.iotConf;
    /*
     * In Linux, use DAS_IOC_EDO_CONFIG to replace DAS_IOC_DO_CONFIG
     * to perform 922x DO configurations.
     */
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_EDO_CONFIG, &(m9222ao))
        !=DAS_ERR_NO)
    {
        return m9222ao.iotConf.wErrorCode;
    }
    //CurrentCard.DO_Status.CurrentConfig.wErrorCode = NoError;

    return NoError;
}


I16 DO_9223_Config(
    U16     wCardNumber,
    U16     ConfigCtrl,
    U16     TrigCtrl,
    U32     ReTrgCnt,
    U32     DLY1Cnt,
    U32     DLY2Cnt,
    BOOLEAN AutoResetBuf
    )
{
    return DO_9222_Config(wCardNumber, ConfigCtrl, TrigCtrl, ReTrgCnt,
        DLY1Cnt, DLY2Cnt, AutoResetBuf);
}


//10/14/08
/*
 * 7300B DO will output unexpected pulse While set diocsr.Enable to 0
 * So, add this function for user to set whether diocsr.Enable will be
 * set to 0 or not.
 * 0: Disable DO in DO_AsyncClear (Default)
 * 1: Does not disable DO in DO_AsyncClear
 */
I16 DO_7300B_SetDODisableMode(
    U16 wCardNumber,
    U16 wMode
    )
{
    DAS_IOT_ASYNC iotAsync;

    memset(&iotAsync, '\0', sizeof(DAS_IOT_ASYNC));

    iotAsync.wStopped = (0x8000|wMode);
    //iotAsync.wErrorCode = ErrorAsyncSetIoctl;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_ASY_CLEAR, &(iotAsync))
        !=DAS_ERR_NO)
    {
        return iotAsync.wErrorCode;
    }

    return NoError;
}


I16 DO_7350_Config(
    U16 wCardNumber,
    U16 wDOPortWidth,
    U16 wDOMode,
    U16 wDOWaitStatus,
    U16 wDOClkConfig
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;

    iotConf = CurrentCard.DO_Status.CurrentConfig;

    /*
     * Error Check
     */
    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber, 0, NULL,
        0L, FALSE))!=NoError)
    {
        return err;
    }

    if(CurrentCard.wCardType!=PCI_7350)
        return ErrorFuncNotSupport;

    if((wDOPortWidth!=8)&&(wDOPortWidth!=16)
        &&(wDOPortWidth!=24)&&(wDOPortWidth!=32))
    {
        return ErrorInvalidPortWidth;
    }
    else{
        U32 vi;
        U16 DOPortCount = 0;
        for(vi=0; vi<4; vi++){
            /*Check DIO Port Directions*/
            if(((iotConf.mData.m7350DIO.PortConfig>>vi)&0x1)==0){
                /*Port vi is enabled*/
                if(((iotConf.mData.m7350DIO.PortConfig>>(4+vi))&0x1)==1){
                    /*Port vi is configured as DO*/
                    DOPortCount++;
                }
            }
        }
        if(wDOPortWidth!=DOPortCount*8)
            return ErrorDIODataWidthError;
    }

    if(wDOMode>P7350_BurstHandShake)
        return ErrorInvalidTriggerMode;

    if(wDOWaitStatus>P7350_WAIT_SOFTTRG)
        return ErrorInvalidTriggerType;

    /*
     * Setup Structure to be Passed to Kernel
     */
    //iotConf.wErrorCode = ErrorConfigIoctl;
    //Let Kernel Know DO_7350_Config will be performed
    iotConf.mData.m7350DIO.ConfigMode = 0;
    /*DO Port Width*/
    //0->8bits; 1->16bits; 2->24bits; 3->32bits
    iotConf.mData.m7350DIO.PortCount = wDOPortWidth/8-1;
    /*DO Mode*/
    iotConf.mData.m7350DIO.OPMode = wDOMode;
    /*Trigger Status*/
    if(wDOWaitStatus==P7350_WAIT_SOFTTRG){
        iotConf.mData.m7350DIO.WaitTrig = 1;
        iotConf.mData.m7350DIO.WaitSTrig = 1;
    }
    else if(wDOWaitStatus==P7350_WAIT_EXTTRG){
        iotConf.mData.m7350DIO.WaitTrig = 1;
        iotConf.mData.m7350DIO.WaitSTrig = 0;
    }
    else{
        iotConf.mData.m7350DIO.WaitTrig = 0;
        iotConf.mData.m7350DIO.WaitSTrig = 0;
    }
    /*Sampled Clock*/
    iotConf.mData.m7350DIO.ExtClk = wDOClkConfig&0x1;
    iotConf.mData.m7350DIO.ClkEdge = (wDOClkConfig>>1)&0x1;
    //04/23/09
    /*Export Sampled Clock?*/
    iotConf.mData.m7350DIO.en_expClk = (wDOClkConfig>>2)&0x1;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }
    /*
     * Re-store Configurations to CurrentConfig
     */
    CurrentCard.DO_Status.CurrentConfig = iotConf;
    //CurrentCard.DO_Status.CurrentConfig.wErrorCode = NoError;

    return NoError;
}


I16 DO_7350_TrigHSConfig(
    U16 wCardNumber,
    U16 wTrigConfig,
    U16 wDO_IPOL,
    U16 wDO_REQSrc,
    U16 wDO_ACKSrc,
    U16 wDO_TRIGSrc,
    U16 wStartTrigSrc,
    U16 wPauseTrigSrc,
    U16 wSoftTrigOutSrc,
    U32 dwSoftTrigOutLength,
    U32 dwTrigCount
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;

    iotConf = CurrentCard.DO_Status.CurrentConfig;

    /*
     * Error Check
     */
    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber, 0, NULL,
        0L, FALSE))!=NoError)
    {
        return err;
    }

    if(CurrentCard.wCardType!=PCI_7350)
        return ErrorFuncNotSupport;

    if((wDO_REQSrc>P7350_AFI_7)||
        (wDO_ACKSrc>P7350_AFI_7)||
        (wDO_TRIGSrc>P7350_AFI_7)||
        (wStartTrigSrc>P7350_AFI_7)||
        (wPauseTrigSrc>P7350_AFI_7)||
        (wSoftTrigOutSrc>P7350_AFI_7))
    {
        return ErrorInvalidDioPort;
    }

    /*
     * Setup Structure to be Passed to Kernel
     */
    //iotConf.wErrorCode = ErrorConfigIoctl;
    //Let Kernel Know DO_7350_TrigHSConfig will be performed
    iotConf.mData.m7350DIO.ConfigMode = 1;
    /*Trigger Configuration*/
    iotConf.mData.m7350DIO.PauseTrig = wTrigConfig&0x1;
    //Only Valid for Waiting Software Trigger
    iotConf.mData.m7350DIO.SoftTrigOut = (wTrigConfig>>1)&0x1;
    /*Trigger and HS Polarity*/
    iotConf.mData.m7350DIO.HSREQPol = wDO_IPOL&0x1;
    iotConf.mData.m7350DIO.HSACKPol = (wDO_IPOL>>1)&0x1;
    iotConf.mData.m7350DIO.HSTRIGPol = (wDO_IPOL>>2)&0x1;
    iotConf.mData.m7350DIO.StartTrigPol = (wDO_IPOL>>3)&0x1;
    iotConf.mData.m7350DIO.PauseTrigPol = (wDO_IPOL>>4)&0x1;
    /*Trigger and HS Source*/
    iotConf.mData.m7350DIO.HSREQSrc = wDO_REQSrc;
    iotConf.mData.m7350DIO.HSACKSrc = wDO_ACKSrc;
    iotConf.mData.m7350DIO.HSTRIGSrc = wDO_TRIGSrc;
    iotConf.mData.m7350DIO.StartTrigSrc = wStartTrigSrc;
    iotConf.mData.m7350DIO.PauseTrigSrc = wPauseTrigSrc;
    iotConf.mData.m7350DIO.SoftTrigOutSrc = wSoftTrigOutSrc;
    //Solve the issue of the size of the configuration structure is too small
    {
        //Get the Mapped IO Memory
        U32 LcrAddr = (U32)(CurrentCard_AOFunc.pView[0]);
        *((U32 *)(LcrAddr+0x114)) = dwSoftTrigOutLength;
        *((U32 *)(LcrAddr+0xEC)) = dwTrigCount;
        /*borrow from ao structure to store them*/
        CurrentCard.AO_Status.CurrentConfig.mData.dwValue[0] = \
            dwSoftTrigOutLength;
        CurrentCard.AO_Status.CurrentConfig.mData.dwValue[1] = \
            dwTrigCount;
    }
    //02/25/09
    if((iotConf.mData.m7350DIO.WaitTrig)&&(!iotConf.mData.m7350DIO.WaitSTrig)){
        /*Wait Hardware Trigger*/
        if(dwTrigCount==0){
            /*Infinite Triggers*/
            iotConf.mData.m7350DIO.fReTrig = 1;
            iotConf.mData.m7350DIO.fInfReTrig = 1;
        }
        else if(dwTrigCount==1){
            /*One Trigger*/
            iotConf.mData.m7350DIO.fReTrig = 0;
            iotConf.mData.m7350DIO.fInfReTrig = 0;
        }
        else{
            /*N Triggers*/
            iotConf.mData.m7350DIO.fReTrig = 1;
            iotConf.mData.m7350DIO.fInfReTrig = 0;
        }
    }
    else{
        iotConf.mData.m7350DIO.fReTrig = 0;
        iotConf.mData.m7350DIO.fInfReTrig = 0;
    }
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }
    /*
     * Re-store Configurations to CurrentConfig
     */
    CurrentCard.DO_Status.CurrentConfig = iotConf;
    //CurrentCard.DO_Status.CurrentConfig.wErrorCode = NoError;

    return NoError;
}


I16 DO_7350_ExtSampCLKConfig(
    U16 wCardNumber,
    U16 wCLK_Src,
    U16 wCLK_DDAMode,
    U16 wCLK_DPAMode,
    U16 wCLK_DDAVlaue,
    U16 wCLK_DPAVlaue
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;

    iotConf = CurrentCard.DO_Status.CurrentConfig;

    /*
     * Error Check
     */
    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber, 0, NULL,
        0L, FALSE))!=NoError)
    {
        return err;
    }

    if(CurrentCard.wCardType!=PCI_7350)
        return ErrorFuncNotSupport;

    if(!iotConf.mData.m7350DIO.ExtClk)
        return ErrorFuncNotSupport;

    if((wCLK_Src!=P7350_ECLK_IN)&&(wCLK_Src!=P7350_AFI_6))
        return ErrorInvalidDioPort;

    if(wCLK_DDAMode&0x1){
        if(wCLK_DDAVlaue>P7350_DDA_1R04NS)
            return ErrorUndefinedParameter;
    }
    if(wCLK_DPAMode&0x1){
        if(wCLK_DPAVlaue>P7350_DPA_337R5DG)
            return ErrorUndefinedParameter;
    }

    /*
     * Setup Structure to be passed to Kernel
     */
    //iotConf.wErrorCode = ErrorConfigIoctl;
    //Let Kernel Know that DO_7350_ExtSampCLKConfig will be performed
    iotConf.mData.m7350DIO.ConfigMode = 2;
    iotConf.mData.m7350DIO.ExtClkSrc = (wCLK_Src==P7350_ECLK_IN)? 0:1;
    iotConf.mData.m7350DIO.DDAMode = wCLK_DDAMode;
    iotConf.mData.m7350DIO.DPAMode = wCLK_DPAMode;
    iotConf.mData.m7350DIO.DDAValue = wCLK_DDAVlaue;
    iotConf.mData.m7350DIO.DPAValue = wCLK_DPAVlaue;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }
    /*
     * Re-store Configurations to CurrentConfig
     */
    CurrentCard.DO_Status.CurrentConfig = iotConf;
    //CurrentCard.DO_Status.CurrentConfig.wErrorCode = NoError;

    return NoError;
}


I16 DO_7350_ExportSampCLKConfig(
    U16 wCardNumber,
    U16 wCLK_Src,
    U16 wCLK_DPAMode,
    U16 wCLK_DPAVlaue
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;

    iotConf = CurrentCard.DO_Status.CurrentConfig;

    /*
     * Error Check
     */
    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber, 0, NULL,
        0L, FALSE))!=NoError)
    {
        return err;
    }

    if(CurrentCard.wCardType!=PCI_7350)
        return ErrorFuncNotSupport;

    //04/23/09
    if(!iotConf.mData.m7350DIO.en_expClk)
        return ErrorFuncNotSupport;

    if((wCLK_Src!=P7350_ECLK_OUT)&&
        (wCLK_Src!=P7350_AFI_6)&&
        (wCLK_Src!=(P7350_ECLK_OUT|P7350_AFI_6)))
    {
        return ErrorInvalidDioPort;
    }

    if(wCLK_DPAMode&0x1){
        if(wCLK_DPAVlaue>P7350_DPA_337R5DG)
            return ErrorUndefinedParameter;
    }

    /*
     * Set Structure to be Passed to Kernel
     */
    //iotConf.wErrorCode = ErrorConfigIoctl;
    //Let Kernel Know that DO_7350_ExportSampCLKConfig will be performed
    iotConf.mData.m7350DIO.ConfigMode = 3;
    if(wCLK_Src==P7350_ECLK_OUT){
        iotConf.mData.m7350DIO.expClkSrc = 1;
    }
    else if(wCLK_Src==P7350_AFI_6){
        iotConf.mData.m7350DIO.expClkSrc = 2;
    }
    else{
        //Both
        iotConf.mData.m7350DIO.expClkSrc = 3;
    }
    iotConf.mData.m7350DIO.eDPAMode = wCLK_DPAMode;
    iotConf.mData.m7350DIO.eDPAValue = wCLK_DPAVlaue;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_CONFIG, &(iotConf))
        !=DAS_ERR_NO)
    {
        return iotConf.wErrorCode;
    }
    /*
     * Re-store Configurations to CurrentConfig
     */
    CurrentCard.DO_Status.CurrentConfig = iotConf;
    //CurrentCard.DO_Status.CurrentConfig.wErrorCode = NoError;

    return NoError;
}


I16 DO_ContWritePortEx(
    U16  wCardNumber,
    U16  wPort,
    void *pwBuffer,
    U32  dwWriteCount,
    U16  wIterations,
    F64  fSampleRate,
    U16  wSyncMode
    )
{
    I16 err;
    DAS_IOT_CONFIG iotConf;
    DAS_IOT_CONT iotCont;
    DAS_IOT_MBUF iotMbuf;
    U32 byte_width;
    U8 needlock = 0;

    memset(&iotConf, '\0', sizeof(DAS_IOT_CONFIG));
    memset(&iotCont, '\0', sizeof(DAS_IOT_CONT));
    memset(&iotMbuf, '\0', sizeof(DAS_IOT_MBUF));

    iotConf = CurrentCard.DO_Status.CurrentConfig;

    if(CurrentCard.wCardType==PCI_7350){
        byte_width = iotConf.mData.m7350DIO.PortCount+1;
        if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber, 1,
            &wPort, dwWriteCount*byte_width, TRUE))!=NoError)
        {
            return err;
        }

        err = mlock((const void *)pwBuffer, dwWriteCount*byte_width);
        if(err!=0){ //errno defined in errno.h
            needlock = 1;
            goto lockio;
            if(errno==ENOMEM)
                printf("memory lock failed : errno = ENOMEM , please check the memory size\n");
            else if(errno==EPERM)
                printf("memory lock failed : error =  EPERM, mlock system call is only for ROOT\n");
            else if(errno==EINVAL)
                printf("memory lock failed : error = EINVAL, Invalate value\n");
            else
                printf("memory lock failed : errno=%d\n", errno);
            return ErrorLockMemory;
        }
lockio:
        //record the user-level buffer for unlock in DO_AsyncClear
        CurrentCard.DO_Status.p7300buf = pwBuffer;

        if(CurrentCard.DO_Status.bDblBufferMode)
            dwWriteCount /= 2;

        //7350 FIFO Size = 9k samples
        if((dwWriteCount*(iotConf.mData.m7350DIO.PortCount+1))>9*1024*4)
            return ErrorTransferCountTooLarge;

        //06/10/09
        if((iotConf.mData.m7350DIO.OPMode!=1)&&(!iotConf.mData.m7350DIO.ExtClk))
        {
            if(fSampleRate<1000000.0)
                return ErrorInvalidSampleRate;
            //Set PacerValue and TimerCascade
            if(!get_pacervalue(wCardNumber, fSampleRate, &iotCont.dwPacerValue))
                return ErrorInvalidSampleRate;
        }
        else if(iotConf.mData.m7350DIO.OPMode==1){
            iotCont.dwPacerValue = 2;
        }

        CurrentCard.DO_Status.dwTransferCount = dwWriteCount;

        //Setup buffer to build DMA Descriptor
        iotMbuf.dwBufferAddr = (U32)pwBuffer;
        iotMbuf.dwAccessCount = dwWriteCount;
        iotMbuf.wBufferId = 1|0x8000; //0: DI; 1:DO; //0x8000 single buffer
        //iotMbuf.wErrorCode = ErrorConfigIoctl;
        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_MBF_SET, &(iotMbuf))
            !=DAS_ERR_NO)
        {
            munlock((const void *)pwBuffer, dwWriteCount*byte_width);
            //yuan add 04/16/01
            CurrentCard.DO_Status.p7300buf = NULL;
            return iotMbuf.wErrorCode;
        }

        //iotCont.wErrorCode = ErrorContOPIoctl;
        iotCont.dwAccessCount = dwWriteCount;
        iotCont.wAutoScan = 1; //Let Kernel Know DO_ContWritePortEx is performed
        iotCont.wIterations = wIterations;
        iotCont.wSyncMode = wSyncMode;
        iotCont.wChannelCount = 1;
        iotCont.dwChannelGain[0].wChannel = wPort;
        iotCont.dwChannelGain[0].wGain = needlock;//0;
        if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_CON_WRITE, &(iotCont))
            !=DAS_ERR_NO)
        {
            munlock((const void *)pwBuffer, dwWriteCount*byte_width);
            CurrentCard.DO_Status.p7300buf = NULL;
            return iotCont.wErrorCode;
        }

        //00/06/02 jeffrey, for SyncMode , Unlock the user-level memory
        if(wSyncMode==SYNCH_OP){
            munlock((const void *)pwBuffer, dwWriteCount*byte_width);
            CurrentCard.DO_Status.p7300buf = NULL;
        }
    }

    return NoError;
}


I16 DO_7350_SoftTriggerGen(
    U16 wCardNumber
    )
{
    I16 err;
    U32 LcrAddr;
    U32 RegData;
    DAS_IOT_CONFIG iotConf;

    iotConf = CurrentCard.DO_Status.CurrentConfig;

    /*
     * Error Check
     */
    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber, 0, NULL,
        0L, FALSE))!=NoError)
    {
        return err;
    }
    if(CurrentCard.wCardType!=PCI_7350)
        return ErrorFuncNotSupport;
    if(!iotConf.mData.m7350DIO.WaitSTrig)
        return ErrorInvalidTriggerType;

    LcrAddr = (U32)(CurrentCard_AOFunc.pView[0]);
    /*
     * Set DO Soft Trigger
     * Should mask DO_WAIT_TRIG (bit5) while setting software trigger (bit7)
     */
    //05/06/09 remove mask
    RegData = *((volatile U32 *)(LcrAddr+0x108));
    //RegData &= 0xffffffdfL;
    RegData |= 0x80L;
    *((U32 *)(LcrAddr+0x108)) = RegData;
    //RegData &= ~(0x1L<<7);
    //*((U32 *)(LcrAddr+0x108)) = RegData;

    return NoError;
}


/*
 * 08/26/09
 * Add to avoid burst hand shake issue
 */
I16 DO_7350_BurstHandShakeDelay(
    U16 wCardNumber,
    U8  bDelay
    )
{
    I16 err;
    U32 LcrAddr;
    U32 RegData;
    DAS_IOT_CONFIG iotConf;

    iotConf = CurrentCard.DO_Status.CurrentConfig;

    /*
     * Error Check
     */
    if((err = fundamental_check(&CurrentCard_DOFunc, wCardNumber, 0, NULL,
        0L, FALSE))!=NoError)
    {
        return err;
    }
    if(CurrentCard.wCardType!=PCI_7350)
        return ErrorFuncNotSupport;
    if(iotConf.mData.m7350DIO.OPMode!=P7350_BurstHandShake)
        return ErrorInvalisOperationMode;

    LcrAddr = (U32)(CurrentCard_AOFunc.pView[0]);
    /*
     * Set DO burst hand shake delay count
     * offset: 0x104 - bit 24 ~ 31
     */
    RegData = *((volatile U32 *)(LcrAddr+0x104));
    RegData &= 0x00ffffffL;
    RegData |= ((bDelay<<24)&0xff000000);
    *((U32 *)(LcrAddr+0x104)) = RegData;

    return NoError;
}
