#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <sys/mman.h> //for mmap(...)
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/user.h>
#include "dask.h"
#include "Adlsys.h"
#include "dask_p.h"


//Driver Name
char *DriverName[] = {
    "",
    "PCI6208",
    "PCI6208",
    "PCI6308",
    "PCI6308",
    "PCI7200",
    "PCI7230",
    "PCI7233",
    "PCI7234",
    "PCI7248",
    "PCI7249",
    "PCI7250",
    "PCI7252",
    "PCI7296",
    "PCI7300A",
    "PCI7300B",
    "PCI7432",
    "PCI7433",
    "PCI7434",
    "PCI8554",
    "PCI9111",
    "PCI9111",
    "PCI9112",
    "PCI9113",
    "PCI9114",
    "PCI9114",
    "PCI9118",
    "PCI9118",
    "PCI9118",
    "PCI9812",
    "PCI9812",
    "PCI7396",
    "PCI9116",
    "PCI7256",
    "PCI7258",
    "PCI7260",
    "PCI7452",
    "PCI7442",
    "PCI7443",
    "PCI7444",
    "PCI9221",
    "PCI9524",
    "PCI6202",
    "PCI9222",
    "PCI9223",
    "PCI7433C",
    "PCI7434C",
    "PCI922A",
    "PCI7350"
};

REGISTER_CARD_TYPE RegisterCards[MAX_CARD];


I16 Register_Card(
    U16 wCardType,
    U16 card_num
    )
{
    U16 i;
    int hDev;
    I16 cardNumber=-1;
    char driverStr[20];
    DAS_IOT_CARD_DESP iotCard;
    DAS_IOT_FUNC_BUFSIZE iotBufSize;
    size_t page_size;
    DAS_IOT_RESRC iotRes;
    void* mapping_addr;
    memset(&iotCard, '\0', sizeof(DAS_IOT_CARD_DESP));
    memset(&iotRes, '\0', sizeof(DAS_IOT_RESRC));
    memset(&iotBufSize, 0, sizeof(DAS_IOT_FUNC_BUFSIZE));

    if(wCardType>NUM_CARD_TYPE)
        return ErrorUnknownCardType;

    //Find an unused RegisterCards entry
    for(i=0; i<MAX_CARD; i++){
        if(!RegisterCards[i].bBeingUsed){
            cardNumber = i;
            break;
        }
    }
    if(cardNumber==-1)
        return ErrorTooManyCardRegistered;

    //Get the right driver name
    sprintf(driverStr,"/dev/%sW%1d",DriverName[wCardType], card_num);
    //Open the device driver
    hDev = open(driverStr, O_RDWR);
    if(hDev<0){
        printf("hDev: %d\n", hDev);
        return ErrorOpenDriverFailed;
    }

    //mmap
    if((wCardType==PCI_9221)||(wCardType==PCI_6202)||(wCardType==PCI_9222)||
        (wCardType==PCI_9223)||(wCardType==PCI_922A)||(wCardType==PCI_7350))
    {
        DAS_IOT_CALCONST_9221 *pCal;
        page_size = getpagesize();
        mapping_addr = mmap(0, page_size, PROT_WRITE|PROT_READ,
            MAP_FILE|MAP_SHARED, hDev, 0
        );
        if(mapping_addr==MAP_FAILED){
            if(hDev>=0)
                close(hDev);
            //01/11/08 modified
            //RegisterCards[cardNumber].CAL_Const = NULL;
            RegisterCards[cardNumber].pCAL_Const = NULL;
            return -212;
        }
        //01/11/08 modified
        //RegisterCards[cardNumber].CAL_Const = (DAS_IOT_CALCONST*)mapping_addr;
        RegisterCards[cardNumber].pCAL_Const = (void *)mapping_addr;
        pCal = (DAS_IOT_CALCONST_9221 *)RegisterCards[cardNumber].pCAL_Const;
        //RegisterCards[cardNumber].CAL_Const->IOP_mem = 1; //for map IO_MEM
        pCal->IOP_mem = 1; //for map IO_MEM
        mapping_addr = mmap(0, page_size, PROT_WRITE|PROT_READ,
            MAP_FILE|MAP_SHARED, hDev, 0
        );
        if(mapping_addr==MAP_FAILED){
            //01/11/08 modified
            //munmap(RegisterCards[cardNumber].CAL_Const, page_size);
            munmap(RegisterCards[cardNumber].pCAL_Const, page_size);
            //RegisterCards[cardNumber].CAL_Const->IOP_mem = 0L;
            pCal->IOP_mem = 0L;
            //RegisterCards[cardNumber].CAL_Const = NULL;
            RegisterCards[cardNumber].pCAL_Const = NULL;
            if(hDev>=0)
                close(hDev);
            return -212;
        }
        //01/11/08 modified
        //RegisterCards[cardNumber].CAL_Const->IOP_mem = (U32)mapping_addr;
        pCal->IOP_mem = (U32)mapping_addr;
    }
    else if(wCardType==PCI_9524){
        //07/23/09
        //DAS_IOT_CALCONST_9524 *pCal;
        size_t total_pgsize;
        page_size = getpagesize();
        total_pgsize = \
            (sizeof(DAS_IOT_CALCONST_9524)+page_size-1)/page_size*page_size;
        mapping_addr = mmap(0, total_pgsize, PROT_WRITE|PROT_READ,
            MAP_FILE|MAP_SHARED, hDev, 0
        );
        if(mapping_addr==MAP_FAILED){
            if(hDev>=0)
                close(hDev);
            RegisterCards[cardNumber].pCAL_Const = NULL;
            return -212;
        }
        RegisterCards[cardNumber].pCAL_Const = (void *)mapping_addr;
        /*
        pCal = (DAS_IOT_CALCONST_9524 *)RegisterCards[cardNumber].pCAL_Const;
        pCal->IOP_mem = 1; //for map IO_MEM
        mapping_addr = mmap(0, page_size, PROT_WRITE|PROT_READ,
            MAP_FILE|MAP_SHARED, hDev, 0
        );
        if(mapping_addr==MAP_FAILED){
            munmap(RegisterCards[cardNumber].pCAL_Const, page_size);
            pCal->IOP_mem = 0L;
            RegisterCards[cardNumber].pCAL_Const = NULL;
            if(hDev>=0)
                close(hDev);
            return -212;
        }
        pCal->IOP_mem = (U32)mapping_addr;
        */
    }
    else{
        //map address for continuous operation
        //jeffery add for BufSize getting
        ioctl(hDev, DAS_IOC_GET_BUFSIZE, &(iotBufSize));
        //yuan modify 06/18/04
        if(wCardType!=PCI_7300A_RevA/*&&wCardType!=PCI_7300A_RevB*/){
            page_size = getpagesize();
            for(i=0; i<DAS_FUNC_COUNT; i++){
                if(iotBufSize.BufSize[i]>0){
                    //yuan add 06/18/04
                    if((wCardType==PCI_7300A_RevB)&&(i!=2))
                        continue;
                    //jeffrey added, for SuSE 7.0 compatible
                    mmap(0, iotBufSize.BufSize[i]+DbxORunStructSize,
                        PROT_WRITE, MAP_FILE|MAP_SHARED, hDev, (i*page_size)
                    );
                    //mmap(0, 0, PROT_WRITE, MAP_FILE | MAP_SHARED, hDev, i);
                }
            }
        }
    }

    //Get the Descriptor...
    if(ioctl(hDev, DAS_IOC_GET_DESP, &iotCard)<0)
        return iotCard.wErrorCode;
    //Copy acquired DAS_IOT_CARD_DESP to this free entry
    memcpy(&(RegisterCards[cardNumber].CardInfo), &iotCard,
        sizeof(DAS_IOT_CARD_DESP)
    );

    //06/09/00 jeffrey, set the file owner of the devices , for linux only
    if(iotCard.fDualInterruptEventNotify){
        //Dual INT SIGNAL is need , set the file owner
        fcntl(hDev, F_SETOWN, getpid());
    }

    /*
     * The resolution of PCI-9113 AI is 12,
     * but for cont AI, every data contains 4 bytes
     * The resolution of PCI-9114 AI is 16,
     * but for cont AI, every data contains 4 bytes
     */
    if((wCardType==PCI_9113)||(wCardType==PCI_9114DG)||(wCardType==PCI_9114HG))
        RegisterCards[cardNumber].CardInfo.mFunc[DAS_FUNC_AI].wResoluation = 32;

    //*wBus = RegisterCards[cardNumber].CardInfo.wBus;
    //*wSlot = RegisterCards[cardNumber].CardInfo.wSlot;
    RegisterCards[cardNumber].bBeingUsed = TRUE;
    RegisterCards[cardNumber].wCardType = wCardType;
    RegisterCards[cardNumber].wDevIndex = card_num;
    RegisterCards[cardNumber].hDevDriver = hDev;
    RegisterCards[cardNumber].AI_Status.bDblBufferMode = FALSE;
    RegisterCards[cardNumber].DI_Status.bDblBufferMode = FALSE;
    RegisterCards[cardNumber].AO_Status.bDblBufferMode = FALSE;
    RegisterCards[cardNumber].DO_Status.bDblBufferMode = FALSE;
    RegisterCards[cardNumber].CTR_Status.CurrentConfig.mData.wValue[0] = 0;
    RegisterCards[cardNumber].CTR_Status.CurrentConfig.mData.wValue[1] = 0;
    //yuan 09/13/00 for pci9812/9810 pacer value
    if((wCardType==PCI_9810)||(wCardType==PCI_9812)){
        RegisterCards[cardNumber].AI_Status.dwPacerValue = 2;
    }

    //yuan 04/11/01 for pci7300 multi-buffer address
    if((wCardType==PCI_7300A_RevA)||(wCardType==PCI_7300A_RevB)){
        //yuan add 04/16/01
        RegisterCards[cardNumber].DI_Status.p7300buf = NULL;
        RegisterCards[cardNumber].DO_Status.p7300buf = NULL;
        //yuan add 04/11/01 for multibuffer unlock
        for(i=0; i<8; i++){
            RegisterCards[cardNumber].DI_Status.p7300mbuf[i].dwBufferAddr = \
                (U32)NULL;
            RegisterCards[cardNumber].DI_Status.p7300mbuf[i].dwAccessCount = 0;
            RegisterCards[cardNumber].DI_Status.p7300mbuf[i].wBufferId = -1;
            RegisterCards[cardNumber].DO_Status.p7300mbuf[i].dwBufferAddr = \
                (U32)NULL;
            RegisterCards[cardNumber].DO_Status.p7300mbuf[i].dwAccessCount = 0;
            RegisterCards[cardNumber].DO_Status.p7300mbuf[i].wBufferId = -1;
        }
        RegisterCards[cardNumber].DI_Status.p7300mbufCnt = 0;
        RegisterCards[cardNumber].DO_Status.p7300mbufCnt = 0;
    }
    if(wCardType==PCI_7350){
        RegisterCards[cardNumber].DI_Status.p7300buf = NULL;
        RegisterCards[cardNumber].DO_Status.p7300buf = NULL;
        for(i=0; i<64; i++){
            RegisterCards[cardNumber].DI_Status.p7350mbuf[i].dwBufferAddr = \
                (U32)NULL;
            RegisterCards[cardNumber].DI_Status.p7350mbuf[i].dwAccessCount = 0;
            RegisterCards[cardNumber].DI_Status.p7350mbuf[i].wBufferId = -1;
            RegisterCards[cardNumber].DO_Status.p7350mbuf[i].dwBufferAddr = \
                (U32)NULL;
            RegisterCards[cardNumber].DO_Status.p7350mbuf[i].dwAccessCount = 0;
            RegisterCards[cardNumber].DO_Status.p7350mbuf[i].wBufferId = -1;
        }
        RegisterCards[cardNumber].DI_Status.p7300mbufCnt = 0;
        RegisterCards[cardNumber].DO_Status.p7300mbufCnt = 0;
    }
    if((wCardType==PCI_9221)||(wCardType==PCI_922A)||(wCardType==PCI_9524)){
        RegisterCards[cardNumber].AI_Status.pDAQbuf[0] = NULL;
        RegisterCards[cardNumber].AI_Status.pDAQbuf[1] = NULL;
    }
    if(wCardType==PCI_6202){
        RegisterCards[cardNumber].AO_Status.pDAQbuf[0] = NULL;
        RegisterCards[cardNumber].AO_Status.pDAQbuf[1] = NULL;
    }
    if((wCardType==PCI_9222)||(wCardType==PCI_9223)){
        RegisterCards[cardNumber].AI_Status.pDAQbuf[0] = NULL;
        RegisterCards[cardNumber].AI_Status.pDAQbuf[1] = NULL;
        RegisterCards[cardNumber].AO_Status.pDAQbuf[0] = NULL;
        RegisterCards[cardNumber].AO_Status.pDAQbuf[1] = NULL;
        RegisterCards[cardNumber].DI_Status.pDAQbuf[0] = NULL;
        RegisterCards[cardNumber].DI_Status.pDAQbuf[1] = NULL;
        RegisterCards[cardNumber].DO_Status.pDAQbuf[0] = NULL;
        RegisterCards[cardNumber].DO_Status.pDAQbuf[1] = NULL;
    }

    RegisterCards[cardNumber].DIOINT_Status.bInt1Enable = 0;
    RegisterCards[cardNumber].DIOINT_Status.bInt2Enable = 0;
    //09/14/00 jeffrey , modify with the source with YUAN
    if(wCardType==PCI_9112){ //default internal ref -5V
        RegisterCards[cardNumber].RefVoltage[0] = \
        RegisterCards[cardNumber].RefVoltage[1] = -5;
    }
    else{ //default internal ref 10V for PCI6308V
        RegisterCards[cardNumber].RefVoltage[0] = \
        RegisterCards[cardNumber].RefVoltage[1] = 10;
    }

    //06/09/00 jeffrey,
    /*
     * initial the signal handler,
     * and ignore the SIGEVENT1 & SIGEVENT2 for linux only
     */
    RegisterCards[cardNumber].sig_act1.sa_handler = SIG_IGN;
    RegisterCards[cardNumber].sig_act2.sa_handler = SIG_IGN;
    if(sigaction(SIGEVENT1, &(RegisterCards[cardNumber].sig_act1), NULL)==-1)
        return ErrorOpenEventFailed;
    if(sigaction(SIGEVENT2, &(RegisterCards[cardNumber].sig_act2), NULL)==-1)
        return ErrorOpenEventFailed;

    //yuan add 06/29/04
    //Initial the signal handler, for linux only
    RegisterCards[cardNumber].i_sig_act.sa_handler = SIG_IGN;
    RegisterCards[cardNumber].o_sig_act.sa_handler = SIG_IGN;
    RegisterCards[cardNumber].i_dbsig_act.sa_handler = SIG_IGN;
    RegisterCards[cardNumber].o_dbsig_act.sa_handler = SIG_IGN;

    if(sigaction(SIGEVENT3, &(RegisterCards[cardNumber].i_sig_act), NULL)==-1)
        return ErrorOpenEventFailed;
    if(sigaction(SIGEVENT4, &(RegisterCards[cardNumber].o_sig_act), NULL)==-1)
        return ErrorOpenEventFailed;

    //yuan add 07/25/01
    ioctl(hDev, DAS_IOC_GET_RESRC, &iotRes);
    for(i=0; i<4; i++)
        RegisterCards[cardNumber].PortAddr[i] = iotRes.PortAddr[i];

    /*
     * reset IOP_mem for another CAL mapped
     */
    if((wCardType==PCI_9221)||(wCardType==PCI_6202)||(wCardType==PCI_9222)||
        (wCardType==PCI_9223)||(wCardType==PCI_922A))
    {
        //01/11/08 modified
        DAS_IOT_CALCONST_9221 *pCal = \
            (DAS_IOT_CALCONST_9221 *)RegisterCards[cardNumber].pCAL_Const;
        RegisterCards[cardNumber].CardInfo.mFunc[DAS_FUNC_AI].pView[1] = \
            //(void *)(RegisterCards[cardNumber].CAL_Const->IOP_mem);
            (void *)(pCal->IOP_mem);
        //RegisterCards[cardNumber].CAL_Const->IOP_mem = 0; //for another CAL mapped
        pCal->IOP_mem = 0;
        RegisterCards[cardNumber].CardInfo.mFunc[DAS_FUNC_AI].pView[1] += \
            iotRes.PortAddr[0]&~PAGE_MASK;
    }
    //12/07/09
    else if(wCardType==PCI_7350){
        U32 *IOP_mem = (U32 *)RegisterCards[cardNumber].pCAL_Const;
        RegisterCards[cardNumber].CardInfo.mFunc[DAS_FUNC_AO].pView[0] = \
            (void *)(*IOP_mem);
        *IOP_mem = 0;
        RegisterCards[cardNumber].CardInfo.mFunc[DAS_FUNC_AO].pView[0] += \
            iotRes.PortAddr[0]&~PAGE_MASK;
    }

    return cardNumber;
}


I16 Release_Card(
    U16 wCardNumber
    )
{
    I16 vi, err;
    U32 dwAccessCnt;
    size_t page_size;

    if((CurrentCard.wCardType==PCI_9221)||(CurrentCard.wCardType==PCI_6202)||
        (CurrentCard.wCardType==PCI_9222)||(CurrentCard.wCardType==PCI_9223)||
        (CurrentCard.wCardType==PCI_922A)||(CurrentCard.wCardType==PCI_7350))
    {
        //12/07/09
        //DAS_IOT_CALCONST_9221 *pCal = (DAS_IOT_CALCONST_9221 *)CurrentCard.pCAL_Const;
        U32 *IOP_mem;
        page_size = getpagesize();
        //12/07/09
        /*
        //if(CurrentCard.CAL_Const->IOP_mem != 0L){
        if(pCal->IOP_mem!=0L){
            //munmap((void*)((CurrentCard.CAL_Const->IOP_mem)&PAGE_MASK), page_size);
            munmap((void*)((pCal->IOP_mem)&PAGE_MASK), page_size);
            //CurrentCard.CAL_Const->IOP_mem=0;
            pCal->IOP_mem=0;
        }

        //if(CurrentCard.CAL_Const != NULL){
        if(CurrentCard.pCAL_Const!=NULL){
            //munmap((void*)(CurrentCard.CAL_Const), page_size);
            munmap((void*)pCal, page_size);
            //CurrentCard.CAL_Const=NULL;
            CurrentCard.pCAL_Const = NULL;
        }
        */
        if(CurrentCard.wCardType==PCI_7350)
            IOP_mem = CurrentCard_AOFunc.pView[0];
        else
            IOP_mem = CurrentCard_AIFunc.pView[1];
        if((IOP_mem)!=0L){
            //printf("[pdask] Release_Card: unmap IO-MEM, Addr: 0x%x\n", (unsigned int)IOP_mem);
            munmap((void *)(((U32)IOP_mem)&PAGE_MASK), page_size);
        }

        if(CurrentCard.pCAL_Const!=NULL){
            //printf("[pdask] Release_Card: unmap CAL-MEM, Addr: 0x%x\n", (unsigned int)CurrentCard.pCAL_Const);
            munmap(CurrentCard.pCAL_Const, page_size);
            CurrentCard.pCAL_Const = NULL;
        }
    }
    else if(CurrentCard.wCardType==PCI_9524){
        DAS_IOT_CALCONST_9524 *pCal = \
            (DAS_IOT_CALCONST_9524 *)CurrentCard.pCAL_Const;
        page_size = getpagesize();
        if(CurrentCard.pCAL_Const!=NULL){
            size_t total_pgsize = \
                (sizeof(DAS_IOT_CALCONST_9524)+page_size-1)/page_size*page_size;
            munmap((void *)pCal, total_pgsize);
            CurrentCard.pCAL_Const = NULL;
        }
    }
    //01/10/08 added
    else if((CurrentCard.wCardType==PCI_7200)||
        (CurrentCard.wCardType==PCI_7300A_RevB))
    {   //yuan modify 12/05/07
        munmap(CurrentCard_AOFunc.pView[0],
            CurrentCardInfo.mFunc[2].dwBufferSize+DbxORunStructSize
        );
    }
    else{
        for(vi=0; vi<DAS_FUNC_COUNT; vi++){
            if((CurrentCardInfo.mFunc[vi].wFlag.bFunction)&&
                (CurrentCardInfo.mFunc[vi].pView[0]!=0))
            {
                if(vi==1)
                    continue; //Only for PCIS-DASK, skip the AO_Function pView
                //Used for DBXORUN
                munmap(CurrentCardInfo.mFunc[vi].pView[0]-DbxORunStructSize,
                    CurrentCardInfo.mFunc[vi].dwBufferSize+DbxORunStructSize
                );
            }
        }
    }

    if((CurrentCard.wCardType==PCI_7300A_RevA)||
        (CurrentCard.wCardType==PCI_7300A_RevB))
    {
        /*
        if(CurrentCard.wCardType==PCI_7300A_RevA)
            byte_width = pwidth[\
            CurrentCard.DI_Status.CurrentConfig.mData.m7300DIO.fPortWidth]/8;
        else
            byte_width = pwidth[\
            CurrentCard.DI_Status.CurrentConfig.mData.m7300bDIO.fPortWidth]/8;

        if(CurrentCard.DI_Status.p7300buf){
            err = munlock((const void*)CurrentCard.DI_Status.p7300buf, \
            CurrentCard.DI_Status.dwTransferCount*byte_width);
            //printf("%d\n", err);
            CurrentCard.DI_Status.p7300buf = NULL;

            //yuan add 04/11/01 for multibuffer unlock
            for(vi=0; vi<CurrentCard.DI_Status.p7300mbufCnt; vi++){
                if(CurrentCard.DI_Status.p7300mbuf[vi].dwBufferAddr)
                    err = munlock((const void*)\
                    CurrentCard.DI_Status.p7300mbuf[vi].dwBufferAddr, \
                    CurrentCard.DI_Status.p7300mbuf[vi].dwAccessCount* \
                    byte_width);
                CurrentCard.DI_Status.p7300mbuf[vi].dwBufferAddr = (U32)NULL;
                CurrentCard.DI_Status.p7300mbuf[vi].dwAccessCount = 0;
                CurrentCard.DI_Status.p7300mbuf[vi].wBufferId = -1;
            }
            CurrentCard.DI_Status.p7300mbufCnt = 0;
        */
        if((CurrentCard.DI_Status.p7300buf)||
            (CurrentCard.DI_Status.p7300mbufCnt))
        {
            err = DI_AsyncClear(wCardNumber, &dwAccessCnt);
            //printf("err: %d, %ld\n", err, dwAccessCnt);
        }
    }

    CurrentCard.bBeingUsed = FALSE;
    close(CurrentCard.hDevDriver);

    return NoError;
}


I16 AI_GetView(
    U16 wCardNumber,
    U32 *pView
    )
{
    *pView = (U32)(CurrentCard_AIFunc.pView[0]);

    return NoError;
}


I16 DI_GetView(
    U16 wCardNumber,
    U32 *pView
    )
{
    *pView = (U32)(CurrentCard_DIFunc.pView[0]);

    return NoError;
}


I16 DO_GetView(
    U16 wCardNumber,
    U32 *pView
    )
{
    *pView = (U32)(CurrentCard_DOFunc.pView[0]);

    return NoError;
}


I16 GetBaseAddr(
    U16 wCardNumber,
    U32 *BaseAddr,
    U32 *BaseAddr2
    )
{
    if((CurrentCard.wCardType==PCI_9221)||(CurrentCard.wCardType==PCI_6202)||
        (CurrentCard.wCardType==PCI_9222)||(CurrentCard.wCardType==PCI_9223)||
        (CurrentCard.wCardType==PCI_922A))
    {
        *BaseAddr = (U32)(CurrentCard.PortAddr[1]); //IO ADDR
        *BaseAddr2 = (U32)(CurrentCard_AIFunc.pView[0]); //MEM PHY ADDR
    }
    else if(CurrentCard.wCardType==PCI_7350){
        *BaseAddr = (U32)(CurrentCard_AIFunc.pView[0]);
        *BaseAddr2 = (U32)(CurrentCard_AIFunc.pView[1]);
    }
    else{
        *BaseAddr = (U32)(CurrentCard.PortAddr[1]);
        *BaseAddr2 = (U32)(CurrentCard.PortAddr[2]);
    }

    return NoError;
}


I16 GetLCRAddr(
    U16 wCardNumber,
    U32 *LcrAddr
    )
{
    if((CurrentCard.wCardType==PCI_9221)||(CurrentCard.wCardType==PCI_6202)||
        (CurrentCard.wCardType==PCI_9222)||(CurrentCard.wCardType==PCI_9223)||
        (CurrentCard.wCardType==PCI_922A))
    {
        *LcrAddr = (U32)(CurrentCard_AIFunc.pView[1]);
        //*LcrAddr = CurrentCard.CAL_Const->IOP_mem; //MEM MAPPED ADDR
    }
    else if(CurrentCard.wCardType==PCI_7350){
        *LcrAddr = (U32)(CurrentCard_AOFunc.pView[0]);
    }
    else
        *LcrAddr = (U32)(CurrentCard.PortAddr[0]);

    return NoError;
}


I16 Get_CardDesc(
    U16               wCardNumber,
    DAS_IOT_CARD_DESP *cardDesc
    )
{
    memcpy(cardDesc, &(CurrentCard.CardInfo), sizeof(DAS_IOT_CARD_DESP));

    return NoError;
}


I16 GetCardType(
    U16 wCardNumber,
    U16 *cardType
    )
{
    *cardType = CurrentCard.wCardType;

    return NoError;
}


I16 GetCardIndexFromID(
    U16 wCardNumber,
    U16 *cardType,
    U16 *cardIndex
    )
{
    *cardType = CurrentCard.wCardType;
    *cardIndex = CurrentCard.wDevIndex;

    return NoError;
}


I16 EMGShutDownControl(
    U16 wCardNumber,
    U8  enable
    )
{
    DAS_IOT_PIO iotPio;

    memset(&iotPio, '\0', sizeof(DAS_IOT_PIO));

    if(CurrentCard.wCardType!=PCI_7260)
        return ErrorFuncNotSupport;

    //yuan 07/31/00
    //iotPio.wErrorCode = ErrorPIOIoctl;
    iotPio.wByteCount = 0x200;
    iotPio.wChannel = enable;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_PIO_WRITE, &iotPio)<0)
        return iotPio.wErrorCode;

    return NoError;
}


I16 EMGShutDownStatus(
    U16 wCardNumber,
    U8  *sts
    )
{
    DAS_IOT_CTR iotCtr;

    memset(&iotCtr, '\0', sizeof(DAS_IOT_CTR));

    if(CurrentCard.wCardType!=PCI_7260)
        return ErrorFuncNotSupport;
    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    //iotCtr.wErrorCode = ErrorCTRIoctl;
    iotCtr.wMode = 1;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_CTR_READ, &iotCtr)<0)
        return iotCtr.wErrorCode;
    *sts = (U8)(iotCtr.dwCounterValue&0x01);

    return NoError;
}


I16 HotResetHoldControl(
    U16 wCardNumber,
    U8  enable
    )
{
    DAS_IOT_PIO iotPio;

    memset(&iotPio, '\0', sizeof(DAS_IOT_PIO));

    if((CurrentCard.wCardType!=PCI_7442)&&(CurrentCard.wCardType!=PCI_7444))
        return ErrorFuncNotSupport;

    iotPio.wByteCount = 0x200;
    iotPio.wChannel = enable;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_PIO_WRITE, &iotPio)<0)
        return iotPio.wErrorCode;

    return NoError;
}


I16 HotResetHoldStatus(
    U16 wCardNumber,
    U8  *sts
    )
{
    DAS_IOT_CTR iotCtr;

    memset(&iotCtr, '\0', sizeof(DAS_IOT_CTR));

    if((CurrentCard.wCardType!=PCI_7442)&&(CurrentCard.wCardType!=PCI_7444))
        return ErrorFuncNotSupport;

    iotCtr.wMode = 1;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_CTR_READ, &iotCtr)<0)
        return iotCtr.wErrorCode;
    *sts = (U8)(iotCtr.dwCounterValue&0x01);

    return NoError;
}


I16 GetInitPattern(
    U16 wCardNumber,
    U8  patID,
    U32 *pattern
    )
{
    I16 err = 0;
    DAS_IOT_PIO iotPio;
    memset(&iotPio, '\0', sizeof(DAS_IOT_PIO));

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    switch(CurrentCard.wCardType){
        case PCI_7260:
            iotPio.wChannel = patID+2;
            break;
        case PCI_7442:
            if(patID==0||patID==1) //0->4; 1->5
                patID += 4;
            else if(patID==4||patID==5) //4->6; 5->7
                patID += 2;
            iotPio.wChannel = patID;
            break;
        case PCI_7444:
            iotPio.wChannel = patID+6;
            break;
        default:
            return ErrorFuncNotSupport;
    }
    iotPio.wByteCount = (CurrentCard_DIFunc.wResoluation+7)/8;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DI_PIO_READ, &iotPio)<0)
        return iotPio.wErrorCode;
    *pattern = iotPio.dwValue;

    return NoError;
}


I16 SetInitPattern(
    U16 wCardNumber,
    U8  patID,
    U32 pattern
    )
{
    I16 err = 0;
    DAS_IOT_PIO iotPio;

    memset(&iotPio, '\0', sizeof(DAS_IOT_PIO));

    if((err = fundamental_check(&CurrentCard_DIFunc, wCardNumber, 0, NULL, 0L,
        FALSE))!=NoError)
    {
        return err;
    }

    switch(CurrentCard.wCardType){
        case PCI_7442:
            if(patID==0||patID==1) //0->4; 1->5
                patID += 4;
            else if(patID==4||patID==5) //4->6; 5->7
                patID += 2;
            iotPio.wChannel = patID;
            break;
        case PCI_7444:
            iotPio.wChannel = patID+6;
            break;
        default:
            return ErrorFuncNotSupport;
    }
    iotPio.wByteCount = (CurrentCard_DIFunc.wResoluation+7)/8;
    iotPio.dwValue = pattern;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_PIO_WRITE, &iotPio)<0)
        return iotPio.wErrorCode;

    return NoError;
}


I16 IdentifyLED_Control(
    U16 wCardNumber,
    U8  ctrl
    )
{
    DAS_IOT_PIO iotPio;

    memset(&iotPio, '\0', sizeof(DAS_IOT_PIO));

    if(CurrentCard.wCardType!=PCI_7260)
        return ErrorFuncNotSupport;

    //yuan 07/31/00
    //iotPio.wErrorCode = ErrorPIOIoctl;
    iotPio.wByteCount = 0x100;
    iotPio.wChannel = ctrl;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_DO_PIO_WRITE, &iotPio)<0)
        return iotPio.wErrorCode;

    return NoError;
}


I16 SSI_SourceConn(
    U16 wCardNumber,
    U16 sigCode
    )
{
    DAS_IOT_SSI iotSSI;

    memset(&iotSSI, '\0', sizeof(DAS_IOT_SSI));

    iotSSI.wSigCode = (U8)sigCode;
    iotSSI.wDir = 1;
    //in driver check if it's a slave
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_SSI_CTL, &iotSSI)<0)
        return iotSSI.wErrorCode;

    return NoError;
}


I16 SSI_SourceDisConn(
    U16 wCardNumber,
    U16 sigCode
    )
{
    DAS_IOT_SSI iotSSI;

    memset(&iotSSI, '\0', sizeof(DAS_IOT_SSI));

    iotSSI.wSigCode = (U8)sigCode;
    iotSSI.wDir = 0;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_SSI_CTL, &iotSSI)<0)
        return iotSSI.wErrorCode;

    return NoError;
}


I16 SSI_SourceClear(
    U16 wCardNumber
    )
{
    DAS_IOT_SSI iotSSI;

    memset(&iotSSI, '\0', sizeof(DAS_IOT_SSI));

    iotSSI.wDir = 2;
    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_SSI_CTL, &iotSSI)<0)
        return iotSSI.wErrorCode;

    return NoError;
}


I16 I2C_Setup(
    U16 wCardNumber,
    U16 wI2C_Port,
    U16 wI2C_Config,
    U32 dwI2C_SetupValue1,
    U32 dwI2C_SetupValue2
    )
{
    DAS_IOT_ISQUAREDC iotI2C;

    memset(&iotI2C, '\0', sizeof(DAS_IOT_ISQUAREDC));

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    switch(CurrentCard.wCardType){
        case PCI_7350:
            /*Only one I2C controller is supported for PCI-7350*/
            if(wI2C_Port>0)
                return ErrorUndefinedParameter;
            iotI2C.bI2C_Port = wI2C_Port;
            /*Pre-Clk should be set for I2C of PCI-7350*/
            iotI2C.dwI2C_REG1 = dwI2C_SetupValue1&0xff;
            /*Let kernel know I2C_Setup is performed*/
            iotI2C.bI2C_Mode = 0;
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_I2C_CONFIG, &iotI2C)<0)
                return iotI2C.wErrorCode;
            break;
        default:
            return ErrorFuncNotSupport;
    }

    return NoError;
}


I16 I2C_Control(
    U16 wCardNumber,
    U16 wI2C_Port,
    U16 wI2C_CtrlParam,
    U32 dwI2C_CtrlValue
    )
{
    DAS_IOT_ISQUAREDC iotI2C;

    memset(&iotI2C, '\0', sizeof(DAS_IOT_ISQUAREDC));

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    switch(CurrentCard.wCardType){
        case PCI_7350:
            /*Only one I2C controller is supported for PCI-7350*/
            if(wI2C_Port>0)
                return ErrorUndefinedParameter;
            /*only Enable and Stop operation items are valid*/
            if(wI2C_CtrlParam>I2C_STOP)
                return ErrorUndefinedParameter;
            iotI2C.bI2C_Port = wI2C_Port;
            /*Control item - 0:Enable; 1:Stop*/
            iotI2C.dwI2C_REG1 = wI2C_CtrlParam;
            /*
             * On or Off control item
             * Enable -
             *   0 : Disable I2C
             *   1 : Enable I2C
             * Stop -
             *   0 : Do nothing
             *   1 : Stop I2C
             */
            iotI2C.dwI2C_REG2 = dwI2C_CtrlValue&0x1;
            /*Let kernel know I2C_Control is performed*/
            iotI2C.bI2C_Mode = 1;
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_I2C_CONFIG, &iotI2C)<0)
                return iotI2C.wErrorCode;
            break;
        default:
            return ErrorFuncNotSupport;
    }

    return NoError;
}


I16 I2C_Status(
    U16 wCardNumber,
    U16 wI2C_Port,
    U32 *dwI2C_Status
    )
{
    DAS_IOT_ISQUAREDC iotI2C;

    memset(&iotI2C, '\0', sizeof(DAS_IOT_ISQUAREDC));

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    switch(CurrentCard.wCardType){
        case PCI_7350:
            /*Only one I2C controller is supported for PCI-7350*/
            if(wI2C_Port>0)
                return ErrorUndefinedParameter;
            iotI2C.bI2C_Port = wI2C_Port;
            /*Let kernel know I2C_Status is performed*/
            iotI2C.bI2C_Mode = 1;
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_I2C_READ, &iotI2C)<0)
                return iotI2C.wErrorCode;
            /*
             * bit 0: Interrupt flag (not export)
             *     1: I2C busy
             *     2: I2C SCL slow
             *     3: I2C SCL conflict flag
             *     4: I2C SDA conflict flag
             *     5: I2C RxACK
             */
            *dwI2C_Status = iotI2C.dwI2C_REG1;
            break;
        default:
            return ErrorFuncNotSupport;
    }

    return NoError;
}


I16 I2C_Read(
    U16 wCardNumber,
    U16 wI2C_Port,
    U16 wI2C_SlaveAddr,
    U16 wI2C_CmdAddrBytes,
    U16 wI2C_DataBytes,
    //VOID *pdwI2C_CmdAddr,
    //VOID *pdwI2C_Data
    U32 dwI2C_CmdAddr,
    U32 *pdwI2C_Data
    )
{
    DAS_IOT_ISQUAREDC iotI2C;

    memset(&iotI2C, '\0', sizeof(DAS_IOT_ISQUAREDC));

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    switch(CurrentCard.wCardType){
        case PCI_7350:
            /*Only one I2C controller is supported for PCI-7350*/
            if(wI2C_Port>0)
                return ErrorUndefinedParameter;
            if(wI2C_SlaveAddr>0xff)
                return ErrorUndefinedParameter;
            if((wI2C_DataBytes==0)||(wI2C_DataBytes==3)||(wI2C_DataBytes>4))
                return ErrorUndefinedParameter;
            if(pdwI2C_Data==NULL)
                return ErrorUndefinedParameter;
            switch(wI2C_CmdAddrBytes){
                case 0:
                    iotI2C.dwI2C_REG1 = 0;
                    break;
                case 1:
                    /*
                    if(pdwI2C_CmdAddr==NULL)
                        return ErrorUndefinedParameter;
                    iotI2C.dwI2C_REG1 = *((UCHAR *)pdwI2C_CmdAddr);
                    */
                    iotI2C.dwI2C_REG1 = dwI2C_CmdAddr&0xff;
                    break;
                case 2:
                    /*
                    if(pdwI2C_CmdAddr==NULL)
                        return ErrorUndefinedParameter;
                    iotI2C.dwI2C_REG1 = *((USHORT *)pdwI2C_CmdAddr);
                    */
                    iotI2C.dwI2C_REG1 = dwI2C_CmdAddr&0xffff;
                    break;
                case 4:
                    /*
                    if(pdwI2C_CmdAddr==NULL)
                        return ErrorUndefinedParameter;
                    iotI2C.dwI2C_REG1 = *((ULONG *)pdwI2C_CmdAddr);
                    */
                    iotI2C.dwI2C_REG1 = dwI2C_CmdAddr;
                    break;
                default:
                    return ErrorUndefinedParameter;
            }
            iotI2C.bI2C_Port = wI2C_Port;
            iotI2C.bI2C_CmdAddrBytes = wI2C_CmdAddrBytes;
            iotI2C.bI2C_DataBytes = wI2C_DataBytes;
            iotI2C.bI2C_SlaveAddr = wI2C_SlaveAddr;
            /*Let kernel know I2C_Read is performed*/
            iotI2C.bI2C_Mode = 0;
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_I2C_READ, &iotI2C)<0)
                return iotI2C.wErrorCode;
            if(wI2C_DataBytes==1)
                //*((UCHAR *)pdwI2C_Data) = (UCHAR)(iotI2C.dwI2C_REG2&0xff);
                *pdwI2C_Data = iotI2C.dwI2C_REG2&0xff;
            else if(wI2C_DataBytes==2)
                //*((USHORT *)pdwI2C_Data) = (USHORT)(iotI2C.dwI2C_REG2&0xffff);
                *pdwI2C_Data = iotI2C.dwI2C_REG2&0xffff;
            else
                //*((ULONG *)pdwI2C_Data) = iotI2C.dwI2C_REG2;
                *pdwI2C_Data = iotI2C.dwI2C_REG2;
            break;
        default:
            return ErrorFuncNotSupport;
    }

    return NoError;
}


I16 I2C_Write(
    U16 wCardNumber,
    U16 wI2C_Port,
    U16 wI2C_SlaveAddr,
    U16 wI2C_CmdAddrBytes,
    U16 wI2C_DataBytes,
    //VOID *pdwI2C_CmdAddr,
    //VOID *pdwI2C_Data
    U32 dwI2C_CmdAddr,
    U32 dwI2C_Data
    )
{
    DAS_IOT_ISQUAREDC iotI2C;

    memset(&iotI2C, '\0', sizeof(DAS_IOT_ISQUAREDC));

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    switch(CurrentCard.wCardType){
        case PCI_7350:
            /*Only one I2C controller is supported for PCI-7350*/
            if(wI2C_Port>0)
                return ErrorUndefinedParameter;
            if(wI2C_SlaveAddr>0xff)
                return ErrorUndefinedParameter;

            switch(wI2C_CmdAddrBytes){
                case 1:
                    /*
                    if(pdwI2C_CmdAddr==NULL)
                        return ErrorUndefinedParameter;
                    iotI2C.dwI2C_REG1 = *((UCHAR *)pdwI2C_CmdAddr);
                    */
                    iotI2C.dwI2C_REG1 = dwI2C_CmdAddr&0xff;
                    break;
                case 2:
                    /*
                    if(pdwI2C_CmdAddr==NULL)
                        return ErrorUndefinedParameter;
                    iotI2C.dwI2C_REG1 = *((USHORT *)pdwI2C_CmdAddr);
                    */
                    iotI2C.dwI2C_REG1 = dwI2C_CmdAddr&0xffff;
                    break;
                case 4:
                    /*
                    if(pdwI2C_CmdAddr==NULL)
                        return ErrorUndefinedParameter;
                    iotI2C.dwI2C_REG1 = *((ULONG *)pdwI2C_CmdAddr);
                    */
                    iotI2C.dwI2C_REG1 = dwI2C_CmdAddr&0xff;
                    break;
                default:
                    return ErrorUndefinedParameter;
            }

            switch(wI2C_DataBytes){
                case 1:
                    /*
                    if(pdwI2C_Data==NULL)
                        return ErrorUndefinedParameter;
                    iotI2C.dwI2C_REG2 = *((UCHAR *)pdwI2C_Data);
                    */
                    iotI2C.dwI2C_REG2 = dwI2C_Data&0xff;
                    break;
                case 2:
                    /*
                    if(pdwI2C_Data==NULL)
                        return ErrorUndefinedParameter;
                    iotI2C.dwI2C_REG2 = *((USHORT *)pdwI2C_Data);
                    */
                    iotI2C.dwI2C_REG2 = dwI2C_Data&0xffff;
                    break;
                case 4:
                    /*
                    if(pdwI2C_Data==NULL)
                        return ErrorUndefinedParameter;
                    iotI2C.dwI2C_REG2 = *((ULONG *)pdwI2C_Data);
                    */
                    iotI2C.dwI2C_REG2 = dwI2C_Data;
                    break;
                default:
                    return ErrorUndefinedParameter;
            }

            iotI2C.bI2C_Port = wI2C_Port;
            iotI2C.bI2C_CmdAddrBytes = wI2C_CmdAddrBytes;
            iotI2C.bI2C_DataBytes = wI2C_DataBytes;
            iotI2C.bI2C_SlaveAddr = wI2C_SlaveAddr;
            /*Let kernel know I2C_Write is performed*/
            iotI2C.bI2C_Mode = 0;
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_I2C_WRITE, &iotI2C)<0)
                return iotI2C.wErrorCode;
            break;
        default:
            return ErrorFuncNotSupport;
    }

    return NoError;
}


I16 SPI_Setup(
    U16 wCardNumber,
    U16 wSPI_Port,
    U16 wSPI_Config,
    U32 dwSPI_SetupValue1,
    U32 dwSPI_SetupValue2
    )
{
    DAS_IOT_SPI iotSPI;

    memset(&iotSPI, '\0', sizeof(DAS_IOT_SPI));

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    switch(CurrentCard.wCardType){
        case PCI_7350:
            /*support only one SPI controller in PCI-7350*/
            if(wSPI_Port>0)
                return ErrorUndefinedParameter;
            iotSPI.bSPI_Port = wSPI_Port;
            iotSPI.bSPI_Conf = wSPI_Config&0xf;
            iotSPI.dwSPI_REG1 = dwSPI_SetupValue1&0xff;
            iotSPI.bSPI_Mode = 0; /*Let kernel know SPI_Setup is performed*/
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_SPI_CONFIG, &iotSPI)<0)
                return iotSPI.wErrorCode;
            break;
        default:
            return ErrorFuncNotSupport;
    }

    return NoError;
}


I16 SPI_Control(
    U16 wCardNumber,
    U16 wSPI_Port,
    U16 wSPI_CtrlParam,
    U32 dwSPI_CtrlValue
    )
{
    DAS_IOT_SPI iotSPI;

    memset(&iotSPI, '\0', sizeof(DAS_IOT_SPI));

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    switch(CurrentCard.wCardType){
        case PCI_7350:
            /*support only one SPI controller in PCI-7350*/
            if(wSPI_Port>0)
                return ErrorUndefinedParameter;
            /*only enable operation are valid*/
            if(wSPI_CtrlParam>SPI_ENABLE)
                return ErrorUndefinedParameter;
            iotSPI.bSPI_Port = wSPI_Port;
            iotSPI.bSPI_Mode = 1; /*Let kernel know SPI_Control is performed*/
            iotSPI.dwSPI_REG1 = wSPI_CtrlParam;
            iotSPI.dwSPI_REG2 = dwSPI_CtrlValue&0x1;
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_SPI_CONFIG, &iotSPI)<0)
                return iotSPI.wErrorCode;
            break;
        default:
            return ErrorFuncNotSupport;
    }

    return NoError;
}


I16 SPI_Status(
    U16 wCardNumber,
    U16 wSPI_Port,
    U32 *dwSPI_Status
    )
{
    DAS_IOT_SPI iotSPI;

    memset(&iotSPI, '\0', sizeof(DAS_IOT_SPI));

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    switch(CurrentCard.wCardType){
        case PCI_7350:
            /*support only one SPI controller in PCI-7350*/
            if(wSPI_Port>0)
                return ErrorUndefinedParameter;
            iotSPI.bSPI_Port = wSPI_Port;
            iotSPI.bSPI_Mode = 1; /*Let kernel know SPI_Status is performed*/
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_SPI_READ, &iotSPI)<0)
                return iotSPI.wErrorCode;
            *dwSPI_Status = iotSPI.dwSPI_REG1;
            break;
        default:
            return ErrorFuncNotSupport;
    }

    return NoError;
}


I16 SPI_Read(
    U16 wCardNumber,
    U16 wSPI_Port,
    U16 wSPI_SlaveAddr,
    U16 wSPI_CmdAddrBits,
    U16 wSPI_DataBits,
    U16 wSPI_FrontDummyBits,
    U32 dwSPI_CmdAddr,
    U32 *pdwSPI_Data
    )
{
    DAS_IOT_SPI iotSPI;

    memset(&iotSPI, '\0', sizeof(DAS_IOT_SPI));

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    switch(CurrentCard.wCardType){
        case PCI_7350:
            /*support only one SPI controller in PCI-7350*/
            if(wSPI_Port>0)
                return ErrorUndefinedParameter;
            if(wSPI_CmdAddrBits>32)
                return ErrorUndefinedParameter;
            if(wSPI_DataBits>32)
                return ErrorUndefinedParameter;
            if(wSPI_FrontDummyBits>15)
                return ErrorUndefinedParameter;
            if(wSPI_SlaveAddr>2)
                return ErrorUndefinedParameter;

            iotSPI.bSPI_Port = wSPI_Port;
            iotSPI.bSPI_SlaveAddr = wSPI_SlaveAddr;
            iotSPI.bSPI_CmdAddrBits = wSPI_CmdAddrBits;
            iotSPI.bSPI_RWBits = wSPI_DataBits;
            iotSPI.bSPI_FrontDBits = wSPI_FrontDummyBits;
            iotSPI.dwSPI_REG1 = \
                dwSPI_CmdAddr&(0xffffffffL>>(32-wSPI_CmdAddrBits));
            iotSPI.bSPI_Mode = 0; /*Let kernel know SPI_Read is performed*/
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_SPI_READ, &iotSPI)<0)
                return iotSPI.wErrorCode;
            *pdwSPI_Data = \
                iotSPI.dwSPI_REG2&(0xffffffffL>>(32-wSPI_DataBits));
            break;
        default:
            return ErrorFuncNotSupport;
    }

    return NoError;
}


I16 SPI_Write(
    U16 wCardNumber,
    U16 wSPI_Port,
    U16 wSPI_SlaveAddr,
    U16 wSPI_CmdAddrBits,
    U16 wSPI_DataBits,
    U16 wSPI_FrontDummyBits,
    U16 wSPI_TailDummyBits,
    U32 dwSPI_CmdAddr,
    U32 dwSPI_Data
    )
{
    DAS_IOT_SPI iotSPI;

    memset(&iotSPI, '\0', sizeof(DAS_IOT_SPI));

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    switch(CurrentCard.wCardType){
        case PCI_7350:
            /*support only one SPI controller in PCI-7350*/
            if(wSPI_Port>0)
                return ErrorUndefinedParameter;
            if(wSPI_CmdAddrBits>32)
                return ErrorUndefinedParameter;
            if(wSPI_DataBits>32)
                return ErrorUndefinedParameter;
            if(wSPI_FrontDummyBits>15)
                return ErrorUndefinedParameter;
            if(wSPI_TailDummyBits>15)
                return ErrorUndefinedParameter;
            if(wSPI_SlaveAddr>2)
                return ErrorUndefinedParameter;

            iotSPI.bSPI_Port = wSPI_Port;
            iotSPI.bSPI_SlaveAddr = wSPI_SlaveAddr;
            iotSPI.bSPI_CmdAddrBits = wSPI_CmdAddrBits;
            iotSPI.bSPI_RWBits = wSPI_DataBits;
            iotSPI.bSPI_FrontDBits = wSPI_FrontDummyBits;
            iotSPI.bSPI_TailDBits = wSPI_TailDummyBits;
            iotSPI.dwSPI_REG1 = \
                dwSPI_CmdAddr&(0xffffffffL>>(32-wSPI_CmdAddrBits));
            iotSPI.dwSPI_REG2 = \
                dwSPI_Data&(0xffffffffL>>(32-wSPI_DataBits));
            iotSPI.bSPI_Mode = 0; /*Let kernel know SPI_Write is performed*/
            if(ioctl(CurrentCard.hDevDriver, DAS_IOC_SPI_WRITE, &iotSPI)<0)
                return iotSPI.wErrorCode;
            break;
        default:
            return ErrorFuncNotSupport;
    }

    return NoError;
}
