#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <math.h>
#include <sys/ioctl.h>
#include "dask.h"
#include "Adlsys.h"
#include "dask_p.h"

extern REGISTER_CARD_TYPE RegisterCards[MAX_CARD];


I16 fundamental_check(
    DAS_IOT_FUNC_DESP *pConfig,
    U16               wCardNumber,
    U16               wNumChans,
    U16               *pwChans,
    U32               dwBufferSize,
    BOOLEAN           bContinuousOp
    )
{
    /*
     * Check whether the card registered or not; if yes, check if the given
     * status are valid to this card. The configuration info of the card is
     * stored in pConfig.
     */
    U16 i;

    if(wCardNumber>=MAX_CARD)
        return ErrorInvalidCardNumber;
    Check_If_Card_Registered(wCardNumber);

    if(!pConfig->wFlag.bFunction)
        return ErrorFuncNotSupport;

    for(i=0; i<wNumChans; i++){
        if(pConfig->wChannelCount<=pwChans[i])
            return ErrorInvalidIoChannel;
    }
    if((CurrentCard.wCardType!=PCI_7300A_RevA)&&
        (CurrentCard.wCardType!=PCI_7300A_RevB)&&
        (CurrentCard.wCardType!=PCI_7350)&&
        (CurrentCard.wCardType!=PCI_9221)&&
        (CurrentCard.wCardType!=PCI_6202)&&
        (CurrentCard.wCardType!=PCI_9524)&&
        (CurrentCard.wCardType!=PCI_9222)&&
        (CurrentCard.wCardType!=PCI_9223)&&
        (CurrentCard.wCardType!=PCI_922A))
    {
        if(CurrentCard.wCardType==PCI_9116){
            if(pConfig->dwBufferSize<wNumChans*dwBufferSize)
                return ErrorTransferCountTooLarge;
            //yuan add 04/24/01
            if(bContinuousOp&&(CurrentCard.AI_Status.wNumChans<wNumChans)){
                return ErrorInvalidIoChannel;
            }
        }
        else {
            if(pConfig->dwBufferSize<dwBufferSize)
                return ErrorTransferCountTooLarge;
        }
    }
    else if((CurrentCard.wCardType==PCI_9221)||
        (CurrentCard.wCardType==PCI_6202)||
        (CurrentCard.wCardType==PCI_9222)||
        (CurrentCard.wCardType==PCI_9223)||
        (CurrentCard.wCardType==PCI_922A))
    {
        if((pConfig->dwBufferSize*256<dwBufferSize)||
            (dwBufferSize&0x00000001))
        {
            return ErrorTransferCountTooLarge;
        }
    }
    else if(CurrentCard.wCardType==PCI_7350){
        if(((((pConfig->dwBufferSize-128*20)/5)*1024)<dwBufferSize)||
            (dwBufferSize&0x00000007))
        {
            return ErrorTransferCountTooLarge;
        }
    }
    else{ //Pci7300, Pci9524
        if((pConfig->dwBufferSize*256<dwBufferSize)||
            (dwBufferSize&0x00000003))
        {
            return ErrorTransferCountTooLarge;
        }
    }

    if(bContinuousOp&&(!pConfig->wFlag.bSampleRate))
        return ErrorContIoNotAllowed;

    //00/06/02 jeffrey, because the 7300 without Mmap
    if((CurrentCard.wCardType!=PCI_7300A_RevA)&&
        (CurrentCard.wCardType!=PCI_7300A_RevB)&&
        (CurrentCard.wCardType!=PCI_9221)&&
        (CurrentCard.wCardType!=PCI_6202)&&
        (CurrentCard.wCardType!=PCI_9524)&&
        (CurrentCard.wCardType!=PCI_9222)&&
        (CurrentCard.wCardType!=PCI_9223)&&
        (CurrentCard.wCardType!=PCI_922A)&&
        (CurrentCard.wCardType!=PCI_7350))
    {
        if(bContinuousOp&&(!pConfig->pView[0]))
            return ErrorContIoNotAllowed;
    }

    return NoError;
}


I16 validate_ad_range(
    U16 wCardNumber,
    U16 wAdRange,
    U16 *wGain
    )
{
    I16 i;

    for(i=0; i<CurrentCardInfo.wAdGainCount; i++){
        if(wAdRange==CurrentCardInfo.mAdGain[i].wRange){
            *wGain = CurrentCardInfo.mAdGain[i].wGain;
            return TRUE;
        }
    }

    return FALSE;
}


I16 getpacervalue_9524(
    F64 fSampleRate,
    U32 *dwPacerValue
    )
{
    int i;
    typedef struct _9524_fSample_pacer{
        F64 fSample;
        U32 dwPacerValue;
    }S_9524_samp_pacer;

    S_9524_samp_pacer Trans_tb[16] = {
        {P9524_ADC_30K_SPS, 0xF0},
        {P9524_ADC_15K_SPS, 0xE0},
        {P9524_ADC_7K5_SPS, 0xD0},
        {P9524_ADC_3K75_SPS, 0xC0},
        {P9524_ADC_2K_SPS, 0xB0},
        {P9524_ADC_1K_SPS, 0xA1},
        {P9524_ADC_500_SPS, 0x92},
        {P9524_ADC_100_SPS, 0x82},
        {P9524_ADC_60_SPS, 0x72},
        {P9524_ADC_50_SPS, 0x63},
        {P9524_ADC_30_SPS, 0x53},
        {P9524_ADC_25_SPS, 0x43},
        {P9524_ADC_15_SPS, 0x33},
        {P9524_ADC_10_SPS, 0x23},
        {P9524_ADC_5_SPS, 0x13},
        {P9524_ADC_2R5_SPS, 0x03}
    };

    for(i=0; i<16; i++){
        if(fSampleRate==Trans_tb[i].fSample){
            *dwPacerValue = Trans_tb[i].dwPacerValue;
            break;
        }
    }

    if(i==16)
        return FALSE;

    return TRUE;
}


I16 get_pacervalue(
    I16 wCardNumber,
    F64 fSampleRate,
    U32 *dwPacerValue
    )
{
    F64 divider;
    //U32 intVal;

    switch(CurrentCard.wCardType){
        case PCI_7200:
            if((fSampleRate>2000000L)||(fSampleRate<(F64)(1/1000)))
                return FALSE;
            divider = ((F64)CurrentCardInfo.dwCtrKHz*1000)/fSampleRate;
            if((divider-floor(divider))>((F64)0.5))
                *dwPacerValue = ((U32)divider)+1;
            else
                *dwPacerValue = ((U32)divider);
            break;
        case PCI_7300A_RevA:
        case PCI_7300A_RevB:
            if((fSampleRate>10000000L)||(fSampleRate<(F64)(1/1000)))
                return FALSE;
            divider = ((F64)CurrentCardInfo.dwCtrKHz*1000)/fSampleRate;
            if((divider-floor(divider))>((F64)0.5))
                *dwPacerValue = ((U32)divider)+1;
            else
                *dwPacerValue = ((U32)divider);
            break;
        case PCI_9112:
            if((fSampleRate>110000L)||(fSampleRate<(F64)(1/2000)))
                return FALSE;
            divider = ((F64)CurrentCardInfo.dwCtrKHz*1000)/fSampleRate;
            if((divider-floor(divider))>((F64)0.5))
                *dwPacerValue = ((U32)divider)+1;
            else
                *dwPacerValue = ((U32)divider);
            break;
        case PCI_9111DG:
        case PCI_9111HR:
            if((fSampleRate>100000L)||(fSampleRate<(F64)(1/2000)))
                return FALSE;
            divider = ((F64)CurrentCardInfo.dwCtrKHz*1000)/fSampleRate;
            if((divider-floor(divider))>((F64)0.5))
                *dwPacerValue = ((U32)divider)+1;
            else
                *dwPacerValue = ((U32)divider);
            break;
        case PCI_9114DG:
        case PCI_9114HG:
            if((fSampleRate>250000L)||(fSampleRate<(F64)(1/2000)))
                return FALSE;
            divider = ((F64)CurrentCardInfo.dwCtrKHz*1000)/fSampleRate;
            if((divider-floor(divider))>((F64)0.5))
                *dwPacerValue = ((U32)divider)+1;
            else
                *dwPacerValue = ((U32)divider);
            break;
        case PCI_9113:
            if((fSampleRate>100000L)||(fSampleRate<(F64)(1/2000)))
                return FALSE;
            divider = ((F64)CurrentCardInfo.dwCtrKHz*1000)/fSampleRate;
            if((divider-floor(divider))>((F64)0.5))
                *dwPacerValue = ((U32)divider)+1;
            else
                *dwPacerValue = ((U32)divider);
            break;
        case PCI_9118DG:
        case PCI_9118HG:
        case PCI_9118HR:
            if((fSampleRate>330000L)||(fSampleRate<(F64)(1/2000)))
                return FALSE;
            divider = ((F64)CurrentCardInfo.dwCtrKHz*1000)/fSampleRate;
            if((divider-floor(divider))>((F64)0.5))
                *dwPacerValue = ((U32)divider)+1;
            else
                *dwPacerValue = ((U32)divider);
            break;
        case PCI_9812:
        case PCI_9810:
            if((fSampleRate>20000000L)||(fSampleRate<(F64)(1/20000)))
                return FALSE;
            divider = ((F64)CurrentCardInfo.dwCtrKHz*1000)/fSampleRate;
            if((divider-floor(divider))>((F64)0.5))
                *dwPacerValue = ((U32)divider)+1;
            else
                *dwPacerValue = ((U32)divider);
            break;
        case PCI_9221:
            if((fSampleRate>40000000L)||(fSampleRate<(F64)40000000/(1<<24)))
                return FALSE;
            divider = ((F64)CurrentCardInfo.dwCtrKHz*1000)/fSampleRate;
            if((divider-floor(divider))>((F64)0.5))
                *dwPacerValue = ((U32)divider)+1;
            else
                *dwPacerValue = ((U32)divider);
            break;
        case PCI_9222:
        case PCI_9223:
        case PCI_922A:
            if((fSampleRate>80000000L)||(fSampleRate<(F64)80000000/(1<<24)))
                return FALSE;
            divider = ((F64)CurrentCardInfo.dwCtrKHz*1000)/fSampleRate;
            if((divider-floor(divider))>((F64)0.5))
                *dwPacerValue = ((U32)divider)+1;
            else
                *dwPacerValue = ((U32)divider);
            break;
        case PCI_7350:
            if((fSampleRate>100000000L)||(fSampleRate<(F64)100000000/0xffff))
                return FALSE;
            divider = ((F64)CurrentCardInfo.dwCtrKHz*1000)/fSampleRate;
            if((divider-floor(divider))>((F64)0.5))
                *dwPacerValue = ((U32)divider)+1;
            else
                *dwPacerValue = ((U32)divider);
            break;
        case PCI_9524:
            return getpacervalue_9524(fSampleRate, dwPacerValue);
        default:
            return FALSE;
    }

    return TRUE;
}


U32 AdlSplitePacer(
    U32 dwPacerIn
    )
{
    U32 dwV1, dwV2;

    if(dwPacerIn<4)
        return 0x00020002;

    dwV1 = dwPacerIn>>1;
    dwV2 = 2;
    while(dwV1&0xFFFF0000){
        dwV1 >>= 1;
        dwV2 <<= 1;
    }

    return (dwV1<<16)|dwV2;
}


I16 get_actualrate_9524(
    U32 dwPacerValue,
    F64 *fActualRate
    )
{
    int i;
    typedef struct _9524_pacer_ActualRate{
        U32 dwPacerValue;
        F64 ActRate_Org;
        F64 ActRate_Scan;
    }S_9524_pacer_actrate;

    S_9524_pacer_actrate Trans_tb[16] = {
        {0xF0, 30000.0, 1637.47},
        {0xE0, 15000.0, 1536.81},
        {0xD0,  7500.0, 1407.07},
        {0xC0,  3750.0, 1189.48},
        {0xB0,  2000.0,  925.33},
        {0xA1,  1000.0,  632.63},
        {0x92,   500.0,  387.49},
        {0x82,   100.0,   94.51},
        {0x72,    60.0,   58.00},
        {0x63,    50.0,   48.59},
        {0x53,    30.0,   29.49},
        {0x43,    25.0,   24.64},
        {0x33,    15.0,   14.87},
        {0x23,    10.0,    9.94},
        {0x13,     5.0,    4.99},
        {0x03,     2.5,    2.50}
    };

    for(i=0; i<16; i++){
        if(dwPacerValue==Trans_tb[i].dwPacerValue){
            *fActualRate = Trans_tb[i].ActRate_Org;
            *(fActualRate+1) = Trans_tb[i].ActRate_Scan;
            break;
        }
    }

    if(i==16)
        return FALSE;

    return TRUE;
}


I16 GetActualRate(
    U16 wCardNumber,
    F64 fSampleRate,
    F64 *fActualRate
    )
{
    U32 dwPacerValue;
    U32 dwPacr, c1, c2;
    int err;

    err = get_pacervalue(wCardNumber, fSampleRate, &dwPacerValue);
    if(!err)
        return ErrorInvalidSampleRate;

    switch(CurrentCard.wCardType){
        case PCI_7200:
           if(dwPacerValue>65535){
                dwPacr = AdlSplitePacer(dwPacerValue);
                c1 = dwPacr&0xFFFF;
                c2 = dwPacr>>16;
                *fActualRate = \
                    ((F64)CurrentCardInfo.dwCtrKHz*1000)/((F64)c1)/((F64)c2);
            }
            else{
                *fActualRate = \
                    ((F64)CurrentCardInfo.dwCtrKHz*1000)/((F64)dwPacerValue);
            }
            break;
        case PCI_7300A_RevA:
        case PCI_7300A_RevB:
        case PCI_7350:
            *fActualRate = \
                ((F64)CurrentCardInfo.dwCtrKHz*1000)/((F64)dwPacerValue);
            break;
        case PCI_9112:
        case PCI_9113:
        case PCI_9111DG:
        case PCI_9111HR:
        case PCI_9114DG:
        case PCI_9114HG:
        case PCI_9118DG:
        case PCI_9118HG:
        case PCI_9118HR:
            dwPacr = AdlSplitePacer(dwPacerValue);
            c1 = dwPacr&0xFFFF;
            c2 = dwPacr>>16;
            *fActualRate = \
                ((F64)CurrentCardInfo.dwCtrKHz*1000)/((F64)c1)/((F64)c2);
            break;
        case PCI_9812:
        case PCI_9810:
            dwPacerValue &= 0xfffe;
            *fActualRate = \
                ((F64)CurrentCardInfo.dwCtrKHz*1000)/((F64)dwPacerValue);
            break;
        case PCI_9116:
        case PCI_9221:
        case PCI_9222:
        case PCI_9223:
        case PCI_922A:
            *fActualRate = \
                ((F64)CurrentCardInfo.dwCtrKHz*1000)/((F64)dwPacerValue);
           break;
        default:
            return ErrorFuncNotSupport;
    }

    return NoError;
}


I16 GetActualRate_9524(
    U16 wCardNumber,
    U16 wGroup,
    F64 fSampleRate,
    F64 *fActualRate
    )
{
    I16 err;
    U32 dwPacerValue;
    F64 ADCRate[2];
    DAS_IOT_CONFIG iotConf = {0};
    //int i;

    err = get_pacervalue(wCardNumber, fSampleRate, &dwPacerValue);
    if(!err)
        return ErrorInvalidSampleRate;

    err = get_actualrate_9524(dwPacerValue, ADCRate);
    if(!err)
        return ErrorInvalidSampleRate;

    //02/19/08
    if(wGroup==0){
        iotConf = CurrentCard.AI_Status.CurrentConfig;
        if(iotConf.mData.m9524AI.Reserved1){
            if(iotConf.mData.m9524AI.fAZEn){
                *fActualRate = *(ADCRate+1)/2;
            }
            else{
                if(iotConf.mData.m9524AI.fAutoScan){
                    *fActualRate = *(ADCRate+1);
                }
                else{
                    *fActualRate = *ADCRate;
                }
            }
        }
        else{
            *fActualRate = 0;
        }
    }
    else if(wGroup==1){
        iotConf.mData.dwValue[0] = CurrentCard.AI_Status.dwPacerValue;
        iotConf.mData.dwValue[1] = CurrentCard.AI_Status.dwSampleIntv;

        if(iotConf.mData.m9524AI.Reserved1){
            if(iotConf.mData.m9524AI.fAutoScan){
                *fActualRate = *(ADCRate+1);
            }
            else{
                *fActualRate = *ADCRate;
            }
        }
        else{
            *fActualRate = 0;
        }
    }

    return NoError;
}


I16 P7300bSetLatencyTimer(
    U16 wCardNumber,
    U8  latency
    )
{
    if(CurrentCard.wCardType!=PCI_7300A_RevB){
        return ErrorFuncNotSupport;
    }

    if(ioctl(CurrentCard.hDevDriver, DAS_IOC_SET_LATENCY_TIMER, &(latency))
        !=DAS_ERR_NO)
    {
        return ErrorFuncNotSupport;
    }

    return NoError;
}
