#include "Adlsys.h"

#define NUM_CARD_TYPE   48

#define Check_If_Card_Registered(card_number) \
    if(!RegisterCards[card_number].bBeingUsed) return(ErrorCardNotRegistered)

typedef struct _DASK_STATUS_TYPE
{
    BOOLEAN        bDblBufferMode;
    BOOLEAN        bToFile;
    BOOLEAN        bChanRange;
    U32            dwTransferCount;
    U16            wNumChans;
    void           *pwBuffer;
    void           *p7300buf;
    //void         *p7300mbuf[8];
    U32            pDAQbufHandle;
    U32            p7300bufHandle;
    U16            wDblBufferIndex;
    DAS_IOT_CONFIG CurrentConfig;
    U32            dwPacerValue;  //09/13/00 for 9812 external clock
    U32            dwSampleIntv;
    void           *pDAQbuf[2];   //addr of buffer set
    DAS_IOT_MBUF   p7300mbuf[8];  //04/11/01 yuan add for multi-buffer
    U32            dwBufCount[2]; //buffer size of buffer set
    U16            p7300mbufCnt;
    DAS_IOT_MBUF   p7350mbuf[64];  //07/23/09
}
DASK_STATUS_TYPE;

typedef struct _DASK_EPT_INFO
{
    //HANDLE  hEvent;
    //HANDLE  hCbThread;
    //char    szEventName[DAS_EVENT_NAME_SIZE];
    U16     wCardNumber;
    BOOLEAN bEnabled;
    BOOLEAN bCompleted;
    //HANDLE  windowHandle;
    U32     message;
    void    (*callbackAddr)(void);
    BOOLEAN bCbCompleted;
    U32     dwIntCount;
    //HANDLE  hCountEvent;
}
DASK_EPT_INFO;

typedef struct _EVENT_MESSAGE_TYPE
{
    U16     wCardNumber;
    U16     wInt1or2;   //0:INT1, 1:INT2
    BOOLEAN bEnabled;
    BOOLEAN bCompleted;
}
EVENT_MESSAGE_TYPE;

typedef struct _EVENT_MESSAGE_TYPE_CB
{
    U16     wCardNumber;
    U16     wInt1or2;   //0:INT1, 1:INT2
    BOOLEAN bEnabled;
    BOOLEAN bCompleted;
    //HANDLE  windowHandle;
    U32     message;
    //U32     callbackAddr;
    void    (*callbackAddr)(void);
    BOOLEAN bBeingUsed;
    //HANDLE  threadID;
    //HANDLE  hCountEvent;
}
EVENT_MESSAGE_TYPE_CB;

typedef struct _DAS_IOT_CALCONST_9221
{
    U32         IOP_mem;
    U16         constant_gain_AD[4];
    U16         constant_offset_AD[4];
    U32         constant_DA0[3];
    U32         constant_DA1[3];
    PCI_DBXORUN OvRun;
}
__attribute__((packed))DAS_IOT_CALCONST_9221;

typedef struct _DAS_IOT_CALCONST_922X
{
    U32         IOP_mem;
    U16         CALConst_AD_Offset[8];
    U16         CALConst_AD_Gain[8];
    U16         CALConst_DA_5V;
    U16         CALConst_DA_Linear[2];
    U16         CALConst_DA_Offset[2];
    U16         CALConst_DA_Gain[2];
    PCI_DBXORUN OvRun[4];
}
__attribute__((packed))DAS_IOT_CALCONST_922X;

typedef struct _DAS_IOT_CALCONST_7350
{
    U32             IOP_mem;
    PCI7350_DBXORUN OvRun[2];
}
__attribute__((packed))DAS_IOT_CALCONST_7350;

typedef struct _ROM_DefaultBank{
    U8 EE_SysDefaultBank;
    U8 EE_AI0CH0UserConversionBank;
    U8 EE_AI0CH1UserConversionBank;
    U8 EE_AI0CH2UserConversionBank;
    U8 EE_AI0CH3UserConversionBank;
    U8 EE_AI1CH0UserConversionBank;
    U8 EE_AI1CH1UserConversionBank;
    U8 EE_AI1CH2UserConversionBank;
    U8 EE_AI1CH3UserConversionBank;
}
__attribute__((packed))ROM_DefaultBank;

typedef struct _ROM_SysExtCalConst{
    U32    EE_MMDD;
    U32    EE_YEAR;
    float  EE_CALTEMP;
    U16    EE_5VREF_TRIM_POS;
    double EE_CAL_SRC_10V;
    double EE_CAL_SRC_5V;
    double EE_CAL_SRC_2R5V;
    double EE_CAL_SRC_1R25V;
    double EE_CAL_SRC_BRIDGE_NULL;
    double EE_CAL_SRC_BRIDGE_FSR;
}
__attribute__((packed))ROM_SysExtCalConst;

typedef struct _ROM_SysAI0IntCalConst{
    U32    EE_MMDD;
    U32    EE_YEAR;
    float  EE_CALTEMP;
    U32    EE_ADC_OFC[4][16];
    U32    EE_ADC_FSC[4][16];
    double EE_RESIDUAL_OFFSET[4][16];
    double EE_SCALING[4][16];
}
__attribute__((packed))ROM_SysAI0IntCalConst;

typedef struct _ROM_SysAI1IntCalConst{
    U32    EE_MMDD;
    U32    EE_YEAR;
    float  EE_CALTEMP;
    U32    EE_ADC_OFC[4][16];
    U32    EE_ADC_FSC[4][16];
    double EE_RESIDUAL_OFFSET[4][16];
    double EE_SCALING[4][16];
}
__attribute__((packed))ROM_SysAI1IntCalConst;

typedef struct _ROM_SysAOIntCalConst{
    U32   EE_MMDD;
    U32   EE_YEAR;
    float EE_CALTEMP;
    U8    EE_DAC_OFFSET[2];
    U8    EE_DAC_LINEARITY[2];
    float EE_GAIN_FACTOR[2];
}
__attribute__((packed))ROM_SysAOIntCalConst;

typedef struct _ROM_UserConversion{
    U8     EE_TAG_NAME[16];
    U32    EE_MMDD;
    U32    EE_YEAR;
    float  EE_CALTEMP;
    double EE_OFFSET_REF;
    double EE_GAIN_REF;
    double EE_EU_CONVERSION;
    U8     EE_UNIT_STRING[8];
}
__attribute__((packed))ROM_UserConversion;

typedef struct _ROM_UserString{
    U8 EE_TAG_NAME[16];
    U8 EE_USER_STRING[128];
}
__attribute__((packed))ROM_UserString;

typedef struct _ROM_UserConstant{
    U8 E_TAG_NAME[16];
    U8 EE_FORMAT;
    U8 EE_CONSTANT_STRING[8];
}
__attribute__((packed))ROM_UserConstant;

typedef struct _DAS_IOT_CALCONST_9524
{
    ROM_DefaultBank       DefaultBank;
    ROM_SysExtCalConst    SysExtCalConst;
    ROM_SysAI0IntCalConst SysAI0IntCalConst;
    ROM_SysAI1IntCalConst SysAI1IntCalConst;
    ROM_SysAOIntCalConst  SysAOIntCalConst;
    ROM_UserConversion    UserConversion;
    ROM_UserString        UserString;
    ROM_UserConstant      UserConstant;
    PCI_DBXORUN           OvRun;
}
__attribute__((packed))DAS_IOT_CALCONST_9524;

typedef struct _REGISTER_CARD_TYPE
{
    BOOLEAN            bBeingUsed;
    U16                wCardType;
    U16                wDevIndex;
    int                hDevDriver;
    DAS_IOT_CARD_DESP  CardInfo;
    DASK_STATUS_TYPE   AI_Status;
    DASK_STATUS_TYPE   DI_Status;
    DASK_STATUS_TYPE   AO_Status;
    DASK_STATUS_TYPE   DO_Status;
    DASK_STATUS_TYPE   CTR_Status;
    DASK_EPT_INFO      EPT_INFO[8];   //David Added
    DAS_IOT_DUALINT    DIOINT_Status; //Alex Lee 5/28/99
    EVENT_MESSAGE_TYPE Event_Msg[2];  //Alex Lee 5/28/99
    F64                RefVoltage[2]; //Alex Lee 6/1/99 for 9112 only
    struct sigaction   sig_act1, sig_act2; // jeffrey 6/9/00 for sigaction
    //yuan add 06/29/04
    struct sigaction   i_sig_act, o_sig_act;
    struct sigaction   i_dbsig_act, o_dbsig_act;
    //12/26/09
    struct sigaction   i_db2sig_act, o_db2sig_act;
    U16                PortAddr[4];
    //01/11/08 modified
    //DAS_IOT_CALCONST   *CAL_Const; //add for pci-9221
    void               *pCAL_Const;
}
REGISTER_CARD_TYPE;

/*
 * Error code
 */
#define ErrorPoolAlloc        -101
#define ErrorIoctlCode        -111
#define ErrorIoctlInSize      -112
#define ErrorIoctlOutSize     -113
#define ErrorNoFunction       -114
#define ErrorRunning          -115

#define ErrorDoubleBufferMode -121
#define ErrorSyncMode         -122
#define ErrorTriggerMode      -123

#define ErrorPacerCount       -131
#define ErrorCounterNumber    -132
#define ErrorCounterMode      -133
#define ErrorCounterState     -134

#define ErrorAccessCount      -141
#define ErrorChannelNumber    -142
#define ErrorChannelByteCount -143
#define ErrorChannelLine      -144

/*
 * Marco definitions
 */
#define CurrentCard \
    RegisterCards[wCardNumber]

#define CurrentCardInfo \
    RegisterCards[wCardNumber].CardInfo

#define CurrentCard_AIFunc \
    RegisterCards[wCardNumber].CardInfo.mFunc[DAS_FUNC_AI]

#define CurrentCard_DIFunc \
    RegisterCards[wCardNumber].CardInfo.mFunc[DAS_FUNC_DI]

#define CurrentCard_AOFunc \
    RegisterCards[wCardNumber].CardInfo.mFunc[DAS_FUNC_AO]

#define CurrentCard_DOFunc \
    RegisterCards[wCardNumber].CardInfo.mFunc[DAS_FUNC_DO]

#define CurrentCard_AICurConfig \
    RegisterCards[wCardNumber].AI_Status.CurrentConfig

#define CurrentCard_DICurConfig \
    RegisterCards[wCardNumber].DI_Status.CurrentConfig

#define CurrentCard_AOCurConfig \
    RegisterCards[wCardNumber].AO_Status.CurrentConfig

#define CurrentCard_DOCurConfig \
    RegisterCards[wCardNumber].DO_Status.CurrentConfig

/*
 * Common functions
 */
I16 fundamental_check(DAS_IOT_FUNC_DESP *pConfig, U16 wCardNumber, \
    U16 wNumChans, U16 *pwChans, U32 dwBufferSize, BOOLEAN bContinuousOp);

I16 validate_ad_range(U16 wCardNumber, U16 wAdRange, U16 *wGain);

I16 get_pacervalue(I16 wCardNumber, F64 fSampleRate, U32 *dwPacerValue);

I16 get_actualrate_9524(U32 dwPacerValue, F64 *fActualRate);
/*
I16 DI_ContBufferStart(U16 wCardNumber, U16 wPort, F64 fSampleRate, \
    BOOLEAN singleOrmulti);

I16 DI_ContBufferStop(U16 wCardNumber, U32 *dwAccessCnt, BOOLEAN restart);

I16 DI_ContMultiBufferReset(U16 wCardNumber);
*/
