#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include "DAQHeader.h"

#define PCI_6208V       1
#define PCI_6208A       2
#define PCI_6308V       3
#define PCI_6308A       4
#define PCI_7200        5
#define PCI_7230        6
#define PCI_7233        7
#define PCI_7234        8
#define PCI_7248        9
#define PCI_7249        10
#define PCI_7250        11
#define PCI_7252        12
#define PCI_7296        13
#define PCI_7300A_RevA  14
#define PCI_7300A_RevB  15
#define PCI_7432        16
#define PCI_7433        17
#define PCI_7434        18
#define PCI_8554        19
#define PCI_9111DG      20
#define PCI_9111HR      21
#define PCI_9112        22
#define PCI_9113        23
#define PCI_9114DG      24
#define PCI_9114HG      25
#define PCI_9118DG      26
#define PCI_9118HG      27
#define PCI_9118HR      28
#define PCI_9810        29
#define PCI_9812        30
#define PCI_7396        31
#define PCI_9116        32
#define PCI_7256        33
#define PCI_7258        34
#define PCI_7260        35
#define PCI_7452        36
#define PCI_7442        37
#define PCI_7443        38
#define PCI_7444        39
#define PCI_9221        40
#define PCI_9524        41
#define PCI_6202        42
#define PCI_9222        43
#define PCI_9223        44
#define PCI_7433C       45
#define PCI_7434C       46
#define PCI_922A        47
#define PCI_7350        48

typedef struct{
    unsigned long number;
    short         cardType;
    double        min;
    double        max;
    short         scanCH;
}
stCvtScaledInfo;

char* strCardType[49] = {
    " ",
    "PCI_6208V",      //PCI_6208V      1
    "PCI_6208A",      //PCI_6208A      2
    "PCI_6308V",      //PCI_6308V      3
    "PCI_6308A",      //PCI_6308A      4
    "PCI_7200",       //PCI_7200       5
    "PCI_7230",       //PCI_7230       6
    "PCI_7233",       //PCI_7233       7
    "PCI_7234",       //PCI_7234       8
    "PCI_7248",       //PCI_7248       9
    "PCI_7249",       //PCI_7249       10
    "PCI_7250",       //PCI_7250       11
    "PCI_7252",       //PCI_7252       12
    "PCI_7296",       //PCI_7296       13
    "PCI_7300A_RevA", //PCI_7300A_RevA 14
    "PCI_7300A_RevB", //PCI_7300A_RevB 15
    "PCI_7432",       //PCI_7432       16
    "PCI_7433",       //PCI_7433       17
    "PCI_7433",       //PCI_7434       18
    "PCI_8554",       //PCI_8554       19
    "PCI_9111DG",     //PCI_9111DG     20
    "PCI_9111HR",     //PCI_9111HR     21
    "PCI_9112",       //PCI_9112       22
    "PCI_9113",       //PCI_9113       23
    "PCI_9114DG",     //PCI_9114DG     24
    "PCI_9114HG",     //PCI_9114HG     25
    "PCI_9118DG",     //PCI_9118DG     26
    "PCI_9118HG",     //PCI_9118HG     27
    "PCI_9118HG",     //PCI_9118HR     28
    "PCI_9810",       //PCI_9810       29
    "PCI_9812",       //PCI_9812       30
    "PCI_7396",       //PCI_7396       31
    "PCI_9116",       //PCI_9116       32
    "PCI_7256",       //PCI_7256       33
    "PCI_7258",       //PCI_7258       34
    "PCI_7260",       //PCI_7260       35
    "PCI_7452",       //PCI_7452       36
    "PCI_7442",       //PCI_7442       37
    "PCI_7443",       //PCI_7443       38
    "PCI_7444",       //PCI_7444       39
    "PCI_9221",       //PCI_9221       40
    "PCI-9524",       //PCI-9524       41
    "PCI-6202",       //PCI-6202       42
    "PCI-9222",       //PCI-9222       43
    "PCI-9223",       //PCI-9223       44
    "PCI-7433C",      //PCI-7433       45
    "PCI-7434C",      //PCI-7434       46
    "PCI-922A",       //PCI-922A       47
    "PCI-7350",       //PCI-7350       48
};

char* strADRange[30] = {
    "none",
    "+/-10V",     //AD_B_10_V     1
    "+/-5V",      //AD_B_5_V      2
    "+/-2.5V",    //AD_B_2_5_V    3
    "+/-1.25V",   //AD_B_1_25_V   4
    "+/-0.625V",  //AD_B_0_625_V  5
    "+/-0.3125V", //AD_B_0_3125_V 6
    "+/-0.5V",    //AD_B_0_5_V    7
    "+/-0.05V",   //AD_B_0_05_V   8
    "+/-0.005V",  //AD_B_0_005_V  9
    "+/-1V",      //AD_B_1_V      10
    "+/-0.1V",    //AD_B_0_1_V    11
    "+/-0.01V",   //AD_B_0_01_V   12
    "+/-0.001V",  //AD_B_0_001_V  13
    "0~20V",      //AD_U_20_V     14
    "0~10V",      //AD_U_10_V     15
    "0~5V",       //AD_U_5_V      16
    "0~2.5V",     //AD_U_2_5_V    17
    "0~1.25V",    //AD_U_1_25_V   18
    "0~1V",       //AD_U_1_V      19
    "0~0.1V",     //AD_U_0_1_V    20
    "0~0.01V",    //AD_U_0_01_V   21
    "0~0.001V",   //AD_U_0_001_V  22
    "+/-2V",      //AD_B_2_V      23
    "+/-0.25V",   //AD_B_0_25_V   24
    "+/-0.2V",    //AD_B_0_2_V    25
    "0~4V",       //AD_U_4_V      26
    "0~2V",       //AD_U_2_V      27
    "0~0.5V",     //AD_U_0_5_V    28
    "0~0.4V",     //AD_U_0_4_V    29
};

double RangeValue[30][2] = {
    {0,       0     },
    {-10,     10    }, //AD_B_10_V     1
    {-5,      5     }, //AD_B_5_V      2
    {-2.5,    2.5   }, //AD_B_2_5_V    3
    {-1.25,   1.25  }, //AD_B_1_25_V   4
    {-0.625,  0.625 }, //AD_B_0_625_V  5
    {-0.3125, 0.3125}, //AD_B_0_3125_V 6
    {-0.5,    0.5   }, //AD_B_0_5_V    7
    {-0.05,   0.05  }, //AD_B_0_05_V   8
    {-0.005,  0.005 }, //AD_B_0_005_V  9
    {-1,      1     }, //AD_B_1_V      10
    {-0.1,    0.1   }, //AD_B_0_1_V    11
    {-0.01,   0.01  }, //AD_B_0_01_V   12
    {-0.001,  0.001 }, //AD_B_0_001_V  13
    {0,       20    }, //AD_U_20_V     14
    {0,       10    }, //AD_U_10_V     15
    {0,       5     }, //AD_U_5_V      16
    {0,       2.5   }, //AD_U_2_5_V    17
    {0,       1.25  }, //AD_U_1_25_V   18
    {0,       1     }, //AD_U_1_V      19
    {0,       0.1   }, //AD_U_0_1_V    20
    {0,       0.01  }, //AD_U_0_01_V   21
    {0,       0.001 }, //AD_U_0_001_V  22
    {-2,      2     }, //AD_B_2_V      23
    {-0.25,   0.25  }, //AD_B_0_25_V   24
    {-0.2,    0.2   }, //AD_B_0_2_V    25
    {0,       4     }, //AD_U_4_V      26
    {0,       2     }, //AD_U_2_V      27
    {0,       0.5   }, //AD_U_0_5_V    28
    {0,       0.4   }, //AD_U_0_4_V    29
};

char *strSP[3] = {"  ", "\t", ", "};
char newLine[3] = {0x0A, 0};

DAQFileHeader m_file_header;
DAQChannelRange *m_pChRange;
double m_Compensation[8];
int m_input_fhd;
FILE *m_output_fp = NULL;
char m_channel_num[20], m_scan_num[40], m_data_width[20];
char m_channel_order[20], m_ad_range[20], m_scan_rate[20];
char m_start_date[20], m_start_time[20], m_chrange[20];
char m_input_file[50], m_output_file[50], m_card_type[20];
void *m_pSrc, *m_pDest;
char m_CurInFile[50];
int m_format = 0;
int m_separator = 0;
int m_head = 1;
int m_digital = 0;

int MOpenDAQFile(char *filename);
int LoadFile();
void CvtScaledData(void *src, float *dest, stCvtScaledInfo *sinfo);
void PCvtScaledData(void *src, float *dest, long rCount, DAQChannelRange *ch_range);
void CvtBinaryCodesNCH(void *src, long *pDest, unsigned long num);
void WriteHead();
void ScaledDataToText(void *pSrc);
void ScaledDataToBinary(void *pSrc);
void BinaryCodesToText(void *pSrc);
void StartConvert();


int main(
    int argc,
    char **argv
    )
{
    int i;

    if(argc<2)
        goto print_usage;

    m_input_file[0] = m_output_file[0] = '\0';
    m_input_fhd = -1; m_output_fp = NULL;
    m_format = 0; m_separator = 0; m_head = 1;
    m_pChRange = NULL;

    for(i=1; i<argc; i++){
        if(argv[i][0] == '-'){
            if(!strcmp(argv[i], "-st")){
                m_format = 0;
            }
            else if(!strcmp(argv[i], "-sb")){
                m_format = 1;
            }
            else if(!strcmp(argv[i], "-bt")){
                m_format = 2;
            }
            else if(!strcmp(argv[i], "-sep0")){
                m_separator = 0;
            }
            else if(!strcmp(argv[i], "-sep1")){
                m_separator = 1;
            }
            else if(!strcmp(argv[i], "-sep2")){
                m_separator = 2;
            }
            else if(!strcmp(argv[i], "-nohead")){
                m_head = 0;
            }
            else if(!strcmp(argv[i], "-dec")){
                m_digital = 0;
            }
            else if(!strcmp(argv[i], "-hex")){
                m_digital = 1;
            }
            else{
                goto print_usage;
            }
        }
        else if(m_input_file[0] == '\0'){
            strcpy(m_input_file, argv[i]);
        }
        else if(m_output_file[0] == '\0'){
            strcpy(m_output_file, argv[i]);
        }
    }

    if(m_input_file[0]!='\0'){
        if(m_output_file[0]=='\0'){
            strcpy(m_output_file, m_input_file);
            *(rindex(m_output_file, '.' )) = 0x00;
            if(m_format==1)
                strcat(m_output_file, ".bin");
            else
                strcat(m_output_file, ".cvt");
        }
        if(LoadFile()!=-1){
            StartConvert();
            if(m_pChRange!=NULL)
                free(m_pChRange);
            exit(1);
        }
        else{
            printf(" can not open %s \n", m_input_file );
            exit(-1);
        }
    }
    else{
        printf(" no input file \n");
    }

print_usage:
    printf("Usage: DAQCvt input_file [-st|-sb|-bt] [-sep0|-sep1|-sep2] [-nohead] output_file \n" );
    printf("        input_file : input file name\n\n");
    printf("        [-st|-sb|-bt] : option for conversion format\n");
    printf("                        -st : Scaled data to text file\n");
    printf("                        -sb : Scaled data to binary file (float)\n");
    printf("                        -bt : Binary codes to text file\n");
    printf("              default : -st\n\n");
    printf("        [-sep0|-sep1|-sep2] : option for separator in Text File\n");
    printf("                        -sep0 : add SPACE as separator\n");
    printf("                        -sep1 : add TAB as separator\n");
    printf("                        -sep2 : add COMMA as separator\n");
    printf("              default : -sep0\n\n");
    printf("        [-dec|-hex] : option for separator in Text File\n");
    printf("                        -dec : convert to decimal value\n");
    printf("                        -hex : convert to hexadecimal value\n");
    printf("              default : -dec\n\n");
    printf("        [-nohead] : option for ignore Title/Head\n\n");
    printf("        output_file : output file name\n\n");

    return 1;
}


int MOpenDAQFile(
    char *filename
    )
{
    int fin, i;
    char szFileName[121];
    int file_name_len;

    file_name_len = strlen(filename);

    for(i=0; i<120&&i<file_name_len; i++)
        szFileName[i] = filename[i];
    szFileName[i] = '\0';

    fin = open(filename, O_RDONLY);
    if(fin!=-1){
        long nread;
        int loop;
        nread = read(fin, &m_file_header, sizeof(DAQFileHeader));
        if((nread!=sizeof(DAQFileHeader))||
            (strncmp(m_file_header.ID, "ADLinkDAQ1", 10)!=0))
        {
            printf("Isn't ADLink DAQ File!\n");
            close(fin);
            return -1;
        }

        if(m_pChRange!=NULL)
            free(m_pChRange);
        if(m_file_header.num_of_channel_range>0){
            m_pChRange = (DAQChannelRange *)malloc(m_file_header.num_of_channel_range*sizeof(DAQChannelRange));
            read(fin, m_pChRange, m_file_header.num_of_channel_range*sizeof(DAQChannelRange));
        }
        else
            m_pChRange = NULL;
        //08/25/09
        if(m_file_header.card_type==PCI_9524){
            for(loop=0; loop<m_file_header.num_of_channel; loop++){
                read(fin, &m_Compensation[2*loop], 8);
                read(fin, &m_Compensation[2*loop+1], 8);
            }
        }
        strcpy(m_CurInFile, filename);
    }
    return fin;
}


int LoadFile()
{
    int i;
    char buf[16];
    long nbytes, nread, ndata;

    if(strlen(m_input_file)<=0)
        return -1;

    m_input_fhd = MOpenDAQFile(m_input_file);

    if(m_input_fhd!=-1){
        strcpy(m_card_type, strCardType[m_file_header.card_type]);
        if((m_file_header.num_of_channel==1)&&(m_file_header.channel_order!=2))
            sprintf(m_channel_num, "1 - channel %d", m_file_header.channel_no);
        else
            sprintf( m_channel_num, "%d", m_file_header.num_of_channel);
        sprintf(m_scan_num, "%li", m_file_header.num_of_scan);
        //08/25/09
        if(m_file_header.card_type==PCI_7350)
            strcpy(m_data_width, (m_file_header.data_width==0)? "8 bits":(m_file_header.data_width==1)? "16 bits":(m_file_header.data_width==3)? "24 bits":"32 bits");
        else
            strcpy(m_data_width, (m_file_header.data_width==0)? "8 bits":(m_file_header.data_width==1)? "16 bits":"32 bits");
        strcpy(m_channel_order, (m_file_header.channel_order==0)? "0-1-2-3":(m_file_header.channel_order==1)? "3-2-1-0":"Custom");
        strcpy(m_ad_range, strADRange[m_file_header.ad_range]);
        sprintf(m_scan_rate, "%-10.3f", m_file_header.scan_rate);
        for(i=0; i<8; i++)
            buf[i] = m_file_header.start_date[i];
        buf[8] = '\0';
        strcpy(m_start_date, buf);
        for(i=0; i<8; i++)
            buf[i] = m_file_header.start_time[i];
        buf[8] = '.';
        for(i=0; i<3; i++)
            buf[9+i] = m_file_header.start_millisec[i];
        buf[12] = '\0';
        strcpy(m_start_time, buf);
        sprintf(m_chrange, "%d", m_file_header.num_of_channel_range);

        //------ read raw data
        ndata = m_file_header.num_of_channel*m_file_header.num_of_scan;
        //08/25/09
        if(m_file_header.card_type==PCI_7350)
            nbytes = (m_file_header.data_width==0)? ndata*1:(m_file_header.data_width==1)? ndata*2:(m_file_header.data_width==3)? ndata*3:ndata*4;
        else
            nbytes = (m_file_header.data_width==0)? ndata*1:(m_file_header.data_width==1)? ndata*2:ndata*4;
        m_pSrc = (void *)malloc(nbytes*sizeof(char));
        nread = read(m_input_fhd, m_pSrc, nbytes);
        if(nread!=nbytes){
            printf("Read Input File Error!\n");
            close(m_input_fhd);
            m_input_fhd = -1;
            return -1;
        }

        //08/25/09
        if((m_file_header.card_type<PCI_9111DG)||
            (m_file_header.card_type==PCI_7350))
        {
            m_format = 2;
        }
        else
            m_format = 0;
    }
    else
        return -1;

    return 1;
}


void CvtScaledData(
    void            *src,
    float           *dest,
    stCvtScaledInfo *sinfo
    )
{
    register unsigned int i;
    if(sinfo->cardType==PCI_9112){
        unsigned short data;
        register unsigned short *pSrc = (unsigned short *)src;
        double ratio = (sinfo->max-sinfo->min)/4095;
        if(sinfo->scanCH==1){
            for(i=0; i<sinfo->number; i++){
                data = pSrc[i]>>4;
                dest[i] = (float)(sinfo->min+(data*ratio));
            }
        }
        else{ //sinfo.scanCH > 1
            int j, k;
            short nSCH = sinfo->scanCH;
            {
                for(i=0; i<sinfo->number; i+=nSCH){
                    for(j=nSCH-1, k=0; j>=0; j--, k++){
                        data = pSrc[i+j]>>4;
                        dest[i+k] = (float)(sinfo->min+(data*ratio));
                    }
                }
            }
            m_file_header.channel_order = 0;
        }
    }
    else if(sinfo->cardType==PCI_9111DG){
        short data;
        register short *pSrc = (short *)src;
        double ratio = (sinfo->max-sinfo->min)/4095;
        for(i=0; i<sinfo->number; i++){
            data = (pSrc[i]>>4)+2048;
            dest[i] = (float)(sinfo->min+(data*ratio));
        }
    }
    else if(sinfo->cardType==PCI_9111HR){
        long data;
        register short *pSrc = (short *)src;
        double ratio = (sinfo->max-sinfo->min)/65535;
        for(i=0; i<sinfo->number; i++){
            data = (long)pSrc[i]+32768;
            dest[i] = (float)(sinfo->min+(data*ratio));
        }
    }
    else if(sinfo->cardType==PCI_9113){
        long data;
        register long *pSrc = (long *)src;
        double ratio = (sinfo->max-sinfo->min)/4095;
        for(i=0; i<sinfo->number; i++){
            data = pSrc[i]&0x00000FFF;
            dest[i] = (float)(sinfo->min+(data*ratio));
        }
    }
    else if((sinfo->cardType==PCI_9114DG)||
        (sinfo->cardType==PCI_9114HG))
    {
        long data;
        register long *pSrc = (long *)src;
        double ratio = (sinfo->max-sinfo->min)/65535;
        for(i=0; i<sinfo->number; i++){
            data = (short)(pSrc[i]&0xFFFF)+32768;
            dest[i] = (float)(sinfo->min+(data*ratio));
        }
    }
    else if(sinfo->cardType==PCI_9118HR){
        long data;
        register short *pSrc = (short *)src;
        double ratio = (sinfo->max-sinfo->min)/65535;
        for(i=0; i<sinfo->number; i++){
            data = (long)pSrc[i]+32768;
            dest[i] = (float)(sinfo->min+(data*ratio));
        }
    }
    else if((sinfo->cardType==PCI_9118DG)||
        (sinfo->cardType==PCI_9118HG))
    {
        unsigned short data;
        register unsigned short *pSrc = (unsigned short *)src;
        double ratio = (sinfo->max-sinfo->min)/4095;
        for(i=0; i<sinfo->number; i++){
            data = pSrc[i]>>4;
            dest[i] = (float)(sinfo->min+(data*ratio));
        }
    }
    else if(sinfo->cardType==PCI_9812){
        unsigned short data;
        register short *pSrc = (short *)src;
        double ratio = (sinfo->max-sinfo->min)/4095;
        for(i=0; i<sinfo->number; i++){
            data = (pSrc[i]>>4)+2048;
            dest[i] = (float)(sinfo->min+(data*ratio));
        }
    }
    else if(sinfo->cardType==PCI_9810){
        unsigned short data;
        register short *pSrc = (short *)src;
        double ratio = (sinfo->max-sinfo->min)/1023;
        for(i=0; i<sinfo->number; i++){
            data = (pSrc[i]>>6)+512;
            dest[i] = (float)(sinfo->min+(data*ratio));
        }
    }
    else if(sinfo->cardType==PCI_9116){
        long data;
        register short *pSrc = (short *)src;
        double ratio = (sinfo->max-sinfo->min)/65535;
        for(i=0; i<sinfo->number; i++){
            data = (long)pSrc[i]+32768;
            dest[i] = (float)(sinfo->min+(data*ratio));
        }
    }
    else if(sinfo->cardType==PCI_9221){
        long data;
        register short *pSrc = (short *)src;
        double ratio = (sinfo->max-sinfo->min)/65535;
        for(i=0; i<sinfo->number; i++){
            data = (long)pSrc[i]+32768;
            dest[i] = (float)(sinfo->min+(data*ratio));
        }
    }
    else if((sinfo->cardType==PCI_9222)||(sinfo->cardType==PCI_9223)||
        (sinfo->cardType==PCI_922A))
    {
        long data;
        if(m_file_header.data_width==2){
            register long *pSrc = (long *)src;
            double ratio = (sinfo->max-sinfo->min)/65535;
            for(i=0; i<sinfo->number; i++){
                data = (long)((short)(pSrc[i]&0xffff)+32768);
                dest[i] = (float)(sinfo->min+(data*ratio));
            }
        }
        else{
            register short *pSrc = (short *)src;
            double ratio = (sinfo->max-sinfo->min)/65535;
            for(i=0; i<sinfo->number; i++){
                data = (long)pSrc[i]+32768;
                dest[i] = (float)(sinfo->min+(data*ratio));
            }
        }
    }
    else if(sinfo->cardType==PCI_9524){
        double *dest_d = (double *)dest;
        double res_offset, scaling;
        long data;
        register long *pSrc;

        res_offset = m_Compensation[0];
        scaling = m_Compensation[1];
        pSrc = (long *)src;

        for(i=0; i<sinfo->number; i++){
            data = pSrc[i]>>8;
            dest_d[i] = ((double)(data-res_offset)*scaling)/*/m_Compfactor*/;
        }
    }
}


void PCvtScaledData(
    void            *src,
    float           *dest,
    long            rCount,
    DAQChannelRange *ch_range
    )
{
    register long i, j;
    double ratio[256];
    double min[256];
    short ch_num = m_file_header.num_of_channel;

    if((m_file_header.card_type==PCI_9118DG)||
        (m_file_header.card_type==PCI_9118HG))
    {
        //PCI9118DG/HG
        unsigned short data;
        register unsigned short *pSrc = (unsigned short *)src;
        for(i=0; i<ch_num; i++){
            min[i] = RangeValue[ch_range[i].range][0];
            ratio[i] = (RangeValue[ch_range[i].range][1]-min[i])/4095;
        }
        for(i=0; i<rCount; i++){
            data = pSrc[i]>>4;
            j = i%ch_num;
            dest[i] = (float)(min[j]+(data*ratio[j]));
        }
    }
    //08/25/09
    else if(m_file_header.card_type==PCI_9524){
        double *dest_d = (double *)dest;
        double res_offset[4];
        double scaling[4];
        long data;
        register long *pSrc;

        pSrc = (long *)(src);

        for(i=0; i<ch_num; i++){
            res_offset[i] = m_Compensation[2*i];
            scaling[i] = m_Compensation[2*i+1];
        }

        for(i=0; i<rCount; i++){
            data = pSrc[i]>>8;
            j = i%ch_num;
            dest_d[i] = ((double)(data-res_offset[j])*scaling[j])/*/m_Compfactor*/;
        }
    }
    else if((m_file_header.card_type==PCI_9222)||
        (m_file_header.card_type==PCI_9223)||
        (m_file_header.card_type==PCI_922A))
    {
        long data;

        if(m_file_header.data_width==2){
            register long *pSrc = (long *)src;
            for(i=0; i<ch_num; i++){
                min[i] = RangeValue[ch_range[i].range][0];
                ratio[i] = (RangeValue[ch_range[i].range][1]-min[i])/65535;
            }
            for(i=0; i<rCount; i++){
                data = (long)((short)(pSrc[i]&0xffff)+32768);
                j = i%ch_num;
                dest[i] = (float)(min[j]+(data*ratio[j]));
            }
        }
        else{
            register short *pSrc = (short *)src;
            for(i=0; i<ch_num; i++){
                min[i] = RangeValue[ch_range[i].range][0];
                ratio[i] = (RangeValue[ch_range[i].range][1]-min[i])/65535;
            }
            for(i=0; i<rCount; i++){
                data = (long)pSrc[i]+32768;
                j = i%ch_num;
                dest[i] = (float)(min[j]+(data*ratio[j]));
            }
        }
    }
    else{
        //PCI_9118HR
        long data;
        register short *pSrc = (short *)src;
        for(i=0; i<ch_num; i++){
            min[i] = RangeValue[ch_range[i].range][0];
            ratio[i] = (RangeValue[ch_range[i].range][1]-min[i])/65535;
        }
        for(i=0; i<rCount; i++){
            data = (long)pSrc[i]+32768;
            j = i%ch_num;
            dest[i] = (float)(min[j]+(data*ratio[j]));
        }
    }
}


void CvtBinaryCodesNCH(
    void          *src,
    long          *pDest,
    unsigned long num
    )
{
    register unsigned int i;

    //------ 16bit unsigned
    if(m_file_header.card_type==PCI_9112){
        unsigned short *pSrc = (unsigned short *)src;
        if(m_file_header.num_of_channel==1){
            for(i=0; i<num; i++)
                pDest[i] = pSrc[i]>>4;
        }
        else{  //chNum > 1
            int j, k;
            for(i=0; i<num; i+=m_file_header.num_of_channel)
                for(j=m_file_header.num_of_channel-1, k=0; j>=0; j--, k++)
                    pDest[i+k] = pSrc[i+j]>>4;
        }
    }
    else if((m_file_header.card_type==PCI_9118DG)||
        (m_file_header.card_type==PCI_9118HG))
    {
        unsigned short *pSrc = (unsigned short *)src;
        for(i=0; i<num; i++)
            pDest[i] = pSrc[i]>>4;
    }
    //------ 32bit unsigned -->
    else if( m_file_header.card_type==PCI_9113){
        long *pSrc = (long *)src;
        for(i=0; i<num; i++)
            pDest[i] = (short)(pSrc[i]&0x00000FFF);
    }
    //------ 32bit - 16bit signed data
    else if((m_file_header.card_type==PCI_9114DG)||
        (m_file_header.card_type==PCI_9114HG))
    {
        long *pSrc = (long *)src;
        for(i=0; i<num; i++)
            pDest[i] = (long)((short)(pSrc[i] & 0x0000FFFF));
    }
    //------ 16bit signed
    else if((m_file_header.card_type==PCI_9111DG)||
        (m_file_header.card_type==PCI_9812))
    {
        short *pSrc = (short *)src;
        for(i=0; i<num; i++)
            pDest[i] = (pSrc[i]>>4);
    }
    else if(m_file_header.card_type==PCI_9810){
        short *pSrc = (short *)src;
        for(i=0; i<num; i++)
            pDest[i] = (pSrc[i]>>6);
    }
    else if(m_file_header.card_type==PCI_9118HR){
        short *pSrc = (short *)src;
        for(i=0; i<num; i++)
            pDest[i] = pSrc[i];
    }
    else{  //PCI_9111HR
        short *pSrc = (short *)src;
        for(i=0; i<num; i++)
            pDest[i] = pSrc[i];
    }
}


void WriteHead()
{
    int i;
    char stBuf[20];
    char Buffer[256];
    short ctype = m_file_header.card_type;

    for(i=0; i<20&&strCardType[ctype][i]!='\0'; i++)
        stBuf[i] = strCardType[ctype][i];

    stBuf[i] = '\0';
    sprintf(Buffer, ";ADLink DAQ Data of Card %s", stBuf);
    strcat(Buffer, newLine);
    fwrite(Buffer, strlen(Buffer), 1, m_output_fp);

    if(ctype<PCI_9111DG) //Digital I/O
        return;

    Buffer[0] = '\0';
    if(m_file_header.num_of_channel==1)
        sprintf(stBuf, "Channel_%-3d", m_file_header.channel_no);
    else if(m_file_header.channel_order==0){
        for(i=0; i<m_file_header.num_of_channel; i++){
            sprintf(stBuf, "Channel_%-3d", i);
            strcat(Buffer, stBuf);
            if(i<m_file_header.num_of_channel-1)
                strcat(Buffer, strSP[m_separator]);
        }
    }
    else if(m_file_header.channel_order==1){
        for(i=m_file_header.num_of_channel-1; i>=0; i--){
            sprintf(stBuf, "Channel_%-3d", i);
            strcat(Buffer, stBuf);
            if(i>0)
                strcat(Buffer, strSP[m_separator]);
        }
    }
    else{
        for(i=0; i<m_file_header.num_of_channel; i++){
            sprintf(stBuf, "Channel_%-3d", m_pChRange[i].channel);
            strcat(Buffer, stBuf);
            if(i<m_file_header.num_of_channel-1)
                strcat(Buffer, strSP[m_separator]);
        }
    }
    strcat(Buffer, newLine);
    fwrite( Buffer, strlen(Buffer), 1, m_output_fp);
}


void ScaledDataToText(
    void *pSrc
    )
{
    stCvtScaledInfo csInfo;
    char Buffer[256], buf[16];
    float* pDest;
    int i, j, k;
    csInfo.cardType = m_file_header.card_type;
    csInfo.min = RangeValue[m_file_header.ad_range][0];
    csInfo.max = RangeValue[m_file_header.ad_range][1];
    csInfo.number = m_file_header.num_of_scan*m_file_header.num_of_channel;
    csInfo.scanCH = m_file_header.num_of_channel;
    //08/25/09
    if(m_file_header.card_type==PCI_9524)
        pDest = (float*)malloc(csInfo.number*2*sizeof(float));
    else
        pDest = (float*)malloc(csInfo.number*sizeof(float));

    if(m_file_header.channel_order!=2)
        CvtScaledData(pSrc, pDest, &csInfo);
    else
        PCvtScaledData(pSrc, pDest, csInfo.number, m_pChRange);

    if(m_head)
        WriteHead();

    Buffer[0] = '\0';
    for(i=0, k=0; i<m_file_header.num_of_scan; i++){
        for(j=0; j<m_file_header.num_of_channel; j++){
            //08/25/09
            if(m_file_header.card_type==PCI_9524){
                sprintf(buf, "%13.9f", *((double *)pDest+k));
                k++;
            }
            else
                sprintf(buf, "%11.3f", pDest[k++]);
            strcat(Buffer, buf);
            if(j<m_file_header.num_of_channel-1)
                strcat(Buffer, strSP[m_separator]);
        }
        strcat(Buffer, newLine);
        fwrite(Buffer, strlen(Buffer), 1, m_output_fp);
        Buffer[0] = '\0';
    }
    free(pDest);
}


void ScaledDataToBinary(
    void *pSrc
    )
{
    stCvtScaledInfo csInfo;
    float *pDest;
    csInfo.cardType = m_file_header.card_type;
    csInfo.min = RangeValue[m_file_header.card_type][0];
    csInfo.max = RangeValue[m_file_header.card_type][1];
    csInfo.number = m_file_header.num_of_scan*m_file_header.num_of_channel;
    csInfo.scanCH = m_file_header.num_of_channel;
    //08/25/09
    if(m_file_header.card_type==PCI_9524)
        pDest = (float *)malloc(csInfo.number*2*sizeof(float));
    else
        pDest = (float *)malloc(csInfo.number*sizeof(float));

    if(m_file_header.channel_order!=2)
        CvtScaledData(pSrc, pDest, &csInfo);
    else
        PCvtScaledData(pSrc, pDest, csInfo.number, m_pChRange);

    //08/25/09
    if(m_file_header.card_type==PCI_9524)
        fwrite((char *)pDest, sizeof(float)*2*csInfo.number, 1, m_output_fp);
    else
        fwrite((char *)pDest, sizeof(float)*csInfo.number, 1, m_output_fp);

    free(pDest);
}


void BinaryCodesToText(
    void *pSrc
    )
{
    long nData = m_file_header.num_of_scan*m_file_header.num_of_channel;

    if(m_file_header.card_type==PCI_7350){
        char Buffer[20], *szFormat;
        int i;
        if(m_head)
            WriteHead();

        szFormat = (m_digital==0)? "%12d\n":"%12X\n";
        if(m_file_header.data_width==0){
            char *pData;
            pData = (char *)pSrc;
            for(i=0; i<m_file_header.num_of_scan; i++){
                sprintf(Buffer, szFormat, pData[i]);
                fwrite(Buffer, 13, 1, m_output_fp);
            }
        }
        else if(m_file_header.data_width==1){
            unsigned short *pData = (unsigned short *)pSrc;
            for(i=0; i<m_file_header.num_of_scan; i++){
                sprintf(Buffer, szFormat, pData[i]);
                fwrite(Buffer, 13, 1, m_output_fp);
            }
        }
        else if(m_file_header.data_width==3){
            char *pData = (char *)pSrc;
            for(i=0; i<m_file_header.num_of_scan; i++){
                sprintf(Buffer, szFormat, (*((long *)(pData+i*3)))&0xffffff);
                fwrite(Buffer, 13, 1, m_output_fp);
            }
        }
        else{
            unsigned long *pData = (unsigned long *)pSrc;
            for(i=0; i<m_file_header.num_of_scan; i++){
                sprintf(Buffer, szFormat, pData[i]);
                fwrite(Buffer, 13, 1, m_output_fp);
            }
        }
    }
    else if(m_file_header.card_type>=PCI_9111DG){
        long *pDest;
        char Buffer[256], buf[16];
        int i, j, k;

        pDest = (long *)malloc(nData*sizeof(long));
        CvtBinaryCodesNCH(pSrc, pDest, nData);
        if(m_head)
            WriteHead();

        Buffer[0] = '\0';
        for(i=0, k=0; i<m_file_header.num_of_scan; i++){
            for(j=0; j<m_file_header.num_of_channel; j++){
                sprintf(buf, "%11ld", pDest[k++]);
                strcat(Buffer, buf);
                if(j<m_file_header.num_of_channel-1)
                    strcat(Buffer, strSP[m_separator]);
            }
            strcat(Buffer, newLine);
            fwrite(Buffer, strlen(Buffer), 1, m_output_fp);
            Buffer[0] = '\0';
        }
        free(pDest);
    }
    else{  //Digital I/O
        char Buffer[20], *szFormat;
        int i;
        if(m_head)
            WriteHead();

        szFormat = (m_digital==0)? "%12d\n":"%12X\n";
        if(m_file_header.data_width==0){
            char *pData;
            pData = (char*)pSrc;
            for(i=0; i<m_file_header.num_of_scan; i++){
                sprintf(Buffer, szFormat, pData[i]);
                fwrite(Buffer, 13, 1, m_output_fp);
            }
        }
        else if(m_file_header.data_width==1){
            unsigned short *pData = (unsigned short *)pSrc;
            for(i=0; i<m_file_header.num_of_scan; i++){
                sprintf(Buffer, szFormat, pData[i]);
                fwrite(Buffer, 13, 1, m_output_fp);
            }
        }
        else{
            unsigned long *pData = (unsigned long *)pSrc;
            for(i=0; i<m_file_header.num_of_scan; i++){
                sprintf(Buffer, szFormat, pData[i]);
                fwrite(Buffer, 13, 1, m_output_fp);
            }
        }
    }
}


void StartConvert()
{
    if(m_pSrc==NULL){
        printf("No source data!\n");
        return;
    }

    if(strlen(m_output_file)<=0){
        printf("No output file name!\n");
        return;
    }

    m_output_fp = fopen(m_output_file, "w+b");
    if(m_output_fp==NULL){
        printf("Open Output File Error!\n");
        return;
    }

    if(m_format==0)
        ScaledDataToText(m_pSrc);
    else if(m_format==1)
        ScaledDataToBinary(m_pSrc);
    else //if(m_format==2)
        BinaryCodesToText(m_pSrc);
    //else
        //BinaryCodesToBinary(m_pSrc);

    fclose(m_output_fp);

    printf("\n\nThe Format Conversion is completed , and write data to %s !\n\n", m_output_file);
}
